/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     demo_DecalTextureProgram.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef DEMO_DECAL_TEXTURE_PROGRAM_OBJECT_H_
#define DEMO_DECAL_TEXTURE_PROGRAM_OBJECT_H_

#include "demo/ShaderProgram/demo_ProgramObject.h"

namespace demo
{ 

    /*!
    :private
    
    @brief Shader program class for performing texture decals on polygons.       
    */

    class DecalTextureProgram : public ProgramObject
    {
    public:
        /*!
        :private

        @brief    Constructor.        
        */
        DecalTextureProgram(void);
        
        /*!
        :private

        @brief    Destructor.        
        */
        virtual ~DecalTextureProgram(void);

    public:
        /*!
        :private

        @brief    Performs initialization for the shader program object.        
        */
        virtual bool Initialize(const GLuint shaderId);

        /*!
        :private

        @brief    Performs termination for the shader program object.        
        */
        virtual bool Finalize(void);
    public:
        /*!
        :private

        @brief    Uses shader program object.        
        */
        virtual bool Use(void);
    
    public:
        /*!
        :private

        @brief    Sets a texture object handle.
       
        @param[in]    textureId    Texture object handle
        */
        virtual void SetTextureId(const GLuint textureId);

        /*!
        :private

        @brief    Updates the internal state of the shader program object.       
        */
        virtual void Update(void);

    protected:
        virtual void InitializeVertexAttributes(void);

    protected:
        GLuint m_TextureId;
    };

}

#endif
