﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     arg.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <string.h>

//----------------------------------------------------------------
// Show arguments
static void showArgv( void )
{
    int argc = nn::dbg::CTR::GetArgc();
    NN_LOG("argc = %d\n", argc );

    for( int n=0; n<argc; n++ )
    {
        NN_LOG("argv[%d] = [%s]\n", n, nn::dbg::CTR::GetArgv(n) );
    }
}

//----------------------------------------------------------------
// Show sample for function that corresponds to the getopt function
static void showGetOptSample( void )
{
    int c;
    while( (c = nn::dbg::CTR::GetOpt("AB:C::")) > 0 )
    {
        switch (c)
        {
        case 'A':
            {
                NN_LOG("*** found option A\n");
            }
            break;
        case 'B':
            {
                const char* str = nn::dbg::CTR::GetOptArg();
                NN_UNUSED_VAR(str); // To avoid release build warnings
                NN_LOG("*** found option B with [%s]\n", str? str: "(null)");
            }
            break;
        case 'C':
            {
                const char* str = nn::dbg::CTR::GetOptArg();
                NN_UNUSED_VAR(str); // To avoid release build warnings
                NN_LOG("*** found option C with [%s]\n", str? str: "(null)");
            }
            break;
        case '?':
        default:
            NN_LOG("*** Unknown option '%c'\n", nn::dbg::CTR::GetOptOpt());
            break;
        }
    }
}

//----------------------------------------------------------------
// Binary dump
static void dump( const void* pBuffer, size_t size )
{
    if ( pBuffer )
    {
        const char* p = static_cast<const char*>(pBuffer);
        NN_UNUSED_VAR(p); // To avoid release build warnings
        for ( int n=0; n<size; n++ )
        {
            if ( (n & 0xf) == 0 )
            {
                NN_LOG("%05x : ", n);
            }
            NN_LOG("%02x %s", (*p++ & 0xff), ((n&0xf)==0xf)?"\n":"" );
        }
        NN_LOG("\n");
    }
}

//----------------------------------------------------------------
// Show arguments
static void showArgumentData( void )
{
    // Show argument information
    showArgv();

    // Show binary information
    const void* pArea = nn::dbg::CTR::GetArgBinary();
    size_t size       = nn::dbg::CTR::GetArgBinarySize();
    NN_LOG("Argument binary buffer 0x%x size 0x%x\n", pArea, size);
    dump( pArea, size );
}

//----------------------------------------------------------------
// Sample
extern "C" void nnMain()
{
    NN_LOG("==== Argument sample starts.\n");

    // information display
    showArgumentData();

    // getopt sample
    NN_LOG("---- Getopt sample\n" );
    showGetOptSample();

    NN_LOG("==== Argument sample end.\n");
    while(1){}
}
