﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     device.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_DEMOS_DEMO1_DEVICE_H_
#define NN_SAMPLE_DEMOS_DEMO1_DEVICE_H_

#include "demo.h"

enum DEVICE_STATE
{
    DEVICE_STATE_OFF = 0,
    DEVICE_STATE_STANDBY,
    DEVICE_STATE_ON
};

class Device
{
public:
    Device() : mp_RenderSystem(0), m_State(DEVICE_STATE_OFF) {}
    ~Device() {}

    virtual void Initialize(demo::RenderSystemDrawing* p_RenderSystem)
    {
        mp_RenderSystem = p_RenderSystem;
        m_State = DEVICE_STATE_STANDBY;
    }

    virtual void Finalize(void)
    {
        m_State = DEVICE_STATE_OFF;
    }

    virtual void Start(void)
    {
        m_State = DEVICE_STATE_ON;
    }

    virtual void End(void)
    {
        m_State = DEVICE_STATE_STANDBY;
    }

    DEVICE_STATE GetState(void) const
    {
        return m_State;
    }

    void Draw(void)
    {
        if(m_State == DEVICE_STATE_ON)
        {
            DrawFrame();
        }
    }

    virtual void DrawFrame(void) = 0;

protected:
    demo::RenderSystemDrawing* mp_RenderSystem;
    DEVICE_STATE m_State;
    NN_PADDING3;
};

/* NN_SAMPLE_DEMOS_DEMO1_DEVICE_H_ */
#endif

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
