/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     pad.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>
#include <nn/hid.h>
#include "pad.h"

using namespace nn::hid;

// Display btn_name in the specified position.
// Display green if the button specified with btn is pressed and white if it is not pressed.
// 
#define DISP_BUTTON_WITH_COLOR(pad, btn, disp_x, disp_y, btn_name)   \
    pad.hold & (btn) ?                          \
    mp_RenderSystem->SetColor(0.f, 1.f, 0.f) :  \
    mp_RenderSystem->SetColor(1.f, 1.f, 1.f);   \
    mp_RenderSystem->DrawText(disp_x, disp_y, btn_name)

void PadDemo::DrawFrame(void)
{
    // Get the most recent pad data
    nn::hid::PadStatus padStatus;
    m_PadReader.ReadLatest(&padStatus);

    // Display button status
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_L,        50, 140, "L");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_R,        80, 140, "R");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_UP    | BUTTON_EMULATION_UP,       30, 150, "UP");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_LEFT  | BUTTON_EMULATION_LEFT,      0, 160, "LEFT");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_RIGHT | BUTTON_EMULATION_RIGHT,    40, 160, "RIGHT");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_DOWN  | BUTTON_EMULATION_DOWN,     20, 170, "DOWN");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_X,       100, 150, "X");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_Y,        90, 160, "Y");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_A,       110, 160, "A");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_B,       100, 170, "B");
    DISP_BUTTON_WITH_COLOR(padStatus, BUTTON_START,    50, 180, "START");
}

nn::hid::PadStatus PadDemo::GetLatestPadStatus(void)
{
    // Get the most recent pad data
    nn::hid::PadStatus padStatus;
    m_PadReader.ReadLatest(&padStatus);

    return padStatus;
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
