﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     Gas.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */
#ifndef DMP_GAS_H__
#define DMP_GAS_H__

#include <nn/gx.h>

/*
 * Definitions
 */

#define APP_NAME            "GasColorOptimal"

/* Buffer size */
#define DISPLAY_WIDTH       (nn::gx::DISPLAY0_WIDTH)
#define DISPLAY_HEIGHT      (nn::gx::DISPLAY0_HEIGHT)

/* accumulation buffer can be smaller size than DISPLAY but should be powered by 2. */
#define GAS_ACC_WIDTH       (128)
#define GAS_ACC_HEIGHT      (128)
/* texture should be same size as accumulation buffer */
#define GAS_TEX_WIDTH       (GAS_ACC_WIDTH)
#define GAS_TEX_HEIGHT      (GAS_ACC_HEIGHT)

/* Default buffer ID */
#define DISPLAY_BUFFER      (0)     /* default frame buffer */

/* Particle pattern */
#define PARTICLE_PATTERNS   (1)
#define PARTICLE_FILES       "rom:/resources/ptcl0.tga"


struct gas_data {
    //Gas geometry definition (accumulation pass)
    GLuint          gasgeo_tri_ID;              /* used in accumulation pass. Vertex Index buffer ID */
    GLuint          gasgeo_center_ID;           /* used in accumulation pass. Vertex position buffer ID */
    GLuint          gasgeo_tx0_ID;              /* used in accumulation pass. Vertex Texture coordinates buffer ID */
    GLuint          gasgeo_density_ID;          /* used in accumulation pass. Vertex color buffer ID */
    GLuint          pattern[PARTICLE_PATTERNS]; /* Texture ID applied to each particle */
    GLsizei         gasgeo_size;                /* Number of elements (triangles) defining the gaseous object*/
    GLfloat         _dela_z;                    /* used in accumulation pass, to control final alpha blending*/

    GLint           _autoAcc;                   /* used in post synthesis pass */
    float           _densMax;                   /* used in post synthesis pass. Shading control*/
    GLfloat         _lightDirX;                 /* used in post synthesis pass. Shading control */
    GLfloat         _lightDirY;                 /* used in post synthesis pass. Shading control */
    GLfloat         _LightXY[4];                /* used in post synthesis pass. Shading control */
    GLfloat         _LightZ[4];                 /* used in post synthesis pass. Shading control */
    GLuint          FogLut_ID;                  /* used in post synthesis pass. Look up table ID */
    GLuint          CollectionLUT_ID;           /* used in post synthesis pass. Texture collection ID */
    GLuint          gasTransfert_ID[3];         /* used in post synthesis pass. Lookup table ID for gas color look up table */
    float           RR[16],GG[16],BB[16];       /* used in post synthesis pass. Lookup table for gas color content */
    float           fogTable[256];              /* used in post synthesis pass. Fog look up table for density to alpha conversion*/
    GLenum          shadingDensitySrc;          /* used in post synthesis pass. Density selection for shading input*/
    GLenum          colorLutInput;              /* used in post synthesis pass. Input selection for the gas color look up*/
    //Gas geometry definition / simple quad (post synthesis pass)
    GLuint          quad_index_ID;              /* used in post synthesis pass. Vertex Index buffer ID */
    GLuint          quad_vertBuf_ID;            /* used in post synthesis pass. Vertex position buffer ID */
    GLuint          quad_texBuf_ID;             /* used in post synthesis pass. Vertex Texture coordinates buffer ID */
    GLuint          quad_colBuf_ID;             /* used in post synthesis pass. Vertex color buffer ID */

};

/*
 * Function declarations
 */

/* buffer initialization */
void GasInitialize(void);

/* gaseous object data structure default initialization */
void DefaultGasObject(struct gas_data *gas,float *gasColorTable);

/* accumulation pass */
void GasAccumulation();

/* shading pass */
void GasShading(void);

/* render accumulation Z buffer */
void GasRenderAccumulationZBuffer(void (*drawfunc)(void));

/* Blend shading result to color buffer */
void GasBlendShadingResult(void);

#endif /* DMP_GAS_H__ */
