/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     PartsysSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

/*
 * Comments
 * --------
 *
 * 
 * DMP Particle system sample. 
 * This sample shows various patterns that can be obtained
 * with particle systems.
 *  
 */


#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>
#include "Util.h"

#include <string.h>

#include "Memory.h"

#include "demo.h"

#define APP_NAME "PartsysSimple"

/* buffer id */
GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

/* program id */
GLuint s_ProgramID[1];

/* Procedural texture */
GLuint  s_ProcTexID[8];
int     s_ProcTexSize;

/* 2D texture */
GLuint  s_Texture2DName;

/* Structure containing particle system settings*/
struct _particleSys
{
    nn::math::Matrix44      m_partsys_center;
    nn::math::Matrix44      m_partsys_color;
    nn::math::Matrix44      m_partsys_radius;
    nn::math::Matrix44      m_partsys_aspect;

    float       m_speed;
    float       m_particleCountMax;
    float       m_size_min_max[2];
    float       m_prng[4];                      /* constant for the Pseudo Random Number Generator (modulo based) */
    float       m_random_seed[4];               /* seeds for the PRNG */

    float       simulationTime;                 /* application side. Holds the current time for particle system */
    float       dTime;                          /* application side. Holds the time interval between each frame */
    int         NFrame;                         /* application side. Number of frames for this particle system */
};

#include "PartsysPattern.h"

struct _particleSys pSys[PARTSYS_LAST];
int CurSys = 0;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

demo::RenderSystem s_RenderSystem;

/* Load a 2D texture (DMP Letters)*/
void LoadTexture2D(void)
{
    int _lev = 0;
    glGenTextures(1, &s_Texture2DName);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, s_Texture2DName);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, 0);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    bool bUseAlpha;
    loadTexture( "rom:/resources/dmp.tga", GL_TEXTURE_2D, _lev, bUseAlpha);
}

/* Procedural texture: RGBA texture definition
 a simple white gradient defined as 2 colors */
int Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA)
{
    memset(RR, 0, 512 * sizeof(GLfloat));
    memset(GG, 0, 512 * sizeof(GLfloat));
    memset(BB, 0, 512 * sizeof(GLfloat));
    memset(AA, 0, 512 * sizeof(GLfloat));

    RR[0] = 0xFF / 255.0f;  GG[0] = 0xFF / 255.0f;  BB[0] = 0xFF / 255.0f;  AA[0] = 0xFF / 255.0f;
    RR[1] = 0x00 / 255.0f;  GG[1] = 0x00 / 255.0f;  BB[1] = 0x00 / 255.0f;  AA[1] = 0x00 / 255.0f;

    for (int x = 0; x < 255; x++)
    {
        RR[256 + x] = RR[x + 1] - RR[x];
        GG[256 + x] = GG[x + 1] - GG[x];
        BB[256 + x] = BB[x + 1] - BB[x];
        AA[256 + x] = AA[x + 1] - AA[x];
    }
    RR[255] = 0.0f;
    GG[255] = 0.0f;
    BB[255] = 0.0f;
    AA[255] = 0.0f;
    
    return 2;
}

/* Procedural texture: mapping function
 bias function in [0.0.5]. zero otherwise */
void LoadBias(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float aa = 1.0f / p;
    float t;
    for (unsigned int i = 0; i < 128; i++)
    {
        t = x;
        if (t >= 1.0f)
            map_1D2D_LUT[i] = 1.0f;
        else
            map_1D2D_LUT[i] = t / ((aa - 2.0f) * (1.0f - t) + 1.0f);
        x += dx;
    }
    for (unsigned int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

/* Procedural texture: mapping function
 potential function in [0.0.5]. zero otherwise */
void LoadPotential(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float f = 0, d2;
    
    for (unsigned int i = 0; i < 128; i++)
    {
        f = 0;
        d2 = x;
        if (d2 < 1.0f)
        {
            if (d2 < 0.25f)
                f = 1.0f - ((3.0f * d2) * (3.0f * d2)) / (p + (4.5f - 4.0f * p) * d2);
            else 
                f = ((1.0f - d2) * (1.0f - d2)) / (0.75f - p + (1.5f + 4.0f * p) * d2);
        }
        
        map_1D2D_LUT[i] = f;
        x += dx;
    }
    
    for (unsigned int i = 0; i < 126; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

void LoadProceduralTexture(GLuint pgID)
{
    (void)pgID;
    glGenTextures(8, &s_ProcTexID[0]);
    
    /* Binding of the procedural texture as a texture collection */
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_ProcTexID[0]);
    
    /* Declaration */
    float F_FunctionRGB[256], F_FunctionZero[256];
    LoadBias(F_FunctionRGB, 0.8f);
    memset(F_FunctionZero, 0, 256 * sizeof(float));

    /* Texture collection */
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_ProcTexID[1]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionRGB);

    glBindTexture(GL_LUT_TEXTURE1_DMP, s_ProcTexID[2]);
    glTexImage1D(GL_LUT_TEXTURE1_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionZero);
    
    glBindTexture(GL_LUT_TEXTURE2_DMP, s_ProcTexID[3]);
    glTexImage1D(GL_LUT_TEXTURE2_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionZero);

    const int NN = 512;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    s_ProcTexSize = Set1DTex(RR,GG,BB,AA);

    glBindTexture(GL_LUT_TEXTURE3_DMP, s_ProcTexID[4]);
    glTexImage1D(GL_LUT_TEXTURE3_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, RR);

    glBindTexture(GL_LUT_TEXTURE4_DMP, s_ProcTexID[5]);
    glTexImage1D(GL_LUT_TEXTURE4_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, GG);

    glBindTexture(GL_LUT_TEXTURE5_DMP, s_ProcTexID[6]);
    glTexImage1D(GL_LUT_TEXTURE5_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, BB);

    glBindTexture(GL_LUT_TEXTURE6_DMP, s_ProcTexID[7]);
    glTexImage1D(GL_LUT_TEXTURE6_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, AA);
}

void ValidateParticleSystem(GLuint pgID, struct _particleSys *pp, unsigned int _width, unsigned int _height)
{
    glUniformMatrix4fv(glGetUniformLocation(pgID, "uCenter"), 1, GL_TRUE, static_cast<f32*>(pp->m_partsys_center));

    glUniformMatrix4fv(glGetUniformLocation(pgID, "uRadii"), 1, GL_TRUE, static_cast<f32*>(pp->m_partsys_radius));

    glUniformMatrix4fv(glGetUniformLocation(pgID, "dmp_PartSys.color"), 1, GL_TRUE, static_cast<f32*>(pp->m_partsys_color));

    glUniformMatrix4fv(glGetUniformLocation(pgID, "dmp_PartSys.aspect"), 1, GL_TRUE, static_cast<f32*>(pp->m_partsys_aspect));

    GLfloat viewport_size[2] = {0.0f, 0.0f};
    viewport_size[0] = 1.f / (GLfloat)_width;
    viewport_size[1] = 1.f / (GLfloat)_height;
    glUniform2fv(glGetUniformLocation(pgID, "dmp_PartSys.viewport"), 1, &viewport_size[0]);
    
    GLfloat distance_attenuation[3] = {0.0f, 0.0f, 0.0f};
    distance_attenuation[0] = 0.0f;
    distance_attenuation[1] = 0.0f;
    distance_attenuation[2] = 1.0f / (float)(_width * _height);
    glUniform3fv(glGetUniformLocation(pgID, "dmp_PartSys.distanceAttenuation"), 1, &distance_attenuation[0]);
    
    glUniform1fv(glGetUniformLocation(pgID, "dmp_PartSys.countMax"), 1, &pp->m_particleCountMax);
    glUniform2fv(glGetUniformLocation(pgID, "dmp_PartSys.pointSize"), 1, &pp->m_size_min_max[0]);
    glUniform1fv(glGetUniformLocation(pgID, "dmp_PartSys.speed"), 1, &pp->m_speed);

    glUniform4fv(glGetUniformLocation(pgID, "dmp_PartSys.randomCore"), 1, &pp->m_prng[0]);
    glUniform4fv(glGetUniformLocation(pgID, "dmp_PartSys.randSeed"), 1, &pp->m_random_seed[0]);
}

void ValidateProceduralTexture(GLuint pgID)
{
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerRgbMap"), 0);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerAlphaMap"), 1);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerNoiseMap"), 2);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerR"), 3);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerG"), 4);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerB"), 5);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptSamplerA"), 6);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptMinFilter"), GL_LINEAR);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptTexWidth"), s_ProcTexSize);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptTexOffset"), 0);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptAlphaSeparate"), GL_FALSE);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptClampU"), GL_CLAMP_TO_EDGE);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptClampV"), GL_CLAMP_TO_EDGE);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptShiftU"), GL_NONE_DMP);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptShiftV"), GL_NONE_DMP);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptRgbMap"), GL_PROCTEX_MAX_DMP);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].ptAlphaMap"), GL_PROCTEX_MAX_DMP);
}

void ValidateTextureBlenderProcTex(GLuint pgID)
{
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_ProcTexID[0]);
    glUniform1i ( glGetUniformLocation(pgID, "dmp_Texture[3].samplerType"), GL_TEXTURE_PROCEDURAL_DMP);
    glUniform1i ( glGetUniformLocation(pgID, "dmp_Texture[3].texcoord"), GL_TEXTURE2);
    glUniform1i ( glGetUniformLocation(pgID, "dmp_Texture[0].samplerType"), GL_FALSE);
    glUniform1i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].combineRgb"), GL_MODULATE);
    glUniform1i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].combineAlpha"), GL_MODULATE);
    glUniform3i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE3, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i ( glGetUniformLocation(pgID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE3, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
}

void ValidateTextureBlenderTX0(GLuint pgID)
{
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, s_Texture2DName);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[3].samplerType"), GL_FALSE);
    glUniform1i(glGetUniformLocation(pgID, "dmp_Texture[0].samplerType"), GL_TEXTURE_2D);
    glUniform1i(glGetUniformLocation(pgID, "dmp_TexEnv[0].combineRgb"), GL_MODULATE);
    glUniform1i(glGetUniformLocation(pgID, "dmp_TexEnv[0].combineAlpha"), GL_MODULATE);
    glUniform3i(glGetUniformLocation(pgID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(pgID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(pgID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE0, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(pgID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE0, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
}

void DrawParticleSystem(GLuint pgID, struct _particleSys *pp, float _Psys_time)
{
    (void)pp;
    glUniform1fv(glGetUniformLocation(pgID, "dmp_PartSys.time"), 1, &_Psys_time);
    glDrawElements(GL_GEOMETRY_PRIMITIVE_DMP, 4, GL_UNSIGNED_SHORT, 0);
}

int DrawFrame(void)
{
    static int frame = 0;
    static bool switchTexture = false;
    static bool useProctex = false;
    static float t = 0.f;
    
    nn::math::Matrix44 proj, m;
    nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);
    
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.0f, 0.0f, 50.0f);
    nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
    nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    if (frame == 0)
    {
        /* reset uniforms */
        ValidateParticleSystem(s_ProgramID[0], &pSys[CurSys], nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID[0], "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));
        
        m = nn::math::Matrix44(mv);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID[0], "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));
    }
    
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();
    
    if (CurSys == PARTSYS_EXPLOSION2)
    {
        /* As an illustration, we draw 2 times the particle system. */
        m = nn::math::Matrix44(mv);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID[0], "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));
        
        DrawParticleSystem(s_ProgramID[0], &pSys[CurSys], t);
        
        nn::math::Matrix34 mv_rot;
        nn::math::MTX34RotXYZDeg(&mv_rot, 0.0f, 0.0f, 45.0f);
        nn::math::MTX34Mult(&mv_rot, &mv, &mv_rot);
        m = nn::math::Matrix44(mv_rot);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID[0], "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));
        DrawParticleSystem(s_ProgramID[0], &pSys[CurSys], t);
    }
    else
        DrawParticleSystem(s_ProgramID[0], &pSys[CurSys], t);
    
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();
    
    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    if (++frame == pSys[CurSys].NFrame)
    {
        /* Changing 2D texture / procedural texture */
        if (switchTexture)
        {
            if (!useProctex)
            {
                ValidateTextureBlenderTX0(s_ProgramID[0]);
                useProctex = true;
            }
            else
            {
                ValidateProceduralTexture(s_ProgramID[0]);
                ValidateTextureBlenderProcTex(s_ProgramID[0]);
                useProctex = false;
            }

            switchTexture = false;
        }
        /* change to next particle system */
        CurSys = ++CurSys % PARTSYS_LAST;
        frame = 0;
        t = 0.f;

        /* Changing blending mode for the spinning pattern */
        glBlendFunc(GL_SRC_ALPHA, (CurSys == PARTSYS_SPIN) ? GL_ONE_MINUS_SRC_ALPHA : GL_ONE);
        switchTexture = (CurSys == PARTSYS_LAST - 1);
    }
    else
        t += pSys[CurSys].dTime;
    
    return !glGetError();
}

/* initialization */
static int Initialize(void)
{
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);
    
    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClearDepthf(1.f);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    glDisable(GL_CULL_FACE);

    /* create program and load & attach vertex shader */
    GLuint shaders[2] ;
    shaders[0] = glCreateShader(GL_VERTEX_SHADER);
    shaders[1] = glCreateShader(GL_GEOMETRY_SHADER_DMP);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);
    
    s32 read = file.Read(buf, fileSize);
    glShaderBinary(2, shaders, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    s_ProgramID[0] = glCreateProgram();
    glAttachShader(s_ProgramID[0], shaders[0]);
    glAttachShader(s_ProgramID[0], shaders[1]);
    glAttachShader(s_ProgramID[0], GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_ProgramID[0], 0, "attrCtrPointIndex");

    glLinkProgram(s_ProgramID[0]);
    glValidateProgram(s_ProgramID[0]);
    glUseProgram(s_ProgramID[0]);
    
    LoadProceduralTexture(s_ProgramID[0]);
    LoadTexture2D();
    LoadParticlePattern(&pSys[PARTSYS_DEFAULT], PARTSYS_DEFAULT);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION1], PARTSYS_EXPLOSION1);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION2], PARTSYS_EXPLOSION2);
    LoadParticlePattern(&pSys[PARTSYS_SNOW], PARTSYS_SNOW);
    LoadParticlePattern(&pSys[PARTSYS_SPIN], PARTSYS_SPIN);
    
    /* make vertex buffer */
    GLfloat vertex[4] = {0.0f, 1.0f, 2.0f, 3.0f};
    GLushort index[4] = {0, 1, 2, 3};
    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat), &vertex[0], GL_STATIC_DRAW);
    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 1, GL_FLOAT, GL_FALSE, 0, 0);
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * sizeof(GLushort), &index[0], GL_STATIC_DRAW);

    /* set texture configuration */
    //validateTextureBlender_TX0(s_ProgramID[0]);
    ValidateProceduralTexture(s_ProgramID[0]);
    ValidateTextureBlenderProcTex(s_ProgramID[0]);
    
    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            (void)DrawFrame();
        }
    }
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

