/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProctexClamp.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */
#include <string.h>

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "demo.h"

// buffer id
GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

// program id
GLuint s_PgID;

// shader id
GLuint s_ShID;

/*Specific to the sample*/
GLuint  s_TexID[8];

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

demo::RenderSystem s_RenderSystem;

/* generate simple object */
static void LoadObjects(void)
{
        /* Setting Vertex Coordinates */
    GLfloat vertex[16] =
    {
        -1.0f,-1.0f,0.0f,1.0f,
        -1.0f, 1.0f,0.0f,1.0f,
         1.0f, 1.0f,0.0f,1.0f,
         1.0f,-1.0f,0.0f,1.0f
    };
    /* Setting Texture Coordinates */
    GLfloat t1 = 0.0f, t2 = 4.0f;
    GLfloat texCoord[8] =
    {
        t1, t1,
        t1, t2,
        t2, t2,
        t2, t1,
    };
    /* Setting Indices */
    GLushort _quadIndex[6] = {0, 2, 1, 0, 3, 2};

    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, 28 * sizeof(GLfloat), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, 16 * sizeof(GLfloat), vertex);
    glBufferSubData(GL_ARRAY_BUFFER, 16 * sizeof(GLfloat), 8 * sizeof(GLfloat), texCoord);
    
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6 * sizeof(GLfloat), _quadIndex, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (GLvoid*)(16 * sizeof(GLfloat)));
}

void Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA)
{
    memset(RR, 0, 256 * sizeof(GLfloat));
    memset(GG, 0, 256 * sizeof(GLfloat));
    memset(BB, 0, 256 * sizeof(GLfloat));
    memset(AA, 0, 256 * sizeof(GLfloat));

    RR[7]=0xFF/255.0f;  GG[7]=0xFF/255.0f;  BB[7]=0xFF/255.0f;  AA[7]=0xFF/255.0f;
    RR[6]=0xFF/255.0f;  GG[6]=0xFF/255.0f;  BB[6]=0x00/255.0f;  AA[6]=0xFF/255.0f;
    RR[5]=0x00/255.0f;  GG[5]=0xFF/255.0f;  BB[5]=0xFF/255.0f;  AA[5]=0xFF/255.0f;
    RR[4]=0x00/255.0f;  GG[4]=0xFF/255.0f;  BB[4]=0x00/255.0f;  AA[4]=0xFF/255.0f;
    RR[3]=0xFF/255.0f;  GG[3]=0x00/255.0f;  BB[3]=0xFF/255.0f;  AA[3]=0xFF/255.0f;
    RR[2]=0xFF/255.0f;  GG[2]=0x00/255.0f;  BB[2]=0x00/255.0f;  AA[2]=0xFF/255.0f;
    RR[1]=0x00/255.0f;  GG[1]=0x00/255.0f;  BB[1]=0xFF/255.0f;  AA[1]=0xFF/255.0f;
    RR[0]=0x00/255.0f;  GG[0]=0x00/255.0f;  BB[0]=0x00/255.0f;  AA[0]=0xFF/255.0f;

    /*RR[7]=0xFFFFFFFF; GG[7]=0xFFFFFFFF;   BB[7]=0xFFFFFFFF;   AA[7]=0xFFFFFFFF;
    RR[6]=0xFF00FFFF;   GG[6]=0xFF00FFFF;   BB[6]=0xFF00FFFF;   AA[6]=0xFF00FFFF;
    RR[5]=0xFFFFFF00;   GG[5]=0xFFFFFF00;   BB[5]=0xFFFFFF00;   AA[5]=0xFFFFFF00;
    RR[4]=0xFF00FF00;   GG[4]=0xFF00FF00;   BB[4]=0xFF00FF00;   AA[4]=0xFF00FF00;
    RR[3]=0xFFFF00FF;   GG[3]=0xFFFF00FF;   BB[3]=0xFFFF00FF;   AA[3]=0xFFFF00FF;
    RR[2]=0xFF0000FF;   GG[2]=0xFF0000FF;   BB[2]=0xFF0000FF;   AA[2]=0xFF0000FF;
    RR[1]=0xFFFF0000;   GG[1]=0xFFFF0000;   BB[1]=0xFFFF0000;   AA[1]=0xFFFF0000;
    RR[0]=0xFF000000;   GG[0]=0xFF000000;   BB[0]=0xFF000000;   AA[0]=0xFF000000;*/

    for (int i = 0; i < 7; i++)
    {
        RR[i + 256] = RR[i + 1] - RR[i];
        GG[i + 256] = GG[i + 1] - GG[i];
        BB[i + 256] = BB[i + 1] - BB[i];
        AA[i + 256] = AA[i + 1] - AA[i];
    }
    RR[7 + 256] = 0;
    GG[7 + 256] = 0;
    BB[7 + 256] = 0;
    AA[7 + 256] = 0;
}

void LoadLamp(float map_1D2D_LUT[256])
{
    for (int i = 0; i < 128; i++)
        map_1D2D_LUT[i] = 1.f - i / 128.f;

    for (int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[127 + 128] = (1.f - 128.f / 128.f) - map_1D2D_LUT[127];
}

void LoadFlat(float map_1D2D_LUT[256], float a)
{
    for (int i = 0; i < 128; i++)
    {
        map_1D2D_LUT[i] = a;
        map_1D2D_LUT[i + 128] = 0;
    }
}

void LoadProceduralTexture(void)
{
    glGenTextures(8, &s_TexID[0]);

    /*Binding of the procedural texture as a texture collection*/
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_TexID[0]);
    
    /* Declaration */
    float F_FunctionRGB[256], F_FunctionA[256], F_Noise[256];
    LoadLamp(F_FunctionRGB);
    LoadFlat(F_FunctionA, 1.0f);
    memset(F_Noise, 0, 256 * sizeof(float));

    //Texture collection
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_TexID[1]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionRGB);

    glBindTexture(GL_LUT_TEXTURE1_DMP, s_TexID[2]);
    glTexImage1D(GL_LUT_TEXTURE1_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionA);
    
    glBindTexture(GL_LUT_TEXTURE2_DMP, s_TexID[3]);
    glTexImage1D(GL_LUT_TEXTURE2_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_Noise);

    const int NN = 8 + 256;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    Set1DTex(RR, GG, BB, AA);

    glBindTexture(GL_LUT_TEXTURE3_DMP, s_TexID[4]);
    glTexImage1D(GL_LUT_TEXTURE3_DMP, 0, GL_LUMINANCEF_DMP, NN, 0, GL_LUMINANCEF_DMP, GL_FLOAT, RR);

    glBindTexture(GL_LUT_TEXTURE4_DMP, s_TexID[5]);
    glTexImage1D(GL_LUT_TEXTURE4_DMP, 0, GL_LUMINANCEF_DMP, NN, 0, GL_LUMINANCEF_DMP, GL_FLOAT, GG);

    glBindTexture(GL_LUT_TEXTURE5_DMP, s_TexID[6]);
    glTexImage1D(GL_LUT_TEXTURE5_DMP, 0, GL_LUMINANCEF_DMP, NN, 0, GL_LUMINANCEF_DMP, GL_FLOAT, BB);

    glBindTexture(GL_LUT_TEXTURE6_DMP, s_TexID[7]);
    glTexImage1D(GL_LUT_TEXTURE6_DMP, 0, GL_LUMINANCEF_DMP, NN, 0, GL_LUMINANCEF_DMP, GL_FLOAT, AA);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerRgbMap"), 0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerAlphaMap"), 1);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerNoiseMap"), 2);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerR"), 3);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerG"), 4);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerB"), 5);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerA"), 6);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].samplerType"), GL_TEXTURE_PROCEDURAL_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].texcoord"), GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptMinFilter"), GL_LINEAR);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexWidth"), 8);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexOffset"), 0);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptRgbMap"), GL_PROCTEX_MAX_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaMap"), GL_PROCTEX_MAX_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaSeparate"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[0].samplerType"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);

    float black[4]={0.3f, 0.2f, 0.2f, 1.0f};
    glUniform4fv(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].constRgba"), 1, black);
}

void TestClamp(GLint fclamp, GLint shiftU, GLint shiftV)
{
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampU"), fclamp);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampV"), fclamp);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftU"), shiftU);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftV"), shiftV);
}

int DrawFrame(void)
{
    static int f = 0;
    GLenum clamp;

    /* Setting Screen */
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();

    nn::math::Matrix44 proj;
    nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 0.f, 25.0f);
    nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
    nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    
    glDisable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    
    /* variable clamp must be GL_SYMMETRICAL_REPEAT_DMP or GL_MIRRORED_REPEAT */
    clamp = ((f / 100) % 2) ? GL_SYMMETRICAL_REPEAT_DMP : GL_MIRRORED_REPEAT;
    
    GLenum _shift[9][2]=
    {
        { GL_NONE_DMP,  GL_NONE_DMP },
        { GL_NONE_DMP,  GL_ODD_DMP  },
        { GL_NONE_DMP,  GL_EVEN_DMP },
        { GL_ODD_DMP,   GL_NONE_DMP },
        { GL_ODD_DMP,   GL_ODD_DMP  },
        { GL_ODD_DMP,   GL_EVEN_DMP },
        { GL_EVEN_DMP,  GL_NONE_DMP },
        { GL_EVEN_DMP,  GL_ODD_DMP  },
        { GL_EVEN_DMP,  GL_EVEN_DMP }
    };
    
    /* Drawing 9 Rectangles */
    for(int i = 0; i < 9; i++)
    {
        GLfloat x = ((GLfloat)(i % 3) - 1.0f) * 4.f;
        GLfloat y = ((GLfloat)(i / 3) - 1.0f) * 4.f;
        nn::math::Matrix34 mv_tr, tmp[2];
        nn::math::Vector3 trans(x, y, 0.0f);
        nn::math::Vector3 scale(1.7f, 1.7f, 1.7f);
        
        nn::math::MTX34Translate(&tmp[0], &trans);
        nn::math::MTX34Scale(&tmp[1], &scale);
        nn::math::MTX34Mult(&mv_tr, &mv, &tmp[0]);
        nn::math::MTX34Mult(&mv_tr, &mv_tr, &tmp[1]);
        nn::math::Matrix44 m(mv_tr);
        glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));
        
        TestClamp(clamp, _shift[i][0], _shift[i][1]);
        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);
    }

    glFinish();

    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH); 

    f++;
    return !glGetError();
}

/* initialization */
static int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);
    
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aTexCoord");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    glClearColor(0.36f, 0.42f, 0.5f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    LoadProceduralTexture();
    LoadObjects();
    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

