/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProctexRandom.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */
#include <string.h>

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "demo.h"

/* buffer id */
GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

/* program id */
GLuint s_PgID;

/* shader id */
GLuint s_ShID;

/*Specific to the sample*/
GLuint  s_TexID[8];

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

demo::RenderSystem s_RenderSystem;

/* generate simple object */
static void LoadObjects(void)
{
    /* Setting Vertex Coordinates */
    GLfloat vertex[16] =
    {
        -1.0f,-1.0f,0.0f,1.0f,
        -1.0f, 1.0f,0.0f,1.0f,
         1.0f, 1.0f,0.0f,1.0f,
         1.0f,-1.0f,0.0f,1.0f
    };
    /* Setting Texture Coordinates */
    GLfloat t1 = -1.0f, t2 = 1.0f;
    GLfloat texCoord[8] =
    {
        t1, t1,
        t1, t2,
        t2, t2,
        t2, t1,
    }; 
    /* Setting Indices */
    GLushort _quadIndex[6]={0, 2, 1, 0, 3, 2};

    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, 28 * sizeof(GLfloat), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, 16 * sizeof(GLfloat), vertex);
    glBufferSubData(GL_ARRAY_BUFFER, 16 * sizeof(GLfloat), 8 * sizeof(GLfloat), texCoord);
    
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6 * sizeof(GLfloat), _quadIndex, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (GLvoid*)(16 * sizeof(GLfloat)));
}

int Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA, int N)
{
    memset(RR, 0, 512 * sizeof(GLfloat));
    memset(GG, 0, 512 * sizeof(GLfloat));
    memset(BB, 0, 512 * sizeof(GLfloat));
    memset(AA, 0, 512 * sizeof(GLfloat));

    RR[7*N]=0xFF/255.0f;    GG[7*N]=0xFF/255.0f;    BB[7*N]=0xFF/255.0f;    AA[7*N]=0xFF/255.0f;
    RR[6*N]=0xFF/255.0f;    GG[6*N]=0xFF/255.0f;    BB[6*N]=0x00/255.0f;    AA[6*N]=0xFF/255.0f;
    RR[5*N]=0x00/255.0f;    GG[5*N]=0xFF/255.0f;    BB[5*N]=0xFF/255.0f;    AA[5*N]=0xFF/255.0f;
    RR[4*N]=0x00/255.0f;    GG[4*N]=0xFF/255.0f;    BB[4*N]=0x00/255.0f;    AA[4*N]=0xFF/255.0f;
    RR[3*N]=0xFF/255.0f;    GG[3*N]=0x00/255.0f;    BB[3*N]=0xFF/255.0f;    AA[3*N]=0xFF/255.0f;
    RR[2*N]=0xFF/255.0f;    GG[2*N]=0x00/255.0f;    BB[2*N]=0x00/255.0f;    AA[2*N]=0xFF/255.0f;
    RR[1*N]=0x00/255.0f;    GG[1*N]=0x00/255.0f;    BB[1*N]=0xFF/255.0f;    AA[1*N]=0xFF/255.0f;
    RR[0*N]=0x00/255.0f;    GG[0*N]=0x00/255.0f;    BB[0*N]=0x00/255.0f;    AA[0*N]=0xFF/255.0f;

    for (int x = 0; x < 255; x++)
    {
        RR[256 + x] = RR[x + 1] - RR[x];
        GG[256 + x] = GG[x + 1] - GG[x];
        BB[256 + x] = BB[x + 1] - BB[x];
        AA[256 + x] = AA[x + 1] - AA[x];
    }
    RR[255] = 0.0;
    GG[255] = 0.0;
    BB[255] = 0.0;
    AA[255] = 0.0;

    return 8 * N;
}

void LoadLamp(float map_1D2D_LUT[256])
{
    for (int i = 0; i < 128; i++)
        map_1D2D_LUT[i] = 1.f - i / 127.f;

    for (int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];

    map_1D2D_LUT[255] = 0.f;
}

void LoadFlat(float map_1D2D_LUT[256], float a)
{
    for (int i = 0; i < 128; i++)
    {
        map_1D2D_LUT[i] = a;
        map_1D2D_LUT[i + 128] = 0;
    }
}

void LoadC3(float F_Noise[256])
{
    float x = 0;
    float t, t_last;
    float dx = 1.0f / 128.f;
    for (unsigned int kk = 0; kk < 128; kk++)
    {
        t = x * x * (3 - 2 * x);    /* c3 */
        F_Noise[kk] = t;
        x += dx;
    }
    t_last = x * x * (3 - 2 * x);

    for (int i = 0; i < 127; i++)
        F_Noise[i + 128] = F_Noise[i + 1] - F_Noise[i];
    F_Noise[255] = t_last - F_Noise[127];
}

void LoadProceduralTexture(void)
{
    glGenTextures(8, &s_TexID[0]);

    /* Binding of the procedural texture as a texture collection */
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_TexID[0]);
    
    /* Declaration */
    float F_FunctionRGB[256], F_FunctionA[256], F_Noise[256];
    LoadLamp(F_FunctionRGB);
    LoadFlat(F_FunctionA, 1.0f);
    LoadC3(F_Noise);

    /* Texture collection */
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_TexID[1]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionRGB);

    glBindTexture(GL_LUT_TEXTURE1_DMP, s_TexID[2]);
    glTexImage1D(GL_LUT_TEXTURE1_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_FunctionA);
    
    glBindTexture(GL_LUT_TEXTURE2_DMP, s_TexID[3]);
    glTexImage1D(GL_LUT_TEXTURE2_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, GL_LUMINANCEF_DMP, GL_FLOAT, F_Noise);

    const int NN = 512;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    
    int texSize = Set1DTex(RR, GG, BB, AA, 2);

    glBindTexture(GL_LUT_TEXTURE3_DMP, s_TexID[4]);
    glTexImage1D(GL_LUT_TEXTURE3_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, RR);

    glBindTexture(GL_LUT_TEXTURE4_DMP, s_TexID[5]);
    glTexImage1D(GL_LUT_TEXTURE4_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, GG);

    glBindTexture(GL_LUT_TEXTURE5_DMP, s_TexID[6]);
    glTexImage1D(GL_LUT_TEXTURE5_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, BB);

    glBindTexture(GL_LUT_TEXTURE6_DMP, s_TexID[7]);
    glTexImage1D(GL_LUT_TEXTURE6_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, AA);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerRgbMap"), 0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerAlphaMap"), 1);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerNoiseMap"), 2);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerR"), 3);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerG"), 4);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerB"), 5);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerA"), 6);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].samplerType"), GL_TEXTURE_PROCEDURAL_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].texcoord"), GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptMinFilter"), GL_LINEAR);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexWidth"), texSize);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexOffset"), 0);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptRgbMap"), GL_PROCTEX_ADDSQRT2_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaMap"), GL_PROCTEX_U_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaSeparate"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampU"), GL_CLAMP_TO_EDGE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampV"), GL_CLAMP_TO_EDGE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftU"), GL_NONE_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftV"), GL_NONE_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseEnable"), GL_TRUE);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[0].samplerType"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
}

int DrawFrame(void)
{
    static int f = 0;
    GLfloat t = (f % 1000) / 100.0f;
    
    /* Setting Screen */
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();   

    nn::math::Matrix44 proj;
    nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.0f, 0.0f, 25.0f);
    nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
    nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE);
    
    GLfloat noiseValue[9][3] =
    {
        { 0.2f, t     , 0.1f },
        { 0.5f, t     , 0.1f },
        { 1.0f, t     , 0.1f },
        { 0.5f, t*0.1f, 0.1f },
        { 0.5f, t*2.0f, 0.1f },
        { 0.5f, t*4.0f, 0.1f },
        { 0.5f, t     , 0.1f },
        { 0.5f, t     , 0.4f },
        { 0.5f, t     , 0.6f }
    };
     
    /* Drawing 9 Rectangles */
    for(int i = 0; i < 9; i++)
    {
        GLfloat x = ((GLfloat)(i % 3) - 1.0f) * 4.f;
        GLfloat y = ((GLfloat)(i / 3) - 1.0f) * 4.f;
        nn::math::Matrix34 mv_tr, tmp[2];
        nn::math::Vector3 trans(x, y, 0.f);
        nn::math::Vector3 scale(1.7f, 1.7f, 1.7f);
        nn::math::MTX34Translate(&tmp[0], &trans);
        nn::math::MTX34Scale(&tmp[1], &scale);
        nn::math::MTX34Mult(&mv_tr, &mv, &tmp[0]);
        nn::math::MTX34Mult(&mv_tr, &mv_tr, &tmp[1]);
        nn::math::Matrix44 m(mv_tr);
        glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));
        glUniform3fv(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseU"), 1, noiseValue[i]);
        glUniform3fv(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseV"), 1, noiseValue[i]);

        glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);
    }

    glFinish();

    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
    
    f++;
    return !glGetError();
}

/* initialization */
static int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aTexCoord");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    LoadProceduralTexture();
    LoadObjects();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

