/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProjectionTexture.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2011 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>

#include "demo.h"

#include "Util.h"
#include "Loader.h"

#include <string.h>
#include <math.h>

#include "Memory.h"

#define APP_NAME "ProjectionTexture"

#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

/* obj file loader class object */
dat_t robot;

/* program id */
GLuint s_ProgramID;

/* shader id */
GLuint s_ShaderID;

/* cube model buffer */
GLuint s_CubeArrayBufferID;
GLuint s_CubeElementArrayBufferID;
GLsizei s_CubeNumIndex;
GLsizei s_CubePosOffset;
GLsizei s_CubeNrmOffset;

/* projection texture id */
GLuint s_ProjTexID;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

demo::RenderSystem s_RenderSystem;

/* generate simple object */
static void LoadObjects(void)
{
    /* load obj file geometry and diffuse texture */
    loadDAT( "rom:/resources/robot1.dat", &robot);

    /* load cube model */
    GLfloat pos[] = {
        -1.f,-1.f,1.f,  1.f,1.f,1.f, -1.f,1.f,1.f, 1.f,-1.f,1.f,        // Front
        -1.f,-1.f,-1.f,  1.f,1.f,-1.f, -1.f,1.f,-1.f, 1.f,-1.f,-1.f,    // Back
        1.f,-1.f,-1.f,  1.f,1.f,1.f, 1.f,-1.f,1.f, 1.f,1.f,-1.f,        // Right
        -1.f,-1.f,-1.f,  -1.f,1.f,1.f, -1.f,-1.f,1.f, -1.f,1.f,-1.f,    // Left
        -1.f,-1.f,-1.f, 1.f,-1.f,1.f, -1.f,-1.f,1.f, 1.f,-1.f,-1.f,     // Bottom
        -1.f,1.f,-1.f, 1.f,1.f,1.f, -1.f,1.f,1.f, 1.f,1.f,-1.f          // Top
    };
    GLfloat nrm[] = {
        0.f,0.f,1.f, 0.f,0.f,1.f, 0.f,0.f,1.f, 0.f,0.f,1.f,
        0.f,0.f,-1.f, 0.f,0.f,-1.f, 0.f,0.f,-1.f, 0.f,0.f,-1.f,
        1.f,0.f,0.f, 1.f,0.f,0.f, 1.f,0.f,0.f, 1.f,0.f,0.f, 
        -1.f,0.f,0.f, -1.f,0.f,0.f, -1.f,0.f,0.f, -1.f,0.f,0.f, 
        0.f,-1.f,0.f, 0.f,-1.f,0.f, 0.f,-1.f,0.f, 0.f,-1.f,0.f, 
        0.f,1.f,0.f, 0.f,1.f,0.f, 0.f,1.f,0.f, 0.f,1.f,0.f
    };
    GLushort idx[] = {
        0,1,2, 0,3,1,
        4,6,5, 4,5,7,
        8,9,10, 8,11,9,
        12,14,13, 12,13,15,
        16,17,18, 16,19,17,
        20,22,21, 20,21,23
    };
    glGenBuffers(1, &s_CubeArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_CubeArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, sizeof(pos)+sizeof(nrm), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(pos), pos);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(pos), sizeof(nrm), nrm);
    s_CubePosOffset = 0;
    s_CubeNrmOffset = sizeof(pos);

    glGenBuffers(1, &s_CubeElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_CubeElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);
    s_CubeNumIndex = sizeof(idx) / sizeof(GLushort);
}

static void UnloadObjects(void)
{
    unloadDAT(&robot);
    
    glDeleteBuffers(1, &s_CubeArrayBufferID);
    glDeleteBuffers(1, &s_CubeElementArrayBufferID);
    glDeleteTextures(1, &s_ProjTexID);
    
    return;
}

int DrawFrame(void)
{
    static int f = 0;
    
    nn::math::Matrix44 proj, tmp44;
    nn::math::Matrix34 mv, view, texview, model, tmp34;
    nn::math::Vector3 vec3;

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();  

    /* setup projection matrix */
    nn::math::MTX44Frustum(&proj, -0.06f, 0.06f, -0.06f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                 0.06f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    /* setup projection matrix for projection texture */
    nn::math::MTX44Frustum(&proj, -0.04f, 0.04f, -0.04f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                 0.04f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    vec3.Set(0.5f, 0.5f, 1.f);
    nn::math::MTX34Scale(&tmp34, &vec3);
    tmp44 = nn::math::Matrix44(tmp34);
    nn::math::MTX44Mult(&proj, &tmp44, &proj);
    vec3.Set(0.5f, 0.5f, 0.f);
    nn::math::MTX34Translate(&tmp34, &vec3);
    tmp44 = nn::math::Matrix44(tmp34);
    nn::math::MTX44Mult(&proj, &tmp44, &proj);

    nn::math::Vector3 camPos(2.5f, 0.5f, 6.f);
    nn::math::Vector3 camUp(-1.f, 0.f, 0.f);
    nn::math::Vector3 target(-0.2f, 0.f, 0.f);
    nn::math::MTX34LookAt(&view, &camPos, &camUp, &target);

    camPos.Set(0.f, 0.f, 4.f);
    camUp.Set(0.f, 1.f, 0.f);
    target.Set(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&texview, &camPos, &camUp, &target);

    /* setup light direction */
    nn::math::Vector4 lpos0(0.f, 0.f, 1.f, 0.f);
    nn::math::Vector4 view0(view.m[0][0], view.m[0][1], view.m[0][2], view.m[0][3]);
    nn::math::Vector4 view1(view.m[1][0], view.m[1][1], view.m[1][2], view.m[1][3]);
    nn::math::Vector4 view2(view.m[2][0], view.m[2][1], view.m[2][2], view.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&view0, &lpos0), nn::math::VEC4Dot(&view1, &lpos0), nn::math::VEC4Dot(&view2, &lpos0), lpos0.w);
    glUniform4fv(glGetUniformLocation(s_ProgramID, "dmp_FragmentLightSource[0].position"), 1, static_cast<f32*>(lpos));

    /* draw robot */
    {
        /* setup modelview matrix */
        int aniframe = f % 360;
        int anitype = (f / 360) % 3;
        
        vec3.Set(0.f, 0.f, 2.f);
        nn::math::MTX34Translate(&model, &vec3);
        switch (anitype)
        {
            case 0: /* rotation */
                nn::math::MTX34RotXYZDeg(&tmp34, 0.f, static_cast<f32>(-aniframe), 0.f);
                nn::math::MTX34Mult(&model, &model, &tmp34);
                break;
            case 1: /* translate to front and back */
                vec3.Set(0.f, 0.f, 0.6f * sin(DMP_PI/180.f*static_cast<f32>(aniframe)));
                nn::math::MTX34Translate(&tmp34, &vec3);
                nn::math::MTX34Mult(&model, &model, &tmp34);
                break;
            case 2: /* translate to right and left */
                vec3.Set(0.6f * sin(DMP_PI/180.f*static_cast<f32>(aniframe)), 0.f, 0.f);
                nn::math::MTX34Translate(&tmp34, &vec3);
                nn::math::MTX34Mult(&model, &model, &tmp34);
                break;
        }
        vec3.Set(0.65f, 0.65f, 0.65f);
        nn::math::MTX34Scale(&tmp34, &vec3);
        nn::math::MTX34Mult(&model, &model, &tmp34);
        
        nn::math::MTX34Mult(&mv, &view, &model);
        nn::math::Matrix44 m(mv);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

        /* setup modelview projection matrix for projection texture */
        nn::math::MTX34Mult(&mv, &texview, &model);
        nn::math::Matrix44 mvp(mv);
        nn::math::MTX44Mult(&mvp, &proj, &mvp);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uTexModelViewProjection"), 1, GL_TRUE, static_cast<f32*>(mvp));

        /* draw objects */
        for (int i = 0; i < robot.obj_num; i++)
        {
            glBindBuffer(GL_ARRAY_BUFFER, robot.posVB);
            glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)robot.obj[i].vtx_offset);
            glBindBuffer(GL_ARRAY_BUFFER, robot.normVB);
            glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)robot.obj[i].nrm_offset);

            for (unsigned j = robot.obj[i].patch_offset; j < robot.obj[i].patch_size + robot.obj[i].patch_offset; j++)
            {
                glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, robot.idxVB);
                glDrawElements(GL_TRIANGLES, robot.patch[j].elm_size,
                    GL_UNSIGNED_SHORT, (GLvoid*)(robot.patch[j].elm_offset + robot.obj[i].elm_offset));
            }
        }
    }

    /* draw cube */
    {
        /* setup modelview matrix */
        vec3.Set(0.f, 0.f, -3.5f);
        nn::math::MTX34Translate(&model, &vec3);
        vec3.Set(1.f, 1.f, 0.f);
        nn::math::MTX34RotAxisFIdx(&tmp34, &vec3, static_cast<f32>(-f)/2.f);
        nn::math::MTX34Mult(&model, &model, &tmp34);
        vec3.Set(1.75f, 1.75f, 1.75f);
        nn::math::MTX34Scale(&tmp34, &vec3);
        nn::math::MTX34Mult(&model, &model, &tmp34);

        nn::math::MTX34Mult(&mv, &view, &model);
        nn::math::Matrix44 m(mv);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

        /* setup modelview projection matrix for projection texture */
        nn::math::MTX34Mult(&mv, &texview, &model);
        nn::math::Matrix44 mvp(mv);
        nn::math::MTX44Mult(&mvp, &proj, &mvp);
        glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uTexModelViewProjection"), 1, GL_TRUE, static_cast<f32*>(mvp));

        glBindBuffer(GL_ARRAY_BUFFER, s_CubeArrayBufferID);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)s_CubePosOffset);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)s_CubeNrmOffset);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_CubeElementArrayBufferID);
        glDrawElements(GL_TRIANGLES, s_CubeNumIndex, GL_UNSIGNED_SHORT, 0);
    }

    glFinish();

    s_RenderSystem.SwapBuffers();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);   

    f++;

    return !glGetError();
}

/* initialization */
static int Initialize(void)
{
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);
    
    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_GxHeapSize);

    s_ProgramID = glCreateProgram();
    s_ShaderID = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);
    
    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShaderID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_ProgramID, s_ShaderID);
    glAttachShader(s_ProgramID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_ProgramID, 0, "aPosition");
    glBindAttribLocation(s_ProgramID, 1, "aNormal");

    glLinkProgram(s_ProgramID);
    glValidateProgram(s_ProgramID);
    glUseProgram(s_ProgramID);

    glClearColor(0.36f, 0.42f, 0.5f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    LoadObjects();
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    /* enable DMP fragment lighting */
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_FragmentLighting.enabled"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_FragmentLightSource[0].enabled"), GL_TRUE);

    /* use only primary color */
    GLfloat gray[] = {0.5f, 0.5f, 0.5f, 1.f};
    glUniform4fv(glGetUniformLocation(s_ProgramID, "dmp_FragmentLightSource[0].diffuse"), 1, gray);
    glUniform4fv(glGetUniformLocation(s_ProgramID, "dmp_FragmentLighting.ambient"), 1, gray);

    /* enable projection texture */
    glActiveTexture(GL_TEXTURE0);
    glGenTextures(1, &s_ProjTexID);
    glBindTexture(GL_TEXTURE_2D, s_ProjTexID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER); /* border color is black by default. */
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
    bool bUseAlpha ;
    loadTexture("rom:/resources/dmp_logo256.tga", GL_TEXTURE_2D, 0, bUseAlpha);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Texture[0].samplerType"), GL_TEXTURE_PROJECTION_DMP);

    /* set combiner: projection texture * lighting + lighting */
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineRgb"), GL_MULT_ADD_DMP);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE0, GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_FRAGMENT_PRIMARY_COLOR_DMP);
    
    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    UnloadObjects();
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
