﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowKnot.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

#include <nn/os.h>
#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>

#include "demo.h"

#include "Util.h"
#include "Loader.h"

#include <string.h>
#include <math.h>

#include "Memory.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

/* program id */
GLuint s_LrProg, s_SdwProg, s_SdwSilhProg;

/* shader id */
GLuint s_Shaders[4];

/* obj loader class objet */
dat_t knot, knotsilh, plane;

/* GL object name */
GLuint s_ShadowTexName;
GLuint s_FboName;

/* frame number */
unsigned int s_FrameNo = 0;
/* 1st pass (shadow accumulation pass) frustum argument */
const GLfloat n = 0.1f, f = 10.f, r = 0.1f, t = 0.1f;
/* Light position */
const GLfloat lpos0[] = {1.f, 2.f, -1.f, 1.f};

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
u32 s_GxHeapSize = 0x400000;
u32 s_HeapSize = 0x200000;

demo::RenderSystem s_RenderSystem;

/* shadow setting */
#define SBWIDTH 512
#define SBHEIGHT 512
/* shadow artifact suppression SETTINGS 0 - lowest 2 - highest */
#define SDW_SETTINGS 0
const float fSZBiasScale[] = {1.2f, 1.3f, 1.3f};

/* if this is defined, soft shadow can be used */
#define SOFT_SHADOW

/* draw hard shadow to shadow buffer */
static void RenderObjectsShadow(dat_t& _objs)
{
    glDisableVertexAttribArray(1);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);

        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_TRIANGLES, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }
}

/* draw soft shadow to shadow buffer */
static void RenderObjectsShadowSilhouette(dat_t& _objs)
{
    glEnableVertexAttribArray(1);

    glDisable(GL_CULL_FACE);
    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);
        glBindBuffer(GL_ARRAY_BUFFER, _objs.normVB);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].nrm_offset);

        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_GEOMETRY_PRIMITIVE_DMP, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }
    glEnable(GL_CULL_FACE);
}

static void RenderObjectsShadowed(dat_t& _objs)
{
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, s_ShadowTexName);
    glEnableVertexAttribArray(1);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, _objs.posVB);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].vtx_offset);
        glBindBuffer(GL_ARRAY_BUFFER, _objs.normVB);
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void*)_objs.obj[i].nrm_offset);

        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            GLfloat ma[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat md[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat ms[4] = {0.f, 0.f, 0.f, 1.f};
            for (int col = 0; col < 3; col++)
            {
                ma[col] = _objs.patch[j].ambient[col];
                md[col] = _objs.patch[j].diffuse[col];
                ms[col] = _objs.patch[j].specular[col];
            }
            glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.ambient"), 1, ma);
            glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.diffuse"), 1, md);
            glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.specular0"), 1, ms);
            glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.specular1"), 1, ms);

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _objs.idxVB);
            glDrawElements(GL_TRIANGLES, _objs.patch[j].elm_size,
                GL_UNSIGNED_SHORT, (GLvoid*)(_objs.patch[j].elm_offset + _objs.obj[i].elm_offset));
        }
    }
}

/* load objects */
static void LoadObjects(void)
{
    loadDAT( "rom:/resources/knot.dat", &knot);
    loadDAT( "rom:/resources/knotsilh.dat", &knotsilh);
    loadDAT( "rom:/resources/plane.dat", &plane);
}

/* unload objects */
static void UnloadObjects(void)
{
    unloadDAT(&knot);
    unloadDAT(&knotsilh);
    unloadDAT(&plane);
}

/* draw object depth to shadow buffer */
static void Draw1stPass(void)
{
    /* set shadow accumulation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, s_FboName);

    /* set camera position and view frustum in light position */
    glViewport(0, 0, SBWIDTH, SBHEIGHT);
    glBindTexture(GL_TEXTURE_2D, 0);

    /* shadow buffer clear color needs to be (1, 1, 1)color */
    glClearColor(1.f, 1.f, 1.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT);

    glUseProgram(s_SdwProg);
    
    nn::math::Matrix34 mv, rot;
    nn::math::Vector3 camPos(lpos0[0], lpos0[1], lpos0[2]);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 axis(1.f, 1.f, 1.f);
    
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 tmp(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(tmp));

    RenderObjectsShadow(plane);

    nn::math::MTX34RotAxisDeg(&rot, &axis, static_cast<f32>(s_FrameNo));
    nn::math::MTX34Mult(&mv, &mv, &rot);
    nn::math::Matrix44 tmp2(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(tmp2));

    RenderObjectsShadow(knot);

    glFinish();

#ifdef SOFT_SHADOW
    glUseProgram(s_SdwSilhProg);

    glUniformMatrix4fv(glGetUniformLocation(s_SdwSilhProg, "uModelView"), 1, GL_TRUE, tmp2);

    RenderObjectsShadowSilhouette(knotsilh);
#endif
    glFinish();

    /* restore original(GL) fragment operation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

static void Draw2ndPass(void)
{
    glClearColor(0.36f, 0.42f, 0.5f, 1.f);
    glClearDepthf(1.f);

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();

    glUseProgram(s_LrProg);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(3.f * sinf(s_FrameNo * static_cast<f32>(M_PI)/180.f),
        1.f,
                   3.f * cosf(s_FrameNo * static_cast<f32>(M_PI)/180.f));
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 tmp(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(tmp));

    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    
    nn::math::Vector4 vlpos0(lpos0);
    nn::math::Vector4 l(nn::math::VEC4Dot(&mv0, &vlpos0), nn::math::VEC4Dot(&mv1, &vlpos0), nn::math::VEC4Dot(&mv2, &vlpos0), vlpos0.w);
    glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].position"), 1, static_cast<f32*>(l));

    nn::math::Matrix34 textmp;
    //Create a projection matrix
    nn::math::MTX34ShadowProjectionFrustum(&textmp,  -r, r, -t, t, n, f );
    nn::math::Matrix44 tex(textmp);

    nn::math::Vector3 camPos2(lpos0[0], lpos0[1], lpos0[2]);

    nn::math::Matrix34 tmpmv;
    nn::math::MTX34LookAt(&tmpmv, &camPos2, &camUp, &target);
    tmp = nn::math::Matrix44(tmpmv);
    nn::math::MTX44Mult(&tex, &tex, &tmp);

    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uTexture"), 1, GL_TRUE, static_cast<f32*>(tex));

    RenderObjectsShadowed(plane);

    glFinish();

    nn::math::Matrix34 rot;
    nn::math::Vector3 axis(1.f, 1.f, 1.f);
    
    nn::math::MTX34RotAxisDeg(&rot, &axis, static_cast<f32>(s_FrameNo));
    nn::math::MTX34Mult(&mv, &mv, &rot);

    tmp = nn::math::Matrix44(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(tmp));

    tmp = nn::math::Matrix44(rot);
    nn::math::MTX44Mult(&tex, &tex, &tmp);
    
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uTexture"), 1, GL_TRUE, static_cast<f32*>(tex));

    RenderObjectsShadowed(knot);

    glFinish();

    s_RenderSystem.SwapBuffers();
}

int DrawFrame(void)
{
    /* shadow accumulation pass */
    Draw1stPass();
    /* shadow test pass */
    Draw2ndPass();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    s_FrameNo++;

    return !glGetError();
}

static void InitializeShadowedProgram(void)
{
    /* Initialize shadowed draw program */
    s_LrProg = glCreateProgram();

    glAttachShader(s_LrProg, s_Shaders[2]);
    glAttachShader(s_LrProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_LrProg, 0, "aPosition");
    glBindAttribLocation(s_LrProg, 1, "aNormal");

    glLinkProgram(s_LrProg);
    glUseProgram(s_LrProg);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_Texture[0].samplerType"), GL_TEXTURE_SHADOW_2D_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragmentLighting.enabled"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].enabled"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].shadowed"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.lutInputRR"), GL_LIGHT_ENV_NH_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.lutInputFR"), GL_LIGHT_ENV_LN_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.lutEnabledRefl"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.config"), GL_LIGHT_ENV_LAYER_CONFIG1_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.fresnelSelector"), GL_LIGHT_ENV_PRI_SEC_ALPHA_FRESNEL_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.shadowAlpha"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_LightEnv.invertShadow"), GL_TRUE);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.samplerRR"), 0);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragmentMaterial.samplerFR"), 1);

    GLuint lutids[3];

    glGenTextures(3, lutids);
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, lutids[0]);
    
    GLfloat lut[512];
    memset(lut, 0, sizeof(lut));

    int j;

    for (j = 1; j < 128; j++)
    {
        lut[j] = powf((float)j/127.f, 30.f);
        lut[j+255] = lut[j] - lut[j-1];
    }

    glBindTexture(GL_LUT_TEXTURE0_DMP, lutids[1]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, lut);

    memset(lut, 0, sizeof(lut));
    for (j = 1; j < 128; j++)
    {
        lut[j] = powf((float)j / 127.f, 2.0f);
        lut[j+255] = lut[j] - lut[j-1];
    }

    glBindTexture(GL_LUT_TEXTURE1_DMP, lutids[2]);
    glTexImage1D(GL_LUT_TEXTURE1_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, GL_LUMINANCEF_DMP, GL_FLOAT, lut);

    GLfloat ld0[] = {1.f,1.f,1.f,1.f};
    GLfloat ls0[] = {0.f,0.f,0.f,0.f};
    GLfloat ls1[] = {1.f,1.f,1.f,1.f};
    GLfloat la0[] = {1.f,1.f,1.f,1.f};

    glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].diffuse"), 1, ld0);
    glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].specular0"), 1, ls0);
    glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].specular1"), 1, ls1);
    glUniform4fv(glGetUniformLocation(s_LrProg, "dmp_FragmentLightSource[0].ambient"), 1, la0);

    glUniform1f(glGetUniformLocation(s_LrProg, "dmp_Texture[0].shadowZBias"), n * fSZBiasScale[SDW_SETTINGS] / (f - n));
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_Texture[0].perspectiveShadow"), GL_TRUE);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].combineRgb"), GL_ADD_MULT_DMP);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);

    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_ONE_MINUS_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].srcRgb"), GL_FRAGMENT_PRIMARY_COLOR_DMP, GL_FRAGMENT_SECONDARY_COLOR_DMP, GL_FRAGMENT_PRIMARY_COLOR_DMP);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].srcAlpha"), GL_FRAGMENT_SECONDARY_COLOR_DMP, GL_CONSTANT, GL_CONSTANT);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);

    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform1f(glGetUniformLocation(s_LrProg, "dmp_FragOperation.wScale"), 0.f);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_GL_DMP);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.1f, 0.1f, -0.1f, 0.1f, 0.2f, 10.f);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_LrProg);
}

static void InitializeShadowProgram(void)
{
    /* Initialize shadow draw program */
    s_SdwProg = glCreateProgram();

    glAttachShader(s_SdwProg, s_Shaders[0]);
    glAttachShader(s_SdwProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_SdwProg, 0, "aPosition");

    glLinkProgram(s_SdwProg);
    glUseProgram(s_SdwProg);

    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_Texture[0].samplerType"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_FragmentLighting.enabled"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_FragmentLightSource[0].enabled"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].srcRgb"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].srcAlpha"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);

    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform1f(glGetUniformLocation(s_SdwProg, "dmp_FragOperation.wScale"), 1.f / f);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_SHADOW_DMP);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -r, r, -t, t, n, f);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_SdwProg);
}

static void InitializeSilhouetteProgram(void)
{
    /* Initialize silhouette draw program */
    s_SdwSilhProg = glCreateProgram();

    glAttachShader(s_SdwSilhProg, s_Shaders[1]);
    glAttachShader(s_SdwSilhProg, s_Shaders[3]);
    glAttachShader(s_SdwSilhProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_SdwSilhProg, 0, "aPosition");
    glBindAttribLocation(s_SdwSilhProg, 1, "aNormal");

    glLinkProgram(s_SdwSilhProg);
    glUseProgram(s_SdwSilhProg);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Texture[0].samplerType"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_FragmentLighting.enabled"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_FragmentLightSource[0].enabled"), GL_FALSE);

    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.penumbraScale"), 0.0f);
    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.penumbraBias"), 1.0f);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);

    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.wScale"), 1.f / f);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_SHADOW_DMP);

    GLfloat silhwidth[2];
    silhwidth[0] = 0.05f;
    silhwidth[1] = silhwidth[0] * static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT);
    glUniform2fv(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.width"), 1, silhwidth);
    
    GLfloat silhcolor[4] = {1.f, 1.f, 1.f, 1.f};
    glUniform4fv(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.color"), 1, silhcolor);
    
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.frontFaceCCW"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.scaleByW"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.acceptEmptyTriangles"), GL_FALSE);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -r, r, -t, t, n, f);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwSilhProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_SdwSilhProg);

}

/* initialization */
static int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    s_Shaders[0] = glCreateShader(GL_VERTEX_SHADER);          /* for 1st pass hard shadow rendering */
    s_Shaders[1] = glCreateShader(GL_VERTEX_SHADER);          /* for 1st pass soft shadow rendering */
    s_Shaders[2] = glCreateShader(GL_VERTEX_SHADER);          /* for 2nd pass shadowed rendering */
    s_Shaders[3] = glCreateShader(GL_GEOMETRY_SHADER_DMP);    /* for 1st pass soft shadow rendering */

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(4, s_Shaders, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    InitializeShadowedProgram();
    InitializeShadowProgram();
    InitializeSilhouetteProgram();

    LoadObjects();

    glEnableVertexAttribArray(0);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    /* generate shadow texture */
    glGenTextures(1, &s_ShadowTexName);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, s_ShadowTexName);
    GLfloat bcolor[] = {1.f, 1.f, 1.f, 1.f};
    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, bcolor);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, 0);

    glTexImage2D(
        GL_TEXTURE_2D, 0, GL_SHADOW_DMP, SBWIDTH, SBHEIGHT,
        0, GL_SHADOW_DMP, GL_UNSIGNED_INT, 0);
    glBindTexture(GL_TEXTURE_2D, 0);

    /* generate frame buffer object name */
    glGenFramebuffers(1, &s_FboName);

    /* set shadow accumulation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, s_FboName);
    /* note attachment point of shadow texture */
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, s_ShadowTexName, 0);

    return 0;
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    UnloadObjects();
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
