﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "demo.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

/* program id */
GLuint s_LrProg, s_SdwProg, s_SdwSilhProg;

/* shader id */
GLuint s_Shaders[4];

/* triangle and plane object vertex attribute */
GLfloat s_Vertices[] =
{
    -0.4f, -0.3f, -0.39f, 1.f,
    -0.3f, 0.3f, -0.1f, 1.f,
    0.0f, 0.3f, -0.0f, 1.f,
    -0.75f, 0.75f, -0.4f, 1.f,
    0.75f, 0.75f, -0.4f, 1.f,
    0.75f, -0.75f, -0.4f, 1.f,
    -0.75f, -0.75f, -0.4f, 1.f,

    -0.4f, -0.3f, -0.4f, 1.f,
    -0.3f, 0.3f, -0.11f, 1.f,
    0.0f, 0.3f, -0.01f, 1.f,
};

GLfloat s_Colors[] =
{
    0.7f, 0.7f, 0.7f, 1.f,
    0.7f, 0.7f, 0.7f, 1.f,
    0.7f, 0.7f, 0.7f, 1.f,
    0.0f, 0.0f, 1.0f, 1.f,
    0.0f, 0.0f, 1.0f, 1.f,
    0.0f, 0.0f, 1.0f, 1.f,
    0.0f, 0.0f, 1.0f, 1.f,
};

GLfloat s_Normals[] =
{
    -0.0f,-1.622f,-0.906f,
    -0.902f, 0.89f, 0.114f,
     0.8f, 0.732f, 0.092f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
     0.f, 0.f, 0.f,
};

GLushort s_OriginDices[] =
{
    2, 1, 0, 6, 5, 3, 4,
};

GLushort s_SilhIndices[] =
{
    2, 1, 7, 0, 8, 9,
};

/* 1st pass (shadow accumulation pass) frustum argument */
const GLfloat n = 0.1f, f = 10.f, sw2 = 0.5f, sh2 = 0.5f;

/* Light position */
const GLfloat s_LPos0[] = {0.f, 0.f, 0.15f, 1.f};

/* GL object name */
GLuint s_ShadowTexName;
GLuint s_FboName;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x800000;

demo::RenderSystem s_RenderSystem;

/* shadow parameter */
#define SBWIDTH 256
#define SBHEIGHT 256
#define SZBIASSCALE 1.3f

/* use soft shadow */
#define SOFT_SHADOW

/* draw object depth to shadow buffer */
static void Draw1stPass()
{
    /* set shadow accumulation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, s_FboName);

    /* set camera position and view frustum in light position */
    glViewport(0, 0, SBWIDTH, SBHEIGHT);
    glBindTexture(GL_TEXTURE_2D, 0);

    /* shadow buffer clear color needs to be (1, 1, 1)color */
    glClearColor(1.f, 1.f, 1.f, 1.f);
    glClear(GL_COLOR_BUFFER_BIT);

    glUseProgram(s_SdwProg);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(s_LPos0);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 m(mv);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glEnableVertexAttribArray(0);
    glDisableVertexAttribArray(1);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, s_Vertices);

    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, s_OriginDices);
    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, s_OriginDices+3);
    glFinish();

    /* draw object silhouette line to shadow buffer for soft shadow */
#ifdef SOFT_SHADOW
    glUseProgram(s_SdwSilhProg);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwSilhProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, s_Normals);

    glDisable(GL_CULL_FACE);

    glDrawElements(GL_GEOMETRY_PRIMITIVE_DMP, 6, GL_UNSIGNED_SHORT, s_SilhIndices);

    glEnable(GL_CULL_FACE);
#endif

    glFinish();

    /* restore original(GL) fragment operation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

static void Draw2ndPass()
{
    glClearColor(0.36f, 0.42f, 0.5f, 1.f);
    glClearDepthf(1.f);

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();
    
    glUseProgram(s_LrProg);

    nn::math::Matrix34 mv, mvl, tmp[2];
    nn::math::Vector3 trans(0.f, 0.f, -0.1f);
    nn::math::MTX34RotXYZDeg(&mv, 0.f, 0.f, 270.f);
    nn::math::MTX34Translate(&tmp[0], &trans);
    nn::math::MTX34RotXYZDeg(&tmp[1], -35.f, 0.f, 0.f);
    nn::math::MTX34Mult(&mvl, &mv, &tmp[0]);
    nn::math::MTX34Mult(&mvl, &mvl, &tmp[1]);
    nn::math::Matrix44 m(mvl);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uModelView"), 1, GL_TRUE, static_cast<f32*>(m));

    nn::math::Matrix34 textmp;
    //Create a projection matrix
    nn::math::MTX34ShadowProjectionFrustum(&textmp,  -sw2, sw2, -sh2, sh2, n, f );
    nn::math::Matrix44 tex(textmp);

    nn::math::Vector3 camPos(s_LPos0);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);

    nn::math::MTX34LookAt(&tmp[1], &camPos, &camUp, &target);
    nn::math::Matrix44 eye(tmp[1]);

    nn::math::MTX44Mult(&tex, &tex, &eye);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uTexture"), 1, GL_TRUE, static_cast<f32*>(tex));

    glBindTexture(GL_TEXTURE_2D, s_ShadowTexName);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, s_Vertices);
    glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, s_Colors);

    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, s_OriginDices);
    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, s_OriginDices+3);
    glFinish();

    s_RenderSystem.SwapBuffers();
}

int DrawFrame(void)
{
    static int f = 0;

    /* shadow accumulation pass */
    Draw1stPass();
    /* shadow test pass */
    Draw2ndPass();

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    f++;

    return !glGetError();
}

static void InitializeShadowedProgram(void)
{
    /* Initialize shadowed draw program */
    s_LrProg = glCreateProgram();

    glAttachShader(s_LrProg, s_Shaders[2]);
    glAttachShader(s_LrProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_LrProg, 0, "aPosition");

    glLinkProgram(s_LrProg);
    glUseProgram(s_LrProg);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_Texture[0].samplerType"), GL_TEXTURE_SHADOW_2D_DMP);

    glUniform1f(glGetUniformLocation(s_LrProg, "dmp_Texture[0].shadowZBias"), n * SZBIASSCALE / (f - n));
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_Texture[0].perspectiveShadow"), GL_TRUE);

    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].combineRgb"), GL_MODULATE);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].combineAlpha"), GL_MODULATE);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE0, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_LrProg, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE0, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

    glUniform1f(glGetUniformLocation(s_LrProg, "dmp_FragOperation.wScale"), 0.f);
    glUniform1i(glGetUniformLocation(s_LrProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_GL_DMP);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.7f, 0.7f, -0.7f, 0.7f, 0.2f, 10.f);
    glUniformMatrix4fv(glGetUniformLocation(s_LrProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_LrProg);
}

static void InitializeShadowProgram(void)
{
    /* Initialize shadow draw program */
    s_SdwProg = glCreateProgram();

    glAttachShader(s_SdwProg, s_Shaders[0]);
    glAttachShader(s_SdwProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_SdwProg, 0, "aPosition");

    glLinkProgram(s_SdwProg);
    glUseProgram(s_SdwProg);

    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_Texture[0].samplerType"), GL_FALSE);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].srcRgb"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_SdwProg, "dmp_TexEnv[0].srcAlpha"), GL_CONSTANT, GL_CONSTANT, GL_CONSTANT);

    glUniform1f(glGetUniformLocation(s_SdwProg, "dmp_FragOperation.wScale"), 1.f / f);
    glUniform1i(glGetUniformLocation(s_SdwProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_SHADOW_DMP);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -sw2, sw2, -sh2, sh2, n, f);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_SdwProg);
}

static void InitializeSilhouetteProgram(void)
{
    /* Initialize silhouette draw program */
    s_SdwSilhProg = glCreateProgram();

    glAttachShader(s_SdwSilhProg, s_Shaders[1]);
    glAttachShader(s_SdwSilhProg, s_Shaders[3]);
    glAttachShader(s_SdwSilhProg, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_SdwSilhProg, 0, "aPosition");
    glBindAttribLocation(s_SdwSilhProg, 1, "aNormal");

    glLinkProgram(s_SdwSilhProg);
    glUseProgram(s_SdwSilhProg);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Texture[0].samplerType"), GL_FALSE);

    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.penumbraScale"), 0.0f);
    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.penumbraBias"), 1.0f);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_SdwSilhProg, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

    glUniform1f(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.wScale"), 1.f / f);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_FragOperation.mode"), GL_FRAGOP_MODE_SHADOW_DMP);

    GLfloat silhwidth[2];
    silhwidth[0] = 0.05f;
    silhwidth[1] = silhwidth[0] * (float)nn::gx::DISPLAY0_WIDTH / (float)nn::gx::DISPLAY0_HEIGHT;
    glUniform2fv(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.width"), 1, silhwidth);
    
    GLfloat silhcolor[4] = {1.f, 1.f, 1.f, 1.f};
    glUniform4fv(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.color"), 1, silhcolor);
    
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.frontFaceCCW"), GL_TRUE);
    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.scaleByW"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_SdwSilhProg, "dmp_Silhouette.acceptEmptyTriangles"), GL_FALSE);

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -sw2, sw2, -sh2, sh2, n, f);
    glUniformMatrix4fv(glGetUniformLocation(s_SdwSilhProg, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    glValidateProgram(s_SdwSilhProg);
}

/* initialization */
static int Initialize(void)
{
    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));
    
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    s_Shaders[0] = glCreateShader(GL_VERTEX_SHADER);            /* for 1st pass hard shadow rendering */
    s_Shaders[1] = glCreateShader(GL_VERTEX_SHADER);            /* for 1st pass soft shadow rendering */
    s_Shaders[2] = glCreateShader(GL_VERTEX_SHADER);            /* for 2nd pass shadowed rendering */
    s_Shaders[3] = glCreateShader(GL_GEOMETRY_SHADER_DMP);  /* for 1st pass soft shadow rendering */

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(4, s_Shaders, GL_PLATFORM_BINARY_DMP, buf, read);    
    file.Finalize();
    s_AppHeap.Free(buf);

    InitializeShadowedProgram();
    InitializeShadowProgram();
    InitializeSilhouetteProgram();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    /* generate shadow texture */
    glGenTextures(1, &s_ShadowTexName);

    glBindTexture(GL_TEXTURE_2D, s_ShadowTexName);
    GLfloat bcolor[] = {1.f, 1.f, 1.f, 1.f};
    glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, bcolor);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, 0);

    glTexImage2D(
        GL_TEXTURE_2D, 0, GL_SHADOW_DMP, SBWIDTH, SBHEIGHT,
        0, GL_SHADOW_DMP, GL_UNSIGNED_INT, 0);
    glBindTexture(GL_TEXTURE_2D, 0);

    /* generate frame buffer object name */
    glGenFramebuffers(1, &s_FboName);

    /* set shadow accumulation render state */
    glBindFramebuffer(GL_FRAMEBUFFER, s_FboName);
    /* note attachment point of shadow texture */
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, s_ShadowTexName, 0);

    return 0;
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    /* initialization */
    if (Initialize() >= 0)
    {
        while (1)
        {
            (void)DrawFrame();
        }
    }
    
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
