﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     LightingSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gdApplicationSample.h"
#define APP_NAME "LightingToonApple"

/* program id */
static nn::gd::ShaderBinary*   shaderBinary = NULL;
static nn::gd::Shader*         s_VSProgramID = NULL;
static nn::gd::ShaderPipeline* SPProgramID = NULL;
static nn::gd::InputLayout*    ILProgramID = NULL;

extern demo::RenderSystem s_RenderSystem;

/* shader id */
//GLuint s_ShaderID;

/* env texture id */
// GLuint s_EnvTexName;
/* obj file loader class object */
datGD_t sphere;

/* texture collection id */
GLuint s_TexColl;

extern nn::fnd::ExpHeap s_AppHeap;

static void InitRenderStateLightingGD(void);

/* if USE_LN is not defined, half-vector is used as the vector controlling both diffuse illumination and highlight position */
#define USE_LN
/* if WINDOW is defined, a window highlight is used */
#define WINDOW

static nn::gd::DepthStencilState* depthStateDrawObject = 0;
static nn::gd::CombinerState *texLight = NULL;
/* Texture 2D resource object*/
static nn::gd::Texture2DResource* s_texture2DRes[6];
static nn::gd::TextureCube* texCube = NULL;
static nn::gd::SamplerState *sampler = NULL;

static nn::gd::UniformLocation ProgShader_Projection; 
static nn::gd::UniformLocation ProgShader_ModelView;
static nn::gd::UniformLocation ProgShader_World;  

using namespace nn::math;

#if 1
unsigned char* load_Texture(const char * _name, bool& _useAlpha, unsigned *_width, unsigned *_height, unsigned *_type, unsigned _forceformat)
{
    unsigned char* pixels;
    unsigned width, height, format, type, orientation;

    (void)_useAlpha;
    pixels = (unsigned char*)dmpLoadTGA(_name, &width, &height, &format, &type, &orientation);
    if (!pixels)
        return (unsigned char *)0;

    switch (format)
    {
        case IMAGE_RGB:     format = GL_RGB;    break;
        case IMAGE_RGBA:    format = GL_RGBA;   break;
        default:
            if (pixels)
                free(pixels);
            return (unsigned char *)0;
    }

    /* force to convert format */
    if (_forceformat != 0 && format != _forceformat)
    {
        unsigned i;
        unsigned char* newpixels = 0;

        switch (format)
        {
            case GL_RGB:
                switch (_forceformat)
                {
                    case GL_RGBA:
                        newpixels = reinterpret_cast<unsigned char*>(malloc(width * height * 4));
                        if (!newpixels)
                        {
                            free(pixels);
                            return (unsigned char *)0;
                        }
                        for (i = 0; i < width * height; i++)
                        {
                            newpixels[i * 4 + 0] = pixels[i * 3 + 0];
                            newpixels[i * 4 + 1] = pixels[i * 3 + 1];
                            newpixels[i * 4 + 2] = pixels[i * 3 + 2];
                            newpixels[i * 4 + 3] = 0xff;
                        }
                        break;
                }
                break;
            case GL_RGBA:
                switch (_forceformat)
                {
                    case GL_RGB:
                        newpixels = reinterpret_cast<unsigned char*>(malloc(width * height * 3));
                        if (!newpixels)
                        {
                            free(pixels);
                            return (unsigned char *)0;
                        }
                        for (i = 0; i < width * height; i++)
                        {
                            newpixels[i * 3 + 0] = pixels[i * 4 + 0];
                            newpixels[i * 3 + 1] = pixels[i * 4 + 1];
                            newpixels[i * 3 + 2] = pixels[i * 4 + 2];
                        }
                        break;
                }
                break;
        }

        if (newpixels)
        {
            format = _forceformat;
            free(pixels);
            pixels = newpixels;
        }
    }

    switch (type)
    {
        case IMAGE_UNSIGNED_BYTE:   type = GL_UNSIGNED_BYTE;    break;
        default:
            if (pixels)
                free(pixels);
            return (unsigned char *)0;
    }

    if (_width)
        (*_width) = width;
    if (_height) *_height = height;
    if (_type)   *_type =  type;


    return pixels;
}
#endif


static void SetLutTable()
{
    GLuint luts[4];
    glGenTextures(4, luts);

    const float outline = 0.15f;
    const float highlight_eps = 0.01f;
    float delta[] = {1.f, 0.7f, 0.5f, -1.f};

    GLfloat lut[512];
    int j;
    memset(lut, 0, sizeof(lut));
    float previous = 0, LN;
    int i = 0;

    /* f(D) = D is used as the initial shading function - see "Illumination Models in Maestro" for details
    first specify the steps in the [0, 1) interval */
    for (j = 127; j >= 0; j--)
    {
        LN = (float)j/127.9375f;
        if (LN > delta[i]) lut[j] = previous;
        else {
            lut[j] = LN;
            previous = lut[j];
            i++;
        }
    }
    for (j = 0; j < 127; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[127 + 256] = 0.f;

    /* now specify the steps in the [-1, 0) interval */
    for (j = 255; j  >= 128; j--)
    {
        LN = (float)(j - 256) /128.f;
        if (LN > delta[i]) lut[j] = previous;
        else {
            lut[j] = LN;
            previous = lut[j];
            i++;
        }
    }

    /* now calculate differences (some are 0, some are not) */
    for (j = 128; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = lut[0] - lut[255];

    //for (j = 0; j < 256; j++) lut[j] *= 0.5;
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RR, 0, lut, lut+256, 256);


    for (j = 0; j < 256; j++)
        if ((float)j/255.9375f <= 1.f - highlight_eps) lut[j] = 0.f;
        else  lut[j] = 1.f;
    for (j = 0; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = 0.f;

    //for (j = 0; j < 256; j++) lut[j] *= 0.5;
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D0, 0, lut, lut+256, 256);


    for (j = 0; j < 256; j++)
        if ((float)j/255.9375f < outline)  lut[j] = 0.f;
        else lut[j] = 1.f;
    for (j = 0; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = 1.f - lut[255];

    //for (j = 0; j < 256; j++) lut[j] *= 0.5;
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D1, 0, lut, lut+256, 256);


    for (j = 0; j < 256; j++) lut[j] = 1.f;
    for (j = 0; j < 256; j++)lut[j + 256] = 0.f;

    //for (j = 0; j < 256; j++) lut[j] *= 0.5;
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_FR, 0, lut, lut+256, 256);
}

static void InitialiseLightingGD()
{
    using namespace nn::gd;
    //LightingStage::EnableFragmentLighting(GD_TRUE);
    for (int i=0; i<8; ++i) LightingStage::light[i].EnableLight(GD_TRUE);

    LightingStage::SetGlobalColorAmbient(0, 0, 0);

    InitRenderStateLightingGD();
}

/* set lighting render state */
static void UpdateLightingRenderState(int f)
{
    float coeff = cosf(2.f * f / 60.f);
    if (coeff < 0) coeff = -coeff;
    coeff = 0.5f + coeff*0.5f;

 //   GLfloat ld[] = {1.f*coeff, 1.f*coeff, 1.f*coeff, 1.f};    /* light diffuse */
 //   GLfloat ls[] = {1.f*coeff, 1.f*coeff, 1.f*coeff, 1.f};    /* light specular */
//    GLfloat ls2[] = {1.f*coeff, 0.2f*coeff, 0.f*coeff, 1.f};  /* light specular2 */
//    GLfloat la[] = {1.f*coeff, 1.f*coeff, 1.f*coeff, 1.f};    /* light ambient */

    /* light direction setting */
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 5.75f, 22.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Vector4 lpos0(10.5f, 9.5f, 14.f, 1.f);
    lpos0.y -= 3.f*cosf(2.f * f / 60.f);
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&mv0, &lpos0), nn::math::VEC4Dot(&mv1, &lpos0), nn::math::VEC4Dot(&mv2, &lpos0), lpos0.w);

    using namespace nn::gd;
    static bool firstAPI = true;
    if (firstAPI)
    {
        InitialiseLightingGD();
        firstAPI = false;
    }

    for (int i=0; i<8; ++i)
    {
        LightingStage::light[i].SetColorAmbient(255*coeff, 255*coeff, 255*coeff);
        LightingStage::light[i].SetColorDiffuse(255*coeff, 255*coeff, 255*coeff);
        LightingStage::light[i].SetColorSpecular0(255*coeff, 255*coeff, 255*coeff);
        LightingStage::light[i].SetColorSpecular1(255*coeff, 51*coeff, 0*coeff);

            LightingStage::light[i].SetPosition(lpos.x - coeff*i*2.f, lpos.y - coeff*i*2.f, lpos.z - coeff*i*2.f);
    }

    LightingStage::light[0].EnableShadowed(GD_FALSE);
    LightingStage::EnableLookUpTableD0(GD_TRUE);
    LightingStage::EnableLookUpTableD1(GD_TRUE);
    LightingStage::EnableLookUpTableReflection(GD_TRUE);
    LightingStage::SetFresnelSelector(LightingStage::FRESNEL_SELECTOR_TYPE_PRIMARY_SECONDARY_ALPHA_FRESNEL);

    LightingStage::SetClampHightlight(GD_FALSE);
    LightingStage::SetLayerConfiguration(LightingStage::LAYER_CONFIGURATION_6);
    LightingStage::SetBumpMode(LightingStage::BUMPMODE_NOT_USED, GD_TRUE);
}

static void InitRenderStateCommon()
{
//   glUseProgram(s_ProgramID);
    nn::gd::ShaderStage::SetShaderPipeline(SPProgramID);

    nn::math::Matrix44 proj;
    /* set projection matrix */
    nn::math::MTX44Frustum(&proj, -0.015f, 0.015f, -0.015f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                 0.015f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_Projection, static_cast<f32*>(proj));

    /* set texture setting */
    /* output color is EnvTexture_rgb * PriColor_alpha + SecColor_rgb
    EnvTexture_rgb is environment highlight specular, PriColor_alpha is frensel reflection,
    and SecColor_rgb is toon shading diffuse color. */

    nn::gd::CombinerStage::SetTextureCombinerState(texLight);

    //modelview settings
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 5.75f, 40.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(-1.5f, 3.5f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    nn::math::Matrix34 rot;
    nn::math::MTX34RotXYZDeg(&rot, 0.f, 0.f, 0.f);
    nn::math::MTX34Mult(&mv, &mv, &rot);
    nn::math::Matrix44 m(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_ModelView, static_cast<f32*>(m));

    nn::math::Matrix44 id;
    nn::math::MTX44Identity(&id);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_World, static_cast<f32*>(id));

}

/* set lighting render state */
static void InitRenderStateLightingGD()
{
    using namespace nn::gd;
    LightingStage::SetLookUpTableAbsInput(LightingStage::LOOKUP_TABLE_ID_D0, GD_TRUE);
    LightingStage::SetLookUpTableAbsInput(LightingStage::LOOKUP_TABLE_ID_D1, GD_TRUE);
    LightingStage::SetLookUpTableAbsInput(LightingStage::LOOKUP_TABLE_ID_FR, GD_TRUE);
    LightingStage::SetLookUpTableAbsInput(LightingStage::LOOKUP_TABLE_ID_RR, GD_FALSE);

#ifdef USE_LN
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_RR, LightingStage::INPUT_VALUE_LN);
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_D0, LightingStage::INPUT_VALUE_LN);
#else
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_RR, LightingStage::INPUT_VALUE_NH);
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_D0, LightingStage::INPUT_VALUE_NH);
#endif
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_D1, LightingStage::INPUT_VALUE_LN);
    LightingStage::SetLookUpTableInputValue(LightingStage::LOOKUP_TABLE_ID_FR, LightingStage::INPUT_VALUE_LN);

    LightingStage::light[0].EnableShadowed(GD_FALSE);
    LightingStage::light[0].EnableTwoSideDiffuse(GD_FALSE);
    LightingStage::light[0].EnableGeomFactor0(GD_FALSE);
    LightingStage::light[0].EnableGeomFactor1(GD_FALSE);

    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_D0, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_D1, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_SP, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_FR, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_RB, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_RG, LightingStage::OUTPUT_SCALE_VALUE_1);
    LightingStage::SetLookUpTableOutputScaling(LightingStage::LOOKUP_TABLE_ID_RR, LightingStage::OUTPUT_SCALE_VALUE_1);

    LightingStage::EnableLookUpTableReflection(GD_TRUE);
    LightingStage::SetLayerConfiguration(LightingStage::LAYER_CONFIGURATION_6);
    LightingStage::SetBumpMode(LightingStage::BUMPMODE_NOT_USED, GD_TRUE);

    LightingStage::EnableLookUpTableD0(GD_TRUE);
    LightingStage::EnableLookUpTableD1(GD_TRUE);
    LightingStage::SetClampHightlight(GD_FALSE);
    LightingStage::SetFresnelSelector(LightingStage::FRESNEL_SELECTOR_TYPE_PRIMARY_SECONDARY_ALPHA_FRESNEL);

    SetLutTable();
}

/* load objects */
static void LoadObjects_GD(void)
{
    /* load obj file geometory and diffuse texture */
    loadDAT_GD( "rom:/resources_light/sphere.dat", &sphere);

    /* if you environment highlight specular, load environment cube texture */
#ifdef WINDOW
    char *env_tex[] =
    {
         "rom:/resources_light/wright.tga",
         "rom:/resources_light/wleft.tga",
         "rom:/resources_light/wtop.tga",
         "rom:/resources_light/wbottom.tga",
         "rom:/resources_light/wback.tga",
         "rom:/resources_light/wfront.tga",
    };

    for (int face = 0; face < 6; face++)
    {
        bool use_alpha;
        unsigned  width, height, type;
        unsigned char* pixels;
        nn::gd::Resource::NativeFormat nativeFormat;
        nnResult res;

        pixels = load_Texture(env_tex[face], use_alpha, &width, &height, &type, GL_RGBA);
        if (!pixels) { NN_ASSERTMSG( 0, "Error loading texture"); }

        //case GL_RGB:    res = nn::gd::TX_FORMAT_RGB_888;      break;
        //case GL_RGBA:   res = nn::gd::TX_FORMAT_RGBA_8888;    break;
        // img_format = convertGLFormatTypeToGDFormat(format, type);

        res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(nn::gd::Resource::FORMAT_RGBA_8888, width, height, pixels, pixels, &nativeFormat);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

         nn::gd::Texture2DResourceDescription Text2DResDesc = {width, height, 0, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
         res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, pixels, GD_FALSE, &s_texture2DRes[face]);
         if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

         free(pixels);
    }
#endif
}

static void UnloadObjects(void)
{
    unloadDAT_GD(&sphere);
    return;
}

void drawSphere()
{
    /* draw objects */
    for (int i = 0; i < sphere.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(ILProgramID);
        nn::gd::VertexBufferResource* buffers[] = { sphere.posVB, sphere.normVB };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

        for (unsigned j = sphere.obj[i].patch_offset; j < sphere.obj[i].patch_size + sphere.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(sphere.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(sphere.patch[j].elm_size, (u32)(sphere.patch[j].elm_offset + sphere.obj[i].elm_offset));
        }
    }
}

int DrawFrame()
{
    static int f = 0;

    nn::gd::OutputStage::SetDepthStencilState(depthStateDrawObject, 0);
    nn::gd::OutputStage::SetBlendState(NULL);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    nn::gd::TextureStage::SetTextureUnit0(*texCube);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, sampler);

    UpdateLightingRenderState(f);

    drawSphere();

    f++;

    return 0;
//    return !glGetError();
}

/* initialization */
int LightInitialize()
{
    nnResult res;

    nn::fs::FileReader file(L"rom:/romfiles_light/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);

    file.Finalize();
    s_AppHeap.Free(buf);

    nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &s_VSProgramID);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSProgramID, NULL, &SPProgramID);

    nn::gd::InputElementDescription descs1[] = {{ 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
                                               { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}};
    u32 strides1[] = {3 * sizeof(float), 3 * sizeof(float)};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs1, 2, strides1, s_VSProgramID, &ILProgramID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        // Get shader uniform address
     ProgShader_Projection = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uProjection");
     ProgShader_ModelView  = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uModelView");
     ProgShader_World      = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uWorld");
    /* create and bind texture collection object to capture subsequent modifications in the texture state
     both lookup table-related one and ordinary (2D, cube) textures */
    LoadObjects_GD();

    nn::gd::TextureCubeDescription desc;

    res = nn::gd::TextureStage::CreateTextureCube(s_texture2DRes, &desc, &texCube);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::SamplerStateDescription samplerdesc;

    samplerdesc.ToDefault();
    samplerdesc.m_MinLod = 0;
    samplerdesc.m_MinFilter = nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR;
    samplerdesc.m_MagFilter = nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR;

    nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &sampler);

    // texture combiner
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0,   nn::gd::CombinerStage::COMBINE_RGB_MULT_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0,   nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
//  combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::TC_OPERAND_ALPHA_SRC_ALPHA,           nn::gd::TC_OPERAND_ALPHA_SRC_ALPHA,           nn::gd::TC_OPERAND_ALPHA_SRC_ALPHA); //default value
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0,    nn::gd::CombinerStage::SOURCE_TEXTURE0,       nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texLight); // create dmp_TexEnv[0] combiner object

    nn::gd::DepthStencilStateDescription Depthdesc;
    Depthdesc.ToDefault();
    Depthdesc.m_DepthEnable = GD_TRUE;
    Depthdesc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ALL;
    Depthdesc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&Depthdesc, &depthStateDrawObject);
    return 0;
}

void LightFinalize()
{
    UnloadObjects();
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);
    if (s_VSProgramID) nn::gd::ShaderStage::ReleaseShader(s_VSProgramID);
    if (SPProgramID) nn::gd::ShaderStage::ReleaseShaderPipeline(SPProgramID);
    if (ILProgramID) nn::gd::VertexInputStage::ReleaseInputLayout(ILProgramID);
    if (texLight) nn::gd::CombinerStage::ReleaseTextureCombinerState(texLight);
    for (int i = 0; i < 6; i++)
    {
        if (s_texture2DRes[i] != NULL) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes[i]);
    }
    if (texCube) nn::gd::TextureStage::ReleaseTextureCube(texCube);
    if (sampler) nn::gd::TextureStage::ReleaseSamplerState(sampler);
    if (depthStateDrawObject) nn::gd::OutputStage::ReleaseDepthStencilState(depthStateDrawObject);
}

void DrawLights()
{
    InitRenderStateCommon();
    DrawFrame();
}

