﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     PartsysSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gdApplicationSample.h"

#define APP_NAME "PartsysSimple"

/* buffer id */
nn::gd::VertexBufferResource* s_ArrayBufferID;
nn::gd::VertexBufferResource* s_ElementArrayBufferID;

/* program id */
static nn::gd::ShaderBinary*   shaderBinary = NULL;
static nn::gd::Shader*         s_VSProgramID = NULL;
static nn::gd::Shader*         s_GS         = NULL;
static nn::gd::ShaderPipeline* SPProgramID = NULL;
static nn::gd::InputLayout*    ILProgramID = NULL;

/* Procedural texture */
static int     s_ProcTexSize;

/* 2D texture */
nn::gd::Texture2D*         s_Texture2DName = NULL;
nn::gd::Texture2DResource* s_Texture2DRes = NULL;
nn::gd::SamplerState*      samplerTex2D = NULL;

/* Structure containing particle system settings*/
struct _particleSys
{
    nn::math::Matrix44      m_partsys_center;
    nn::math::Matrix44      m_partsys_color;
    nn::math::Matrix44      m_partsys_radius;
    nn::math::Matrix44      m_partsys_aspect;

    float       m_speed;
    float       m_particleCountMax;
    float       m_size_min_max[2];
    float       m_prng[4];                      /* constant for the Pseudo Random Number Generator (modulo based) */
    float       m_random_seed[4];               /* seeds for the PRNG */

    float       simulationTime;                 /* application side. Holds the current time for particle system */
    float       dTime;                          /* application side. Holds the time interval between each frame */
    int         NFrame;                         /* application side. Number of frames for this particle system */
};

/* Texture Combiner State object */
nn::gd::CombinerState *texProcTex = NULL;
nn::gd::CombinerState *texTX0 = NULL;

extern unsigned char* loadTGA_Texture(const char * _name, unsigned int& width, unsigned int& height, unsigned int& format, unsigned int& type);
extern unsigned int convertGLFormatTypeToGDFormat(unsigned int format, unsigned int type);

#include "PartsysPattern.h"

struct _particleSys pSys[PARTSYS_LAST];
int CurSys = 0;

/* ExpHeap for app. */
extern nn::fnd::ExpHeap s_AppHeap;
extern demo::RenderSystem s_RenderSystem;

//API callback function
extern void callbackErrorFunction(nnResult result);

static nn::gd::DepthStencilState* depthStateDrawObject = 0;
static nn::gd::BlendState* blendState_SrcAlpha = 0;
static nn::gd::BlendState* blendState_One = 0;

static nn::gd::UniformLocation ProgShader_Center;
static nn::gd::UniformLocation ProgShader_Radii;
static nn::gd::UniformLocation ProgShader_PartsysColor;
static nn::gd::UniformLocation ProgShader_PartsysAspect;
static nn::gd::UniformLocation ProgShader_PartsysViewport;
static nn::gd::UniformLocation ProgShader_PartsysDistanceAttenuation;
static nn::gd::UniformLocation ProgShader_PartsysCountMax;
static nn::gd::UniformLocation ProgShader_PartsysPointSize;
static nn::gd::UniformLocation ProgShader_PartsysSpeed;
static nn::gd::UniformLocation ProgShader_PartsysRandomCore;
static nn::gd::UniformLocation ProgShader_PartsysRandSeed;
static nn::gd::UniformLocation ProgShader_PartsysTime;
static nn::gd::UniformLocation ProgShader_Projection;
static nn::gd::UniformLocation ProgShader_ModuleView;

/* Load a 2D texture (DMP Letters)*/
void LoadTexture2D(void)
{
    unsigned int width, height, format, type, img_format;
    unsigned char* pixels;
    nn::gd::Resource::NativeFormat nativeFormat;
    nnResult res;

    pixels = loadTGA_Texture("rom:/resources_PartsysSimple/dmp.tga", width, height, format, type);
    if (!pixels){ NN_ASSERTMSG( 0, "Error loading texture"); }

    img_format = convertGLFormatTypeToGDFormat(format, type);
    res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(
            (nn::gd::Resource::Format)img_format, width, height, pixels, pixels, &nativeFormat);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DResourceDescription Text2DResDesc = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, pixels, GD_FALSE, &s_Texture2DRes);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DDescription tex2ddesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_Texture2DRes, &tex2ddesc, &s_Texture2DName);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

	free(pixels);

    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.ToDefault();
    samplerdesc.m_MinFilter = nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR;
    samplerdesc.m_MagFilter = nn::gd::TextureStage::SAMPLER_MAG_FILTER_NEAREST;
    samplerdesc.m_WrapS     = nn::gd::TextureStage::SAMPLER_WRAP_CLAMP_TO_EDGE;
    samplerdesc.m_WrapT     = nn::gd::TextureStage::SAMPLER_WRAP_CLAMP_TO_EDGE;
    nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &samplerTex2D);
}

/* Procedural texture: RGBA texture definition
 a simple white gradient defined as 2 colors */
int Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA)
{
    memset(RR, 0, 512 * sizeof(GLfloat));
    memset(GG, 0, 512 * sizeof(GLfloat));
    memset(BB, 0, 512 * sizeof(GLfloat));
    memset(AA, 0, 512 * sizeof(GLfloat));

    RR[0] = 0xFF / 255.0f;  GG[0] = 0xFF / 255.0f;  BB[0] = 0xFF / 255.0f;  AA[0] = 0xFF / 255.0f;
    RR[1] = 0x00 / 255.0f;  GG[1] = 0x00 / 255.0f;  BB[1] = 0x00 / 255.0f;  AA[1] = 0x00 / 255.0f;

    for (int x = 0; x < 255; x++)
    {
        RR[256 + x] = RR[x + 1] - RR[x];
        GG[256 + x] = GG[x + 1] - GG[x];
        BB[256 + x] = BB[x + 1] - BB[x];
        AA[256 + x] = AA[x + 1] - AA[x];
    }
    RR[255] = 0.0f;
    GG[255] = 0.0f;
    BB[255] = 0.0f;
    AA[255] = 0.0f;

    return 2;
}

/* Procedural texture: mapping function
 bias function in [0.0.5]. zero otherwise */
void LoadBias(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float aa = 1.0f / p;
    float t;
    for (unsigned int i = 0; i < 128; i++)
    {
        t = x;
        if (t >= 1.0f)
            map_1D2D_LUT[i] = 1.0f;
        else
            map_1D2D_LUT[i] = t / ((aa - 2.0f) * (1.0f - t) + 1.0f);
        x += dx;
    }
    for (unsigned int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

/* Procedural texture: mapping function
 potential function in [0.0.5]. zero otherwise */
void LoadPotential(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float f = 0, d2;

    for (unsigned int i = 0; i < 128; i++)
    {
        f = 0;
        d2 = x;
        if (d2 < 1.0f)
        {
            if (d2 < 0.25f)
                f = 1.0f - ((3.0f * d2) * (3.0f * d2)) / (p + (4.5f - 4.0f * p) * d2);
            else
                f = ((1.0f - d2) * (1.0f - d2)) / (0.75f - p + (1.5f + 4.0f * p) * d2);
        }

        map_1D2D_LUT[i] = f;
        x += dx;
    }

    for (unsigned int i = 0; i < 126; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

void LoadProceduralTexture()
{
    /* Declaration */
    float F_FunctionRGB[256], F_FunctionZero[256];
    LoadBias(F_FunctionRGB, 0.8f);
    memset(F_FunctionZero, 0, 256 * sizeof(float));

    nn::gd::ProceduralTextureStage::UploadLookUpTableRgbMapFloat(0, F_FunctionRGB, F_FunctionRGB + 128, 128);

    const int NN = 512;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    s_ProcTexSize = Set1DTex(RR,GG,BB,AA);

    f32 *colorLUT[4];
    colorLUT[0] = RR;
    colorLUT[1] = GG;
    colorLUT[2] = BB;
    colorLUT[3] = AA;

    f32 *colorDeltaLUT[4];
    colorDeltaLUT[0] = RR+256;
    colorDeltaLUT[1] = GG+256;
    colorDeltaLUT[2] = BB+256;
    colorDeltaLUT[3] = AA+256;

    nn::gd::ProceduralTextureStage::UploadLookUpTableColorMapFloat(0, (f32 **)&colorLUT, (f32 **)&colorDeltaLUT, 256);
}

void ValidateParticleSystem(nn::gd::ShaderPipeline* pgID, struct _particleSys *pp, unsigned int _width, unsigned int _height)
{
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_Center, (f32 *)(pp->m_partsys_center));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_Radii, (f32 *)(pp->m_partsys_radius));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysColor, (f32 *)(pp->m_partsys_color));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysAspect, (f32 *)(pp->m_partsys_aspect));

    GLfloat viewport_size[2] = {0.0f, 0.0f};
    viewport_size[0] = 1.f / (GLfloat)_width;
    viewport_size[1] = 1.f / (GLfloat)_height;
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysViewport, &viewport_size[0]);

    GLfloat distance_attenuation[3] = {0.0f, 0.0f, 0.0f};
    distance_attenuation[0] = 0.0f;
    distance_attenuation[1] = 0.0f;
    distance_attenuation[2] = 1.0f / (float)(_width * _height);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysDistanceAttenuation, &distance_attenuation[0]);

    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysCountMax, &pp->m_particleCountMax);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysPointSize, &pp->m_size_min_max[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysSpeed, &pp->m_speed);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysRandomCore, &pp->m_prng[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysRandSeed, &pp->m_random_seed[0]);
}

void ValidateProceduralTexture()
{
    nn::gd::ProceduralTextureStage::SetMinFilter(nn::gd::ProceduralTextureStage::MIN_FILTER_LINEAR);
    nn::gd::ProceduralTextureStage::SetTexWidth(s_ProcTexSize);
    nn::gd::ProceduralTextureStage::SetTexOffset(0);
    nn::gd::ProceduralTextureStage::SetAlphaSeparate(GD_FALSE);
    nn::gd::ProceduralTextureStage::SetClampUV(nn::gd::ProceduralTextureStage::CLAMP_TO_EDGE, nn::gd::ProceduralTextureStage::CLAMP_TO_EDGE);
    nn::gd::ProceduralTextureStage::SetShiftUV(nn::gd::ProceduralTextureStage::SHIFT_NONE, nn::gd::ProceduralTextureStage::SHIFT_NONE);
    nn::gd::ProceduralTextureStage::SetRgbMap(nn::gd::ProceduralTextureStage::UV_MAP_MAX);
    nn::gd::ProceduralTextureStage::SetAlphaMap(nn::gd::ProceduralTextureStage::UV_MAP_MAX);
}

void ValidateTextureBlenderProcTex()
{
    nn::gd::TextureStage::SetTextureCoordinateSourceForUnit3Procedural(nn::gd::CTR::TextureStage::UNIT3_PROCEDURAL_TEXTURE_COORDINATE_FROM_UNIT_2);
    nn::gd::CombinerStage::SetTextureCombinerState(texProcTex);
}

void ValidateTextureBlenderTX0()
{
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_Texture2DName);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, samplerTex2D);
    nn::gd::CombinerStage::SetTextureCombinerState(texTX0);
}

void DrawParticleSystem(nn::gd::ShaderPipeline* pgID, float _Psys_time)
{
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(pgID, ProgShader_PartsysTime, &_Psys_time);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
    nn::gd::System::DrawIndexed(4, 0);
}

int DrawPartsysSimple()
{
    static int frame = 0;
    static bool switchTexture = false;
    static bool useProctex = true;
    static float t = 0.f;
    static int firstTime = 0;

    nn::gd::ShaderStage::SetShaderPipeline(SPProgramID);

    nn::math::Matrix44 proj, m;
    nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.0f, 0.0f, 50.0f);
    nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
    nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    nn::gd::OutputStage::SetDepthStencilState(depthStateDrawObject, 0);
    if (CurSys == PARTSYS_SPIN) {
        nn::gd::OutputStage::SetBlendState(blendState_SrcAlpha);
    } else {
        nn::gd::OutputStage::SetBlendState(blendState_One);
    }
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    // LoadTexture2D
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_Texture2DName);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, samplerTex2D);

    nn::gd::VertexInputStage::SetInputLayout(ILProgramID);

    nn::gd::VertexBufferResource* buffers[] = { s_ArrayBufferID };
    u32 offsets[] = { 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
    nn::gd::VertexInputStage::SetIndexBuffer(s_ElementArrayBufferID, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    if (firstTime == 0 ) {
          LoadProceduralTexture();  /* set texture configuration */
          firstTime = 1;
    }

    if (!useProctex)
    {
        ValidateTextureBlenderTX0();
    } else {
        ValidateProceduralTexture();
        ValidateTextureBlenderProcTex();
    }

    if (frame == 0)
    {
        /* reset uniforms */
        ValidateParticleSystem(SPProgramID, &pSys[CurSys], nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_Projection, static_cast<f32*>(proj));

        m = nn::math::Matrix44(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_ModuleView, static_cast<f32*>(m));
    }

    if (CurSys == PARTSYS_EXPLOSION2)
    {
        /* As an illustration, we draw 2 times the particle system. */
        m = nn::math::Matrix44(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_ModuleView, static_cast<f32*>(m));

        DrawParticleSystem(SPProgramID, t);

        nn::math::Matrix34 mv_rot;
        nn::math::MTX34RotXYZDeg(&mv_rot, 0.0f, 0.0f, 45.0f);
        nn::math::MTX34Mult(&mv_rot, &mv, &mv_rot);
        m = nn::math::Matrix44(mv_rot);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, ProgShader_ModuleView, static_cast<f32*>(m));
        DrawParticleSystem(SPProgramID, t);
    }
    else
        DrawParticleSystem(SPProgramID, t);

    if (++frame == pSys[CurSys].NFrame)
    {
        /* Changing 2D texture / procedural texture */
        if (switchTexture)
        {
            useProctex = (!useProctex) ? true : false;
            switchTexture = false;
        }
        /* change to next particle system */
        CurSys = ++CurSys % PARTSYS_LAST;
        frame = 0;
        t = 0.f;

        /* Changing blending mode for the spinming pattern */
        switchTexture = (CurSys == PARTSYS_LAST - 1);
    }
    else
        t += pSys[CurSys].dTime;


    if (useProctex == true) nn::gd::TextureStage::SetTextureCoordinateSourceForUnit3Procedural(nn::gd::CTR::TextureStage::UNIT3_PROCEDURAL_TEXTURE_COORDINATE_FROM_UNIT_0);

    return 0;
//    return !glGetError();
}

/* initialization */
int PartsysSimpleInitialize()
{
    /* create program and load & attach vertex shader */

    nn::fs::FileReader file(L"rom:/romfiles_PartsysSimple/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);
    file.Finalize();
    s_AppHeap.Free(buf);

    nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &s_VSProgramID);
    nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &s_GS);

    nn::gd::ShaderStage::CreateShaderPipeline(s_VSProgramID, s_GS, &SPProgramID);

    nn::gd::InputElementDescription descs1[] = { 0, "attrCtrPointIndex", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 1, 0};
    u32 strides1[] = {1 * sizeof(float)} ;
    nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs1, 1, strides1, s_VSProgramID, &ILProgramID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    LoadTexture2D();
    LoadParticlePattern(&pSys[PARTSYS_DEFAULT], PARTSYS_DEFAULT);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION1], PARTSYS_EXPLOSION1);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION2], PARTSYS_EXPLOSION2);
    LoadParticlePattern(&pSys[PARTSYS_SNOW], PARTSYS_SNOW);
    LoadParticlePattern(&pSys[PARTSYS_SPIN], PARTSYS_SPIN);

    /* make vertex buffer */
    GLfloat vertex[4] = {0.0f, 1.0f, 2.0f, 3.0f};
    GLushort index[4] = {0, 1, 2, 3};

    nn::gd::VertexBufferResourceDescription desc;

    desc.m_ByteSize = 4 * sizeof(float);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, vertex, &s_ArrayBufferID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = 4 * sizeof(short);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, index, &s_ElementArrayBufferID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    // texture combiner
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse( nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texProcTex); // create dmp_TexEnv[0] combiner object

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texTX0); // create dmp_TexEnv[0] combiner object

    nn::gd::DepthStencilStateDescription depthdesc;
    depthdesc.ToDefault();
    depthdesc.m_DepthEnable = GD_FALSE;
    depthdesc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ALL;
    depthdesc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LEQUAL;
    nn::gd::OutputStage::CreateDepthStencilState(&depthdesc, &depthStateDrawObject);

    nn::gd::BlendStateDescription blendesc;
    blendesc.ToDefault();
    blendesc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
    blendesc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE_MINUS_SRC_ALPHA, nn::gd::OutputStage::BLEND_EQUATION_ADD);
    nn::gd::OutputStage::CreateBlendState(&blendesc, &blendState_SrcAlpha);

    blendesc.ToDefault();
    blendesc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
    blendesc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE, nn::gd::OutputStage::BLEND_EQUATION_ADD);
    nn::gd::OutputStage::CreateBlendState(&blendesc, &blendState_One);

    ProgShader_Center = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uCenter");
    ProgShader_Radii = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uRadii");
    ProgShader_PartsysColor = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.color");
    ProgShader_PartsysAspect = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.aspect");
    ProgShader_PartsysViewport = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.viewport");
    ProgShader_PartsysDistanceAttenuation = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.distanceAttenuation");
    ProgShader_PartsysCountMax = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.countMax");
    ProgShader_PartsysPointSize = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.pointSize");
    ProgShader_PartsysSpeed = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.speed");
    ProgShader_PartsysRandomCore = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.randomCore");
    ProgShader_PartsysRandSeed = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.randSeed");
    ProgShader_PartsysTime = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "dmp_PartSys.time");
    ProgShader_Projection = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uProjection");
    ProgShader_ModuleView = nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uModelView");

    return 0;
}

void PartsysSimpleFinalize()
{
    if (s_ArrayBufferID) nn::gd::Resource::ReleaseVertexBufferResource(s_ArrayBufferID);
    if (s_ElementArrayBufferID) nn::gd::Resource::ReleaseVertexBufferResource(s_ElementArrayBufferID);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);
    if (s_VSProgramID) nn::gd::ShaderStage::ReleaseShader(s_VSProgramID);
    if (s_GS) nn::gd::ShaderStage::ReleaseShader(s_GS);
    if (SPProgramID) nn::gd::ShaderStage::ReleaseShaderPipeline(SPProgramID);
    if (ILProgramID) nn::gd::VertexInputStage::ReleaseInputLayout(ILProgramID);
    if (s_Texture2DName) nn::gd::TextureStage::ReleaseTexture2D(s_Texture2DName);
    if (s_Texture2DRes) nn::gd::Resource::ReleaseTexture2DResource(s_Texture2DRes);
    if (samplerTex2D) nn::gd::TextureStage::ReleaseSamplerState(samplerTex2D);
    if (texProcTex) nn::gd::CombinerStage::ReleaseTextureCombinerState(texProcTex);
    if (texTX0) nn::gd::CombinerStage::ReleaseTextureCombinerState(texTX0);
    if (depthStateDrawObject) nn::gd::OutputStage::ReleaseDepthStencilState(depthStateDrawObject);
    if (blendState_SrcAlpha) nn::gd::OutputStage::ReleaseBlendState(blendState_SrcAlpha);
    if (blendState_One) nn::gd::OutputStage::ReleaseBlendState(blendState_One);
}
