﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     FogAndLight.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <string.h>
#include <math.h>
#include <nn/applet.h>

#include "demo.h"

#include "Util.h"
#include "Loader.h"
#include "Memory.h"

//gdAPI include
#include <nn/gd.h>

// sample tools
#include "meshUtil.h"
#include "binFileUtil.h"

#include "tree.h"
#include "light.h"
#include "ground.h"

//#define WAIT_VSYNC

//test using recording packet for jump
//#define USE_JUMP_PACKET_COMMAND

#define PRINT(__msg) NN_LOG __msg

// shaders
const u8 MESH_SHADER_INDEX = 0;
const u8 LIGHT_SHADER_INDEX = 1;

// lights
const u8 LIGHT_COUNT = 3;
nn::math::Vector3 g_lightPos[3];
int g_lightColor[3][3] =  { {255,0,0}, {0,255,0}, {0,0,255} };

// Scene models
const u8 TREE_ROW_COUNT = 6;
const u8 TREE_COLUMN_COUNT = 6;
const f32 TREE_INTERSPACE = 3.5f;

CSampleMesh *g_treeMesh = NULL;
CSampleMesh *g_groundMesh = NULL;
CSampleMesh *g_lightMesh = NULL;

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;
static nn::gd::TargetProperties mainRenderTargetProperty;

/* program id */
static nn::gd::ShaderBinary*   shaderBinary = NULL;
static nn::gd::Shader*         s_VSMesh = NULL;
static nn::gd::Shader*         s_VSLight         = NULL;
static nn::gd::ShaderPipeline* SPMesh = NULL;
static nn::gd::ShaderPipeline* SPLight = NULL;
static nn::gd::InputLayout*    ILMesh = NULL;
static nn::gd::InputLayout*    ILLight = NULL;

static nn::gd::Texture2DResource* g_tree2DResource = 0;
static nn::gd::Texture2D* g_treetexture2D = 0;
static nn::gd::Texture2DResource* g_ground2DResource = 0;
static nn::gd::Texture2D* g_groundtexture2D = 0;
static nn::gd::SamplerState* g_samplerLinear = 0;

static nn::gd::CombinerState* g_texComStateWithTexture = 0;
static nn::gd::CombinerState* g_texComStateWithColor = 0;

static nn::gd::DepthStencilState* g_depthStateDrawMeshes = 0;
static nn::gd::BlendState* g_blendStateDrawMeshes = 0;

#ifdef USE_JUMP_PACKET_COMMAND
static nn::gd::RecordedPacketId* g_recordedPacketForJump = 0;
#endif

u32 s_CmdlistID;

// Camera
nn::math::Vector3 g_cameraPosition;
nn::math::Matrix44 g_projectionMatrix;

// Fog
#define FOG_MODE            (FOG_MODE_EXP)   // fog mode
#define FOG_MODE_LINEAR     (0)                 // GL_LINEAR mode in OpenGL ES1.1 compatible
#define FOG_MODE_EXP        (1)                 // GL_EXP mode in OpenGL ES1.1 compatible

float g_fogStart = 5.f;
float g_fogEnd = 60.f;
float g_fogDensity = 0.08f;

// particles
extern int PartsysSimpleInitialize();
extern void PartsysSimpleFinalize();
extern int DrawPartsysSimple();

u8 g_clearColor[3] = { 0.36f*255, 0.42f*255, 0.5f*255 };

// ExpHeap for app.
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x400000;

demo::RenderSystem s_RenderSystem;

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

GLfloat g_light_lut[512];
u32 frameCounter = 0;

// ===================  TOOLS TEXTURES  ==================================

typedef __packed struct BitmapFileHeader
{
    u8 bfType[2];
    u16 bfSize[2];
    u16 bfReserved1;
    u16 bfReserved2;
    u16 bfOffBits[2];    
} BitmapFileHeader;

typedef __packed struct BitmapInfoHeader
{
    u32 biSize;
    s32 biWidth;
    s32 biHeight;
    u16 biPlanes;
    u16 biBitCount;
    u32 biCompression;
    u32 biSizeImage;
    s32 biXPixPerMeter;
    s32 biYPixPerMeter;
    u32 biClrUsed;
    u32 biClrImportant;    
} BitmapInfoHeader;


u8* GetBmpFileData(u8* rawDataBuffer, u32& bmpWidth, u32& bmpHeight)
{
    BitmapInfoHeader* bmp_info_header_ptr = (BitmapInfoHeader*)(rawDataBuffer + sizeof(BitmapFileHeader));
    bmpWidth = bmp_info_header_ptr->biWidth;
    bmpHeight = bmp_info_header_ptr->biHeight;
    u8* bmpDataBuffer = rawDataBuffer + sizeof(BitmapFileHeader) + sizeof(BitmapInfoHeader);

    return bmpDataBuffer;
}

u32 GetTextureLength(const u32& imageLength)
{        
    u32 textureLength = 8;

    // 8, 16, 32, 64, 128, 256, 512, 1024
    for (u32 i = 0; i < 7; i++)
    {
        if ( imageLength > textureLength )
        {
            textureLength *= 2;
        }
        else
        {
            return textureLength;
        }
    }

    return 1024;
}

//#define CONVERT_TEXTURE_TO_NATIVE_FORMAT_USING_GD
u8* GetTextureDataFromBmpFileData(const u32& bmpWidth, const u32& bmpHeight, u8* bmpDataBuffer, u32& textureWidth, u32& textureHeight, bool keepInitilaResolution)
{
    // Convert BMP to OpenGL RGB format
    textureWidth = GetTextureLength(bmpWidth);
    textureHeight = GetTextureLength(bmpHeight);
    u8* textureGLDataBuffer = reinterpret_cast<u8*>( s_AppHeap.Allocate(3 * textureWidth * textureHeight) );

    if (keepInitilaResolution)
    {
        textureWidth = bmpWidth;
        textureHeight = bmpHeight;
    }

    for (u32 y = 0; y < textureHeight; y++)
    {        
        for (u32 x = 0; x < textureWidth; x++)
        {
            u8* textureDataPtr = textureGLDataBuffer;
            textureDataPtr += 3 * ((static_cast<u32>(textureWidth) * y) + x);
            if ( ( x >= bmpWidth ) || ( y >= bmpHeight ) )
            {
                (*textureDataPtr) = 0x00;
                textureDataPtr += 1;
                
                (*textureDataPtr) = 0x00;
                textureDataPtr += 1;
                
                (*textureDataPtr) = 0x00;
            }
            else
            {
                u8* bmpDataPtr = bmpDataBuffer;
                bmpDataPtr += 3 * (bmpWidth * y + x);
                
                (*textureDataPtr) = (*(bmpDataPtr + 2));
                textureDataPtr += 1;
                (*textureDataPtr) = (*(bmpDataPtr + 1));
                textureDataPtr += 1;
                
                (*textureDataPtr) = (*(bmpDataPtr + 0));
            }
        }
    }

    u8* textureDataBuffer = reinterpret_cast<u8*>( s_AppHeap.Allocate(3 * textureWidth * textureHeight) );

#ifdef CONVERT_TEXTURE_TO_NATIVE_FORMAT_USING_GD
    nn::gd::Texture::ConvertTextureResourceToNativeFormat(nn::gd::TX_FORMAT_RGB_888, textureWidth, textureHeight, textureGLDataBuffer, textureDataBuffer);
#else
    // Convert OpenGL RGB format to PICA Native RGB format
    GLenum format = GL_RGB_NATIVE_DMP;
    bool result = demo::ConvertGLTextureToNative(format, textureWidth, textureHeight, textureGLDataBuffer, textureDataBuffer);
    if ( result )
    {
        NN_LOG("  Convert to GL_RGB_NATIVE_DMP succeeded.\n");
    }
    else
    {
        NN_PANIC("  Convert to GL_RGB_NATIVE_DMP failed.\n");
    }
#endif

    s_AppHeap.Free(textureGLDataBuffer);

    return textureDataBuffer;
}

u8* GetTextureData(const wchar_t* bmpRomFilename, u32& bmpWidth, u32& bmpHeight, u32& textureWidth, u32& textureHeight, bool keepInitilaResolution)
{   
    NN_LOG("\nReading data form ROMFS...\n");

    nn::fs::FileReader file(bmpRomFilename);
    
    size_t fileSize = file.GetSize();
    NN_LOG("  fileSize = %d (Byte)\n", fileSize);
    if ( fileSize == 0 )
    {
        NN_PANIC("Failed to open bmp file.\n");
        return NULL;
    }

    void* buf = s_AppHeap.Allocate(fileSize);

    s32 readSize = file.Read(buf, fileSize);
    NN_LOG("  file readSize = %d (Byte)\n", readSize);
    if ( readSize == 0 )
    {
        NN_PANIC("Failed to open bmp file.\n");
        return NULL;
    }

    u8* bmpDataBuffer = NULL;
    bmpDataBuffer = GetBmpFileData((u8*)buf, bmpWidth, bmpHeight);
    NN_LOG("  bmpWidth = %d, bmpHeight = %d\n", bmpWidth, bmpHeight);

    u8* textureDataBuffer = NULL;
    textureDataBuffer = GetTextureDataFromBmpFileData(bmpWidth, bmpHeight, bmpDataBuffer, textureWidth, textureHeight, keepInitilaResolution);
    NN_LOG("  textureWidth = %d, textureHeight = %d\n", textureWidth, textureHeight);

    file.Finalize();
    s_AppHeap.Free(buf);

    return textureDataBuffer;
}

// =================   Fog-related functions  ========================


// This function returns fog coefficient for input value c
float FogCoef(float fog_c)
{
    // GL_LINEAR compatible
    if (FOG_MODE == FOG_MODE_LINEAR)
    {
        if (fog_c < g_fogStart)
            return 1.0f;
        else if (fog_c > g_fogEnd)
            return 0.0f;
        else
            return (g_fogEnd - fog_c) / (g_fogEnd - g_fogStart);
    }
    // GL_EXP compatible
    else if (FOG_MODE == FOG_MODE_EXP)
    {
        return exp(-g_fogDensity * fog_c);
    }
    // Error
    else
        return -1.0;
}

// hard-coded LUT
void uploadFogLUT()
{
    u32 rawData[] = {
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffffd0,
        0x00fa1fa0, 0x00ee3f8e, 0x00dfff75, 0x00cebf54,
        0x00b93f24, 0x009ddede, 0x00799e70, 0x0047bdc2,
        0x00000000, 0x00000000, 0x00000000, 0x00000000
    };
    nn::gd::FogStage::UploadFogLookUpTableNative(0, rawData, 128);
}

void InitializeFogTable(nn::math::Matrix44 *MP)
{
    static float Fog_LUT[256];      // Fog lut contents
    static float Fog_c[128 + 1];    // c (distance from origin in eye coordinate)

    int i;
    nn::math::Matrix44 invPM;
    nn::math::Vector4 v_clip(0.0f, 0.0f, 0.0f, 1.0f);
    nn::math::Vector4 v_eye;

    // If projection matrix is DMPGL2.0 one, USE code following
    nn::math::MTX44Inverse(&invPM, MP);
    nn::math::Vector4 v0(invPM.m[0]);
    nn::math::Vector4 v1(invPM.m[1]);
    nn::math::Vector4 v2(invPM.m[2]);
    nn::math::Vector4 v3(invPM.m[3]);
    for (i = 0; i <= 128; i++)
    {
        v_clip.z = -(static_cast<f32>(i)) / 128;
        v_eye.x = nn::math::VEC4Dot(&v0, &v_clip);
        v_eye.y = nn::math::VEC4Dot(&v1, &v_clip);
        v_eye.z = nn::math::VEC4Dot(&v2, &v_clip);
        v_eye.w = nn::math::VEC4Dot(&v3, &v_clip);
        Fog_c[i] = -(v_eye.z / v_eye.w);
    }

    for (i = 0; i < 128; i++)
    {
        Fog_LUT[i] = FogCoef(Fog_c[i]);
        Fog_LUT[128 + i]= FogCoef(Fog_c[i + 1]) - FogCoef(Fog_c[i]);
    }
    
    nn::gd::FogStage::UploadFogLookUpTableFloat(0, Fog_LUT, &(Fog_LUT[128]), 128);
}

static void InitLightLUT()
{
    int j;
    memset(g_light_lut, 0, sizeof(g_light_lut));
    for (j = 0; j < 256; j++)
    {
        float x = (float)j / 255.f;
        float y = 1 - x;
        g_light_lut[j] = y;
    }
    for (j = 0; j < 255; j++) g_light_lut[j + 256] = g_light_lut[j + 1] - g_light_lut[j];
    g_light_lut[255 + 256] = 0.f - g_light_lut[255];
}

#ifdef USE_JUMP_PACKET_COMMAND
static void CreateJumpPacket()
{
    //we're recording some command into a packet in order to replay them using a jump
    //all the following gd functions are immediate, so they will immediately been inserted in the command buffer
    nn::gd::System::StartRecordingPackets(NULL, nn::gd::System::RECORD_3D_COMMAND_BUFFER_FOR_JUMP);

    for (int i = 0; i < LIGHT_COUNT; i++)
    {
        nn::gd::LightingStage::SetGlobalColorAmbient(100,100,100);
        
        nn::gd::LightingStage::light[i].SetColorAmbient(g_lightColor[i][0], g_lightColor[i][1], g_lightColor[i][2]);
        nn::gd::LightingStage::light[i].SetColorSpecular0(255, 255, 255);
        nn::gd::LightingStage::light[i].SetColorSpecular1(255, 255, 255);
        nn::gd::LightingStage::light[i].SetColorDiffuse(g_lightColor[i][0], g_lightColor[i][1], g_lightColor[i][2]);
        nn::gd::LightingStage::light[i].SetSpotDirection(0.0f, -1.0f, 0.0f);

        float scale; float bias;
        nn::gd::Utils::ConvertStartEndToScaleBias(0.f, 3.f, scale, bias);
        nn::gd::LightingStage::light[i].SetDistanceAttenuationScaleBias(scale, bias);

        nn::gd::LightingStage::light[i].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_DA, 0, g_light_lut, g_light_lut+256, 256);
    }

    nn::gd::System::StopRecordingPackets(&g_recordedPacketForJump);

    //we clear the command list after having recorded the packet (we don't necessary want to play it at the beginning)
    nngxClearCmdlist();
    nngxRunCmdlist();
}
#endif

// load sample meshes
static void LoadObjects(void)
{
    g_treeMesh = loadMesh(tree_mesh_tree_PNU, "rom:/");
    g_groundMesh = loadMesh(ground_mesh_ground_PNU, "rom:/");
    g_lightMesh = loadMesh(light_mesh_light_P, "rom:/");   
}


void LoadTextures_GD(wchar_t* bmpRomFilename, nn::gd::Texture2DResource** texResPtr, nn::gd::Texture2D** texPtr)
{
    if (!bmpRomFilename)    return;

    u32 width = 0;
    u32 height = 0;
    u32 textureWidth = 0;
    u32 textureHeight = 0;

    u8* textureDataPtr = GetTextureData(bmpRomFilename, width, height, textureWidth, textureHeight, true);

    nn::gd::Texture2DResourceDescription Text2DResDesc = {textureWidth, textureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, texResPtr);
    s_AppHeap.Free(textureDataPtr);
    if ( GD_NNRESULT_IS_FAILURE(res) )
    {
        NN_PANIC("\n");
    }

    nn::gd::Texture2DDescription tex2DDesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(*texResPtr, &tex2DDesc, texPtr);

    if ( GD_NNRESULT_IS_FAILURE(res) )
    {
        NN_PANIC("\n");
    }
}

static void UnloadObjects(void)
{
    if(g_treeMesh) delete g_treeMesh;
    if(g_groundMesh) delete g_groundMesh;
    if(g_lightMesh) delete g_lightMesh;

    return;
}

// ==================== FRAME RECORDING ========================
//FPS recording
#define FPSTABCOUNT 500
static double fpsTab[FPSTABCOUNT];
static int curFpsPos = 0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}


int DrawFrame(void)
{
    s64 startTime;

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);
    }

    startTime = GetTimeMicroSeconds();

    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, g_clearColor, 1.f, 0);
    nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);
         
    // camera position
    nn::math::Matrix34 mv, rot;
    nn::math::Vector3 camPos(0.f, 15.f, 20.f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34RotXYZDeg(&rot, 0.f, 0.f, -90.0f);
    nn::math::MTX34LookAt(&mv, &g_cameraPosition, &camUp, &target);
    nn::math::MTX34Mult(&mv, &rot, &mv);

    // setup lights
    for (int i = 0; i < LIGHT_COUNT; i++)
    {
        nn::math::Vector4 lpos0(g_lightPos[i][0], g_lightPos[i][1], g_lightPos[i][2], 1.f);
        nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
        nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
        nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);

        nn::math::Vector4 lpos(VEC4Dot(&mv0, &lpos0), VEC4Dot(&mv1, &lpos0), VEC4Dot(&mv2, &lpos0), lpos0.w);
        nn::gd::LightingStage::light[i].SetPosition(lpos.x, lpos.y, lpos.z);

#ifdef USE_JUMP_PACKET_COMMAND
        nn::gd::System::ReplayPackets(g_recordedPacketForJump, NULL);   //we just replay the packet
#else
        nn::gd::LightingStage::SetGlobalColorAmbient(100,100,100);
        
        nn::gd::LightingStage::light[i].SetColorAmbient(g_lightColor[i][0], g_lightColor[i][1], g_lightColor[i][2]);
        nn::gd::LightingStage::light[i].SetColorSpecular0(255, 255, 255);
        nn::gd::LightingStage::light[i].SetColorSpecular1(255, 255, 255);
        nn::gd::LightingStage::light[i].SetColorDiffuse(g_lightColor[i][0], g_lightColor[i][1], g_lightColor[i][2]);
        nn::gd::LightingStage::light[i].SetSpotDirection(0.0f, -1.0f, 0.0f);

        float scale; float bias;
        nn::gd::Utils::ConvertStartEndToScaleBias(0.f, 3.f, scale, bias);
        nn::gd::LightingStage::light[i].SetDistanceAttenuationScaleBias(scale, bias);
        
        nn::gd::LightingStage::light[i].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_DA, 0, g_light_lut, g_light_lut+256, 256);
#endif
     }

    // set rendering states for the objects in the scene
    nn::gd::OutputStage::SetDepthStencilState(g_depthStateDrawMeshes, 0);
    nn::gd::OutputStage::SetBlendState(g_blendStateDrawMeshes);
    nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_FOG);
      
    // draw the trees
    nn::gd::ShaderStage::SetShaderPipeline(SPMesh);

    // setup texture unit
    nn::gd::CombinerStage::SetTextureCombinerState(g_texComStateWithTexture);
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *g_treetexture2D);
 
    // setup projection matrix
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPMesh, nn::gd::ShaderStage::GetShaderUniformLocation(SPMesh, "uProjection"), static_cast<f32*>(g_projectionMatrix));

    int xstart = TREE_ROW_COUNT / 2;
    int zstart = TREE_COLUMN_COUNT / 2;
    for(int z = -zstart; z <= zstart; z++)
    {
        for(int x = -xstart; x <= xstart; x++)
        {
            nn::math::Vector3 trans( x * TREE_INTERSPACE, 0.0f,  z * TREE_INTERSPACE);
            nn::math::Matrix34 mv_tr;
            nn::math::MTX34Translate(&mv_tr, &trans);      
            MTX34Mult(&mv_tr, &mv, &mv_tr);
            nn::math::Matrix44 m(mv_tr);

            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPMesh, nn::gd::ShaderStage::GetShaderUniformLocation(SPMesh, "uModelView"), static_cast<f32*>(m));
 
            nn::gd::VertexInputStage::SetInputLayout(ILMesh);
            g_treeMesh->drawMesh(0, 1, 2);
        }
    }

    // draw the ground
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *g_groundtexture2D);
    nn::gd::VertexInputStage::SetInputLayout(ILMesh);
    g_groundMesh->drawMesh(0, 1, 2);

    // draw the lights
    nn::gd::ShaderStage::SetShaderPipeline(SPLight);
    nn::gd::CombinerStage::SetTextureCombinerState(g_texComStateWithColor);
    
    // setup projection matrix for the new shader
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLight, nn::gd::ShaderStage::GetShaderUniformLocation(SPLight, "uProjection"), static_cast<f32*>(g_projectionMatrix));
   
    for (int i = 0; i < 3; i++)
    {
        nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT0, g_lightColor[i][0], g_lightColor[i][1], g_lightColor[i][2], 255);

        nn::math::Matrix34 mv_tr;
        nn::math::MTX34Translate(&mv_tr, &(g_lightPos[i]));
        MTX34Mult(&mv_tr, &mv, &mv_tr);
        nn::math::Matrix44 m(mv_tr);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLight, nn::gd::ShaderStage::GetShaderUniformLocation(SPLight, "uModelView"), static_cast<f32*>(m));

        nn::gd::VertexInputStage::SetInputLayout(ILLight);
        g_lightMesh->drawMesh(0);
    }

      nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_NONE);

    // draw the particles
    DrawPartsysSimple();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (frameCounter == 0)
    {
        nngxStartLcdDisplay();
    }

    frameCounter++;

    //FPS management
    fpsTab[curFpsPos++] = GetTimeMicroSeconds() - startTime;
    if (curFpsPos == FPSTABCOUNT)
    {
        double avgtime = 0;
        for (int i=1; i<FPSTABCOUNT; ++i)
        {
            avgtime += fpsTab[i];
            //PRINT(("Time: %0.5f    (fps: %0.3f).\n", fpsTab[i], (1. / fpsTab[i])));
        }
        avgtime /= FPSTABCOUNT-1;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        curFpsPos = 0;
    }

    return !glGetError();
}


int updateMeshPositions(void)
{
    f32 fs = frameCounter / 300.0f;    // a rough notion of time (fps dependant)
    const f32 T[3] = { 2.24f, 1.79f, 1.47f };

    nn::math::Vector3 rotAxis(0.f, 1.f, 0.f);
    nn::math::Matrix34 rot;

    for(int i = 0; i < 3; i++)
    {
        f32 r =  ::nn::math::F_PI *  T[i];
        nn::math::MTX34RotAxisDeg(&rot, &rotAxis,  fs * NN_MATH_RAD_TO_DEG(r));
        nn::math::Vector3 p(0.0f, T[i], T[i] * 2.0f);
        nn::math:: VEC3Transform(&g_lightPos[i],& rot, &p);
    }

    // camera position
    float angle = fs * NN_MATH_RAD_TO_DEG( ::nn::math::F_PI * 0.25f );
    nn::math::MTX34RotAxisDeg(&rot, &rotAxis,  angle);
    float sinVal = sinf(fs) + 1.3f;
    nn::math::Vector3 p(0.0f, sinVal * 2.0f, 15.0f);
    nn::math:: VEC3Transform(&g_cameraPosition,& rot, &p);
 
    return 1;
}


/* initialization */
static int Initialize(void)
{
     const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    //---------------------  gdAPI call ---------------------------

    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;

    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    InitLightLUT();

#ifdef USE_JUMP_PACKET_COMMAND
    CreateJumpPacket();
#endif

    //------------------ Init main frame buffers ------------------
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

    // shader settings
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);
    
    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);
    file.Finalize();
    s_AppHeap.Free(buf);

    // create mesh shader
    nn::gd::ShaderStage::CreateShader(shaderBinary, MESH_SHADER_INDEX, &s_VSMesh);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSMesh, NULL, &SPMesh);
    nn::gd::InputElementDescription descs0[] = {{ 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
                                               { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
                                               { 2, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, 0}};
    u32 strides0[] = {3 * sizeof(float), 3 * sizeof(float), 2 * sizeof(float)} ;
    res = nn::gd::VertexInputStage::CreateInputLayout(descs0, 3, strides0, s_VSMesh, &ILMesh);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    // create lighting shader
    nn::gd::ShaderStage::CreateShader(shaderBinary, LIGHT_SHADER_INDEX, &s_VSLight);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSLight, NULL, &SPLight);
    nn::gd::InputElementDescription descs1[] = { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0};
     u32 strides1[] = {3 * sizeof(float)};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs1, 1, strides1, s_VSLight, &ILLight);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    DEMO_ASSERT_GL_ERROR();
    
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    // load textures
    LoadObjects();
    LoadTextures_GD(L"rom:/tree.bmp", &g_tree2DResource, &g_treetexture2D );
    LoadTextures_GD(L"rom:/ground.bmp", &g_ground2DResource, &g_groundtexture2D );

    //init some light settings
    for (int i = 0; i < LIGHT_COUNT; i++)
    {
        nn::gd::LightingStage::light[i].EnableLight(GD_TRUE);
        nn::gd::LightingStage::light[i].SetLightType(nn::gd::Light::SOURCE_TYPE_POINT);
        nn::gd::LightingStage::light[i].EnableDistanceAttenuation(GD_TRUE);
        nn::gd::LightingStage::light[i].EnableGeomFactor1( GD_TRUE );
    }
    
    // texture combiner settings
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MULT_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &g_texComStateWithTexture);

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &g_texComStateWithColor);
    //nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT_0, 255, 0, 0, 255);

    nn::gd::CombinerStage::SetTextureCombinerState(g_texComStateWithTexture);
   
    // sampler state: linear filtering
    nn::gd::SamplerStateDescription descLinear;
    descLinear.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR);
    descLinear.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR);
    descLinear.SetWrapS(nn::gd::TextureStage::SAMPLER_WRAP_REPEAT);
    descLinear.SetWrapT(nn::gd::TextureStage::SAMPLER_WRAP_REPEAT);
    nn::gd::TextureStage::CreateSamplerState(&descLinear, &g_samplerLinear);
    //nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, g_samplerLinear);

    //---------------------  gdAPI call ---------------------------

    // compute projection matrix
    nn::math::MTX44Frustum(&g_projectionMatrix, -0.06f, 0.06f, -0.06f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH,
                           0.06f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);

    // fog setup
    InitializeFogTable(&g_projectionMatrix);
    nn::gd::FogStage::SetFogColor(g_clearColor[0], g_clearColor[1], g_clearColor[2]);
//    nn::gd::FogStage::SetFogColor(g_clearColor[0] * 255, g_clearColor[1] * 255, g_clearColor[2] * 255);
    nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_FOG);

    // particle setup
    PartsysSimpleInitialize();

    // render states
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ALL;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &g_depthStateDrawMeshes);

    nn::gd::BlendStateDescription descBlend;
    descBlend.ToDefault();
    nn::gd::OutputStage::CreateBlendState(&descBlend, &g_blendStateDrawMeshes);

    return 0;
}
static void Finalize()
{
    nngxStopCmdlist();

    // Finalize PartsysSimple 
    PartsysSimpleFinalize();

    if (g_treetexture2D) nn::gd::TextureStage::ReleaseTexture2D(g_treetexture2D);
    if (g_tree2DResource) nn::gd::Resource::ReleaseTexture2DResource(g_tree2DResource);
    if (g_groundtexture2D) nn::gd::TextureStage::ReleaseTexture2D(g_groundtexture2D);
    if (g_ground2DResource) nn::gd::Resource::ReleaseTexture2DResource(g_ground2DResource);
    if (g_samplerLinear) nn::gd::TextureStage::ReleaseSamplerState(g_samplerLinear);
    if (g_texComStateWithTexture) nn::gd::CombinerStage::ReleaseTextureCombinerState(g_texComStateWithTexture);
    if (g_texComStateWithColor) nn::gd::CombinerStage::ReleaseTextureCombinerState(g_texComStateWithColor);
    if (g_depthStateDrawMeshes) nn::gd::OutputStage::ReleaseDepthStencilState(g_depthStateDrawMeshes);
    if (g_blendStateDrawMeshes) nn::gd::OutputStage::ReleaseBlendState(g_blendStateDrawMeshes);

    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);
    if (s_VSMesh) nn::gd::ShaderStage::ReleaseShader(s_VSMesh);
    if (s_VSLight) nn::gd::ShaderStage::ReleaseShader(s_VSLight);
    if (SPMesh) nn::gd::ShaderStage::ReleaseShaderPipeline(SPMesh);
    if (SPLight) nn::gd::ShaderStage::ReleaseShaderPipeline(SPLight);
    if (ILMesh) nn::gd::VertexInputStage::ReleaseInputLayout(ILMesh);
    if (ILLight) nn::gd::VertexInputStage::ReleaseInputLayout(ILLight);

    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);

#ifdef USE_JUMP_PACKET_COMMAND
    if (g_recordedPacketForJump)  nn::gd::System::ReleasePackets(g_recordedPacketForJump);
#endif

    nn::gd::System::Finalize();
    //---------------------  gdAPI call ---------------------------

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

  // Stop command list execution.
    nngxStopCmdlist();

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();
    
    //Release memory
    demo::FinalizeMemoryManager();
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    nn::fs::Initialize();
    
    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            updateMeshPositions();
            (void)DrawFrame();
        }
    }
    UnloadObjects();
    Finalize();
    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}
