﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     binFileUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "stdio.h"
#include <string.h>
#include <ctype.h>

#include <nn/fs.h>
#include "../Common/Memory.h"
#include "binFileUtil.h"

//----------------------------------------------------------------------------
CBinalyFile::CBinalyFile() :
	m_data(NULL),
	m_ownMemory(NULL),
	m_size(0)
{}

//----------------------------------------------------------------------------
CBinalyFile::~CBinalyFile()
{
	if(m_ownMemory) free(m_ownMemory);
}

//----------------------------------------------------------------------------

bool CBinalyFile::load(const char *filePath)
{
    if (!filePath)    return false;

    if(m_ownMemory) free(m_ownMemory);
	m_ownMemory = NULL;
	m_data = NULL;

    bool res = true;
    wchar_t* wfname = 0;
    nn::fs::FileReader file;

    do
    {
        // convert the filepath to wide char
        wfname = static_cast<wchar_t*>(malloc( (strlen(filePath)+1)*2 ));
        if (!wfname)  
        {
            res = false; 
            break;
        }
      
        (void)std::mbstowcs( wfname, filePath, strlen(filePath)+1 );
        file.Initialize(wfname);
        m_size = file.GetSize();
        m_ownMemory = static_cast<unsigned char*>(malloc(m_size));

        if(!m_ownMemory)
        {
            res = false; 
            break;
        }

        m_data = (void *)m_ownMemory;
        m_size = file.Read(m_ownMemory, m_size);
    } while (0);

    // cleaning up
    if (wfname) free(wfname);
    file.Finalize();

    if(!res)
	{
		m_data = NULL;
		free(m_ownMemory);
		m_ownMemory = NULL;
	}

    return res;
}
