﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     meshUtil.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef __MESHUTIL_H__
#define __MESHUTIL_H__

#include <GLES2/gl2.h>

#define GD_USE_API
#include <nn/gd.h>

//----------------------------------------------------------------------------
// float[16] array indicating the matrix.
// This sample stores the matrix in memory as column major. 
struct float16
{
    // Matrix element.
    float val[16];

    // Access to elements.
    float &operator [](int i) { return val[i]; }
    const float &operator [](int i) const { return val[i]; }

    // Cast to the float pointer.
    operator float*() { return val; }
    operator const float*() const { return val; }

    // Cast to the Matrix.
//  operator const ASPMatrix4 () const
//  {
//      return ASPMatrix4(val);
//  }
};

//----------------------------------------------------------------------------
// Bone matrix.
struct Bone
{
    const char   *name;             // Bone name
    int          animeOffset;       // Offset for the animation to be applied
    float16      bindMatrix;        // Matrix to convert vertex coordinates to the bone coordinate system
    float16      poseMatrix;        // Initial posture matrix
};

//----------------------------------------------------------------------------
// Data structure in the binary file
// Unused field are reported with -1
struct MeshDeclaration
{
    const char   *fileName;        // filename
    unsigned int numVertices;      // Vertex count
    unsigned int numIndices;       // index count
    unsigned int stride;           // Number of byte per vertex
    int          position;         // Coords offset
    int          normal;           // Normals offset
    int          color;            // Vertex color offset
    int          texcoord;         // UVs offset
    int          tangent;          // Tangent vector offset
    int          binormal;         // Binormal offset
    int          index;            // Skinning matrix index offset
    int          weight;           // Skinning weight offset

    Bone         *bones;           // Array of bones
    int          numBones;         // Number of bones

    float16      *mtxInstance;     // Instance's posture matrix
    int          *animeOffsets;    // Offset array to each animation to be applied for each instance
    unsigned int numInstance;      // Number of instances
};

//----------------------------------------------------------------------------
// Sample mesh can be loaded from a binary file
class CSampleMesh
{
public:
    CSampleMesh(void);
    ~CSampleMesh(void);

    // Generate the sample mesh and upload VBO to GPU memory
    bool createMesh(const MeshDeclaration &mdecl, const void *meshData);

    // Render the sample mesh
    void drawMesh(GLint positionAttr = 0, GLint normalAttr = -1, GLint textAttr = -1,
                  GLint colorAttr = -1, GLint tangentAttr = -1, GLint binormalAttr = -1,
                  GLint skinningIndexAttr = -1, GLint skinningWeightAttr = -1);

    // Execute rendering for morphing.
    //bool drawMorphMesh(ASPContext dev, CSampleMesh *target[3]);

protected:
    MeshDeclaration     m_mdecl;        // Mesh structure
#if defined(GD_USE_API)
    nn::gd::VertexBufferResource* m_vbo1[8];
#else
    GLuint              m_vbo[8];       // Index of the generic vertex attribute
#endif
};


// Load the mesh
CSampleMesh *loadMesh(  const MeshDeclaration &mdecl,   const char* szDataPath);


//----------------------------------------------------------------------------
// Motion data

/*
struct CMotion
{
    CMotion() :
        m_numInstances(0),
        m_numStep(0),
        m_data(NULL)
    {
    }

    ~CMotion(){}

    // Initialize from the data in memory.
    void init(const void *data, unsigned int size, int numInstances);

    // Get the matrix that specifies the node number and step count.
    const ASPMatrix4 getMatrix(int index, int step, int loopStep=0) const;

    // Number of instances
    int     m_numInstances;

    // Number of steps
    int     m_numStep;

    // Motion data
    const float16   *m_data;
};

// Load motion data.
CMotion *loadMotion(
    const char *szFilePath,
    int numInstances,
    const char *szDataPath = NULL);
*/
#endif // __MESHUTIL_H__
