﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     Fog.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/fnd/fnd_ExpHeap.h>
#include <nn/applet.h>

#include "TexUtils.h"

#include <nn/gd.h>
#include "demo.h"

#define WAIT_VSYNC

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;

static nn::gd::VertexBufferResource* s_QuadFogIndexBufID = 0;
static nn::gd::VertexBufferResource* s_QuadFogVertBufID = 0;
static nn::gd::VertexBufferResource* s_QuadFogColBufID = 0;

static nn::gd::VertexBufferResource* s_QuadAllIndexBufID = 0;
static nn::gd::VertexBufferResource* s_QuadAllVertBufID = 0;
static nn::gd::VertexBufferResource* s_QuadAllColBufID = 0;

nn::gd::VertexBufferResource* buffersFog[2];
nn::gd::VertexBufferResource* buffersAll[2];

//Uniforms
static nn::gd::UniformLocation s_shaderVariable_proj;
static nn::gd::UniformLocation s_shaderVariable_view;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

/* fog color */
GLfloat s_FogColor[4]={1.0f, 1.0f, 1.0f, 1.0f};

/* fog LUT number */
#define FOG_LUT_NUM     (0)

/* fog mode settings */
#define FOG_MODE            (FOG_MODE_LINEAR)   /* Selection of fog mode */
#define FOG_MODE_LINEAR     (0)                 /* GL_LINEAR mode in OpenGL ES1.1 compatible */
#define FOG_MODE_EXP        (1)                 /* GL_EXP mode in OpenGL ES1.1 compatible */

float s_FogStart = 10.f;        /* GL_FOG_START in OpenGL ES1.1 compatible settings */
float s_FogEnd = 30.f;      /* GL_FOG_END in OpenGL ES1.1 compatible settings */
float s_FogDensity = 0.1f;  /* GL_FOG_DENSITY in OpenGL ES1.1 compatible settings */

float s_FogAfter = 500.f;

float s_Width = 1000.f;
float s_Height = 0.f;

/* Vertex attribute related */
f32 s_VertexFog[12];
const u16 s_QuadIndexAll[18] =
{
    0, 1, 2, 0, 2, 3,
    4, 5, 6, 4, 6, 7,
    8, 9, 10, 8, 9, 11
};
const u16 s_QuadIndexFog[6] = {0, 1, 2, 0, 2, 3};

f32 s_VertexAll[36];
f32 s_VertexAllColor[48];
f32 s_VertexFogColor[48];

u32 s_FogLutID;
u32 s_CollectionLUTID;

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

demo::RenderSystem s_RenderSystem;

/* This function returns fog coefficient for input value c */
float FogCoef(float fog_c)
{
    /* GL_LINEAR compatible */
    if (FOG_MODE == FOG_MODE_LINEAR)
    {
        if (fog_c < s_FogStart)
            return 1.0f;
        else if (fog_c > s_FogEnd)
            return 0.0f;
        else
            return (s_FogEnd - fog_c) / (s_FogEnd - s_FogStart);
    }
    /* GL_EXP compatible */
    else if (FOG_MODE == FOG_MODE_EXP)
    {
        return std::exp(-s_FogDensity * fog_c);
    }
    /* Error */
    else
        return -1.0;
}


void uploadFogLUT()
{
    u32 rawData[] = {
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffe000,
        0x00ffe000, 0x00ffe000, 0x00ffe000, 0x00ffffd0,
        0x00fa1fa0, 0x00ee3f8e, 0x00dfff75, 0x00cebf54,
        0x00b93f24, 0x009ddede, 0x00799e70, 0x0047bdc2,
        0x00000000, 0x00000000, 0x00000000, 0x00000000
    };
    nn::gd::FogStage::UploadFogLookUpTableNative(0, rawData, 128);
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    s_shaderVariable_proj = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    s_shaderVariable_view = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    NN_ASSERT(s_shaderVariable_proj.IsValid());
    NN_ASSERT(s_shaderVariable_view.IsValid());

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{

    s_VertexAll[0] = -s_Width;
    s_VertexAll[1] = s_Height;
    s_VertexAll[2] = 0;

    s_VertexAll[3] = s_Width;
    s_VertexAll[4] = s_Height;
    s_VertexAll[5] = 0;

    s_VertexAll[6] = s_Width;
    s_VertexAll[7] = s_Height;
    s_VertexAll[8] = s_FogStart;

    s_VertexAll[9]  = -s_Width;
    s_VertexAll[10] = s_Height;
    s_VertexAll[11] = s_FogStart;

    // Bottom red
    for (int i = 0; i < 4; i++)
    {
        s_VertexAllColor[4 * i + 0] = 1.0f;
        s_VertexAllColor[4 * i + 1] = 0.0f;
        s_VertexAllColor[4 * i + 2] = 0.0f;
        s_VertexAllColor[4 * i + 3] = 1.0f;
    }

    s_VertexAll[12 + 0] = -s_Width;
    s_VertexAll[12 + 1] = s_Height;
    s_VertexAll[12 + 2] = s_FogStart;

    s_VertexAll[12 + 3] = s_Width;
    s_VertexAll[12 + 4] = s_Height;
    s_VertexAll[12 + 5] = s_FogStart;

    s_VertexAll[12 + 6] = s_Width;
    s_VertexAll[12 + 7] = s_Height;
    s_VertexAll[12 + 8] = s_FogEnd;

    s_VertexAll[12 + 9]  = -s_Width;
    s_VertexAll[12 + 10] = s_Height;
    s_VertexAll[12 + 11] = s_FogEnd;
    // Bottom yellow
    for (int i = 0; i < 4; i++)
    {
        s_VertexAllColor[16 + 4 * i + 0] = 1.0f;
        s_VertexAllColor[16 + 4 * i + 1] = 1.0f;
        s_VertexAllColor[16 + 4 * i + 2] = 0.0f;
        s_VertexAllColor[16 + 4 * i + 3] = 1.0f;
    }

    s_VertexAll[24 + 0] = -s_Width;
    s_VertexAll[24 + 1] = s_Height;
    s_VertexAll[24 + 2] = s_FogEnd;

    s_VertexAll[24 + 3] = s_Width;
    s_VertexAll[24 + 4] = s_Height;
    s_VertexAll[24 + 5] = s_FogEnd;

    s_VertexAll[24 + 6] = s_Width;
    s_VertexAll[24 + 7] = s_Height;
    s_VertexAll[24 + 8] = s_FogEnd+s_FogAfter;

    s_VertexAll[24 + 9]  = -s_Width;
    s_VertexAll[24 + 10] = s_Height;
    s_VertexAll[24 + 11] = s_FogEnd+s_FogAfter;
    // Bottom purple
    for (int i = 0; i < 4; i++)
    {
        s_VertexAllColor[32 + 4 * i + 0] = 1.0f;
        s_VertexAllColor[32 + 4 * i + 1] = 0.0f;
        s_VertexAllColor[32 + 4 * i + 2] = 1.0f;
        s_VertexAllColor[32 + 4 * i + 3] = 1.0f;
    }

    s_VertexFog[0] = -s_Width;
    s_VertexFog[1] = s_Height;
    s_VertexFog[2] = 0;

    s_VertexFog[3] = s_Width;
    s_VertexFog[4] = s_Height;
    s_VertexFog[5] = 0;

    s_VertexFog[6] = s_Width;
    s_VertexFog[7] = s_Height;
    s_VertexFog[8] = s_FogEnd + s_FogAfter;

    s_VertexFog[9]  = -s_Width;
    s_VertexFog[10] = s_Height;
    s_VertexFog[11] = s_FogEnd + s_FogAfter;
    //Fog part
    for (int i = 0; i < 4; i++)
    {
        s_VertexFogColor[4 * i + 0] = 0.0f;
        s_VertexFogColor[4 * i + 1] = 0.0f;
        s_VertexFogColor[4 * i + 2] = 1.0f;
        s_VertexFogColor[4 * i + 3] = 1.0f;
    }

    /* --------------------------------------------------------------*/

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    nnResult res;

    //Index
    desc.m_ByteSize = 6 * sizeof(u16);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_QuadIndexFog, &s_QuadFogIndexBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    desc.m_ByteSize = 18 * sizeof(u16);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_QuadIndexAll, &s_QuadAllIndexBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Vertices
    desc.m_ByteSize = 12 * sizeof(f32);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_VertexFog, &s_QuadFogVertBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    desc.m_ByteSize = 36 * sizeof(f32);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_VertexAll, &s_QuadAllVertBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Color
    desc.m_ByteSize = 16 * sizeof(f32);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_VertexFogColor, &s_QuadFogColBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    desc.m_ByteSize = 48 * sizeof(f32);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, s_VertexAllColor, &s_QuadAllColBufID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Input layout
    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0}
    };
    u32 strides[] = { sizeof(float) * 3, sizeof(float) * 4};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &inputLayout);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    nn::gd::VertexInputStage::SetInputLayout(inputLayout);

    //Vertex buffers
    buffersFog[0] = s_QuadFogVertBufID;
    buffersFog[1] = s_QuadFogColBufID;
    buffersAll[0] = s_QuadAllVertBufID;
    buffersAll[1] = s_QuadAllColBufID;

}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);
}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    static int frame = 0;
    static const int step = 128;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Perspective(&proj, 65.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 500.0f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(proj));
    }

    //static int firstTimeModelView = true;
    //if (firstTimeModelView)
    {
        //Modelview
        nn::math::Matrix34 eye;
        nn::math::Vector3 camPos(0.0f, 2.0f, 0.0f);
        nn::math::Vector3 camUp(1.0f, 0.0f, 0.0f);
        nn::math::Vector3 target(0.0f, 2.0f, 10.0f);
    
        nn::math::MTX34Identity(&eye);
        nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);
        nn::math::Matrix44 mv(eye);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_view, static_cast<f32*>(mv));

        //firstTimeModelView = false;
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);
        
        /* --------------------------------------------------------- */

        float dx = frame * (1.0f / (float)(step - 1));
        s_FogColor[0] = 0.2f + dx * (1.0f - 0.2f);
        s_FogColor[1] = 0.4f + dx * (1.0f - 0.4f);
        s_FogColor[2] = 0.6f + dx * (1.0f - 0.6f);
        int fogColor0 = (int)(s_FogColor[0] * 255);
        int fogColor1 = (int)(s_FogColor[1] * 255);
        int fogColor2 = (int)(s_FogColor[2] * 255);
        nn::gd::FogStage::SetFogColor(fogColor0, fogColor1, fogColor2);

        //Draw stripes
        //Scissor
        nn::gd::RasterizerStage::EnableScissor(GD_TRUE);
        nn::gd::Viewport scissor(0, 0, mainRenderTargetProperty.m_Width / 2, mainRenderTargetProperty.m_Height / 2);
        nn::gd::RasterizerStage::SetScissor(scissor);

        u32 offsets[] = { 0, 0 };

        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffersAll, offsets);
        nn::gd::VertexInputStage::SetIndexBuffer(s_QuadAllIndexBufID, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

        nn::gd::System::DrawIndexed(18, 0);
        // End draw stripes
        
        // Draw fog part
        nn::gd::Viewport scissorFog(0, mainRenderTargetProperty.m_Height / 2, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height / 2);
        nn::gd::RasterizerStage::SetScissor(scissorFog);
        nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_FOG);

        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffersFog, offsets);
        nn::gd::VertexInputStage::SetIndexBuffer(s_QuadFogIndexBufID, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

        nn::gd::System::DrawIndexed(6, 0);

        nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_NONE);
      
        /* --------------------------------------------------------- */

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif
    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    if (frame == step - 1)
    {
        frame=0;
        //s_useGdApi = !s_useGdApi;
        //if (s_useGdApi) PRINT(("Now Using GD...\n")); else PRINT(("Now NOT using GD...\n"));
    }
    else
        frame++;

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (s_QuadFogIndexBufID) nn::gd::Resource::ReleaseVertexBufferResource(s_QuadFogIndexBufID);
    if (s_QuadFogVertBufID)  nn::gd::Resource::ReleaseVertexBufferResource(s_QuadFogVertBufID);
    if (s_QuadFogColBufID)   nn::gd::Resource::ReleaseVertexBufferResource(s_QuadFogColBufID);
    if (s_QuadAllIndexBufID) nn::gd::Resource::ReleaseVertexBufferResource(s_QuadAllIndexBufID);
    if (s_QuadAllVertBufID)  nn::gd::Resource::ReleaseVertexBufferResource(s_QuadAllVertBufID);
    if (s_QuadAllColBufID)   nn::gd::Resource::ReleaseVertexBufferResource(s_QuadAllColBufID);

    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    uploadFogLUT();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

