﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     LightingAniso.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <math.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include "Util.h"
#include "Loader.h"

#include <nn/gd.h>

#define WAIT_VSYNC

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

static nn::gd::UniformLocation s_shaderVariable_view;
static nn::gd::UniformLocation s_shaderVariable_uTangEnabled;

//Vertex input
static nn::gd::InputLayout* s_inputLayout = 0;
static nn::gd::InputLayout* s_inputLayout_tan = 0;

//Textures
static nn::gd::Texture2DResource* s_texture2DResource_0 = 0;
static nn::gd::Texture2D* s_texture2D_0 = 0;
static nn::gd::Texture2DResource* s_texture2DResource_1 = 0;
static nn::gd::Texture2D* s_texture2D_1 = 0;
static nn::gd::SamplerState* sampler_0 = 0;
static nn::gd::SamplerState* sampler_1 = 0;

/* obj loader class object */
typedef struct tag_datGD_t
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    int             obj_num;
} datGD_t;
static datGD_t kimono;

void loadDAT_GD(const char* _name, datGD_t* _dat);
void unloadDAT_GD(datGD_t* dat);

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void SetLutTable()
{
    /* Schlick suggested that the distribution function of anisotropic reflection
    is composed of two functions: one dependent on the dot product NH, and the other dependent on the
    dot product of the angle between the projection of H on the tangent plane and the
    tangent vector. In microfacets theory, the distribution function is essentially not
    dependent on wavelength (R,G,B). This sample distribution term 
    looked up by NH is dependent on wave length; therefore, specular has a little blue. */

 //   GLuint lutids[5] ;

 //   glGenTextures(5, lutids);

    GLfloat lut[512];
    int j;
    memset(lut, 0, sizeof(lut));

    /* to ignore 1st specular term, set 0. */
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D0, 0, lut, lut+256, 256);

    /* set term depended on NH */
    for (j = 0; j < 256; j++)
        lut[j] = 0.7f * z_schlick(0.7f, (float)j/256.f, true);

    for (j = 0; j < 255; j++)
        lut[j + 256] = lut[j+1] - lut[j];

    lut[255 + 256] = 0.7f * z_schlick(0.7f, 1.f, true) - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RR, 0, lut, lut+256, 256);

    for (j = 0; j < 256; j++)
        lut[j] = 0.7f * z_schlick(0.7f, (float)j/256.f, true);

    for (j = 0; j < 255; j++)
        lut[j + 256] = lut[j+1] - lut[j];

    lut[255 + 256] = 0.7f * z_schlick(0.7f, 1.f, true) - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RG, 0, lut, lut+256, 256);

    for (j = 0; j < 256; j++)
        lut[j] = z_schlick(0.5f, (float)j/256.f, true);

    for (j = 0; j < 255; j++)
        lut[j + 256] = lut[j+1] - lut[j];

    lut[255 + 256] = z_schlick(0.5f, 1.f, true) - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RB, 0, lut, lut+256, 256);

    /* set term dependent on dot product between the projection of H on the tangent plane
    and the tangent vector */
    for (j = 0; j < 256; j++)
        lut[j] = a_schlick(0.015f, (float)j/256.f, true);

    for (j = 0; j < 255; j++)
        lut[j + 256] = lut[j + 1] - lut[j];

    lut[255 + 256] = a_schlick(0.015f, 1.f, true) - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D1, 0, lut, lut+256, 256);
}

unsigned int convertGLFormatTypeToGDFormat(unsigned int format, unsigned int type)
{
    unsigned int res = 1000;
    if (type != GL_UNSIGNED_BYTE) return 1000;
    switch (format)
    {
        case GL_RGB:    res = nn::gd::Resource::FORMAT_RGB_888;      break;
        case GL_RGBA:   res = nn::gd::Resource::FORMAT_RGBA_8888;    break;
        default: return 1000;
    }
    return res;
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    s_shaderVariable_view = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    s_shaderVariable_uTangEnabled = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uTangentEnabled");
    NN_ASSERT(s_shaderVariable_view.IsValid());
    NN_ASSERT(s_shaderVariable_uTangEnabled.IsValid());

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    //Load kimono data
    loadDAT_GD( "rom:/resources/kimono.dat", &kimono);

    //Create vertex input layout
    nnResult res;
    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides[] = { sizeof(float) * 3, sizeof(float) * 3};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &s_inputLayout);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::InputElementDescription descs_tan[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 2, "aTang", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 3, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides_tan[] = { sizeof(float) * 3, sizeof(float) * 3, sizeof(float) * 3, sizeof(float) * 3};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs_tan, 4, strides_tan, vertexShader, &s_inputLayout_tan);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Load textures
    u32 width, height, format, type;
    nn::gd::Resource::Format img_format;
    u8* pixels;
    nn::gd::Resource::NativeFormat nativeFormat;

    pixels = GetTextureData_TGA("rom:/resources/KIMONO01.tga", width, height, format, type, &s_AppHeap);
    if ( !pixels ) {NN_PANIC("\n");}
    img_format = (nn::gd::Resource::Format)convertGLFormatTypeToGDFormat(format, type);
    res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(img_format, width, height, pixels, pixels, &nativeFormat);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}
    nn::gd::Texture2DResourceDescription Text2DResDesc = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, pixels, GD_FALSE, &s_texture2DResource_0);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DDescription tex2DDesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResource_0, &tex2DDesc, &s_texture2D_0);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}
    s_AppHeap.Free(pixels);

    nn::gd::SamplerStateDescription descSampler_0;
    descSampler_0.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR);
    descSampler_0.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR);
    descSampler_0.SetMinLod(0);
    nn::gd::TextureStage::CreateSamplerState(&descSampler_0, &sampler_0);

    // load bump texture
    pixels = GetTextureData_TGA("rom:/resources/KIMONOB0.tga", width, height, format, type, &s_AppHeap);
    if (!pixels){ NN_ASSERTMSG( 0, "Error loading texture"); }
    img_format = (nn::gd::Resource::Format)convertGLFormatTypeToGDFormat(format, type);
    res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(img_format, width, height, pixels, pixels, &nativeFormat);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}
    nn::gd::Texture2DResourceDescription Text2DResDesc_1 = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_1, pixels, GD_FALSE, &s_texture2DResource_1);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DDescription tex2DDesc_1 = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResource_1, &tex2DDesc_1, &s_texture2D_1);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}
    s_AppHeap.Free(pixels);

    nn::gd::SamplerStateDescription descSampler_1;
    descSampler_1.SetMinFilter(nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR);
    descSampler_1.SetMagFilter(nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR);
    descSampler_1.SetMinLod(0);
    nn::gd::TextureStage::CreateSamplerState(&descSampler_1, &sampler_1);

    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_0);
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_1, *s_texture2D_1);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, sampler_0);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_1, sampler_1);
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    //Projection matrix
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.01f, 0.01f, -0.01f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.01f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 10.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection"), static_cast<f32*>(proj));

    //Texture combiner
    nn::gd::CombinerDescription combiner;
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MULT_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetCombinerInUse( nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //Initialize lighting
    nn::gd::LightingStage::light[0].EnableLight(GD_TRUE);
    nn::gd::LightingStage::light[1].EnableLight(GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_SP, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_RB, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_RG, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, GD_TRUE);

    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, nn::gd::LightingStage::INPUT_VALUE_CP);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RB, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RG, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::INPUT_VALUE_NH);

    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_SP, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RB, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RG, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);

    /* when anisotropic lighting, distribution function is looked up by
    dot product between projection of half-vector onto tangent plane and
    tangent vector. HW specification require layer configuration setting7
    in this case. another layer configuration setting can not be used. */
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_7);

    nn::gd::LightingStage::SetBumpSelectorTexture(nn::gd::TextureStage::TEXTURE_UNIT_1);
    nn::gd::LightingStage::SetBumpMode(nn::gd::LightingStage::BUMPMODE_AS_BUMP, GD_FALSE);

    nn::gd::LightingStage::EnableLookUpTableD0(GD_TRUE);
    nn::gd::LightingStage::EnableLookUpTableD1(GD_TRUE);

    nn::gd::LightingStage::SetClampHightlight(GD_FALSE);

    nn::gd::LightingStage::light[0].EnableGeomFactor0(GD_FALSE);
    nn::gd::LightingStage::light[0].EnableGeomFactor1(GD_FALSE);
    nn::gd::LightingStage::light[0].EnableTwoSideDiffuse(GD_TRUE);

    nn::gd::LightingStage::light[1].EnableGeomFactor0(GD_FALSE);
    nn::gd::LightingStage::light[1].EnableGeomFactor1(GD_FALSE);
    nn::gd::LightingStage::light[1].EnableTwoSideDiffuse(GD_TRUE);

    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);

    u8 ld0[] = {128, 128, 128} ;
    u8 ls0a[] = {48, 48, 48} ;
    u8 ls0b[] = {179, 179, 179} ;
    u8 ld1[] = {128, 128, 128} ;
    u8 ls1a[] = {48, 48, 48} ;
    u8 ls1b[] = {179, 179, 179} ;

    nn::gd::LightingStage::light[0].SetColorSpecular0(ls0a[0], ls0a[1], ls0a[2]);
    nn::gd::LightingStage::light[0].SetColorSpecular1(ls0b[0], ls0b[1], ls0b[2]);
    nn::gd::LightingStage::light[0].SetColorDiffuse(ld0[0], ld0[1], ld0[2]);

    nn::gd::LightingStage::light[1].SetColorSpecular0(ls1a[0], ls1a[1], ls1a[2]);
    nn::gd::LightingStage::light[1].SetColorSpecular1(ls1b[0], ls1b[1], ls1b[2]);
    nn::gd::LightingStage::light[1].SetColorDiffuse(ld1[0], ld1[1], ld1[2]);

    /* light position settings */
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 0.f, 1.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);

    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    nn::math::Vector4 lpos0(-1.f, 0.5f, 1.f, 1.f);
    nn::math::Vector4 lpos1(1.f, 0.5f,1.f,1.f);
    
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&mv0, &lpos0), nn::math::VEC4Dot(&mv1, &lpos0), nn::math::VEC4Dot(&mv2, &lpos0), lpos0.w);
    
    nn::gd::LightingStage::light[0].SetPosition(lpos.x, lpos.y, lpos.z);

    lpos = nn::math::Vector4(nn::math::VEC4Dot(&mv0, &lpos1), nn::math::VEC4Dot(&mv1, &lpos1), nn::math::VEC4Dot(&mv2, &lpos1), lpos1.w);
    nn::gd::LightingStage::light[1].SetPosition(lpos.x, lpos.y, lpos.z);

    SetLutTable();
}

void DrawObj()
{
    for (int i = 0; i < kimono.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(s_inputLayout_tan);

        nn::gd::VertexBufferResource* buffers[] = { kimono.posVB, kimono.normVB, kimono.tangVB, kimono.texVB };

        u32 offsets[] = { kimono.obj[i].vtx_offset, kimono.obj[i].nrm_offset, kimono.obj[i].tgt_offset, kimono.obj[i].tex_offset };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 4, buffers, offsets);

        int TangEnabled = (kimono.obj[i].tex_size) ? 1 : 0;
        nn::gd::ShaderStage::SetShaderPipelineConstantBoolean(shaderPipeline, s_shaderVariable_uTangEnabled, TangEnabled, 1);
   
        for (unsigned j = kimono.obj[i].patch_offset; j < kimono.obj[i].patch_size + kimono.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(kimono.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, kimono.patch[j].elm_offset + kimono.obj[i].elm_offset);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(kimono.patch[j].elm_size, 0);
        }
        //break;
    }
}
        
//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        u8 clearColor[] = {80, 100, 125, 255};
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

        //modelview
        nn::math::Matrix34 mv, rot;
        nn::math::Vector3 camPos(0.f, 0.f, 1.5f);
        nn::math::Vector3 camUp(0.f, 1.f, 0.f);
        nn::math::Vector3 target(0.f, 0.f, 0.f);
        nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
        nn::math::MTX34RotXYZDeg(&rot, 0.f, static_cast<f32>(-f), 0.f);
        nn::math::MTX34Mult(&mv, &mv, &rot);
        nn::math::Matrix44 tmp(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_view, static_cast<f32*>(tmp));

        //Draw
        DrawObj();

        //if (f == 10) nngxDumpHWRegisters(0x0, 0x2dd);
        //nngxDumpCmdBufRegistries(0x0, 0xfff, false);
        //nngxDumpCmdBufHistory(0x2c0, 0x2d6, true);

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (s_inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout);
    if (s_inputLayout_tan) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout_tan);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    if (s_texture2DResource_0) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_0);
    if (s_texture2DResource_1) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_1);
    if (s_texture2D_0) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_0);
    if (s_texture2D_1) nn::gd::TextureStage::ReleaseTexture2D(s_texture2D_1);
    if (sampler_0) nn::gd::TextureStage::ReleaseSamplerState(sampler_0);
    if (sampler_1) nn::gd::TextureStage::ReleaseSamplerState(sampler_1);

   //release data
    unloadDAT_GD(&kimono);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void loadDAT_GD(const char* _name, datGD_t* _dat)
{
    if (!_name || !_dat)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)s_AppHeap.Allocate(strlen(_name) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, _name);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(s_AppHeap.Allocate( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(_dat, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }

        _dat->obj = (dat_obj_t*)s_AppHeap.Allocate(sizeof(dat_obj_t) * header->obj_num);
        _dat->patch = (dat_patch_t*)s_AppHeap.Allocate(sizeof(dat_patch_t) * header->patch_num);
        if (!_dat->obj || !_dat->patch)
        {
            if (_dat->obj)
                s_AppHeap.Free(_dat->obj);
            if (_dat->patch)
                s_AppHeap.Free(_dat->patch);
            break;
        }
        memcpy(_dat->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(_dat->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        _dat->obj_num = header->obj_num;
    } while (0);

    if (fname)
        s_AppHeap.Free(fname);
    if (wfname)
        s_AppHeap.Free(wfname);
    if (dat)
        s_AppHeap.Free(dat);
    file.Finalize();

    return;
}

void unloadDAT_GD(datGD_t* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            s_AppHeap.Free(dat->obj);
        if (dat->patch)
            s_AppHeap.Free(dat->patch);
    }

    return;
}
