﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     LightingSample2.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 50059 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <math.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include "Util.h"
#include "Loader.h"

#include <nn/gd.h>

#define WAIT_VSYNC

/* if USE_LN is not defined, half-vector is used as the vector controlling both diffuse illumination and highlight position */
#define USE_LN
/* if WINDOW is defined, a window highlight is used */
#define WINDOW

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* s_inputLayout = 0;

//TextureCube
static nn::gd::Texture2DResource* s_texture2DRes_cube[6] = {0, 0, 0, 0, 0, 0};
static nn::gd::TextureCube* s_textureCube = 0;

/* obj loader class object */
typedef struct tag_datGD_t
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    int             obj_num;
} datGD_t;
static datGD_t sphere;

void loadDAT_GD(const char* _name, datGD_t* _dat);
void unloadDAT_GD(datGD_t* dat);

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void SetLutTable()
{
    const float outline = 0.15f; 
    const float highlight_eps = 0.01f; 
    float delta[] = {1.f, 0.7f, 0.5f, -1.f};

    GLfloat lut[512];
    int j;
    memset(lut, 0, sizeof(lut));
    float previous = 0, LN; 
    int i = 0; 

    /* f(D) = D is used as the initial shading function - see "Illumination Models in Maestro" for details
    first specify the steps in the [0, 1) interval */
    for (j = 127; j >= 0; j--)
    {
        LN = (float)j/127.9375f;
        if (LN > delta[i]) lut[j] = previous; 
        else {
            lut[j] = LN; 
            previous = lut[j]; 
            i++; 
        }
    }
    for (j = 0; j < 127; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[127 + 256] = 0.f; 

    /* now specify the steps in the [-1, 0) interval */
    for (j = 255; j  >= 128; j--)
    {
        LN = (float)(j - 256) /128.f;
        if (LN > delta[i]) lut[j] = previous; 
        else {
            lut[j] = LN; 
            previous = lut[j]; 
            i++; 
        }
    }

    /* now calculate differences (some are 0, some are not) */
    for (j = 128; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = lut[0] - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RR, 0, lut, lut+256, 256);

    for (j = 0; j < 256; j++)
        if ((float)j/255.9375f <= 1.f - highlight_eps) lut[j] = 0.f; 
        else  lut[j] = 1.f; 
    for (j = 0; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = 0.f;
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D0, 0, lut, lut+256, 256);

    for (j = 0; j < 256; j++)
        if ((float)j/255.9375f < outline)  lut[j] = 0.f; 
        else lut[j] = 1.f; 
    for (j = 0; j < 255; j++) lut[j + 256] = lut[j+1] - lut[j];
    lut[255 + 256] = 1.f - lut[255];
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D1, 0, lut, lut+256, 256);

    for (j = 0; j < 256; j++) lut[j] = 1.f; 
    for (j = 0; j < 256; j++)lut[j + 256] = 0.f; 
    nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_FR, 0, lut, lut+256, 256);
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    //Load sphere data
    loadDAT_GD( "rom:/resources/sphere.dat", &sphere);

    //Create vertex input layout
    nnResult res;
    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
    };

    u32 strides[] = { sizeof(float) * 3, sizeof(float) * 3};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &s_inputLayout);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Load textures
    /* if you environment highlight specular, load environment cube texture */
#ifdef WINDOW
    char *env_tex[] =
    {
         "rom:/resources/wright.tga",
         "rom:/resources/wleft.tga",
         "rom:/resources/wtop.tga",
         "rom:/resources/wbottom.tga",
         "rom:/resources/wback.tga",
         "rom:/resources/wfront.tga",
    };

    for (int face = 0; face < 6; face++)
    {
        u32 width, height, format, type;
        u8* pixel = GetTextureData_TGA(env_tex[face], width, height, format, type, &s_AppHeap);
        if ( !pixel ) {NN_PANIC("\n");}

        nn::gd::Resource::Format gdformat;
        if (format == IMAGE_RGB && type == IMAGE_UNSIGNED_BYTE) NN_PANIC("\n"); // RGB_888 is not supported for cube textures
        else if (format == IMAGE_RGBA && type == IMAGE_UNSIGNED_BYTE) gdformat = nn::gd::Resource::FORMAT_RGBA_8888;
        else {NN_PANIC("\n");}

        nn::gd::Resource::NativeFormat nativeFormat;
        res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(gdformat, width, height, pixel, pixel, &nativeFormat);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        nn::gd::Texture2DResourceDescription text2DResDesc = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateTexture2DResource(&text2DResDesc, pixel, GD_FALSE, s_texture2DRes_cube + face);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        s_AppHeap.Free(pixel);
    }

    //nn::gd::Texture2DResource* s_texture2DRes_cubeTest[6] = {*s_texture2DRes_cube, *s_texture2DRes_cube, *s_texture2DRes_cube, *s_texture2DRes_cube, *s_texture2DRes_cube, *s_texture2DRes_cube};
    nn::gd::TextureCubeDescription textCubeDesc(0,0);
    res = nn::gd::TextureStage::CreateTextureCube(s_texture2DRes_cube, &textCubeDesc, &s_textureCube);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::TextureStage::SetTextureUnit0(*s_textureCube);
#endif
}

void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    //Projection matrix
    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.015f, 0.015f, -0.015f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.015f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 200.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection"), static_cast<f32*>(proj));

    //modelview matrix
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 5.75f, 22.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    nn::math::Matrix34 rot;
    nn::math::MTX34RotXYZDeg(&rot, 0.f, 0.f, 0.f);
    nn::math::MTX34Mult(&mv, &mv, &rot);
    nn::math::Matrix44 m(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView"), static_cast<f32*>(m));

    //world matrix
    nn::math::Matrix44 id;
    nn::math::MTX44Identity(&id);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uWorld"), static_cast<f32*>(id));

    //Texture combiner
#ifdef WINDOW
    /* output color is EnvTexture_rgb * PriColor_alpha + SecColor_rgb
    EnvTexture_rgb is environment highlight specular, PriColor_alpha is frensel reflection,
    and SecColor_rgb is toon shading diffuse color. */
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MULT_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState); // create dmp_TexEnv[0] combiner object
#else
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.m_combinerUnit[0].SetSourceRGB(nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.m_combinerUnit[0].SetOperandRGB(nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.m_combinerUnit[0].SetCombineRGB(nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.m_combinerUnit[0].SetCombineAlpha(nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.m_CombinerInUse = nn::gd::CombinerStage::UNIT_0;
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState); // create dmp_TexEnv[0] combiner object
#endif

    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //Initialize lighting
    for (int i=0; i<nn::gd::LightingStage::COUNT_LIGHTS; ++i) nn::gd::LightingStage::light[i].EnableLight(GD_TRUE);

    nn::gd::LightingStage::SetGlobalColorAmbient(0, 0, 0);

    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, GD_FALSE);
    
#ifdef USE_LN
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::INPUT_VALUE_LN);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, nn::gd::LightingStage::INPUT_VALUE_LN);
#else
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, nn::gd::LightingStage::INPUT_VALUE_NH);
#endif
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, nn::gd::LightingStage::INPUT_VALUE_NV);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, nn::gd::LightingStage::INPUT_VALUE_NV);

    nn::gd::LightingStage::light[0].EnableShadowed(GD_FALSE);
    nn::gd::LightingStage::light[0].EnableTwoSideDiffuse(GD_FALSE);
    nn::gd::LightingStage::light[0].EnableGeomFactor0(GD_FALSE);
    nn::gd::LightingStage::light[0].EnableGeomFactor1(GD_FALSE);

    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_D1, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_SP, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_FR, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RB, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RG, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);
    nn::gd::LightingStage::SetLookUpTableOutputScaling(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::OUTPUT_SCALE_VALUE_1);

    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_6);
    nn::gd::LightingStage::SetBumpMode(nn::gd::LightingStage::BUMPMODE_NOT_USED, GD_TRUE);

    nn::gd::LightingStage::EnableLookUpTableD0(GD_TRUE);
    nn::gd::LightingStage::EnableLookUpTableD1(GD_TRUE);
    nn::gd::LightingStage::SetClampHightlight(GD_FALSE);
    nn::gd::LightingStage::SetFresnelSelector(nn::gd::LightingStage::FRESNEL_SELECTOR_TYPE_PRIMARY_SECONDARY_ALPHA_FRESNEL);

    SetLutTable();
}

static void UpdateLightingRenderState(int f)
{
    float coeff = cosf(2.f * f / 60.f);
    if (coeff < 0) coeff = -coeff;
    coeff = 0.5f + coeff*0.5f;

    /* light direction setting */
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 5.75f, 22.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Vector4 lpos0(10.5f, 9.5f, 14.f, 1.f);
    lpos0.y -= 3.f*cosf(2.f * f / 60.f);
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&mv0, &lpos0), nn::math::VEC4Dot(&mv1, &lpos0), nn::math::VEC4Dot(&mv2, &lpos0), lpos0.w);

    for (int i=0; i<nn::gd::LightingStage::COUNT_LIGHTS; ++i)
    {
        nn::gd::LightingStage::light[i].SetColorAmbient(255*coeff, 255*coeff, 255*coeff);
        nn::gd::LightingStage::light[i].SetColorDiffuse(255*coeff, 255*coeff, 255*coeff);
        nn::gd::LightingStage::light[i].SetColorSpecular0(255*coeff, 255*coeff, 255*coeff);
        nn::gd::LightingStage::light[i].SetColorSpecular1(255*coeff, 51*coeff, 0*coeff);

        nn::gd::LightingStage::light[i].SetPosition(lpos.x - coeff*i*2.f, lpos.y - coeff*i*2.f, lpos.z - coeff*i*2.f);
    }

    nn::gd::LightingStage::light[0].EnableShadowed(GD_FALSE);
    nn::gd::LightingStage::EnableLookUpTableD0(GD_TRUE);
    nn::gd::LightingStage::EnableLookUpTableD1(GD_TRUE);
    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);
    nn::gd::LightingStage::SetFresnelSelector(nn::gd::LightingStage::FRESNEL_SELECTOR_TYPE_PRIMARY_SECONDARY_ALPHA_FRESNEL);

    nn::gd::LightingStage::SetClampHightlight(GD_FALSE);
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_6);
    nn::gd::LightingStage::SetBumpMode(nn::gd::LightingStage::BUMPMODE_NOT_USED, GD_TRUE);
}

void DrawSphere()
{
    for (int i = 0; i < sphere.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(s_inputLayout);

        nn::gd::VertexBufferResource* buffers[] = { sphere.posVB, sphere.normVB };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);
        for (unsigned j = sphere.obj[i].patch_offset; j < sphere.obj[i].patch_size + sphere.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(sphere.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(sphere.patch[j].elm_size, 0);
        }
    }
}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    s64 startTime = GetTimeMicroSeconds();

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);
    }

    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        u8 clearColor[] = {80, 100, 125, 255};
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

        //Draw
        UpdateLightingRenderState(f);
        DrawSphere();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (s_inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    for (int i=0; i<6; ++i)
    {
        if (s_texture2DRes_cube[i]) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes_cube[i]);
    }
    if (s_textureCube) nn::gd::TextureStage::ReleaseTextureCube(s_textureCube);


    //release data
    unloadDAT_GD(&sphere);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
   // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void loadDAT_GD(const char* _name, datGD_t* _dat)
{
    if (!_name || !_dat)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)s_AppHeap.Allocate(strlen(_name) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, _name);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(s_AppHeap.Allocate( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(_dat, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }

        _dat->obj = (dat_obj_t*)s_AppHeap.Allocate(sizeof(dat_obj_t) * header->obj_num);
        _dat->patch = (dat_patch_t*)s_AppHeap.Allocate(sizeof(dat_patch_t) * header->patch_num);
        if (!_dat->obj || !_dat->patch)
        {
            if (_dat->obj)
                s_AppHeap.Free(_dat->obj);
            if (_dat->patch)
                s_AppHeap.Free(_dat->patch);
            break;
        }
        memcpy(_dat->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(_dat->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        _dat->obj_num = header->obj_num;
    } while (0);

    if (fname)
        s_AppHeap.Free(fname);
    if (wfname)
        s_AppHeap.Free(wfname);
    if (dat)
        s_AppHeap.Free(dat);
    file.Finalize();

    return;
}

void unloadDAT_GD(datGD_t* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            s_AppHeap.Free(dat->obj);
        if (dat->patch)
            s_AppHeap.Free(dat->patch);
    }

    return;
}
