﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     Gas.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <math.h>
#include "Gas.h"
#include "Util.h"

#include "demo.h"
#include <assert.h>

#include <nn/gd.h>
#include "../../../gx/DMPGL/Common/Tga.h"
#include "../../../gx/DMPGL/Common/Memory.h"

extern GLuint pAccId;
extern GLuint pPostId;
extern struct gas_data gas;

/*
 * Local definition
 */
#define SHADING_WIDTH       nn::gx::DISPLAY0_WIDTH
#define SHADING_HEIGHT      nn::gx::DISPLAY0_HEIGHT
// SHADING_BUFFER uses the display buffer created by demo::RenderSystem.
// #define SHADING_BUFFER      DISPLAY_BUFFER

/* buffer id */
static struct tagBufID
{
    nn::gd::RenderTarget*      acc;   /* Color buffer for accumulation */
    nn::gd::Texture2DResource* accColor;
} gasbuf;

/*struct tagCopyParam
{
    GLvoid* copySrcaddr;
    GLsizei copySrcwidth;
    GLsizei copySrcInterval;
    GLvoid* copyDstaddr;
    GLsizei copyDstwidth;
    GLsizei copyDstInterval;
    GLsizei copyTtotalsize;
} copyParam;*/

/* texture id */
// GLuint gastex;

/* particle pattern file names */
static char* particle_files[PARTICLE_PATTERNS] = {PARTICLE_FILES};

/* vertex buffer id */
//GLuint element_array_buffer_id;

extern nn::gd::CombinerState *texGasAcc;
extern nn::gd::CombinerState *texGasShading;

static nn::gd::Texture2DResource* s_texture2DRes=0;
static nn::gd::Texture2D* texGas=0;
static nn::gd::SamplerState *sampler=0;
static nn::gd::VertexBufferResource* bufferIndex = 0;
extern nn::gd::ShaderPipeline* SPAccId;
extern nn::gd::ShaderPipeline* SPPostId;
extern nn::gd::InputLayout* ILAccId;
extern nn::gd::InputLayout* ILPostID;

unsigned char* loadTGA_Texture(const char * _name, unsigned int& width, unsigned int& height, unsigned int& format, unsigned int& type)
{
    unsigned char* pixels;
    unsigned orientation;

    pixels = (unsigned char*)dmpLoadTGA(_name, &width, &height, &format, &type, &orientation);
    if (!pixels) return 0;

    switch (format)
    {
        case IMAGE_RGB:     format = GL_RGB;    break;
        case IMAGE_RGBA:    format = GL_RGBA;   break;
        default:
            if (pixels)
                free(pixels);
            return 0;
    }

    switch (type)
    {
        case IMAGE_UNSIGNED_BYTE:   type = GL_UNSIGNED_BYTE;    break;
        default:
            if (pixels) free(pixels);
            return 0;
    }

    return pixels;
}

unsigned int convertGLFormatTypeToGDFormat(unsigned int format, unsigned int type)
{
    unsigned int res = 1000;
    if (type != GL_UNSIGNED_BYTE) return 1000;
    switch (format)
    {
        case GL_RGB:    res = nn::gd::Resource::FORMAT_RGB_888;      break;
        case GL_RGBA:   res = nn::gd::Resource::FORMAT_RGBA_8888;    break;
        default: return 1000;
    }
    return res;
}

/*=======================================================*/
/* buffer initialization                                 */
/*=======================================================*/
void GasInitialize(void)
{
    /*
     * Generate framebuffers and texture for gas rendering
    */
    nnResult res;

    nn::gd::Texture2DResourceDescription ResDesc = {GAS_TEX_WIDTH, GAS_TEX_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_GAS, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    res = nn::gd::Resource::CreateTexture2DResource(&ResDesc, NULL, GD_FALSE, &s_texture2DRes);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DDescription Texdesc = {0, 0};

    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DRes, &Texdesc, &texGas);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.SetGas();
//        samplerdesc.m_MinLod = 0;
//        samplerdesc.m_WrapS = samplerdesc.m_WrapT = nn::gd::TextureStage::SAMPLER_WRAP_CLAMP_TO_BORDER;
//        samplerdesc.m_MinFilter = nn::gd::TextureStage::SAMPLER_MIN_FILTER_NEAREST;
//        samplerdesc.m_MagFilter = nn::gd::TextureStage::SAMPLER_MAG_FILTER_NEAREST;
    nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &sampler);

    /* generation */
    nn::gd::Texture2DResourceDescription Text2DResDesc = {GAS_ACC_WIDTH, GAS_ACC_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};

    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, 0, GD_FALSE, &gasbuf.accColor);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(gasbuf.accColor, &descRenderTarget, &gasbuf.acc);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    /* element array buffer initialize */
    GLushort index[4] = {0, 1, 2, 3};
    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(index);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, index, &bufferIndex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    // set parameters for accumulation result texture copy
    nn::gd::Texture2DProperties GasTexProp;
    nn::gd::TextureStage::GetTexture2DProperties(texGas, &GasTexProp);
    //u32 bytesize = GasTexProp.m_pixelSize>>3;

    nn::gd::TargetProperties accRenderProp;
    nn::gd::OutputStage::GetRenderTargetProperties(gasbuf.acc, &accRenderProp);

    /*
    u32 srcwidth = accRenderProp.m_Width;
    u32 srcheight = accRenderProp.m_Height;
    u32 dstwidth = GasTexProp.m_Width;
    u32 dstheight = GasTexProp.m_Height;

    copyParam.copySrcaddr = (GLvoid *)accRenderProp.m_surfAddr;
    copyParam.copySrcwidth = copyParam.copyDstwidth = srcwidth * accRenderProp.m_pixelSize;
    copyParam.copySrcInterval = 0;
    copyParam.copyDstaddr = (GLvoid *)(GasTexProp.m_surfAddr+(dstheight-srcheight) * dstwidth * bytesize);
    copyParam.copyDstInterval = (dstheight - srcwidth) * accRenderProp.m_pixelSize;
    copyParam.copyTtotalsize = srcwidth * bytesize * srcheight;*/
}
/*=======================================================*/
/* Setup of default gasesous object data structure       */
/*=======================================================*/
void DefaultGasObject(struct gas_data *gas, float *gasColorTable)
{
    nnResult res;

    gas->_dela_z = 5000.f;
    gas->_autoAcc = GL_TRUE;
    gas->_densMax = 1.0f;
    gas->_lightDirX = 0.0f;
    gas->_lightDirY = 0.0f;

    gas->_LightXY[0] = 0.0f;
    gas->_LightXY[1] = 0.0f;
    gas->_LightXY[2] = 0.0f;
    gas->_LightXY[3] = 0.0f;    /* unsused */

    gas->_LightZ[0] = 1.0f;
    gas->_LightZ[1] = 0.0f;
    gas->_LightZ[2] = 1.0f;
    gas->_LightZ[3] = 1.0f;

    gas->shadingDensitySrc = GL_GAS_DEPTH_DENSITY_DMP;      /* GL_GAS_PLAIN_DENSITY_DMP/GL_GAS_DEPTH_DENSITY_DMP */
    gas->colorLutInput = GL_GAS_DENSITY_DMP;

    float dxt = 1.0f / 127.0f;
    float xt = 0;
    for (int i = 0; i < 128; i++)
    {
        gas->fogTable[i]= 1.0f - exp(-4.0f * xt);
        xt += dxt;
    }

    for (int i = 0; i < 128; i++)
    {
        gas->fogTable[128 + i] = gas->fogTable[i + 1] - gas->fogTable[i];
    }

    gas->fogTable[255] = 0;

    SetGasColorTable(gas,gasColorTable);

    float u0 = 0.0f;
    float v0 = 0.0f;

    float u1 = (GAS_ACC_WIDTH * 1.0f) / (GAS_TEX_WIDTH * 1.0f);
    float v1 = (GAS_ACC_HEIGHT * 1.0f) / (GAS_TEX_HEIGHT * 1.0f);

    GLfloat tex_unit[8]= {u0, v0, u0, v1, u1, v1, u1, v0};

    GLfloat LX0, LY0, LX1, LY1;
    LX0 = 0.0f;
    LY0 = 0.0f;
    LX1 = gas->_lightDirX;
    LY1 = gas->_lightDirY;
    GLfloat vertex_color[16] =
    {
        LX0, 0.0f, 0.0f, LY0,
        LX0, 0.0f, 0.0f, LY1,
        LX1, 0.0f, 0.0f, LY1,
        LX1, 0.0f, 0.0f, LY0
    };

    GLushort _quadIndex[6] = {0, 1, 2, 0, 2, 3};
    GLfloat vertex_unit[16] =
    {
        -1.0f, -1.0f, 0.0f, 1.0f,
        -1.0f,  1.0f, 0.0f, 1.0f,
         1.0f,  1.0f, 0.0f, 1.0f,
         1.0f, -1.0f, 0.0f, 1.0f
    };

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(_quadIndex);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, _quadIndex, &gas->quad_index_ID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(vertex_unit);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, vertex_unit, &gas->quad_vertBuf_ID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(tex_unit);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, tex_unit, &gas->quad_texBuf_ID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(vertex_color);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, vertex_color, &gas->quad_colBuf_ID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    /*
     * Load particle patterns
    */
    unsigned width, height, format, type, img_format;
    unsigned char* pixels=0;
    nn::gd::Resource::NativeFormat nativeFormat;

    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.ToDefault();
    samplerdesc.m_MinFilter = nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR;
    samplerdesc.m_MagFilter = nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR;
    nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &gas->samplerPattern);

    for (int i = 0; i < PARTICLE_PATTERNS; i++)
    {
        pixels = loadTGA_Texture(particle_files[i], width, height, format, type);
        if (!pixels){ NN_ASSERTMSG( 0, "Error loading texture"); }

        img_format = convertGLFormatTypeToGDFormat(format, type);
        res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat((nn::gd::Resource::Format)img_format, width, height, pixels, pixels, &nativeFormat);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

        nn::gd::Texture2DResourceDescription Text2DResDesc = {width, height, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
         res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, pixels, GD_FALSE, &gas->patternResource[i]);
         if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

        nn::gd::Texture2DDescription tex2ddesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(gas->patternResource[i], &tex2ddesc, &gas->pattern[i]);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    }

    if (pixels) free(pixels);
}

void GasFinalize()
{
    if (gasbuf.accColor) nn::gd::Resource::ReleaseTexture2DResource(gasbuf.accColor);
    if (gasbuf.acc)      nn::gd::OutputStage::ReleaseRenderTarget(gasbuf.acc);

    if (s_texture2DRes)  nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes);
    if (texGas)          nn::gd::TextureStage::ReleaseTexture2D(texGas);
    if (sampler)         nn::gd::TextureStage::ReleaseSamplerState(sampler);
    if (bufferIndex)     nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
}

/*=======================================================*/
/* Load Gas color lookup table                           */
/*=======================================================*/
void SetGasColorTable(struct gas_data *gas, float *gasColorTable)
{
//    glBindTexture(GL_TEXTURE_COLLECTION_DMP, gas->CollectionLUT_ID);
    for(int i = 0; i < 8; i++)
    {
        gas->RR[i] = gasColorTable[3 * i + 0];
        gas->GG[i] = gasColorTable[3 * i + 1];
        gas->BB[i] = gasColorTable[3 * i + 2];

        gas->RR[8 + i] = gasColorTable[3 * (i + 1) + 0] - gasColorTable[3 * i + 0];
        gas->GG[8 + i] = gasColorTable[3 * (i + 1) + 1] - gasColorTable[3 * i + 1];
        gas->BB[8 + i] = gasColorTable[3 * (i + 1) + 2] - gasColorTable[3 * i + 2];
    }
    gas->RR[15] = 0.0f;
    gas->GG[15] = 0.0f;
    gas->BB[15] = 0.0f;

#define TMPBUFSIZE 16+16+16
        float tmpbuf[TMPBUFSIZE];
        int bpos = 0;

        for (int i = 0; i < 16; i++)
        {
            tmpbuf[bpos++] = gas->RR[i];
            tmpbuf[bpos++] = gas->GG[i];
            tmpbuf[bpos++] = gas->BB[i];
        }
        nn::gd::FogStage::UploadGasLookUpTableFloat(0, tmpbuf,  &tmpbuf[8*3], 8);

}

/*=======================================================*/
/* accumulation pass                                     */
/*=======================================================*/
void GasAccumulation()
{
    /*
     * In accumulation pass, particles are accumulated into gas accumulation buffer as
     * density information. This buffer is 'gasbuf.acc' in this sample.
     * This buffer is exactly the same size as DISPLAY_BUFFER (possible not to be
     * pow2 size). So, for the next pass, the accumulation result should be copied to pow2 size texture.
    */
//nngxDumpCmdBufAll(0, 0);
    nn::gd::OutputStage::SetRenderTarget(gasbuf.acc);
        nn::gd::Viewport viewPort(0, 0, GAS_ACC_WIDTH, GAS_ACC_HEIGHT);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    u8 color[] = {0, 0, 0, 0};
    nn::gd::Memory::ClearTargets(gasbuf.acc, NULL, color, 1.f, 0);

    /* use gas accumulation related program */
    nn::gd::ShaderStage::SetShaderPipeline(SPAccId);

    u8 stencilRef = 0;
    nn::gd::OutputStage::SetDepthStencilState(NULL, stencilRef);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *gas.pattern[0]);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, gas.samplerPattern);

    nn::gd::CombinerStage::SetTextureCombinerState(texGasAcc);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_GAS_ACCUMULATION);

    nn::gd::FogStage::SetGasDeltaZ(gas._dela_z);
    nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_NONE);

   // nn::math::Matrix44 m;
   // nn::math::MTX44Transpose(&m, &gas.pSys.m_partsys_center);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "uCenter"), (f32 *)&gas.pSys.m_partsys_center);

//    nn::math::MTX44Transpose(&m, &gas.pSys.m_partsys_radius);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "uRadii"), (f32 *)&gas.pSys.m_partsys_radius);

  //  nn::math::MTX44Transpose(&m, &gas.pSys.m_partsys_color);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.color"), (f32 *)&gas.pSys.m_partsys_color);

  //  nn::math::MTX44Transpose(&m, &gas.pSys.m_partsys_aspect);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.aspect"), (f32 *)&gas.pSys.m_partsys_aspect);

    GLfloat viewport_size[2] = {0.0f, 0.0f};
    viewport_size[0] = 1.f / (GLfloat)GAS_ACC_WIDTH;
    viewport_size[1] = 1.f / (GLfloat)GAS_ACC_HEIGHT;
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.viewport"), &viewport_size[0]);

    GLfloat distance_attenuation[3] = {0.0f, 0.0f, 0.0f};
    distance_attenuation[0] = 0.0f;
    distance_attenuation[1] = 0.0f;
    distance_attenuation[2] = 1.0f / (float)(GAS_ACC_WIDTH * GAS_ACC_HEIGHT);

    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.distanceAttenuation"), &distance_attenuation[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.countMax"), &gas.pSys.m_particleCountMax);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.pointSize"), &gas.pSys.m_size_min_max[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.speed"), &gas.pSys.m_speed);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.randomCore"), &gas.pSys.m_prng[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.randSeed"), &gas.pSys.m_random_seed[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "dmp_PartSys.time"), &gas.pSys.simulationTime);

    nn::gd::VertexInputStage::SetInputLayout(ILAccId);

    nn::gd::VertexBufferResource* buffers[] = { gas.pSys.m_pSysPositionID };
    u32 offsets[] = { 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);
    nn::gd::System::DrawIndexed(4, 0);

    /*
     * Copy accumulation result to the texture buffer
    */
    nn::gd::Memory::Rect memRect(0, 0, -1, -1);
    nnResult res = nn::gd::Memory::CopyTextureSubResource(gasbuf.accColor, -1, memRect, s_texture2DRes, -1, 0, 0);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //nngxWaitCmdlistDone();
    //nngxDumpHWRegisters(0x125, 0x125);
    //nngxDumpHWRegisters(0xe5, 0xe5);

    //nngxSplitDrawCmdlist();
    //nngxAddBlockImageCopyCommand(copyParam.copySrcaddr, copyParam.copySrcwidth, copyParam.copySrcInterval,
    //   copyParam.copyDstaddr, copyParam.copyDstwidth, copyParam.copyDstInterval, copyParam.copyTtotalsize);
}

/*=======================================================*/
/* shading pass                                          */
/*=======================================================*/
void GasShading(void)
{
    /*
     * In shading pass, shaded gaseous image is blended to DISPLAY_BUFFER.
     * Accumulated density information is used as a gas texture.
    */

    /* Bind shading buffer and set viewport
     * Shading buffer is DISPLAY_BUFFER */
//    glBindFramebuffer(GL_FRAMEBUFFER, gasbuf.shading);
//    glViewport(0, 0, SHADING_WIDTH, SHADING_HEIGHT);

    /* use gas accumulation related program */
    nn::gd::ShaderStage::SetShaderPipeline(SPPostId);

    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *texGas);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, sampler);

    /* Setup blending unit #0
     * r component of primary color has the influence of LIGHT_X and LIGHT_Y
     * This output is transferred to FOG unit when FOG_MODE is set to GAS_DMP */
    nn::gd::CombinerStage::SetTextureCombinerState(texGasShading);

    nn::gd::FogStage::GasShadingDensitySource shadingDensitySrc =
        (gas.shadingDensitySrc==GL_GAS_PLAIN_DENSITY_DMP) ? nn::gd::FogStage::GAS_SHADING_DENSITY_SOURCE_PLAIN : nn::gd::FogStage::GAS_SHADING_DENSITY_SOURCE_DEPTH;
    nn::gd::FogStage::GasColorLookUpTableInput colorLutInput =
        (nn::gd::FogStage::GasColorLookUpTableInput)((gas.colorLutInput==GL_GAS_DENSITY_DMP) ? nn::gd::FogStage::GAS_COLOR_INPUT_DENSITY : nn::gd::FogStage::GAS_COLOR_INPUT_LIGHT_FACTOR);

    /* setup of gas shading */
    nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_GAS);
    nn::gd::FogStage::SetGasShadingDensity(shadingDensitySrc);
    nn::gd::FogStage::SetGasLightColorLutInput(colorLutInput);
    nn::gd::FogStage::SetGasLightXY(nn::gd::Utils::FloatToUnsignedByte(gas._LightXY[0]), nn::gd::Utils::FloatToUnsignedByte(gas._LightXY[1]), nn::gd::Utils::FloatToUnsignedByte(gas._LightXY[2]));
    nn::gd::FogStage::SetGasLightZ(nn::gd::Utils::FloatToUnsignedByte(gas._LightZ[0]), nn::gd::Utils::FloatToUnsignedByte(gas._LightZ[1]), nn::gd::Utils::FloatToUnsignedByte(gas._LightZ[2]), nn::gd::Utils::FloatToUnsignedByte(gas._LightZ[3]));
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);

    nn::gd::FogStage::SetGasAccumulationMax(gas._densMax);
    //nn::gd::FogStage::SetGasAutoAccumulation(GD_TRUE);

    /*
     * Miscellaneous settings
    */
    static nn::gd::DepthStencilState* depthStateGasShading = 0;
    if (!depthStateGasShading)
    {
        nn::gd::DepthStencilStateDescription desc;
        desc.ToDefault();
        desc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ZERO;
        nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStateGasShading);
    }

    static nn::gd::BlendState* blendStateGasShading = 0;
    if (!blendStateGasShading)
    {
        nn::gd::BlendStateDescription desc;
        nn::gd::OutputStage::BlendFunction src = nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA;
        nn::gd::OutputStage::BlendFunction dst = nn::gd::OutputStage::BLEND_FUNCTION_ONE_MINUS_SRC_ALPHA;
        nn::gd::OutputStage::BlendEquation eq  = nn::gd::OutputStage::BLEND_EQUATION_ADD;

        desc.ToDefault();
        desc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
        desc.SetBlendFunc(src, dst, eq);
        nn::gd::OutputStage::CreateBlendState(&desc, &blendStateGasShading);
    }

    u8 stencilRef = 0;
    nn::gd::OutputStage::SetDepthStencilState(depthStateGasShading, stencilRef);
    nn::gd::OutputStage::SetBlendState(blendStateGasShading);

    nn::gd::VertexInputStage::SetInputLayout(ILPostID);

    nn::gd::VertexBufferResource* buffers[] = { gas.quad_vertBuf_ID, gas.quad_texBuf_ID,  gas.quad_colBuf_ID };
    u32 offsets[] = { 0, 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 3, buffers, offsets);

    nn::gd::VertexInputStage::SetIndexBuffer(gas.quad_index_ID, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    nn::gd::System::DrawIndexed(6, 0);


    nn::gd::FogStage::SetGasFogMode(nn::gd::FogStage::MODE_NONE);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);


    //nngxDumpHWRegisters(0x125, 0x125);
    //nngxDumpHWRegisters(0xe5, 0xe5);
}

