﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     PartsysGas.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>
#include "Util.h"
#include "Loader.h"
#include "Gas.h"

#include <string.h>

#include "Memory.h"

#include "demo.h"
#include <nn/gd.h>

#define WAIT_VSYNC


/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

/* obj loader class object */
typedef struct tag_datGD_t
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    int             obj_num;
} datGD_t;

void loadDAT_GD(const char* _name, datGD_t* _dat);
void unloadDAT_GD(datGD_t* dat);

/* Standard object (triangles, not gaseous object) */
datGD_t plane;
struct gas_data gas;

// Texture Combiner objects
nn::gd::CombinerState *texDrawObject;
nn::gd::CombinerState *texGasAcc;
nn::gd::CombinerState *texGasShading;

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

// Shader objects
static nn::gd::ShaderBinary* shaderBinary = NULL;
static nn::gd::Shader* s_VSGeoId = NULL;
static nn::gd::Shader* s_VSAccId = NULL;
static nn::gd::Shader* s_GS      = NULL;
static nn::gd::Shader* s_VSPostId      = NULL;
static nn::gd::ShaderPipeline* SPGeoID = NULL;
nn::gd::ShaderPipeline* SPAccId = NULL;
nn::gd::ShaderPipeline* SPPostId = NULL;
static nn::gd::InputLayout* ILGeoID = NULL;
nn::gd::InputLayout* ILAccId = NULL;
nn::gd::InputLayout* ILPostID = NULL;

GLuint s_CmdlistID;
/*=======================================================*/
/* Sample specific. Different color variation for gas    */
/*=======================================================*/
#define NUM_COLOR_SAMPLE  6
static float gasColorTable[NUM_COLOR_SAMPLE][24]=
{
    {   /* red flame */
        0.0f,   0.0f,   0.0f,
        0.2f,   0.15f,  0.05f,
        0.6f,   0.25f,  0.15f,
        0.9f,   0.35f,  0.2f,
        0.92f,  0.6f,   0.15f,
        0.95f,  0.85f,  0.05f,
        1.0f,   0.95f,  0.0f,
        1.0f,   1.0f,   1.0f
    },
    {   /* blue flame */
        0.00f,  0.00f,  0.00f,
        0.05f,  0.15f,  0.20f,
        0.15f,  0.25f,  0.60f,
        0.20f,  0.35f,  0.90f,
        0.15f,  0.60f,  0.92f,
        0.05f,  0.85f,  0.95f,
        0.00f,  0.95f,  1.00f,
        1.00f,  1.00f,  1.00f
    },
    {   /* Grey Gradient */
        0.000000f,   0.000000f, 0.000000f,
        0.140000f,   0.140000f, 0.140000f,
        0.280000f,   0.280000f, 0.280000f,
        0.420000f,   0.420000f, 0.420000f,
        0.560000f,   0.560000f, 0.560000f,
        0.700000f,   0.700000f, 0.700000f,
        0.840000f,   0.840000f, 0.840000f,
        1.000000f,   1.000000f, 1.000000f
    },
    {   /* Pastel color */
        255.0f/255.0f,      191.0f/255.0f,  191.0f/255.0f,
        255.0f/255.0f,      238.0f/255.0f,  191.0f/255.0f,
        222.0f/255.0f,      255.0f/255.0f,  191.0f/255.0f,
        191.0f/255.0f,      255.0f/255.0f,  207.0f/255.0f,
        191.0f/255.0f,      254.0f/255.0f,  255.0f/255.0f,
        191.0f/255.0f,      206.0f/255.0f,  255.0f/255.0f,
        223.0f/255.0f,      191.0f/255.0f,  255.0f/255.0f,
        255.0f/255.0f,      191.0f/255.0f,  238.0f/255.0f
    },
    {   /* Greenish */
        125.0f/255.0f,  255.0f/255.0f,  1.0f/255.0f,
        173.0f/255.0f,  255.0f/255.0f,  35.0f/255.0f,
        189.0f/255.0f,  255.0f/255.0f,  46.0f/255.0f,
        204.0f/255.0f,  255.0f/255.0f,  57.0f/255.0f,
        219.0f/255.0f,  255.0f/255.0f,  68.0f/255.0f,
        229.0f/255.0f,  247.0f/255.0f,  65.0f/255.0f,
        234.0f/255.0f,  230.0f/255.0f,  53.0f/255.0f,
        239.0f/255.0f,  212.0f/255.0f,  40.0f/255.0f
    },
    {   /* Red strips */
         0.000000f,     0.000000f,   0.000000f,
         1.000000f,     0.500000f,   0.000000f,
         0.500000f,     0.000000f,   0.000000f,
         1.000000f,     0.500000f,   0.000000f,
         0.500000f,     0.000000f,   0.000000f,
         1.000000f,     0.50000f,    0.000000f,
         0.500000f,     0.000000f,   0.000000f,
         1.000000f,     0.200000f,   0.300000f
    }
};

void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

/*=======================================================*/
/* Static function declaration                           */
/*=======================================================*/
static int DrawFrame(void);

/*=======================================================*/
/* initialization                                        */
/*=======================================================*/

static void Initialize(void)
{
    /* Initialize display */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    /* Initialize display */
    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);

    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);
    
    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    //---------------------  gdAPI call ---------------------------
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return ;

    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);
    //---------------------  gdAPI call ---------------------------

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nnResult res;

    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
     res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

 /********************************************************/
    /* Initialization for gas */
    GasInitialize();
}

static void LoadParticleSystem1(struct _particleSys *pp)
{
    nnResult res;

    GLfloat vertex[4] = {0.0f, 1.0f, 2.0f, 3.0f};
    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(vertex);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;

    res = nn::gd::Resource::CreateVertexBufferResource(&desc, vertex, &pp->m_pSysPositionID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    /*------------------------------------------------------------------------------------*/
    /*--------Definition of the 4 control point positions---------------------------------*/
    /*------------------------------------------------------------------------------------*/
    /*-----------Control Point 1---------------------Control Point 2----------------------*/
    pp->m_partsys_center.m[0][0] = 0.0f;      /*|*/   pp->m_partsys_center.m[1][0] = -5.0f;
    pp->m_partsys_center.m[0][1] = -20.0f;    /*|*/   pp->m_partsys_center.m[1][1] = -20.0f;
    pp->m_partsys_center.m[0][2] = 0.0f;      /*|*/   pp->m_partsys_center.m[1][2] = 0.0f;
    pp->m_partsys_center.m[0][3] = 1.0f;      /*|*/   pp->m_partsys_center.m[1][3] = 1.0f;
    /*-----------Control Point 3---------------------Control Point 4----------------------*/
    pp->m_partsys_center.m[2][0] = 5.0f;      /*|*/   pp->m_partsys_center.m[3][0] = 0.0f;
    pp->m_partsys_center.m[2][1] = 5.0f;      /*|*/   pp->m_partsys_center.m[3][1] = 15.0f;
    pp->m_partsys_center.m[2][2] = 0.0f;      /*|*/   pp->m_partsys_center.m[3][2] = 0.0f;
    pp->m_partsys_center.m[2][3] = 1.0f;      /*|*/   pp->m_partsys_center.m[3][3] = 1.0f;
    /*------------------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------------------*/

    /*------------------------------------------------------------------------------------*/
    /*--------Definition of the 4 control point colors------------------------------------*/
    /*------------------------------------------------------------------------------------*/
    /*Color is not used in this sample.                                                   */
    /*To modify the alpha value, please use pp->m_partsys_aspect[k][3]                    */
    /*-----------Control Point 1---------------------Control Point 2----------------------*/
    pp->m_partsys_color.m[0][0] = 0.0f;       /*|*/   pp->m_partsys_color.m[1][0] = 1.0f;
    pp->m_partsys_color.m[0][1] = 0.0f;       /*|*/   pp->m_partsys_color.m[1][1] = 0.0f;
    pp->m_partsys_color.m[0][2] = 0.0f;       /*|*/   pp->m_partsys_color.m[1][2] = 0.0f;
    pp->m_partsys_color.m[0][3] = 1.0f;       /*|*/   pp->m_partsys_color.m[1][3] = 1.0f;
    /*-----------Control Point 3---------------------Control Point 4----------------------*/
    pp->m_partsys_color.m[2][0] = 0.0f;       /*|*/   pp->m_partsys_color.m[3][0] = 0.0f;
    pp->m_partsys_color.m[2][1] = 1.0f;       /*|*/   pp->m_partsys_color.m[3][1] = 0.0f;
    pp->m_partsys_color.m[2][2] = 0.0f;       /*|*/   pp->m_partsys_color.m[3][2] = 1.0f;
    pp->m_partsys_color.m[2][3] = 0.5f;       /*|*/   pp->m_partsys_color.m[3][3] = 0.0f;
    /*------------------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------------------*/

    /*------------------------------------------------------------------------------------*/
    /*--------Definition of the 4 bounding boxes------------------------------------------*/
    /*--------Radii in X, Y and Z direction. W not used ----------------------------------*/
    /*------------------------------------------------------------------------------------*/
    /*-----------Control Point 1---------------------Control Point 2----------------------*/
    pp->m_partsys_radius.m[0][0] = 0.0f;      /*|*/   pp->m_partsys_radius.m[1][0] = 0.2f;
    pp->m_partsys_radius.m[0][1] = 0.0f;      /*|*/   pp->m_partsys_radius.m[1][1] = 2.0f;
    pp->m_partsys_radius.m[0][2] = 0.0f;      /*|*/   pp->m_partsys_radius.m[1][2] = 0.2f;
    pp->m_partsys_radius.m[0][3] = 1.0f;      /*|*/   pp->m_partsys_radius.m[1][3] = 1.0f;
    /*-----------Control Point 3---------------------Control Point 4----------------------*/
    pp->m_partsys_radius.m[2][0] = 1.0f;      /*|*/   pp->m_partsys_radius.m[3][0] = 5.0f;
    pp->m_partsys_radius.m[2][1] = 10.0f;     /*|*/   pp->m_partsys_radius.m[3][1] = 0.0f;
    pp->m_partsys_radius.m[2][2] = 1.0f;      /*|*/   pp->m_partsys_radius.m[3][2] = 5.0f;
    pp->m_partsys_radius.m[2][3] = 1.0f;      /*|*/   pp->m_partsys_radius.m[3][3] = 1.0f;
    /*------------------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------------------*/

    /*------------------------------------------------------------------------------------*/
    /*--------Definition of the particle aspect-------------------------------------------*/
    /*--------Particle size, texture coordinates !!!??TBD   ------------------------------*/
    /*------------------------------------------------------------------------------------*/
    /*Rotation disabled: Size, unused,unused,unused/alpha                                 */
    /*Rotation Enabled:  Size,  Angle, Radius, unused/alpha                                */
    /*-----------Control Point 1---------------------Control Point 2----------------------*/
    pp->m_partsys_aspect.m[0][0] = 0.0f;      /*|*/   pp->m_partsys_aspect.m[1][0] = 4.0f;
    pp->m_partsys_aspect.m[0][1] = 0.0f;      /*|*/   pp->m_partsys_aspect.m[1][1] = 0.0f;
    pp->m_partsys_aspect.m[0][2] = 0.9f;      /*|*/   pp->m_partsys_aspect.m[1][2] = 0.9f;
    pp->m_partsys_aspect.m[0][3] = 1.0f;      /*|*/   pp->m_partsys_aspect.m[1][3] = 1.0f;
    /*-----------Control Point 3---------------------Control Point 4----------------------*/
    pp->m_partsys_aspect.m[2][0] = 6.0f;      /*|*/   pp->m_partsys_aspect.m[3][0] = 30.0f;
    pp->m_partsys_aspect.m[2][1] = 0.0f;      /*|*/   pp->m_partsys_aspect.m[3][1] = 0.0f;
    pp->m_partsys_aspect.m[2][2] = 0.9f;      /*|*/   pp->m_partsys_aspect.m[3][2] = 0.9f;
    pp->m_partsys_aspect.m[2][3] = 0.0f;      /*|*/   pp->m_partsys_aspect.m[3][3] = 0.0f;
    /*------------------------------------------------------------------------------------*/
    /*------------------------------------------------------------------------------------*/
    pp->m_size_min_max[0] = 1.0f;
    pp->m_size_min_max[1] = static_cast<GLfloat>(NN_GX_DISPLAY0_HEIGHT) / 2.0f;

    pp->m_random_seed[0] = 45.0f;
    pp->m_random_seed[1] = 2.564f + pp->m_random_seed[0];
    pp->m_random_seed[2] = 1.15f + pp->m_random_seed[0];
    pp->m_random_seed[3] = 0.9875f + pp->m_random_seed[0];

    /* Core value for random
     Algorithm is based on pseudo-random number generators: X(n+1) = (a*Xn+b) mod m
     Expected values for uniform are
     [0] = a, [1] = b, [2] = m, [3] = 1/m
    */
    pp->m_prng[0] = 17.0f;                  /* a */
    pp->m_prng[1] = 37.0f;                  /* b */
    pp->m_prng[2] = 65535.0f;               /* m */
    pp->m_prng[3] = 1.0f / (pp->m_prng[2]); /* 1/m */

    pp->m_speed = 1.0f;             /*  emitter->particle_speed  */
    pp->m_particleCountMax = 60.0f; /*  particle number          */

    /* Application only */
    pp->simulationTime = 2.0f;
    pp->NFrame = 512;
    pp->dTime = pp->simulationTime / pp->NFrame;
}

static void SceneObjectInitialize(void)
{
    /* Load object (non-gaseous object) */
    loadDAT_GD( "rom:/resources/planeXY3.dat", &plane);

    /* Load Gaseous object */
    DefaultGasObject(&gas, gasColorTable[0]);

    /* Define the gaseous object geometry as a particle system */
    LoadParticleSystem1(&gas.pSys);
}

static void UnloadObjects(void)
{
    unloadDAT_GD(&plane);
    return;
}


/*=======================================================*/
/* Initialization of shaders                            */
/*=======================================================*/
int ShaderInitialize(void)
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);
    file.Finalize();
    s_AppHeap.Free(buf);

    /* for Drawing pass */
    nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &s_VSGeoId);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSGeoId, NULL, &SPGeoID);

    nn::gd::InputElementDescription descs[] = { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0};
    u32 strides[] = {3 * sizeof(float)} ;
    nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, s_VSGeoId, &ILGeoID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    /* for gas accumulation pass */
    nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &s_VSAccId);
    nn::gd::ShaderStage::CreateShader(shaderBinary, 3, &s_GS);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSAccId, s_GS, &SPAccId);

    nn::gd::InputElementDescription descs1[] = { 0, "attrCtrPointIndex", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 1, 0};
    u32 strides1[] = {1 * sizeof(float)} ;
    res = nn::gd::VertexInputStage::CreateInputLayout(descs1, 1, strides1, s_VSAccId, &ILAccId);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    /* for gas shading/pos synthesis pass */
    nn::gd::ShaderStage::CreateShader(shaderBinary, 2, &s_VSPostId);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSPostId, NULL, &SPPostId);

    nn::gd::InputElementDescription descs2[] =
    {{ 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0},
    { 1, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, 0},
    { 2, "aColor", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0}};
    u32 strides2[3] = {4 * sizeof(float), 2 * sizeof(float), 4 * sizeof(float)} ;
    res = nn::gd::VertexInputStage::CreateInputLayout(descs2, 3, strides2, s_VSPostId, &ILPostID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    // Texture combiner
    nn::gd::CombinerDescription combiner;

    /* for Drawing pass */
    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texDrawObject); // create dmp_TexEnv[0] combiner object

    /* for gas accumulation pass */
    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR); //default value
//  combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA); //default value
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texGasAcc); // create dmp_TexEnv[0] combiner object

    /* for gas shading/pos synthesis pass */
    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_ADD);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_ADD);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR); //default value
//      combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA); //default value
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT5, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_TEXTURE0);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT5, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_TEXTURE0);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT5, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texGasShading); // create dmp_TexEnv[0] combiner object
    return 0;
}

static void Finalize()
{
    //Release GD objects
    GasFinalize();

    if (gas.quad_index_ID)   nn::gd::Resource::ReleaseVertexBufferResource(gas.quad_index_ID);
    if (gas.quad_vertBuf_ID) nn::gd::Resource::ReleaseVertexBufferResource(gas.quad_vertBuf_ID);
    if (gas.quad_texBuf_ID)  nn::gd::Resource::ReleaseVertexBufferResource(gas.quad_texBuf_ID);
    if (gas.quad_colBuf_ID)  nn::gd::Resource::ReleaseVertexBufferResource(gas.quad_colBuf_ID);
    if (gas.samplerPattern)  nn::gd::TextureStage::ReleaseSamplerState(gas.samplerPattern);

    for (int i = 0; i < PARTICLE_PATTERNS; i++)
    {
        if (gas.patternResource[i]) nn::gd::Resource::ReleaseTexture2DResource(gas.patternResource[i]);
        if (gas.pattern[i])         nn::gd::TextureStage::ReleaseTexture2D(gas.pattern[i]);
    }
    if (gas.pSys.m_pSysPositionID) nn::gd::Resource::ReleaseVertexBufferResource(gas.pSys.m_pSysPositionID);

    if (texDrawObject) nn::gd::CombinerStage::ReleaseTextureCombinerState(texDrawObject);
    if (texGasAcc)     nn::gd::CombinerStage::ReleaseTextureCombinerState(texGasAcc);
    if (texGasShading) nn::gd::CombinerStage::ReleaseTextureCombinerState(texGasShading);  

    if (ILGeoID)  nn::gd::VertexInputStage::ReleaseInputLayout(ILGeoID);
    if (ILAccId)  nn::gd::VertexInputStage::ReleaseInputLayout(ILAccId);
    if (ILPostID) nn::gd::VertexInputStage::ReleaseInputLayout(ILPostID);

    if (SPGeoID)      nn::gd::ShaderStage::ReleaseShaderPipeline(SPGeoID);
    if (SPPostId)     nn::gd::ShaderStage::ReleaseShaderPipeline(SPPostId);
    if (SPAccId)      nn::gd::ShaderStage::ReleaseShaderPipeline(SPAccId);
    if (s_VSGeoId)    nn::gd::ShaderStage::ReleaseShader(s_VSGeoId);
    if (s_VSAccId)    nn::gd::ShaderStage::ReleaseShader(s_VSAccId);
    if (s_GS)         nn::gd::ShaderStage::ReleaseShader(s_GS);
    if (s_VSPostId)   nn::gd::ShaderStage::ReleaseShader(s_VSPostId);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget)                         nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer)        nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget)                   nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    UnloadObjects();

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}
/*=======================================================*/
/* draw object                                           */
/*=======================================================*/
static void DrawObject(void)
{
    /* This function draws only one quadrangle. */
    nn::gd::ShaderStage::SetShaderPipeline(SPGeoID);

    //* Set fragment operation mode to GL_FRAGOP_MODE_GL_DMP */
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);
    /*
     * Setup texture and blending unit
     * The color of the object is defined as a constant color from blending unit 0
    */
    nn::gd::CombinerStage::SetTextureCombinerState(texDrawObject);
    u8 almostblack[4] = {nn::gd::Utils::FloatToUnsignedByte(0.15f), nn::gd::Utils::FloatToUnsignedByte(0.15f), nn::gd::Utils::FloatToUnsignedByte(0.15f), nn::gd::Utils::FloatToUnsignedByte(0.0f)};
    nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT0, almostblack[0], almostblack[1], almostblack[2], almostblack[3]);

    /*
     * Miscellaneous settings
    */
    static nn::gd::DepthStencilState* depthStateDrawObject = 0;
    if (!depthStateDrawObject)
    {
        nn::gd::DepthStencilStateDescription desc;
        desc.ToDefault();
        desc.m_DepthEnable = GD_TRUE;
        desc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ALL;
        desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LEQUAL;
        nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStateDrawObject);
    }
    u8 stencilRef = 0;
    nn::gd::OutputStage::SetDepthStencilState(depthStateDrawObject, stencilRef);
    nn::gd::OutputStage::SetBlendState(NULL);

    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    for (int i = 0; i < plane.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(ILGeoID);

        nn::gd::VertexBufferResource* buffers[] = { plane.posVB };
        u32 offsets[] = { 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
        for (unsigned j = plane.obj[i].patch_offset; j < plane.obj[i].patch_size + plane.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(plane.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
            nn::gd::System::DrawIndexed(plane.patch[j].elm_size, 0);
        }
    }

    /*
     * Finalization
    */
}

/*=======================================================*/
/* draw frame                                            */
/*=======================================================*/
static int DrawFrame(void)
{
    static int f = 0;
    static int pattern = 0;
    static int AtFirst = 0;


    {
        if (!AtFirst)   // Set Fog LUT table
        {
            nn::gd::FogStage::UploadFogLookUpTableFloat(0, &gas.fogTable[0], &gas.fogTable[128], 128);
            AtFirst = 1;
        }
        if (f == 0)
        {
            /* set next gas table */
            SetGasColorTable(&gas,gasColorTable[pattern % NUM_COLOR_SAMPLE]);
            gas.pSys.simulationTime = 0.f;
            pattern++;
        }
    }


    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;

    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);
    }

    /* Clear display buffer */
    nn::gd::ShaderStage::SetShaderPipeline(SPGeoID);

#define GD_UTL_CLAMP(x)     (((x) <= 0.f) ? 0.f : ((x) >= 1.f) ? 1.f : (x))
    u8 color[] = {GD_UTL_CLAMP(0.2f) * 0xff, GD_UTL_CLAMP(0.4f) * 0xff, GD_UTL_CLAMP(0.7f) * 0xff, GD_UTL_CLAMP(1.f) * 0xff};

    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, color, 1.f, 0);

    nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    nn::math::Matrix44 proj, m;
    nn::math::MTX44Perspective(&proj, 35.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.f);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.f, 0.f, 80.0f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    {
        /* Draw the Object */
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPGeoID, nn::gd::ShaderStage::GetShaderUniformLocation(SPGeoID, "uProjection"), static_cast<f32*>(proj));

        nn::math::Matrix34 mv_rot, tmp;
        nn::math::MTX34RotXYZDeg(&tmp, 0.f, 45.0f, 0.f);
        nn::math::MTX34Mult(&mv_rot, &mv, &tmp);

        nn::math::Vector3 scale(0.5f, 0.5f, 0.5f);
        nn::math::MTX34MultScale(&mv_rot, &mv_rot, &scale);

        m = nn::math::Matrix44(mv_rot);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPGeoID, nn::gd::ShaderStage::GetShaderUniformLocation(SPGeoID, "uModelView"), static_cast<f32*>(m));
        DrawObject();
    }

    {
        /* Gas accumulation pass */
        nn::gd::ShaderStage::SetShaderPipeline(SPAccId);

        nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "uProjection"), static_cast<f32*>(proj));

        m = nn::math::Matrix44(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPAccId, nn::gd::ShaderStage::GetShaderUniformLocation(SPAccId, "uModelView"), static_cast<f32*>(m));
        GasAccumulation();
    }

    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);
    nn::gd::RasterizerStage::SetViewport(viewPort);


    /* Gas shading pass */
    GasShading();

    gas.pSys.simulationTime += gas.pSys.dTime;


    //Transfer framebuffer to display buffer
    int dstAddr;
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
    nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
        s_texture2DResource_MainColorBuffer,                    //source
        0,                                                      //sourceMipLevelIndex
        0,                                                      //srcOffsetY
        (u8*)dstAddr,                                           //dstAddr
        disp0_Width,                                            //dstWidth
        disp0_Height,                                           //dstHeight
        nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
        nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
        GD_FALSE                                                //yFlip
    );

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }
     /* return 0 when 16 frames done */

    if (++f == gas.pSys.NFrame)
        f = 0;  /* go to next gas pattern */
        /* return 0 when 16 frames done */

    return !glGetError();
}

/* ======================================================= */
/* main function                                           */
/* ======================================================= */
void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* Initialize the framebuffer and gas buffer*/
    Initialize();

    /* Initialization of shaders */
    if (ShaderInitialize() >= 0)
    {
        /*Initialize standard and gaseous objects */
        SceneObjectInitialize();

        /* Enter loop */
        while (1)
        {
            DrawFrame();
        }
    }

 
    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void loadDAT_GD(const char* _name, datGD_t* _dat)
{
    if (!_name || !_dat)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)malloc(strlen(_name) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, _name);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(malloc( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(malloc(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(_dat, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(malloc(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            free(attribdata);
            attribfile.Finalize();
        }

        _dat->obj = (dat_obj_t*)malloc(sizeof(dat_obj_t) * header->obj_num);
        _dat->patch = (dat_patch_t*)malloc(sizeof(dat_patch_t) * header->patch_num);
        if (!_dat->obj || !_dat->patch)
        {
            if (_dat->obj)
                free(_dat->obj);
            if (_dat->patch)
                free(_dat->patch);
            break;
        }
        memcpy(_dat->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(_dat->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        _dat->obj_num = header->obj_num;
    } while (0);

    if (fname)
        free(fname);
    if (wfname)
        free(wfname);
    if (dat)
        free(dat);
    file.Finalize();

    return;
}

void unloadDAT_GD(datGD_t* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            free(dat->obj);
        if (dat->patch)
            free(dat->patch);
    }

    return;
}

