﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     PartsysSimple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"
#include "demo.h"

#include <string.h>
#include <math.h>
#include "Memory.h"

#include <nn/gd.h>

#define WAIT_VSYNC

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;
static nn::gd::Shader* geometryShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferColor = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;

static nn::gd::UniformLocation uniformLoc_projection;
static nn::gd::UniformLocation uniformLoc_modelview;

//2D textures
//static nn::gd::Texture2DResource* s_texture2D_res = 0;
//static nn::gd::Texture2D* s_texture2D_BMP = 0;

//Depth, blend states
static nn::gd::DepthStencilState* depthStateDrawObject = 0;
static nn::gd::BlendState* blendState_SrcAlpha = 0;
static nn::gd::BlendState* blendState_One = 0;

/* Texture Combiner State object */
static nn::gd::CombinerState *texProcTex = 0;
static nn::gd::CombinerState *texTX0 = 0;


/* Structure containing particle system settings*/
struct _particleSys
{
    nn::math::Matrix44      m_partsys_center;
    nn::math::Matrix44      m_partsys_color;
    nn::math::Matrix44      m_partsys_radius;
    nn::math::Matrix44      m_partsys_aspect;

    float       m_speed;
    float       m_particleCountMax;
    float       m_size_min_max[2];
    float       m_prng[4];                      /* constant for the Pseudo Random Number Generator (modulo based) */
    float       m_random_seed[4];               /* seeds for the PRNG */

    float       simulationTime;                 /* application side. Holds the current time for particle system */
    float       dTime;                          /* application side. Holds the time interval between each frame */
    int         NFrame;                         /* application side. Number of frames for this particle system */
};

#include "PartsysPattern.h"

struct _particleSys pSys[PARTSYS_LAST];
int CurSys = 0;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

GLuint s_CmdlistID;

/* Procedural texture */
static int     s_ProcTexSize;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

/* Load a 2D texture (DMP Letters)*/
void LoadTexture2D(void)
{
    /*
    int _lev = 0;
    glGenTextures(1, &s_Texture2DName);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, s_Texture2DName);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, 0);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    bool bUseAlpha;
    loadTexture( "rom:/resources/dmp.tga", GL_TEXTURE_2D, _lev, bUseAlpha);
    */
    /*
    //TODO load TGA
    //Texture 2D
    wchar_t* bmpRomFilename = L"rom:/resources/dmp.tga";

    u32 width = 0; u32 height = 0; u32 textureWidth = 0; u32 textureHeight = 0;
    //Get the texture data in native format
    u8* textureDataPtr = GetTextureData(bmpRomFilename, width, height, textureWidth, textureHeight, false, &s_AppHeap);
    
    nn::gd::Texture2DResourceDescription Text2DResDesc = {textureWidth, textureHeight, 1, nn::gd::Resource::NATIVE_FORMAT_RGB_888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, &s_texture2D_res);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::Texture2DDescription tex2DDesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2D_res, &tex2DDesc, &s_texture2D_BMP);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    
    //Texture unit 0
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_BMP);
    */
}

/* Procedural texture: RGBA texture definition
 a simple white gradient defined as 2 colors */
int Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA)
{
    memset(RR, 0, 512 * sizeof(GLfloat));
    memset(GG, 0, 512 * sizeof(GLfloat));
    memset(BB, 0, 512 * sizeof(GLfloat));
    memset(AA, 0, 512 * sizeof(GLfloat));

    RR[0] = 0xFF / 255.0f;  GG[0] = 0xFF / 255.0f;  BB[0] = 0xFF / 255.0f;  AA[0] = 0xFF / 255.0f;
    RR[1] = 0x00 / 255.0f;  GG[1] = 0x00 / 255.0f;  BB[1] = 0x00 / 255.0f;  AA[1] = 0x00 / 255.0f;

    for (int x = 0; x < 255; x++)
    {
        RR[256 + x] = RR[x + 1] - RR[x];
        GG[256 + x] = GG[x + 1] - GG[x];
        BB[256 + x] = BB[x + 1] - BB[x];
        AA[256 + x] = AA[x + 1] - AA[x];
    }
    RR[255] = 0.0f;
    GG[255] = 0.0f;
    BB[255] = 0.0f;
    AA[255] = 0.0f;
    
    return 2;
}

/* Procedural texture: mapping function
 bias function in [0.0.5]. zero otherwise */
void LoadBias(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float aa = 1.0f / p;
    float t;
    for (unsigned int i = 0; i < 128; i++)
    {
        t = x;
        if (t >= 1.0f)
            map_1D2D_LUT[i] = 1.0f;
        else
            map_1D2D_LUT[i] = t / ((aa - 2.0f) * (1.0f - t) + 1.0f);
        x += dx;
    }
    for (unsigned int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

/* Procedural texture: mapping function
 potential function in [0.0.5]. zero otherwise */
void LoadPotential(float *map_1D2D_LUT, float p)
{
    float x = 0.0f;
    float dx = 2.0f / 127.0f;
    float f = 0, d2;
    
    for (unsigned int i = 0; i < 128; i++)
    {
        f = 0;
        d2 = x;
        if (d2 < 1.0f)
        {
            if (d2 < 0.25f)
                f = 1.0f - ((3.0f * d2) * (3.0f * d2)) / (p + (4.5f - 4.0f * p) * d2);
            else 
                f = ((1.0f - d2) * (1.0f - d2)) / (0.75f - p + (1.5f + 4.0f * p) * d2);
        }
        
        map_1D2D_LUT[i] = f;
        x += dx;
    }
    
    for (unsigned int i = 0; i < 126; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    map_1D2D_LUT[255] = 0.0f;
}

void LoadProceduralTexture()
{   
    /* Declaration */
    float F_FunctionRGB[256], F_FunctionZero[256];
    LoadBias(F_FunctionRGB, 0.8f);
    memset(F_FunctionZero, 0, 256 * sizeof(float));

    nn::gd::ProceduralTextureStage::UploadLookUpTableRgbMapFloat(0, F_FunctionRGB, F_FunctionRGB + 128, 128);

    const int NN = 512;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    s_ProcTexSize = Set1DTex(RR,GG,BB,AA);

    f32 *colorLUT[4];
    colorLUT[0] = RR;
    colorLUT[1] = GG;
    colorLUT[2] = BB;
    colorLUT[3] = AA;

    f32 *colorDeltaLUT[4];
    colorDeltaLUT[0] = RR+256;
    colorDeltaLUT[1] = GG+256;
    colorDeltaLUT[2] = BB+256;
    colorDeltaLUT[3] = AA+256;

    nn::gd::ProceduralTextureStage::UploadLookUpTableColorMapFloat(0, (f32 **)&colorLUT, (f32 **)&colorDeltaLUT, 256);
}

void ValidateParticleSystem(struct _particleSys *pp, unsigned int _width, unsigned int _height)
{
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uCenter"), static_cast<f32*>(pp->m_partsys_center));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uRadii"), static_cast<f32*>(pp->m_partsys_radius));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.color"), static_cast<f32*>(pp->m_partsys_color));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.aspect"), static_cast<f32*>(pp->m_partsys_aspect));

    f32 viewport_size[2] = {0.0f, 0.0f};
    viewport_size[0] = 1.f / (f32)_width;
    viewport_size[1] = 1.f / (f32)_height;
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.viewport"), &viewport_size[0]);

    f32 distance_attenuation[3] = {0.0f, 0.0f, 0.0f};
    distance_attenuation[0] = 0.0f;
    distance_attenuation[1] = 0.0f;
    distance_attenuation[2] = 1.0f / (float)(_width * _height);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.distanceAttenuation"), &distance_attenuation[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.countMax"), &pp->m_particleCountMax);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.pointSize"), &pp->m_size_min_max[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.speed"), &pp->m_speed);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.randomCore"), &pp->m_prng[0]);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.randSeed"), &pp->m_random_seed[0]);
}

void ValidateProceduralTexture()
{
    nn::gd::ProceduralTextureStage::SetMinFilter(nn::gd::ProceduralTextureStage::MIN_FILTER_LINEAR);
    nn::gd::ProceduralTextureStage::SetTexWidth(s_ProcTexSize);
    nn::gd::ProceduralTextureStage::SetTexOffset(0);
    nn::gd::ProceduralTextureStage::SetAlphaSeparate(GD_FALSE);
    nn::gd::ProceduralTextureStage::SetClampUV(nn::gd::ProceduralTextureStage::CLAMP_TO_EDGE, nn::gd::ProceduralTextureStage::CLAMP_TO_EDGE);
    nn::gd::ProceduralTextureStage::SetShiftUV(nn::gd::ProceduralTextureStage::SHIFT_NONE, nn::gd::ProceduralTextureStage::SHIFT_NONE);
    nn::gd::ProceduralTextureStage::SetRgbMap(nn::gd::ProceduralTextureStage::UV_MAP_MAX);
    nn::gd::ProceduralTextureStage::SetAlphaMap(nn::gd::ProceduralTextureStage::UV_MAP_MAX);
}

void ValidateTextureBlenderProcTex()
{
    nn::gd::TextureStage::SetTextureCoordinateSourceForUnit3Procedural(nn::gd::TextureStage::UNIT3_PROCEDURAL_TEXTURE_COORDINATE_FROM_UNIT_2);
    nn::gd::CombinerStage::SetTextureCombinerState(texProcTex);
}

void ValidateTextureBlenderTX0()
{
    nn::gd::CombinerStage::SetTextureCombinerState(texTX0);
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);                        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &geometryShader);                      if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, geometryShader, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    uniformLoc_projection = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    uniformLoc_modelview = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void){

    /* make vertex buffer */
    f32 vertex[4] = {0.0f, 1.0f, 2.0f, 3.0f};
    u16 index[4] = {0, 1, 2, 3};

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "attrCtrPointIndex", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 1, 0 },
    };
    u32 strides[] = { sizeof(float) };
    nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, vertexShader, &inputLayout);

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize =  sizeof(vertex);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    nn::gd::Resource::CreateVertexBufferResource(&desc, vertex, &bufferCoord);

    desc.m_ByteSize = sizeof(index);
    nn::gd::Resource::CreateVertexBufferResource(&desc, index, &bufferIndex);

    u32 offset = 0;
    nn::gd::VertexInputStage::SetVertexBuffers(0, 1, &bufferCoord, &offset);
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_GEOMETRY);

    nn::gd::VertexInputStage::SetInputLayout(inputLayout);

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

}

void Initialize_GD()
{
    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

    if (!depthStateDrawObject)
    {
        nn::gd::DepthStencilStateDescription desc;
        desc.ToDefault();
        desc.m_DepthEnable = GD_FALSE;
        desc.m_DepthMask = nn::gd::OutputStage::DEPTH_WRITE_MASK_ALL;
        desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LEQUAL;
        nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStateDrawObject);
    }
            
    if (!blendState_SrcAlpha)
    {
        nn::gd::BlendStateDescription desc;
        desc.ToDefault();
        desc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
        desc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE_MINUS_SRC_ALPHA, nn::gd::OutputStage::BLEND_EQUATION_ADD);
        nn::gd::OutputStage::CreateBlendState(&desc, &blendState_SrcAlpha);
    }

    if (!blendState_One)
    {
        nn::gd::BlendStateDescription desc;
        desc.ToDefault();
        desc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
        desc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE, nn::gd::OutputStage::BLEND_EQUATION_ADD);
        nn::gd::OutputStage::CreateBlendState(&desc, &blendState_One);
    }
    
    nn::gd::OutputStage::SetDepthStencilState(depthStateDrawObject, 0);

    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);
}

void DrawParticleSystem(struct _particleSys *pp, float _Psys_time)
{
    NN_UNUSED_VAR(pp);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "dmp_PartSys.time"), &_Psys_time);
    nn::gd::System::DrawIndexed(4, 0);
}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    s64 startTime = GetTimeMicroSeconds();
    
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);

    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(0.0f, 0.0f, 50.0f);
    nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
    nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_projection, static_cast<f32*>(proj));

    }

    /********************************************************/
    /********************** Display 0 ***********************/

    //Clear the render buffers
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);


    // *********************************************************** //

    static int frame = 0;
//    static bool switchTexture = false;
    static bool useProctex = true;
    static float t = 0.f;
    static int firstTime = 0;

    
    if (CurSys == PARTSYS_SPIN) {
        nn::gd::OutputStage::SetBlendState(blendState_SrcAlpha);
    } else {
        nn::gd::OutputStage::SetBlendState(blendState_One);
    }
    
    if (firstTime == 0 ) {
          LoadProceduralTexture();  
          firstTime = 1;
    }

    if (!useProctex)
    {
        ValidateTextureBlenderTX0();
    } else {
        ValidateProceduralTexture();
        ValidateTextureBlenderProcTex();
    }

    if (frame == 0)
    {
        /* reset uniforms */
        ValidateParticleSystem(&pSys[CurSys], nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);

        //Modelview
        nn::math::Matrix44 m(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
    }
    
    if (CurSys == PARTSYS_EXPLOSION2)
    {
        // As an illustration, we draw 2 times the particle system. 

        nn::math::Matrix44 m = nn::math::Matrix44(mv);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
        
        DrawParticleSystem(&pSys[CurSys], t);
        
        nn::math::Matrix34 mv_rot;
        nn::math::MTX34RotXYZDeg(&mv_rot, 0.0f, 0.0f, 45.0f);
        nn::math::MTX34Mult(&mv_rot, &mv, &mv_rot);
        m = nn::math::Matrix44(mv_rot);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, uniformLoc_modelview, static_cast<f32*>(m));
        DrawParticleSystem(&pSys[CurSys], t);
    }
    else
        DrawParticleSystem(&pSys[CurSys], t);

 
    if (++frame == pSys[CurSys].NFrame)
    {
        /* Changing 2D texture / procedural texture */
#if 0  // No switch
        if (switchTexture)
        {
            useProctex = (!useProctex) ? true : false;
            switchTexture = false;
        }
#endif
        /* change to next particle system */
        CurSys = ++CurSys % PARTSYS_LAST;
        frame = 0;
        t = 0.f;

        /* Changing blending mode for the spinning pattern */
        // switchTexture = (CurSys == PARTSYS_LAST - 1);
    }
    else
        t += pSys[CurSys].dTime;
    
    
    if (useProctex == true)
    {
        nn::gd::TextureStage::SetTextureCoordinateSourceForUnit3Procedural(nn::gd::TextureStage::UNIT3_PROCEDURAL_TEXTURE_COORDINATE_FROM_UNIT_0);
    }

    // *********************************************************** //

    //Transfer framebuffer to display buffer
    int dstAddr;
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
    nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
        s_texture2DResource_MainColorBuffer,                    //source
        0,                                                      //sourceMipLevelIndex
        0,                                                      //srcOffsetY
        (u8*)dstAddr,                                           //dstAddr
        disp0_Width,                                            //dstWidth
        disp0_Height,                                           //dstHeight
        nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
        nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
        GD_FALSE                                                //yFlip
    );

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_texComState)        nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);
    if (texProcTex)           nn::gd::CombinerStage::ReleaseTextureCombinerState(texProcTex);
    if (texTX0)               nn::gd::CombinerStage::ReleaseTextureCombinerState(texTX0);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord)          nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferColor)          nn::gd::Resource::ReleaseVertexBufferResource(bufferColor);
    if (bufferIndex)          nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (inputLayout)          nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline)       nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader)         nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (geometryShader)       nn::gd::ShaderStage::ReleaseShader(geometryShader);
    if (shaderBinary)         nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget)   nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    if (depthStateDrawObject) nn::gd::OutputStage::ReleaseDepthStencilState(depthStateDrawObject);
    if (blendState_SrcAlpha)  nn::gd::OutputStage::ReleaseBlendState(blendState_SrcAlpha);
    if (blendState_One)       nn::gd::OutputStage::ReleaseBlendState(blendState_One);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    //Texture combiners
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texProcTex); 

    combiner.ToDefault();
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_MODULATE);
    combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_MODULATE);
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_PRIMARY_COLOR);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texTX0);

    InitShader();

    LoadProceduralTexture();
    //LoadTexture2D();
    LoadParticlePattern(&pSys[PARTSYS_DEFAULT], PARTSYS_DEFAULT);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION1], PARTSYS_EXPLOSION1);
    LoadParticlePattern(&pSys[PARTSYS_EXPLOSION2], PARTSYS_EXPLOSION2);
    LoadParticlePattern(&pSys[PARTSYS_SNOW], PARTSYS_SNOW);
    LoadParticlePattern(&pSys[PARTSYS_SPIN], PARTSYS_SPIN);

    
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

