﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProctexFFunction.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "TexUtils.h"

#include <string.h>

#include <nn/gd.h>

#include "demo.h"

#define WAIT_VSYNC

#define DMP_PI  (3.1415926f)
#define REV_PI  (1.0f/DMP_PI)

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;
u8 s_clearColor[4];

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

//Shader
static nn::gd::ShaderBinary* shaderBinary = 0;
static nn::gd::ShaderPipeline* shaderPipeline = 0;
static nn::gd::Shader* vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* inputLayout = 0;
static nn::gd::VertexBufferResource* bufferCoord = 0;
static nn::gd::VertexBufferResource* bufferTex = 0;
static nn::gd::VertexBufferResource* bufferIndex = 0;

static nn::gd::UniformLocation s_shaderVariable_proj;
static nn::gd::UniformLocation s_shaderVariable_view;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

GLuint s_CmdlistID;

#define PRINT(__msg) NN_LOG __msg

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShaderPipeline(vertexShader, NULL, &shaderPipeline); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    s_shaderVariable_proj = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uProjection");
    s_shaderVariable_view = nn::gd::ShaderStage::GetShaderUniformLocation(shaderPipeline, "uModelView");
    NN_ASSERT(s_shaderVariable_proj.IsValid());
    NN_ASSERT(s_shaderVariable_view.IsValid());

    nn::gd::ShaderStage::SetShaderPipeline(shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

int Set1DTex(GLfloat* RR, GLfloat* GG, GLfloat* BB, GLfloat* AA, int N)
{
    memset(RR, 0, 512 * sizeof(GLfloat));
    memset(GG, 0, 512 * sizeof(GLfloat));
    memset(BB, 0, 512 * sizeof(GLfloat));
    memset(AA, 0, 512 * sizeof(GLfloat));

    RR[7*N]=0x00/255.0f;    GG[7*N]=0x00/255.0f;    BB[7*N]=0x00/255.0f;    AA[7*N]=0xFF/255.0f;
    RR[6*N]=0x00/255.0f;    GG[6*N]=0x00/255.0f;    BB[6*N]=0xFF/255.0f;    AA[6*N]=0xFF/255.0f;
    RR[5*N]=0xFF/255.0f;    GG[5*N]=0x00/255.0f;    BB[5*N]=0x00/255.0f;    AA[5*N]=0xFF/255.0f;
    RR[4*N]=0xFF/255.0f;    GG[4*N]=0x00/255.0f;    BB[4*N]=0xFF/255.0f;    AA[4*N]=0xFF/255.0f;
    RR[3*N]=0x00/255.0f;    GG[3*N]=0xFF/255.0f;    BB[3*N]=0x00/255.0f;    AA[3*N]=0xFF/255.0f;
    RR[2*N]=0x00/255.0f;    GG[2*N]=0xFF/255.0f;    BB[2*N]=0xFF/255.0f;    AA[2*N]=0xFF/255.0f;
    RR[1*N]=0xFF/255.0f;    GG[1*N]=0xFF/255.0f;    BB[1*N]=0x00/255.0f;    AA[1*N]=0xFF/255.0f;
    RR[0*N]=0xFF/255.0f;    GG[0*N]=0xFF/255.0f;    BB[0*N]=0xFF/255.0f;    AA[0*N]=0xFF/255.0f;

    for (int x = 0; x < 255; x++)
    {
        RR[256+x] = RR[x+1] - RR[x];
        GG[256+x] = GG[x+1] - GG[x];
        BB[256+x] = BB[x+1] - BB[x];
        AA[256+x] = AA[x+1] - AA[x];
    }
    RR[255] = 0.0f;
    GG[255] = 0.0f;
    BB[255] = 0.0f;
    AA[255] = 0.0f;

    return N * 8;
}

void LoadSin(float map_1D2D_LUT[256], GLfloat f)
{
    for (int i = 0; i < 128; i++)
        map_1D2D_LUT[i] = std::sin(std::fmod(f * DMP_PI * i/128.f, DMP_PI));

    for (int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];

    map_1D2D_LUT[255] = std::sin(std::fmod(f * DMP_PI, DMP_PI)) - map_1D2D_LUT[127];
}

void LoadPow(float map_1D2D_LUT[256], GLfloat v)
{
    for (int i = 0; i< 128; i++)
        map_1D2D_LUT[i] = std::pow(i / 128.f, v);

    for (int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];

    map_1D2D_LUT[255] = 1.f - map_1D2D_LUT[127];
}

void LoadHat(float map_1D2D_LUT[256], GLfloat h)
{
    for (int i = 0; i < 128; i++)
    {
        if (i < 64)
            map_1D2D_LUT[i] = h * (i / 64.f);
        else
            map_1D2D_LUT[i] = h * ((128 - i) / 64.f);
    }
    for (int i = 0; i < 127; i++)
    {
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];
    }
    map_1D2D_LUT[255] = 0.f - map_1D2D_LUT[127];
}

void LoadLamp(float map_1D2D_LUT[256], GLfloat a)
{
    if (a < 0)
    {
        for (int i = 0; i < 128; i++)
            map_1D2D_LUT[i] = -a * (1.f - i / 128.f);
    }
    else
    {
        for (int i = 0; i < 128; i++)
            map_1D2D_LUT[i] = a * (i / 128.f);
    }

    for (int i = 0; i < 127; i++)
        map_1D2D_LUT[i + 128] = map_1D2D_LUT[i + 1] - map_1D2D_LUT[i];

    map_1D2D_LUT[255] = ((a < 0) ? 0.f : a)  - map_1D2D_LUT[127];
}

void LoadFlat(float map_1D2D_LUT[256], float a)
{
    for (int i = 0; i < 128; i++)
    {
        map_1D2D_LUT[i] = a;
        map_1D2D_LUT[i + 128] = 0;
    }
}

void LoadProceduralTexture(float* F_FunctionRGB)
{
    /* Binding of the procedural texture as a texture collection */
 
    /* Declaration */
    float F_FunctionA[256], F_Noise[256];
    LoadFlat(F_FunctionA, 1.0f);
    memset(F_Noise, 0, 256 * sizeof(float));

    /* Texture collection */
    nn::gd::ProceduralTextureStage::UploadLookUpTableRgbMapFloat(0, F_FunctionRGB, F_FunctionRGB+128, 128);

    const int NN = 512;
    GLfloat RR[NN], GG[NN], BB[NN], AA[NN];
    int texSize = Set1DTex(RR, GG, BB, AA, 1);

    f32 *colorLUT[4];
    colorLUT[0] = RR;
    colorLUT[1] = GG;
    colorLUT[2] = BB;
    colorLUT[3] = AA;

    f32 *colorDeltaLUT[4];
    colorDeltaLUT[0] = RR+256;
    colorDeltaLUT[1] = GG+256;
    colorDeltaLUT[2] = BB+256;
    colorDeltaLUT[3] = AA+256;

    nn::gd::ProceduralTextureStage::UploadLookUpTableColorMapFloat(0, (f32 **)&colorLUT, (f32 **)&colorDeltaLUT, 256);

    nn::gd::ProceduralTextureStage::SetMinFilter(nn::gd::ProceduralTextureStage::MIN_FILTER_LINEAR);
    nn::gd::ProceduralTextureStage::SetTexWidth(texSize);
    nn::gd::ProceduralTextureStage::SetTexOffset(0);

    nn::gd::ProceduralTextureStage::SetRgbMap(nn::gd::ProceduralTextureStage::UV_MAP_ADDSQRT2);
    nn::gd::ProceduralTextureStage::SetAlphaMap(nn::gd::ProceduralTextureStage::UV_MAP_ADDSQRT2);
    nn::gd::ProceduralTextureStage::SetAlphaSeparate(GD_FALSE);
    nn::gd::ProceduralTextureStage::SetClampUV(nn::gd::ProceduralTextureStage::CLAMP_SYMMETRICAL_REPEAT, nn::gd::ProceduralTextureStage::CLAMP_TO_EDGE);
    nn::gd::ProceduralTextureStage::SetShiftUV(nn::gd::ProceduralTextureStage::SHIFT_NONE, nn::gd::ProceduralTextureStage::SHIFT_NONE);

    static nn::gd::CombinerState* s_texComState = 0;
    if (!s_texComState)
    {
        nn::gd::CombinerDescription combiner;
        combiner.ToDefault();
        combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
        combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
        combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
        combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
        combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
        nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    }
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

}


static void InitObjects(void)
{
    //////////////////////////////////////
    float coords[] = {
        -1.0f,-1.0f,0.0f,1.0f,
        -1.0f, 1.0f,0.0f,1.0f,
         1.0f, 1.0f,0.0f,1.0f,
         1.0f,-1.0f,0.0f,1.0f
    };
    f32 t1 = -1.0f, t2 = 1.0f;
    f32 texCoord[8] =
    {
        t1, t1,
        t1, t2,
        t2, t2,
        t2, t1,
    };
    u16 idxs[] = {0, 2, 1, 0, 3, 2};

    nnResult res;

    nn::gd::InputElementDescription descs[] =
    {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0},
        { 1, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, 0}
    };

    u32 strides[] = { sizeof(float) * 4, sizeof(float) * 2};
    res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, vertexShader, &inputLayout);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::VertexBufferResourceDescription desc;
    desc.m_ByteSize = sizeof(coords);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, coords, &bufferCoord);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(texCoord);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, texCoord, &bufferTex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = sizeof(idxs);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, idxs, &bufferIndex);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::VertexBufferResource* buffers[] = { bufferCoord, bufferTex };
    u32 offsets[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

    nn::gd::VertexInputStage::SetInputLayout(inputLayout);
        
    nn::gd::VertexInputStage::SetIndexBuffer(bufferIndex, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
}

void LoadTextureGD(int num, GLfloat t)
{
    GLfloat F_FunctionRGB[256];

    switch(num)
    {
        case 0:
        /* Setting Lower 3 Textures */
            LoadLamp(F_FunctionRGB, 0.6f + std::sin(DMP_PI * t) * 0.4f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 1:
            LoadLamp(F_FunctionRGB, -0.6f + std::sin(DMP_PI * t) * 0.4f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 2:
            LoadHat(F_FunctionRGB, 0.7f + std::sin(DMP_PI * t) * 0.2f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 3:
             /* Setting Middle 3 Textures */
            LoadPow(F_FunctionRGB, 0.7f + std::sin(DMP_PI * t) * 0.2f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 4:
            LoadPow(F_FunctionRGB, 1.5f + std::sin(DMP_PI * t) * 0.5f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 5:
            LoadPow(F_FunctionRGB, 2.0f + std::sin(DMP_PI * t) * 1.2f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 6:
            /* Setting Upper 3 Textures */
            LoadSin(F_FunctionRGB, 1.f + std::sin(DMP_PI * t) * 0.2f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 7:
            LoadSin(F_FunctionRGB, 2.f + std::sin(DMP_PI * t) * 0.5f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
        case 8:
            LoadSin(F_FunctionRGB, 3.f + std::sin(DMP_PI * t) * 1.5f);
            LoadProceduralTexture(F_FunctionRGB);
            break;
    }
}


void Initialize_GD()
{
    //common
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);

    //Clear color
    s_clearColor[0] = s_clearColor[1] = 25; 
    s_clearColor[2] = 122; 
    s_clearColor[3] = 255;

}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

int DrawFrame(void)
{
    s64 startTime = GetTimeMicroSeconds();
    static int f = 0;
    //NN_LOG("\nDrawFrame: %d\n", f);
    
    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    static nn::gd::TargetProperties mainRenderTargetProperty;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);

        //Set the viewport
        nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
        nn::gd::RasterizerStage::SetViewport(viewPort);

        //Projection matrix
        nn::math::Matrix44 proj;
        nn::math::MTX44Perspective(&proj, 45.0f, static_cast<f32>(nn::gx::DISPLAY0_WIDTH) / static_cast<f32>(nn::gx::DISPLAY0_HEIGHT), 1.0f, 200.0f);
        nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(proj));

    }


    /********************************************************/
    /********************** Display 0 ***********************/
    {
        //Clear the render buffers
        nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, s_clearColor, 1.f, 0);

        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        nn::math::Matrix34 mv;
        nn::math::Vector3 camPos(0.f, 0.f, 25.0f);
        nn::math::Vector3 camUp(0.0f, 1.0f, 0.0f);
        nn::math::Vector3 target(0.0f, 0.0f, 0.0f);
        nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);

        /* Preparing texture collection */
        /* One collection for each procedural texture */
        //GLfloat F_FunctionRGB[256];
        GLfloat t = f / 100.f;

        /* Drawing 9 Rectangles */
        for(int i = 0; i < 9; i++)
        {
            GLfloat x = ((GLfloat)(i % 3) - 1.0f) * 4.f;
            GLfloat y = ((GLfloat)(i / 3) - 1.0f) * 4.f;
            nn::math::Matrix34 mv_tr, tmp[2];
            nn::math::Vector3 trans(x, y, 0.0f);
            nn::math::Vector3 scale(1.7f, 1.7f, 1.7f);
            nn::math::MTX34Translate(&tmp[0], &trans);
            nn::math::MTX34Scale(&tmp[1], &scale);
            nn::math::MTX34Mult(&mv_tr, &mv, &tmp[0]);
            nn::math::MTX34Mult(&mv_tr, &mv_tr, &tmp[1]);
            nn::math::Matrix44 m(mv_tr);
            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(shaderPipeline, s_shaderVariable_view, static_cast<f32*>(m));
            LoadTextureGD(i, t);
            nn::gd::System::DrawIndexed(6, 0);

        }


        // ***********************************************************************

        
        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );
    }

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        NN_UNUSED_VAR(avgtime);
        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (bufferCoord) nn::gd::Resource::ReleaseVertexBufferResource(bufferCoord);
    if (bufferTex) nn::gd::Resource::ReleaseVertexBufferResource(bufferTex);
    if (bufferIndex) nn::gd::Resource::ReleaseVertexBufferResource(bufferIndex);
    if (inputLayout) nn::gd::VertexInputStage::ReleaseInputLayout(inputLayout);

    if (shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(shaderPipeline);
    if (vertexShader) nn::gd::ShaderStage::ReleaseShader(vertexShader);
    if (shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    ////InitLUTProceduralTextureGL();


    //float black[4]={0.3f, 0.2f, 0.2f, 1.0f};
    ////glUniform4fv(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].constRgba"), 1, black);

    static nn::gd::CombinerState* s_texComState = 0;
    if (!s_texComState)
    {
        nn::gd::CombinerDescription combiner;
        combiner.ToDefault();
        combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_TEXTURE3, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
        combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
        combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
        combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
        combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
        nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
    }
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

/*
    static nn::gd::BlendState* blendState = 0;
    if (!blendState){
        nn::gd::BlendStateDescription desc;
        desc.ToDefault();
        desc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_LOGICOP;
        desc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE, nn::gd::OutputStage::BLEND_EQUATION_ADD);
        nn::gd::OutputStage::CreateBlendState(&desc, &blendState);
    }
    nn::gd::OutputStage::SetBlendState(blendState);
*/

    InitShader();
    InitObjects();
    Initialize_GD();

    return 0;
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

