﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ProfilingSample.cpp

  Copyright (C)2009-2011 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/fnd/fnd_ExpHeap.h>
#include <nn/ulcd.h>
#include <nn/applet.h>

#include "TexUtils.h"
#include "demo.h"

#include <nn/gd.h>

#include "PrimitiveUtils.h"

//#define WAIT_VSYNC

#define CMD_LIST_MAX_REQUEST_COUNT 128

//If define, call the profiling functions
#define USE_PROFILING_FUNCTIONS
//If define, will print the profiling result (if too many logs are printed it may crash the application, so we display the result only every X frames)
#define PRINT_PROFILING_RESULT

//The demo will use 2 textures and alternate between them
//#define DEMO_ALTERNATE_TEXTURES

#define SINGLE_COMMAND_LIST_PARALLEL 0
#define SINGLE_COMMAND_LIST_SERIE 1
#define DOUBLE_COMMAND_LIST 2

#define USE_COMMAND_LIST DOUBLE_COMMAND_LIST

#define USE_PACKET_RECORDING_NONE 0
#define USE_PACKET_RECORDING_COPY 1
#define USE_PACKET_RECORDING_NOCOPY 2

// SHADER_USE_UNMANAGED_UNIFORM must be enabled when USE_PACKET_RECORDING is enabled
#define PACKET_RECORDING_MODE  USE_PACKET_RECORDING_NONE

#define SHADER_USE_UNMANAGED_UNIFORM
#define DISPLAY_TIMING_PER_SCREEN

#define PRINT(__msg) NN_LOG __msg

static int displayBufferTop_width = nn::gx::DISPLAY0_WIDTH;
static int displayBufferTop_height = nn::gx::DISPLAY0_HEIGHT;
static int displayBufferBottom_width = nn::gx::DISPLAY1_WIDTH;
static int displayBufferBottom_height = nn::gx::DISPLAY1_HEIGHT;
nn::gd::Memory::MemoryLayout frameBuffer_blockSize = nn::gd::Memory::LAYOUT_BLOCK_8;

//Display buffers
GLuint s_displaybuffer_top[2];
GLuint s_displaybuffer_bottom[2];
GLuint s_displaybuffer_ext[2];
GLint s_CurrentDisplaybuffer_top = 0;
GLint s_CurrentDisplaybuffer_bottom = 0;
GLint s_CurrentDisplaybuffer_ext = 0;

//Depth stencil
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;

//Texture combiner
static nn::gd::CombinerState* s_texComState = 0;

//Textures
static nn::gd::Texture2DResource* s_texture2DResourceA = 0;
static nn::gd::Texture2D* s_texture2DA = 0;
static nn::gd::Texture2DResource* s_texture2DResourceB = 0;
static nn::gd::Texture2D* s_texture2DB = 0;

//Sampler states
static nn::gd::SamplerState* s_samplerDefault = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;
static nn::gd::TargetProperties s_mainRenderTargetProperty;

//Shader
static nn::gd::ShaderBinary* s_shaderBinary = 0;
static nn::gd::ShaderPipeline* s_shaderPipeline = 0;
static nn::gd::Shader* s_vertexShader = 0;

//Vertex input
static nn::gd::InputLayout* s_inputLayout_cube = 0;
static nn::gd::VertexBufferResource* s_buffers_cube[4] = {0, 0, 0, 0};
static nn::gd::VertexBufferResource* s_bufferPosition_cube = 0;
static nn::gd::VertexBufferResource* s_bufferTexCoord_cube = 0;
static nn::gd::VertexBufferResource* s_bufferColor_cube = 0;
static nn::gd::VertexBufferResource* s_bufferNormal_cube = 0;
static nn::gd::VertexBufferResource* s_bufferIndex_cube = 0;
static u32 s_sphereVertexCount = 0;
static u32 s_cubeVertexCount = 0;

static nn::gd::InputLayout* s_inputLayout_sphere = 0;
static nn::gd::VertexBufferResource* s_bufferInterleave_sphere = 0;
static nn::gd::VertexBufferResource* s_bufferIndex_sphere = 0;

//Shader variables
static nn::gd::UniformLocation s_shaderVariable_proj;
static nn::gd::UniformLocation s_shaderVariable_view;
static nn::gd::UniformLocation s_shaderVariable_user;

//Others
static nn::math::MTX34  mtx;
static nn::math::MTX44  projectionMtx;
static nn::math::MTX34  viewMtx;
static nn::ulcd::StereoCamera stereo;

static u32  m_Counter = 0;
static f32  m_Sin, m_Cos;
static f32  m_Sin2, m_Cos2;
static f32  m_Sin01, m_Cos01;
enum LightType
{
    LIGHT_TYPE_SPOT,
    LIGHT_TYPE_POINT,
    LIGHT_TYPE_DIRECTIONAL,
    LIGHT_TYPE_NUM
};
u32 m_LightType;

//Texture info
namespace
{
    const int s_TexNum = 5;
    const wchar_t* s_TextureNames[ s_TexNum ] =
    {
        L"rom:/resources/sample_RGBA4_32_32.btex",
        L"rom:/resources/sample_RGBA8_64_64.btex",
        L"rom:/resources/sample_ETC1A4_64_64.btex",
        L"rom:/resources/sample_LA8_256_256.btex",
        L"rom:/resources/sample_bump_HILO8_32_32.btex"
    };
    
    const nn::gd::Resource::NativeFormat s_TextureFormat[ s_TexNum ] =
    {
        nn::gd::Resource::NATIVE_FORMAT_RGBA_4444,
        nn::gd::Resource::NATIVE_FORMAT_RGBA_8888,
        nn::gd::Resource::NATIVE_FORMAT_ETC1_A4,
        nn::gd::Resource::NATIVE_FORMAT_LUMINANCE_ALPHA_88,
        nn::gd::Resource::NATIVE_FORMAT_HILO_88
    };
    const u32 s_TextureWidth[ s_TexNum ] =
    {
        32, 64, 64, 256, 32
    };
}

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
const u32 s_GxHeapSize = 0x400000;

#if USE_COMMAND_LIST == DOUBLE_COMMAND_LIST
GLuint s_CmdlistID[2];
GLuint s_CurrentCommandListIndex = 0;
#else
GLuint s_CmdlistID[1];
#endif

static void RenderScene();
static void SetupLights(nn::math::MTX34& view);


#ifdef USE_PROFILING_FUNCTIONS

static bool isProfiling = false;    //to make sure we don't call startGPUProfiling twice before calling stopGPUProfiling

#if USE_COMMAND_LIST == DOUBLE_COMMAND_LIST
//We need 2 different lists, because we are in double command buffer mode
u8 listIdRequest_0[CMD_LIST_MAX_REQUEST_COUNT + 1];
u8 listIdRequest_1[CMD_LIST_MAX_REQUEST_COUNT + 1];
u8 *listIdRequest_ToBeCalledByCallBackFunction = listIdRequest_0;  // list of requests to be called by the callback function
u8 *listIdRequest_ExecutedByCallBackFunction = listIdRequest_1;    // list of requests called by the callback function
#else
//Both list are similar, because we are in single command buffer mode
u8 listIdRequest_ToBeCalledByCallBackFunction[CMD_LIST_MAX_REQUEST_COUNT + 1];               // list of requests to be called by the callback function
u8 *listIdRequest_ExecutedByCallBackFunction = listIdRequest_ToBeCalledByCallBackFunction;   // list of requests called by the callback function
#endif

void startGPUProfiling()
{
    if (isProfiling) {NN_ASSERTMSG( 0, "Already being profiled"); return; }
    isProfiling = true;

    //get the id of the next request
    int reqCount;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &reqCount);
    int idRequest = reqCount + 1;

    //check that the id is not already used
    if (listIdRequest_ToBeCalledByCallBackFunction[idRequest] != 0) {NN_ASSERTMSG( 0, "Id already exists in listIdRequest_ToBeCalledByCallBackFunction"); return; }

    //insert the id as a start request
    listIdRequest_ToBeCalledByCallBackFunction[idRequest] = 1;  //1 == Start
    
    //the request will call the callback function when its execution is over, so we can start the profiling in the callback function
    nngxEnableCmdlistCallback(idRequest);
    //to make sure that no other requests is send after this request is executed (the callback function will start the command list again)
    nngxReserveStopCmdlist(idRequest);

    //add a dummy command, to make sure we don't insert an empty P3D request
    nn::gd::System::AddDummyCommands(1);
    //insert the p3d request in the request list, then the callback function will be called when the request has been completed
    nn::gd::System::Execute3DCommandList();
}

void stopGPUProfiling()
{
    if (!isProfiling) {NN_ASSERTMSG( 0, "startGPUProfiling Was not called"); return; }
    isProfiling = false;

    //get the id of the next request
    int reqCount;
    nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &reqCount);
    int idRequest = reqCount + 1;

    //check that the id is not already used
    if (listIdRequest_ToBeCalledByCallBackFunction[idRequest] != 0) {NN_ASSERTMSG( 0, "Id already exists in listIdRequest_ToBeCalledByCallBackFunction"); return; }

    //insert the id as a stop request
    listIdRequest_ToBeCalledByCallBackFunction[idRequest] = 2;  //2 == Stop

    //the request will call the callback function when its execution is over, so we can stop the profiling in the callback function
    nngxEnableCmdlistCallback(idRequest);
    //to make sure that no other requests is send after this request is executed (the callback function will start the command list again)
    nngxReserveStopCmdlist(idRequest);

    //add a dummy command, to make sure we don't insert an empty P3D request
    nn::gd::System::AddDummyCommands(1);
    //insert the p3d request in the request list, then the callback function will be called when the request has been completed
    nn::gd::System::Execute3DCommandList();
}

//The following arrays will store the profiling values
u32 buffer_busy[NN_GX_PROFILING_RESULT_BUFSIZE_BUSY];
u32 buffer_vertexCache[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_CACHE];
u32 buffer_vertexShader0[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER0];
u32 buffer_vertexShader1[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER1];
u32 buffer_vertexShader2[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER2];
u32 buffer_vertexShader3[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER3];
u32 buffer_Polygon[NN_GX_PROFILING_RESULT_BUFSIZE_POLYGON];
u32 buffer_Fragment[NN_GX_PROFILING_RESULT_BUFSIZE_FRAGMENT];
u32 buffer_MemoryAccess[NN_GX_PROFILING_RESULT_BUFSIZE_MEMORY_ACCESS];

//We need to store the start value for some counters in order to calculate the difference between stop and start calls
u32 buffer_vertexShader0_start[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER0];
u32 buffer_vertexShader1_start[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER1];
u32 buffer_vertexShader2_start[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER2];
u32 buffer_vertexShader3_start[NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER3];
u32 buffer_Polygon_start[NN_GX_PROFILING_RESULT_BUFSIZE_POLYGON];
u32 buffer_Fragment_start[NN_GX_PROFILING_RESULT_BUFSIZE_FRAGMENT];
u32 buffer_MemoryAccess_start[NN_GX_PROFILING_RESULT_BUFSIZE_MEMORY_ACCESS];

//All nngx profiling calls can be done inside this callback function because we're sure that at this moment no P3D requests are being executed
void cmdListCallbackFunction(int idRequest)
{
    //PRINT(("\n Callback id: %d.  ", idRequest));
    NN_ASSERTMSG(idRequest >= 0 && idRequest <= CMD_LIST_MAX_REQUEST_COUNT, "Invalid idRequest: %d", idRequest);

    if (listIdRequest_ExecutedByCallBackFunction[idRequest] != 1 && listIdRequest_ExecutedByCallBackFunction[idRequest] != 2)
    {
        NN_ASSERTMSG( 0, "Callback function called for no purpose");
        return;
    }

    //It's a start request
    if (listIdRequest_ExecutedByCallBackFunction[idRequest] == 1)   
    {
        //PRINT(("Start profiling\n"));

        listIdRequest_ExecutedByCallBackFunction[idRequest] = 0;

        nngxStartProfiling(NN_GX_PROFILING_BUSY);
        //nngxStartProfiling(NN_GX_PROFILING_VERTEX_CACHE);     //no use (reset at each draw)

        //get the current value of the profiling counter
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER0, buffer_vertexShader0_start);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER1, buffer_vertexShader1_start);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER2, buffer_vertexShader2_start);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER3, buffer_vertexShader3_start);
        nngxGetProfilingResult(NN_GX_PROFILING_POLYGON, buffer_Polygon_start);
        nngxGetProfilingResult(NN_GX_PROFILING_FRAGMENT, buffer_Fragment_start);
        nngxGetProfilingResult(NN_GX_PROFILING_MEMORY_ACCESS, buffer_MemoryAccess_start);

        //restart the command list (it has been stopped so that the request can call the callback function safely)
#if USE_COMMAND_LIST != DOUBLE_COMMAND_LIST
        nngxRunCmdlist();   //only one command list in single command buffer
#else
        //We need to restart the command list which was running (not the current bound command list)
        /*nngxBindCmdlist( s_CmdlistID[1 - s_CurrentCommandListIndex] );
        nngxRunCmdlist();
        nngxBindCmdlist( s_CmdlistID[s_CurrentCommandListIndex] );*/
        nngxRunCmdlistByID( s_CmdlistID[1 - s_CurrentCommandListIndex] );
#endif

        return;
    }

    //It's a stop request
    if (listIdRequest_ExecutedByCallBackFunction[idRequest] == 2)   
    {
        //PRINT(("Stop profiling\n"));

        listIdRequest_ExecutedByCallBackFunction[idRequest] = 0;

        nngxStopProfiling(NN_GX_PROFILING_BUSY);
        //nngxStopProfiling(NN_GX_PROFILING_VERTEX_CACHE);      //no use (this counter is reset at each draw)

        nngxGetProfilingResult(NN_GX_PROFILING_BUSY, buffer_busy);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_CACHE, buffer_vertexCache);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER0, buffer_vertexShader0);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER1, buffer_vertexShader1);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER2, buffer_vertexShader2);
        nngxGetProfilingResult(NN_GX_PROFILING_VERTEX_SHADER3, buffer_vertexShader3);
        nngxGetProfilingResult(NN_GX_PROFILING_POLYGON, buffer_Polygon);
        nngxGetProfilingResult(NN_GX_PROFILING_FRAGMENT, buffer_Fragment);
        nngxGetProfilingResult(NN_GX_PROFILING_MEMORY_ACCESS, buffer_MemoryAccess);

        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER0; n++) buffer_vertexShader0[n] = buffer_vertexShader0[n] - buffer_vertexShader0_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER1; n++) buffer_vertexShader1[n] = buffer_vertexShader1[n] - buffer_vertexShader1_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER2; n++) buffer_vertexShader2[n] = buffer_vertexShader2[n] - buffer_vertexShader2_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_VERTEX_SHADER3; n++) buffer_vertexShader3[n] = buffer_vertexShader3[n] - buffer_vertexShader3_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_POLYGON; n++) buffer_Polygon[n] = buffer_Polygon[n] - buffer_Polygon_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_FRAGMENT; n++) buffer_Fragment[n] = buffer_Fragment[n] - buffer_Fragment_start[n];
        for (int n=0; n<NN_GX_PROFILING_RESULT_BUFSIZE_MEMORY_ACCESS; n++) buffer_MemoryAccess[n] = buffer_MemoryAccess[n] - buffer_MemoryAccess_start[n];

        //==============================================================================
#ifdef PRINT_PROFILING_RESULT
            
        //if too many logs are printed it may crash the application, so we display the result only every X frames
        static int printCounter = 0;
        printCounter++;
        if (printCounter > 100)
        {
            printCounter = 0;

            //Busy counter
            PRINT(("\nProfiling counter: Busy\n"
                "Shader processor 0: %d\n"
                "Command buffer and vertex array load module: %d\n"
                "Rasterization module: %d\n"
                "Triangle setup: %d\n"
                "Fragment lighting: %d\n"
                "Texture unit: %d\n"
                "Per-fragment operations module: %d\n"
                "Texture combiner: %d\n",
                buffer_busy[0] >> 16,
                buffer_busy[0] & 0xffff,
                buffer_busy[1] >> 16,
                buffer_busy[1] & 0xffff,
                buffer_busy[2] >> 16,
                buffer_busy[2] & 0xffff,
                buffer_busy[3] >> 16,
                buffer_busy[3] & 0xffff
            ));

            //==============================================================================
            //Vertex cache 0
            PRINT(("\nProfiling counter: Vertex cache: %d\n", buffer_vertexCache[0]));

            //==============================================================================
            //Vertex Shader 0 to 3
            for (int n=0; n<4; ++n)
            {
                u32* buffer = n==0?buffer_vertexShader0: ( n==1?buffer_vertexShader1: ( n==2?buffer_vertexShader2: buffer_vertexShader3 ) );
                NN_UNUSED_VAR(buffer);  // Warning measure during Release build
                PRINT(("\nProfiling counter: Vertex Shader %d\n"
                    "Program counter transition count: %d\n"
                    "Cycles stalled due to command dependencies: %d\n"
                    "Cycles stalled due to address register updates (mova): %d\n",
                    n,
                    buffer[0],
                    buffer[1],
                    buffer[2]
                ));
                PRINT(("Cycles stalled due to status register updates (cmp): %d\n"
                    "Cycles stalled due to program pre-fetch misses: %d\n",
                    buffer[3],
                    buffer[4]
                ));
            }

            //==============================================================================
            //Polygon
            PRINT(("\nProfiling counter: Polygon\n"
                "Number of vertices input into triangle setup: %d\n"
                "Number of polygons input into triangle setup: %d\n"
                "Number of polygons output from triangle setup: %d\n",
                buffer_Polygon[0],
                buffer_Polygon[1],
                buffer_Polygon[2]
            ));

            //==============================================================================
            //Fragment
            PRINT(("\nProfiling counter: Fragment\n"
                "number of fragments input into the per-fragment operations module: %d\n",
                buffer_Fragment[0]
            ));

            //==============================================================================
            //Memory Access
            PRINT(("\nProfiling counter: Memory Access\n"
                "PICA reads to the VRAM A channel: %d\n"
                "PICA writes to the VRAM A channel: %d\n"
                "PICA reads to the VRAM B channel: %d\n"
                "PICA writes to the VRAM B channel: %d\n",
                buffer_MemoryAccess[0],
                buffer_MemoryAccess[1],
                buffer_MemoryAccess[2],
                buffer_MemoryAccess[3]
            ));
            PRINT(("Reads of command buffers, vertex arrays, and index arrays: %d\n"
                "Texture memory reads by the texture unit: %d\n"
                "Depth and stencil buffer reads by the perfragment operation module: %d\n",
                buffer_MemoryAccess[4],
                buffer_MemoryAccess[5],
                buffer_MemoryAccess[6]
            ));
            PRINT(("Depth and stencil buffer writes by the perfragment operation module: %d\n"
                "Color buffer reads by the per-fragment operations module: %d\n"
                "Color buffer writes by the per-fragment operations module: %d\n",
                buffer_MemoryAccess[7],
                buffer_MemoryAccess[8],
                buffer_MemoryAccess[9]
            ));
            PRINT(("Display buffer reads by the LCD upper screen controller: %d\n"
                "Display buffer reads by the LCD lower screen controller: %d\n"
                "Reads by the post-transfer module: %d\n"
                "Writes by the post-transfer module: %d\n",
                buffer_MemoryAccess[10],
                buffer_MemoryAccess[11],
                buffer_MemoryAccess[12],
                buffer_MemoryAccess[13]
            ));
            PRINT(("Channel 0 buffer writes by the memory fill module: %d\n"
                "Channel 1 buffer writes by the memory fill module: %d\n"
                "VRAM reads by the CPU: %d\n"
                "VRAM writes by the CPU: %d\n",
                buffer_MemoryAccess[14],
                buffer_MemoryAccess[15],
                buffer_MemoryAccess[16],
                buffer_MemoryAccess[17]
            ));
        }
#endif //PRINT_PROFILING_RESULT

        //restart the command list (it has been stopped so that the request can call the callback function safely)
#if USE_COMMAND_LIST != DOUBLE_COMMAND_LIST
        nngxRunCmdlist();   //only one command list in single command buffer
#else
        //We need to restart the command list which was running (not the current bound command list)
        /*nngxBindCmdlist( s_CmdlistID[1 - s_CurrentCommandListIndex] );
        nngxRunCmdlist();
        nngxBindCmdlist( s_CmdlistID[s_CurrentCommandListIndex] );*/
        nngxRunCmdlistByID( s_CmdlistID[1 - s_CurrentCommandListIndex] );
#endif

        return;
    }
}
#endif  //USE_PROFILING_FUNCTIONS

//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

void UpdateScene( void )
{
    nn::math::SinCosFIdx( &m_Sin,   &m_Cos,   m_Counter );
    nn::math::SinCosFIdx( &m_Sin2,  &m_Cos2,  m_Counter * 2 );
    nn::math::SinCosFIdx( &m_Sin01, &m_Cos01, m_Counter * 0.1f );
    ++m_Counter;

    m_LightType = ( m_Counter / 1200 ) % LIGHT_TYPE_NUM;

    nn::math::MTX34Identity( &mtx );

    const nn::math::VEC3 t = nn::math::VEC3( m_Cos2, 0.f, m_Sin2 ) * 15.f;
    const nn::math::VEC3 s = nn::math::VEC3( 1.f, 1.f, 1.f ) * 5.f;

    const f32 opt1 = m_Cos2;
    const f32 opt2 = m_Sin2 * m_Sin2;
    const f32 opt3 = 0;

    const f32 tmp00 = s.x * (m_Cos2);
    const f32 tmp10 = 0;
    const f32 tmp20 = s.x * (-m_Sin2);

    const f32 tmp01 = s.y * ((opt2) - (opt3));
    const f32 tmp11 = s.y * ((opt1));
    const f32 tmp21 = s.y * ((m_Sin2 * m_Cos2));

    const f32 tmp02 = s.z * (opt1 * m_Sin2);
    const f32 tmp12 = s.z * ((opt3 * m_Sin2) - (m_Sin2));
    const f32 tmp22 = s.z * ((m_Cos2 * m_Cos2));

    mtx.f._00 = tmp00;
    mtx.f._01 = tmp01;
    mtx.f._02 = tmp02;
    mtx.f._03 = t.x;

    mtx.f._10 = tmp10;
    mtx.f._11 = tmp11;
    mtx.f._12 = tmp12;
    mtx.f._13 = t.y;

    mtx.f._20 = tmp20;
    mtx.f._21 = tmp21;
    mtx.f._22 = tmp22;
    mtx.f._23 = t.z;
}

static void InitShader()
{
    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nnResult res;
    res = nn::gd::ShaderStage::CreateShaderBinary(buf, read, &s_shaderBinary);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    res = nn::gd::ShaderStage::CreateShader(s_shaderBinary, 0, &s_vertexShader);              if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::ShaderPipelineUnmanagedRegisters *pUnmanagedRegisters = NULL;
#ifdef SHADER_USE_UNMANAGED_UNIFORM

    //nn::gd::ShaderStage::SetUnmanagedRegisters(s_shaderPipeline, nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 8/*s_shaderVariable_user.GetRegister()*/, 3);
    nn::gd::ShaderPipelineUnmanagedRegisters unmanagedRegisters;

    nn::gd::UnmanagedRegistersInterval arrayUnmanagedRegistersInterval[] = {
        {nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 0, 4},
        {nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 4, 3},
        {nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 8, 3}
    };
    unmanagedRegisters.m_ArrayUnmanagedRegistersInterval = arrayUnmanagedRegistersInterval;
    unmanagedRegisters.m_CountUnmanagedRegistersInterval = 3;
    pUnmanagedRegisters = &unmanagedRegisters;
#endif

    res = nn::gd::ShaderStage::CreateShaderPipeline(s_vertexShader, NULL, &s_shaderPipeline, pUnmanagedRegisters); if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    s_shaderVariable_proj = nn::gd::ShaderStage::GetShaderUniformLocation(s_shaderPipeline, "proj");
    s_shaderVariable_view = nn::gd::ShaderStage::GetShaderUniformLocation(s_shaderPipeline, "view");
    s_shaderVariable_user = nn::gd::ShaderStage::GetShaderUniformLocation(s_shaderPipeline, "user");
    NN_ASSERT(s_shaderVariable_proj.IsValid());
    NN_ASSERT(s_shaderVariable_view.IsValid());
    NN_ASSERT(s_shaderVariable_user.IsValid());

    nn::gd::ShaderStage::SetShaderPipeline(s_shaderPipeline);

    file.Finalize();
    s_AppHeap.Free(buf);
}

static void InitObjects(void)
{
    nnResult res;

    //Create cube
    {
        nn::gd::InputElementDescription descs[] =
        {
            { 0, "Position", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            { 1, "TexCoord0", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, 0},
            { 2, "Color", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0},
            { 3, "Normal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}
        };

        u32 strides[] = { sizeof(float) * 3, sizeof(float) * 2, sizeof(float) * 4, sizeof(float) * 3};
        res = nn::gd::VertexInputStage::CreateInputLayout(descs, 4, strides, s_vertexShader, &s_inputLayout_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Position
        nn::gd::VertexBufferResourceDescription descPosCube = {sizeof(s_BoxVtxStreamPos), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descPosCube, s_BoxVtxStreamPos, &s_bufferPosition_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //TexCoord
        nn::gd::VertexBufferResourceDescription descTexCoordCube = {sizeof(s_BoxVtxStreamTex), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descTexCoordCube, s_BoxVtxStreamTex, &s_bufferTexCoord_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Color
        nn::gd::VertexBufferResourceDescription descColCube = {sizeof(s_BoxVtxStreamClr), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descColCube, s_BoxVtxStreamClr, &s_bufferColor_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Normal
        nn::gd::VertexBufferResourceDescription descNormCube = {sizeof(s_BoxVtxStreamNrm), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descNormCube, s_BoxVtxStreamNrm, &s_bufferNormal_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Index
        nn::gd::VertexBufferResourceDescription descIdxCube = {sizeof(s_BoxVtxStreamIdx), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descIdxCube, s_BoxVtxStreamIdx, &s_bufferIndex_cube);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        s_buffers_cube[0] = s_bufferPosition_cube;
        s_buffers_cube[1] = s_bufferTexCoord_cube;
        s_buffers_cube[2] = s_bufferColor_cube;
        s_buffers_cube[3] = s_bufferNormal_cube;

        s_cubeVertexCount = BOX_VERTEX_NUM;
    }

    //Create interleaved sphere
    {
        PrimitiveInterleavedSphere sphere;
        sphere.SetupVtxStream();

        nn::gd::InputElementDescription descs[] =
        {
            //{ 0, "Position", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
            //{ 0, "TexCoord0", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, sizeof(float) * 3},
            //{ 0, "Color", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, sizeof(float) * 3 + sizeof(float) * 2},
            //{ 0, "Normal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, sizeof(float) * 3 + sizeof(float) * 2 + sizeof(float) * 4}

            //Same declaration as above, but using nn::gd::InputElementDescription::Append (this can simplify the code)
            { 0, "Position", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, nn::gd::InputElementDescription::Append },
            { 0, "TexCoord0", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 2, nn::gd::InputElementDescription::Append },
            { 0, "Color", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, nn::gd::InputElementDescription::Append },
            { 0, "Normal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, nn::gd::InputElementDescription::Append }
        };

        u32 strides[] = { sizeof(float) * 12 };
        res = nn::gd::VertexInputStage::CreateInputLayout(descs, 4, strides, s_vertexShader, &s_inputLayout_sphere);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Buffer
        nn::gd::VertexBufferResourceDescription descBuff = {sizeof(sphere.stream), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descBuff, sphere.stream, &s_bufferInterleave_sphere);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        //Index
        nn::gd::VertexBufferResourceDescription descIdx = {sizeof(sphere.streamIdx), nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateVertexBufferResource(&descIdx, sphere.streamIdx, &s_bufferIndex_sphere);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        s_sphereVertexCount = sphere.vertexCount;
    }

}

void Initialize_GD()
{
    nnResult res;

    //DepthStencil
    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    desc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    res = nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthStencil_Enabled);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //2D textures
    {
        int textureIndex = 4;
        nn::fs::FileReader texture_file_reader( s_TextureNames[ textureIndex ] );
        u32 texture_file_size = static_cast<u32>( texture_file_reader.GetSize() );
        u8* textureDataPtr = reinterpret_cast<u8*>( s_AppHeap.Allocate(texture_file_size) );
        texture_file_reader.Read( textureDataPtr, texture_file_size );

        nn::gd::Texture2DResourceDescription Text2DResDesc = {s_TextureWidth[textureIndex], s_TextureWidth[textureIndex], 1, s_TextureFormat[textureIndex], nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, &s_texture2DResourceA);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        nn::gd::Texture2DDescription tex2DDesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResourceA, &tex2DDesc, &s_texture2DA);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        s_AppHeap.Free(textureDataPtr);
        texture_file_reader.Finalize();
    }

#ifdef DEMO_ALTERNATE_TEXTURES
    {
        int textureIndex = 2; 
        nn::fs::FileReader texture_file_reader( s_TextureNames[ textureIndex ] );
        u32 texture_file_size = static_cast<u32>( texture_file_reader.GetSize() );
        u8* textureDataPtr = reinterpret_cast<u8*>( s_AppHeap.Allocate(texture_file_size) );
        texture_file_reader.Read( textureDataPtr, texture_file_size );

        nn::gd::Texture2DResourceDescription Text2DResDesc = {s_TextureWidth[textureIndex], s_TextureWidth[textureIndex], s_TextureFormat[textureIndex], 1, nn::gd::Memory::FCRAM};
        res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, textureDataPtr, GD_FALSE, &s_texture2DResourceB);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        nn::gd::Texture2DDescription tex2DDesc = {0, 0};
        res = nn::gd::TextureStage::CreateTexture2D(s_texture2DResourceB, &tex2DDesc, &s_texture2DB);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

        s_AppHeap.Free(textureDataPtr);
        texture_file_reader.Finalize();
    }
#endif

    //Sampler states
    {
        nn::gd::SamplerStateDescription descDefault;
        descDefault.ToDefault();
        res = nn::gd::TextureStage::CreateSamplerState(&descDefault, &s_samplerDefault);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    }

    //Texture combiner
    {
        nn::gd::CombinerDescription combiner;
        combiner.ToDefault();
        combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_FRAGMENT_SECONDARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
        combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR, nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
        combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
        combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_RGB_ADD);
        combiner.SetCombineAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::COMBINE_ALPHA_REPLACE);
        combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
        res = nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &s_texComState);
        if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
    }

    /*******************************************************************/
    //Set default settings
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    //Texture unit 0
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2DA);

    //nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2D_colorBuffer);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, s_samplerDefault);

    //Texture combiner
    nn::gd::CombinerStage::SetTextureCombinerState(s_texComState);

    //LightingStage
    nn::gd::LightingStage::SetGlobalColorAmbient(0, 0, 0);
    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);

    nn::gd::LightingStage::light[0].SetColorAmbient(16, 0, 0);
    nn::gd::LightingStage::light[0].SetColorDiffuse(255, 0, 0);
    nn::gd::LightingStage::light[0].SetColorSpecular0(255, 255, 255);
    nn::gd::LightingStage::light[0].SetColorSpecular1(255, 255, 255);
    nn::gd::LightingStage::light[0].EnableGeomFactor1( GD_TRUE );

    nn::gd::LightingStage::light[1].SetColorAmbient(0, 16, 0);
    nn::gd::LightingStage::light[1].SetColorDiffuse(0, 255, 0);
    nn::gd::LightingStage::light[1].SetColorSpecular0(255, 255, 255);
    nn::gd::LightingStage::light[1].SetColorSpecular1(255, 255, 255);
    nn::gd::LightingStage::light[1].EnableGeomFactor1( GD_TRUE );

    nn::gd::LightingStage::light[2].SetColorAmbient(0, 0, 16);
    nn::gd::LightingStage::light[2].SetColorDiffuse(0, 0, 255);
    nn::gd::LightingStage::light[2].SetColorSpecular0(255, 255, 255);
    nn::gd::LightingStage::light[2].SetColorSpecular1(255, 255, 255);
    nn::gd::LightingStage::light[2].EnableGeomFactor1( GD_TRUE );

    f32 attenuationScale, attenuationBias;
    nn::gd::Utils::ConvertStartEndToScaleBias(20.f, 40.f, attenuationScale, attenuationBias);
    nn::gd::LightingStage::light[2].SetDistanceAttenuationScaleBias(attenuationScale, attenuationBias);

    // Use texture unit 0 as BUMP
    nn::gd::LightingStage::SetBumpMode(nn::gd::LightingStage::BUMPMODE_AS_BUMP, GD_TRUE);
    nn::gd::LightingStage::SetBumpSelectorTexture(nn::gd::TextureStage::TEXTURE_UNIT_0);

    // Layer config 0
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_0);

    // Enable distribution 0
    nn::gd::LightingStage::EnableLookUpTableD0(GD_TRUE);

    // Set distribution 0 lookup table
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_D0, GD_TRUE);

    // Set spot lookup table
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_SP, nn::gd::LightingStage::INPUT_VALUE_SP);
    nn::gd::LightingStage::SetLookUpTableAbsInput(nn::gd::LightingStage::LOOKUP_TABLE_ID_SP, GD_TRUE);

    //LUT
    {
        // Lookup table (LUT) initialization
        // Save LUT curve in form that can be sent to GPU.
        f32 table_value[ 256 ];
        f32 table_delta[ 256 ];

        //Distance attenuation LUT
        for ( int i = 0; i < 256; ++i )
        {
            table_value[ i ] = ( 256 - i ) / 256.f;
        }
        for ( int i = 0; i < 255; ++i )
        {
            table_delta[i] = table_value[i+1] - table_value[i];
        }
        table_delta[255] = table_value[255];
        nn::gd::LightingStage::light[2].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_DA, 0, table_value, table_delta, 256);

        //Spot LUT
        for ( int i = 0; i < 256; ++i )
        {
            table_value[ i ] = i / 256.f;
            table_value[ i ] *= table_value[ i ];
            table_value[ i ] *= table_value[ i ];
            table_value[ i ] *= table_value[ i ];
        }
        for ( int i = 0; i < 255; ++i )
        {
            table_delta[i] = table_value[i+1] - table_value[i];
        }
        table_delta[255] = table_value[255];
        nn::gd::LightingStage::light[0].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_SP, 0, table_value, table_delta, 256);
        nn::gd::LightingStage::light[1].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_SP, 0, table_value, table_delta, 256);
        nn::gd::LightingStage::light[2].UploadLookUpTableFloat(nn::gd::Light::LOOKUP_TABLE_UPLOAD_SP, 0, table_value, table_delta, 256);

        //Specular LUT
        for ( int i = 0; i < 256; ++i )
        {
            table_value[ i ] *= table_value[ i ];
            table_value[ i ] *= table_value[ i ];
            table_value[ i ] *= table_value[ i ];
            table_value[ i ] *= table_value[ i ];
        }
        for ( int i = 0; i < 255; ++i )
        {
            table_delta[i] = table_value[i+1] - table_value[i];
        }
        table_delta[255] = table_value[255];
        nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_D0, 0, table_value, table_delta, 256);
    }
}

//FPS management
#define FPS_AVG_FRAME_COUNT 300
static int s_curFpsCount = 0;
static double s_tmpFpsSum = 0.0;

s64 GetTimeMicroSeconds()
{
    nn::os::Tick ticks = nn::os::Tick::GetSystemCurrent();
    nn::fnd::TimeSpan TimeSpan = ticks;
    return TimeSpan.GetMicroSeconds();
}

nn::gd::RecordedPacketId* packetId = 0;
#if PACKET_RECORDING_MODE == USE_PACKET_RECORDING_NOCOPY
    nn::gd::System::RecordingPacketUsage recordingPacketUsage = nn::gd::System::RECORD_COMMAND_LIST_NO_COPY;
#else
    nn::gd::System::RecordingPacketUsage recordingPacketUsage = nn::gd::System::RECORD_COMMAND_LIST_COPY;
#endif

void DrawTop(void)
{
    // Calculate view/projection matrices
    { // The camera matrix
        f32 r = 60.f;
        viewMtx.SetupLookAt(
            nn::math::VEC3( r * m_Sin01,  0.0f, r * m_Cos01 ),
            nn::math::VEC3( 0.0f,         1.0f, 0.0f ),
            nn::math::VEC3( 0.0f,         0.0f, 0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        400.f / 240.f,  // aspect
        10.f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    //Shader variables
#ifdef SHADER_USE_UNMANAGED_UNIFORM
    // Use functions that write commands immediately when USE_PACKET_RECORDING is used
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_proj.GetRegister(), 4, static_cast<f32*>(projL));
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_view.GetRegister(), 3, static_cast<f32*>(viewL));
#else
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(projL));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_view, static_cast<f32*>(viewL));
#endif

    //Set the viewport
    nn::gd::Viewport viewPort(0, 0, s_mainRenderTargetProperty.m_Width, s_mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    //Scissor
    nn::gd::RasterizerStage::EnableScissor(GD_TRUE);
    nn::gd::RasterizerStage::SetScissor(viewPort);

    SetupLights(viewL);

#if PACKET_RECORDING_MODE == USE_PACKET_RECORDING_COPY || PACKET_RECORDING_MODE == USE_PACKET_RECORDING_NOCOPY
    if (packetId) nn::gd::System::ReleasePackets(packetId);
    u32 flag = nn::gd::System::MODULE_ALL;
    nn::gd::System::StartRecordingPackets(&flag, recordingPacketUsage);
#endif

    //Clear the render buffers
    u8 clearColor[4] = {51, 51, 51, 51};
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

    RenderScene();

#if PACKET_RECORDING_MODE == USE_PACKET_RECORDING_COPY || PACKET_RECORDING_MODE == USE_PACKET_RECORDING_NOCOPY
    nn::gd::System::StopRecordingPackets(&packetId);
#endif
}

void DrawExt()
{
    // Calculate view/projection matrices
    { // The camera matrix
        f32 r = 60.f;
        viewMtx.SetupLookAt(
            nn::math::VEC3( r * m_Sin01,  0.0f, r * m_Cos01 ),
            nn::math::VEC3( 0.0f,         1.0f, 0.0f ),
            nn::math::VEC3( 0.0f,         0.0f, 0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        400.f / 240.f,  // aspect
        10.f,           // near
        200.f,          // far
        nn::math::PIVOT_NONE );

    nn::math::MTX44 projL, projR;
    nn::math::MTX34 viewL, viewR;
    stereo.SetBaseCamera( &viewMtx );
    stereo.SetBaseFrustum( &projectionMtx );
    stereo.CalculateMatrices( &projL, &viewL, &projR, &viewR, 15.f, 1.f );

    //Shader variables
#ifdef SHADER_USE_UNMANAGED_UNIFORM
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_proj.GetRegister(), 4, static_cast<f32*>(projR));
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_view.GetRegister(), 3, static_cast<f32*>(viewR));
#else
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(projR));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_view, static_cast<f32*>(viewR));
#endif

    //Set the viewport
    nn::gd::Viewport viewPort(0, 0, s_mainRenderTargetProperty.m_Width, s_mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    //Scissor
    nn::gd::RasterizerStage::EnableScissor(GD_TRUE);
    nn::gd::RasterizerStage::SetScissor(viewPort);

    SetupLights(viewR);

#if PACKET_RECORDING_MODE == USE_PACKET_RECORDING_COPY || PACKET_RECORDING_MODE == USE_PACKET_RECORDING_NOCOPY
    u32 flag = nn::gd::System::MODULE_ALL;
    nn::gd::System::ReplayPackets(packetId, &flag);
#else
    //Clear the render buffers
    u8 clearColor[4] = {51, 51, 51, 51};
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

    //RenderScene( projR, viewR );
    RenderScene();
#endif
}

void DrawBottom()
{
    { // The camera matrix
        f32 r = 120.f;

        viewMtx.SetupLookAt(
            nn::math::VEC3( r * m_Sin01, r * m_Sin01, r * m_Cos01 ),
            nn::math::VEC3( 0.0f,        1.0f,        0.0f ),
            nn::math::VEC3( 0.0f,        0.0f,        0.0f ) );
    }

    // Projection matrix
    projectionMtx.SetupPerspective(
        0.66f,          // fovy
        320.f / 240.f,  // aspect
        0.1f,           // near
        20000.f,        // far
        nn::math::PIVOT_UPSIDE_TO_TOP );

    //Shader variables
#ifdef SHADER_USE_UNMANAGED_UNIFORM
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_proj.GetRegister(), 4, static_cast<f32*>(projectionMtx));
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_view.GetRegister(), 3, static_cast<f32*>(viewMtx));
#else
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_proj, static_cast<f32*>(projectionMtx));
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_view, static_cast<f32*>(viewMtx));
#endif

    //Set the viewport
    nn::gd::Viewport viewPort(0, 0, s_mainRenderTargetProperty.m_Width, s_mainRenderTargetProperty.m_Height);

#if 1
    // Change render position every 256 frames (Viewport/Scissor test)
    if ( ( m_Counter % 512 ) < 256 )
    {
        u32 c = ( m_Counter % ( 512 * 4 ) ) / 512;
        s32 x = 120 * ( c % 2 );
        s32 y = 160 * ( c / 2 );
        viewPort.Set( x, y, 120, 160 );
    }
    else
    {
        viewPort.Set( 0, 0, 240, 320 );
    }
#endif
    nn::gd::RasterizerStage::SetViewport(viewPort);
    nn::gd::RasterizerStage::EnableScissor(GD_TRUE);
    nn::gd::RasterizerStage::SetScissor(viewPort);

    SetupLights(viewMtx);

#if PACKET_RECORDING_MODE == USE_PACKET_RECORDING_COPY || PACKET_RECORDING_MODE == USE_PACKET_RECORDING_NOCOPY
    u32 flag = nn::gd::System::MODULE_ALL;
    nn::gd::System::ReplayPackets(packetId, &flag);
#else
    //Clear the render buffers
    u8 clearColor[4] = {0, 0, 0, 0};
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

    RenderScene();
#endif
}

int DrawFrame(void)
{
    static int frameCount = 0;
    //PRINT(("DrawFrame: %d\n", frameCount));
    s64 startTime = GetTimeMicroSeconds();
#ifdef DISPLAY_TIMING_PER_SCREEN
    double timeLast = startTime;
    static double timeTop = 0.;
    static double timeExt = 0.;
    static double timeBottom = 0.;
    static double timeWaitCmdListDone = 0.;
#endif

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer_top[s_CurrentDisplaybuffer_top]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &s_mainRenderTargetProperty);

        //Set the color/depth/stencil targets
        nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
        nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);
    }

#ifdef USE_PROFILING_FUNCTIONS
    startGPUProfiling();
#endif

    /**************************************************************/
    /********************** Display 0 (Top) ***********************/
    {
        DrawTop();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer_top[s_CurrentDisplaybuffer_top]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

#ifdef DISPLAY_TIMING_PER_SCREEN
        s64 ctime = GetTimeMicroSeconds();
        timeTop += ctime - timeLast;
        timeLast = ctime;
#endif

    }

#ifdef USE_PROFILING_FUNCTIONS
    stopGPUProfiling();
#endif

    /**************************************************************/
    /********************** Display 0 (ext) ***********************/
    {
        DrawExt();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0_EXT);
        nngxBindDisplaybuffer(s_displaybuffer_ext[s_CurrentDisplaybuffer_ext]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

#ifdef DISPLAY_TIMING_PER_SCREEN
        s64 ctime = GetTimeMicroSeconds();
        timeExt += ctime - timeLast;
        timeLast = ctime;
#endif
    }

    /***************************************************************/
    /********************* Display 1 (Bottom) **********************/
    {
        DrawBottom();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxBindDisplaybuffer(s_displaybuffer_bottom[s_CurrentDisplaybuffer_bottom]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp1_Width,                                            //dstWidth
            disp1_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

#ifdef DISPLAY_TIMING_PER_SCREEN
        s64 ctime = GetTimeMicroSeconds();
        timeBottom += ctime - timeLast;
        timeLast = ctime;
#endif
    }

/*#ifdef USE_PROFILING_FUNCTIONS
    stopGPUProfiling();
#endif*/

#if USE_COMMAND_LIST == DOUBLE_COMMAND_LIST
    //Wait for the previous command list to be done
    nngxBindCmdlist( s_CmdlistID[1 - s_CurrentCommandListIndex] );

#ifdef USE_PROFILING_FUNCTIONS
    //we need to make sure that the command list is really done (some start and stop profiling requests will stop the cmdList and make it considered as done)
    bool done = false;
    while (!done){
        nngxWaitCmdlistDone();

        int usedReqCount, runReqCount;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &usedReqCount);
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_RUN_REQCOUNT, &runReqCount);

        if (usedReqCount != runReqCount){
            //PRINT(("usedReqCount (%d) != runReqCount (%d)\n", usedReqCount, runReqCount));
        }
        else done = true;
    }

    //switch the lists
    u8* tmpList = listIdRequest_ToBeCalledByCallBackFunction;
    listIdRequest_ToBeCalledByCallBackFunction = listIdRequest_ExecutedByCallBackFunction;
    listIdRequest_ExecutedByCallBackFunction = tmpList;
#else
    nngxWaitCmdlistDone();
#endif
    nngxClearCmdlist();

    //Run the current command list
    nngxBindCmdlist( s_CmdlistID[s_CurrentCommandListIndex] );
    nngxRunCmdlist();

    //Swap and bind the next (previous) command list
    s_CurrentCommandListIndex = 1 - s_CurrentCommandListIndex;
    nngxBindCmdlist( s_CmdlistID[s_CurrentCommandListIndex] );

#else

#if USE_COMMAND_LIST == SINGLE_COMMAND_LIST_SERIE
    nngxRunCmdlist();
#endif

#ifdef USE_PROFILING_FUNCTIONS
    //we need to make sure that the command list is really done (some start and stop profiling requests will stop the cmdList and make it considered as done)
    bool done = false;
    while (!done){
        nngxWaitCmdlistDone();

        int usedReqCount, runReqCount;
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &usedReqCount);
        nngxGetCmdlistParameteri(NN_GX_CMDLIST_RUN_REQCOUNT, &runReqCount);

        if (usedReqCount != runReqCount){
            //PRINT(("usedReqCount (%d) != runReqCount (%d)\n", usedReqCount, runReqCount));
        }
        else done = true;
    }
#else
    nngxWaitCmdlistDone();
#endif

    nngxClearCmdlist();
#if USE_COMMAND_LIST == SINGLE_COMMAND_LIST_PARALLEL
    nngxRunCmdlist();
#else
    nngxStopCmdlist();
#endif

#endif

 
#ifdef DISPLAY_TIMING_PER_SCREEN
    s64 ctime = GetTimeMicroSeconds();
    timeWaitCmdListDone += ctime - timeLast;
    timeLast = ctime;
#endif

    // Swap the display buffers
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer_top = (s_CurrentDisplaybuffer_top == 0 ? 1 : 0);
    s_CurrentDisplaybuffer_ext = (s_CurrentDisplaybuffer_ext == 0 ? 1 : 0);
    s_CurrentDisplaybuffer_bottom = (s_CurrentDisplaybuffer_bottom == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    //nngxClearCmdlist();
    //nngxRunCmdlist();

    if (frameCount == 0)
    {
        nngxStartLcdDisplay();
    }

    frameCount++;

    //FPS recording/display
    s_tmpFpsSum += GetTimeMicroSeconds() - startTime;
    if (++s_curFpsCount == FPS_AVG_FRAME_COUNT)
    {
        double avgtime = s_tmpFpsSum / FPS_AVG_FRAME_COUNT;
        NN_UNUSED_VAR(avgtime);
        PRINT(("AVG Time: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));

#ifdef DISPLAY_TIMING_PER_SCREEN
        double total = 0.f;
        avgtime = timeTop / FPS_AVG_FRAME_COUNT; total+=avgtime;
        PRINT(("Time top: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        avgtime = timeExt / FPS_AVG_FRAME_COUNT; total+=avgtime;
        PRINT(("Time Ext: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        avgtime = timeBottom / FPS_AVG_FRAME_COUNT; total+=avgtime;
        PRINT(("Time bottom: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        avgtime = timeWaitCmdListDone / FPS_AVG_FRAME_COUNT; total+=avgtime;
        PRINT(("Time wait: %0.0f    (fps: %0.3f)\n", avgtime, (1. / (avgtime/1000000.0))));
        timeTop = timeBottom = timeExt = timeWaitCmdListDone = 0.;
        PRINT(("Total: %0.0f    (fps: %0.3f)\n", total, (1. / (total/1000000.0))));
#endif

        s_curFpsCount = 0;
        s_tmpFpsSum = 0.0;
    }

    //if (frameCount == 100) return -1;
    return 0;
}

static void Finalize()
{
    //Release GD objects
    if (s_texComState) nn::gd::CombinerStage::ReleaseTextureCombinerState(s_texComState);

    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);

    if (s_texture2DResourceA) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResourceA);
    if (s_texture2DA) nn::gd::TextureStage::ReleaseTexture2D(s_texture2DA);
    if (s_texture2DResourceB) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResourceB);
    if (s_texture2DB) nn::gd::TextureStage::ReleaseTexture2D(s_texture2DB);

    if (s_samplerDefault) nn::gd::TextureStage::ReleaseSamplerState(s_samplerDefault);

    if (s_bufferPosition_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferPosition_cube);
    if (s_bufferTexCoord_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferTexCoord_cube);
    if (s_bufferColor_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferColor_cube);
    if (s_bufferNormal_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferNormal_cube);
    if (s_bufferIndex_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferIndex_cube);
    if (s_inputLayout_cube) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout_cube);

    if (s_bufferNormal_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferInterleave_sphere);
    if (s_bufferIndex_cube) nn::gd::Resource::ReleaseVertexBufferResource(s_bufferIndex_sphere);
    if (s_inputLayout_cube) nn::gd::VertexInputStage::ReleaseInputLayout(s_inputLayout_sphere);

    if (s_shaderPipeline) nn::gd::ShaderStage::ReleaseShaderPipeline(s_shaderPipeline);
    if (s_vertexShader) nn::gd::ShaderStage::ReleaseShader(s_vertexShader);
    if (s_shaderBinary) nn::gd::ShaderStage::ReleaseShaderBinary(s_shaderBinary);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer_top);
    nngxDeleteDisplaybuffers(2, s_displaybuffer_ext);
    nngxDeleteDisplaybuffers(2, s_displaybuffer_bottom);
    s_displaybuffer_top[0] = s_displaybuffer_top[1] = 0;
    s_displaybuffer_ext[0] = s_displaybuffer_ext[1] = 0;
    s_displaybuffer_bottom[0] = s_displaybuffer_bottom[1] = 0;
    
    //Release command list
#if USE_COMMAND_LIST == DOUBLE_COMMAND_LIST
    nngxDeleteCmdlists(2, s_CmdlistID);
#else
    nngxDeleteCmdlists(1, s_CmdlistID);
#endif

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

/* initialization */
static int Initialize(void)
{
    nn::fs::Initialize();

    /******************************************************/
    //Init memory
    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);

    /******************************************************/
#if USE_COMMAND_LIST == DOUBLE_COMMAND_LIST
    // Initialize double command list
    nngxGenCmdlists(2, s_CmdlistID);

    nngxBindCmdlist(s_CmdlistID[0]);
    nngxCmdlistStorage(0x80000, CMD_LIST_MAX_REQUEST_COUNT);
    nngxSetCmdlistParameteri(NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN);
    nngxClearCmdlist();

#ifdef USE_PROFILING_FUNCTIONS
    nngxSetCmdlistCallback(cmdListCallbackFunction);
#endif

    nngxBindCmdlist(s_CmdlistID[1]);
    nngxCmdlistStorage(0x80000, CMD_LIST_MAX_REQUEST_COUNT);
    nngxSetCmdlistParameteri(NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN);
    nngxClearCmdlist();

#ifdef USE_PROFILING_FUNCTIONS
    nngxSetCmdlistCallback(cmdListCallbackFunction);
#endif

    s_CurrentCommandListIndex = 0;
    nngxBindCmdlist(s_CmdlistID[s_CurrentCommandListIndex]);
#else
    // Initialize single command list
    nngxGenCmdlists(1, s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID[0]);
    nngxCmdlistStorage(0x80000, CMD_LIST_MAX_REQUEST_COUNT);
    nngxClearCmdlist();

#ifdef USE_PROFILING_FUNCTIONS
    nngxSetCmdlistCallback(cmdListCallbackFunction);
#endif

#if USE_COMMAND_LIST == SINGLE_COMMAND_LIST_SERIE
    nngxSetCmdlistParameteri(NN_GX_CMDLIST_RUN_MODE, NN_GX_CMDLIST_SERIAL_RUN);
#else
    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();
#endif
#endif

#ifdef USE_PROFILING_FUNCTIONS
    memset(listIdRequest_ToBeCalledByCallBackFunction, 0, (CMD_LIST_MAX_REQUEST_COUNT + 1) * sizeof(u8));
    memset(listIdRequest_ExecutedByCallBackFunction, 0, (CMD_LIST_MAX_REQUEST_COUNT + 1) * sizeof(u8));

    //set the profiling parameters
    nngxSetProfilingParameter(NN_GX_PROFILING_BUSY_SAMPLING_TIME, 20);
    nngxSetProfilingParameter(NN_GX_PROFILING_BUSY_SAMPLING_COUNT, 0xafff);
#endif


    /******************************************************/
    // Init display buffers
    //s_displayBuffer_upper = DisplayBuffer::Builder().screenKind(UPPER_SCREEN)    .bufferSize(400, 240).Create();
    //s_displayBuffer_lower = DisplayBuffer::Builder().screenKind(LOWER_SCREEN)    .bufferSize(320, 240).Create();
    //s_displayBuffer_exten = DisplayBuffer::Builder().screenKind(EXTENSION_SCREEN).bufferSize(400, 240).Create();
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxGenDisplaybuffers(2, s_displaybuffer_top);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer_top[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferTop_width, displayBufferTop_height, NN_GX_MEM_FCRAM);
    nngxBindDisplaybuffer(s_displaybuffer_top[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferTop_width, displayBufferTop_height, NN_GX_MEM_FCRAM);

    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxGenDisplaybuffers(2, s_displaybuffer_bottom);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer_bottom[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferBottom_width, displayBufferBottom_height, NN_GX_MEM_FCRAM);
    nngxBindDisplaybuffer(s_displaybuffer_bottom[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferBottom_width, displayBufferBottom_height, NN_GX_MEM_FCRAM);

    nngxActiveDisplay(NN_GX_DISPLAY0_EXT);
    nngxGenDisplaybuffers(2, s_displaybuffer_ext);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer_ext[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferTop_width, displayBufferTop_height, NN_GX_MEM_FCRAM);
    nngxBindDisplaybuffer(s_displaybuffer_ext[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, displayBufferTop_width, displayBufferTop_height, NN_GX_MEM_FCRAM);

    /******************************************************/
    //Initialize the GD API
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);

    /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {displayBufferTop_width, displayBufferTop_height, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, frameBuffer_blockSize, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {displayBufferTop_width, displayBufferTop_height, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, frameBuffer_blockSize, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    InitShader();
    InitObjects();
    Initialize_GD();
    UpdateScene();

    stereo.Initialize();
    nngxSetDisplayMode( NN_GX_DISPLAYMODE_STEREO );

    return 0;
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    if (Initialize() >= 0)
    {
        while (1)
        {
            UpdateScene();
            if (DrawFrame() < 0) break;
        }
    }

    Finalize();

    /* shutdown_display */
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void SetupLights(nn::math::MTX34& view)
{
    switch ( m_LightType )
    {
    case LIGHT_TYPE_SPOT :
        { // Spotlight settings
            f32 r = 20.f;    
            nn::math::VEC3 pos[3] =
            {
                nn::math::VEC3( r*m_Sin, r*m_Cos, 0.f ),
                nn::math::VEC3( r*m_Cos, 0.f, r*m_Sin ),
                nn::math::VEC3( 0.f, r*m_Sin, r*m_Cos ),
            };
            nn::math::VEC3 dir[3] =
            {
                nn::math::VEC3( -m_Sin,  -m_Cos,  0.f ),
                nn::math::VEC3(  -m_Cos, 0.f,  -m_Sin ),
                nn::math::VEC3( 0.f,  -m_Sin,  -m_Cos ),
            };

            // Convert to view coordinates and set the lights
            for ( int i = 0; i < 3; ++i )
            {
                nn::gd::LightingStage::light[i].EnableLight(GD_TRUE);
                nn::gd::LightingStage::light[i].EnableSpotLight(GD_TRUE);
                nn::gd::LightingStage::light[i].EnableDistanceAttenuation(GD_FALSE);

                nn::math::VEC3Transform( &pos[i], &view, &pos[i] );
                f32 X = pos[i].x; f32 Y = pos[i].y; f32 Z = pos[i].z;
                nn::gd::LightingStage::light[i].SetPosition(X, Y, Z);
                nn::gd::LightingStage::light[i].SetLightType(nn::gd::Light::SOURCE_TYPE_POINT);

                nn::math::VEC3TransformNormal( &dir[i], &view, &dir[i] );
                nn::gd::LightingStage::light[i].SetSpotDirection(-dir[i].x, -dir[i].y, -dir[i].z);
            }
        }
        break;
    case LIGHT_TYPE_POINT :
        {
            nn::gd::LightingStage::light[0].EnableLight(GD_FALSE);
            nn::gd::LightingStage::light[1].EnableLight(GD_FALSE);
            nn::gd::LightingStage::light[2].EnableLight(GD_TRUE);
            nn::gd::LightingStage::light[2].EnableSpotLight(GD_FALSE);
            nn::gd::LightingStage::light[2].EnableDistanceAttenuation(GD_TRUE);
            
            nn::math::VEC3 pos( mtx.f._03, mtx.f._13, mtx.f._23 );
            nn::math::VEC3Transform( &pos, &view, &pos );
            nn::gd::LightingStage::light[2].SetPosition(pos.x, pos.y, pos.z);
            nn::gd::LightingStage::light[2].SetLightType(nn::gd::Light::SOURCE_TYPE_POINT);
        }
        break;
    case LIGHT_TYPE_DIRECTIONAL :
        {
            nn::gd::LightingStage::light[0].EnableLight(GD_FALSE);
            
            nn::gd::LightingStage::light[1].EnableLight(GD_TRUE);
            nn::gd::LightingStage::light[1].EnableSpotLight(GD_FALSE);
            nn::math::VEC3 dir( -1.f, -1.f, -1.f );
            nn::math::VEC3TransformNormal( &dir, &view, &dir );

            // Directional light = point light positioned at infinite distance
            // Set (light direction) x (-1) as the position
            nn::gd::LightingStage::light[1].SetPosition(-dir.x, -dir.y, -dir.z);
            nn::gd::LightingStage::light[1].SetLightType(nn::gd::Light::SOURCE_TYPE_DIRECTIONAL);

            nn::gd::LightingStage::light[2].EnableLight(GD_FALSE);
        }
        break;
    }
}

void RenderScene()
{
#ifdef DEMO_ALTERNATE_TEXTURES
    nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2DA);
#endif

    u32 counter = m_Counter % 3000;
    u32 counter_d = counter / 5;

    //Draw the main object (cube or sphere)
#ifdef SHADER_USE_UNMANAGED_UNIFORM
    nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, s_shaderVariable_user.GetRegister(), 3, static_cast<f32*>(mtx));
#else
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_user, static_cast<f32*>(mtx));
#endif

    u32 nbVertexIndexToDraw;
    if (counter < 1500)
    {
        //Draw the cube
        u32 offsets[] = { 0, 0, 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 4, s_buffers_cube, offsets);
        nn::gd::VertexInputStage::SetInputLayout(s_inputLayout_cube);
        nn::gd::VertexInputStage::SetIndexBuffer(s_bufferIndex_cube, nn::gd::VertexInputStage::INDEX_FORMAT_UBYTE, 0);
        nbVertexIndexToDraw = s_cubeVertexCount;
    }
    else
    {
        //Draw the sphere
        u32 offset = 0;
        nn::gd::VertexInputStage::SetVertexBuffers(0, 1, &s_bufferInterleave_sphere, &offset);
        nn::gd::VertexInputStage::SetInputLayout(s_inputLayout_sphere);
        nn::gd::VertexInputStage::SetIndexBuffer(s_bufferIndex_sphere, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
        nbVertexIndexToDraw = s_sphereVertexCount;
    }
    nn::gd::System::DrawIndexed(nbVertexIndexToDraw, 0);

    //Draw 300 more elements
    nn::math::MTX34 mtx34( mtx );
    for ( int i = 0; i < 300; ++i )
    {
        // Select primitive
        if ( counter_d == i )
        {
            //Draw the sphere
            u32 offset = 0;
            nn::gd::VertexInputStage::SetVertexBuffers(0, 1, &s_bufferInterleave_sphere, &offset);
            nn::gd::VertexInputStage::SetInputLayout(s_inputLayout_sphere);
            nn::gd::VertexInputStage::SetIndexBuffer(s_bufferIndex_sphere, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);
            nbVertexIndexToDraw = s_sphereVertexCount;
        }
        else if ( counter_d == 300 + i )
        {
            //Draw the cube
            u32 offsets[] = { 0, 0, 0, 0 };
            nn::gd::VertexInputStage::SetVertexBuffers(0, 4, s_buffers_cube, offsets);
            nn::gd::VertexInputStage::SetInputLayout(s_inputLayout_cube);
            nn::gd::VertexInputStage::SetIndexBuffer(s_bufferIndex_cube, nn::gd::VertexInputStage::INDEX_FORMAT_UBYTE, 0);
            nbVertexIndexToDraw = s_cubeVertexCount;
        }
         
        { 
            // Render primitive (upper part)
            mtx34.f._03 = ( i % 100 / 10 - 4.5f ) * 10.f;
            mtx34.f._13 = 15.f + ( ( 299 - i ) / 100 ) * 10.f;
            mtx34.f._23 = ( i % 10 -  4.5f ) * 10.f;

            // Set model matrix and render
#ifdef SHADER_USE_UNMANAGED_UNIFORM
            nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 8, 3, static_cast<f32*>(mtx34));
#else
            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_user, static_cast<f32*>(mtx34));
#endif

#ifdef DEMO_ALTERNATE_TEXTURES
            nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2DA);
#endif

            // glDrawElements()
            nn::gd::System::DrawIndexed(nbVertexIndexToDraw, 0);
        }

        {
            // Render primitive (lower part)
            mtx34.f._13 = - mtx34.f._13;

            // Set model matrix and render
#ifdef SHADER_USE_UNMANAGED_UNIFORM
            nn::gd::ShaderStage::SetFloatConstantBuffer(nn::gd::ShaderStage::SHADER_TYPE_VERTEX, 8, 3, static_cast<f32*>(mtx34));
#else
            nn::gd::ShaderStage::SetShaderPipelineConstantFloat(s_shaderPipeline, s_shaderVariable_user, static_cast<f32*>(mtx34));
#endif

#ifdef DEMO_ALTERNATE_TEXTURES
            nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_texture2DB);
#endif

            // glDrawElements()
            nn::gd::System::DrawIndexed(nbVertexIndexToDraw, 0);

        }
    }

}
