﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ShadowKnotCube.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/math.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/applet.h>
#include "Util.h"
#include "Loader.h"

#include "demo.h"

#include <string.h>
#include <math.h>

#include "Memory.h"


#include <GLES2/gl2.h>
#include <GLES2/gl2ext.h>
#include <string.h>
#include <cstdlib>
#include <nn/dbg.h>

#define WAIT_VSYNC

#define APP_NAME "ShadowKnotCube"
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/* shadow setting */
static const float fSZBiasScale[] = {1.7f, 1.3f, 1.3f};
/* shadow artifact suppression SETTINGS 0 - lowest 2 - highest */
/* Parameters for suppressing artifacts. The larger the number, the greater the effect of the suppression.*/
#define SDW_SETTINGS 0
/* Shadow texture size. Shadow textures used with omnidirectional shadow maps must be in a square-shape size, the same as the cube map.
 * */
#define SBWIDTHHEIGHT 128


/* 1st pass (shadow accumulation pass) frustum argument */
/* With omnidirectional shadow maps, as with the case of dynamic cube maps, the view angle of the camera used with the 1st path must be 90 degrees.
 * 
 * For this reason, the left(r) and top(t) values for the near face must be the same.*/
static GLfloat n = 0.2f, f = 10.f, r = 0.2f, t = 0.2f;

/* Light position */
static GLfloat lpos0[] = {0.9f, 0.f, 0.f, 1.f};

/* frame number */
static unsigned int s_FrameNo = 0;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000;

demo::RenderSystem s_RenderSystem;

#include <nn/gd.h>

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
static nn::gd::RenderTarget* s_MainRenderTarget = 0;
static nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;

static nn::gd::Texture2DResource* s_texture2DRes[6];
static nn::gd::TextureCube* texCubeShadow;
static nn::gd::SamplerState* sampler;
static nn::gd::CombinerState* CombinerSdwProg=0;
static nn::gd::CombinerState* CombinerLrProg=0;
static nn::gd::CombinerState* CombinerLight=0;
static nn::gd::DepthStencilState* depthStencil_Enabled = 0;
static nn::gd::RenderTarget* s_renderTarget[6];
static nn::gd::BlendState* EnableBlend;
static nn::gd::BlendState* DisableBlend;
static nn::gd::Viewport viewPort;
static nn::gd::ShaderBinary* shaderBinary = NULL;
static nn::gd::ShaderPipeline* SPSdwProgName = NULL;
static nn::gd::ShaderPipeline* SPLrProgName = NULL;
static nn::gd::Shader* s_VSSdwProgName = NULL;
static nn::gd::Shader* s_VSLrProgName = NULL;
static nn::gd::InputLayout* ILSdwProgName = NULL;
static nn::gd::InputLayout* ILLrProgName = NULL;

GLuint s_CmdlistID;

void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}

/* obj loader class object */
typedef struct tag_datGD_t
{
    nn::gd::VertexBufferResource*          posVB;
    nn::gd::VertexBufferResource*          normVB;
    nn::gd::VertexBufferResource*          texVB;
    nn::gd::VertexBufferResource*          tangVB;
    nn::gd::VertexBufferResource*          wghtVB;
    nn::gd::VertexBufferResource*          midxVB;
    nn::gd::VertexBufferResource*          idxVB;
    dat_obj_t*      obj;
    dat_patch_t*    patch;
    int             obj_num;
} datGD_t;

static datGD_t knot, cube;
static datGD_t* objects[] = {&knot, &cube, 0};

void loadDAT_GD(const char* _name, datGD_t* _dat);
void unloadDAT_GD(datGD_t* dat);

// Generates a standard OpenGL projection transformation matrix
nn::math::Matrix44*
MTX44FrustumGl(nn::math::Matrix44* pOut, f32 l, f32 r, f32 b, f32 t, f32 n, f32 f)
{
    f32 (*const m)[4] = pOut->m;
    f32 tmp     =  1.0f / (r - l);
    m[0][0] =  (2*n) * tmp;
    m[0][1] =  0.0f;
    m[0][2] =  (r + l) * tmp;
    m[0][3] =  0.0f;

    tmp     =  1.0f / (t - b);
    m[1][0] =  0.0f;
    m[1][1] =  (2*n) * tmp;
    m[1][2] =  (t + b) * tmp;
    m[1][3] =  0.0f;

    m[2][0] =  0.0f;
    m[2][1] =  0.0f;

    tmp = 1.0f / (n - f);

    m[2][2] = (f + n) * tmp;
    m[2][3] = 2.0f * f * n * tmp;

    m[3][0] =  0.0f;
    m[3][1] =  0.0f;
    m[3][2] = -1.0f;
    m[3][3] =  0.0f;

    return pOut;
}

static void RenderObjectsShadow(datGD_t& _objs)
{
    /* The vertex process with the 1st path only performs vertex coordinate conversion, so the only vertex attribute that is enabled is the vertex coordinate.
     * */

    for (int i = 0; i < _objs.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(ILSdwProgName);

        nn::gd::VertexBufferResource* buffers[] = {  _objs.posVB };
        u32 offsets[] = { 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 1, buffers, offsets);
        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(_objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            //---------------------  gdAPI call ---------------------------
                nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
                nn::gd::System::DrawIndexed(_objs.patch[j].elm_size, 0);
        }
    }
}


static void RenderObjectsShadowed(datGD_t& _objs)
{
    /* Normal vectors are needed to perform fragment lighting in the second pass.
     * The vertex attributes to enable are vertex coordinates and normals.*/

    for (int i = 0; i < _objs.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(ILLrProgName);

        nn::gd::VertexBufferResource* buffers[] = { _objs.posVB, _objs.normVB };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            GLfloat ma[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat md[4] = {0.f, 0.f, 0.f, 1.f};
            GLfloat ms[4] = {0.f, 0.f, 0.f, 1.f};

            for (int col = 0; col < 3; col++)
            {
                ma[col] = _objs.patch[j].ambient[col];
                md[col] = _objs.patch[j].diffuse[col];
                ms[col] = _objs.patch[j].specular[col];
            }

           // dmp_FragmentLightSource
            GLfloat ld0[] = {1.f, 1.f, 1.f, 1.f};  // Diffuse
            GLfloat ls0[] = {0.f, 0.f, 0.f, 0.f};  // specular0
            GLfloat ls1[] = {1.f, 1.f, 1.f, 1.f};  // specular1
            GLfloat la0[] = {1.f, 1.f, 1.f, 1.f};  // Ambient

            // dmp_FragmentMaterial.ambient(Acm)×dmp_FragmentLighting.ambient(Acli)
            // dmp_FragmentMaterial.ambient  ma[4]
            // dmp_FragmentLighting.ambient  la0[4]
            u8 Ambinet[3];
            Ambinet[0] = nn::gd::Utils::FloatToUnsignedByte(ma[0] * la0[0]);
            Ambinet[1] = nn::gd::Utils::FloatToUnsignedByte(ma[1] * la0[1]);
            Ambinet[2] = nn::gd::Utils::FloatToUnsignedByte(ma[2] * la0[2]);
            // Ambient : dmp_FragmentMaterial.ambient×dmp_FragmentLightSource[i].ambient
            nn::gd::LightingStage::light[0].SetColorAmbient(Ambinet[0], Ambinet[1], Ambinet[2]);

            // Diffuse : dmp_FragmentMaterial.diffuse(Dcm)×dmp_FragmentLightSource[i].diffuse(Dcli)
            // dmp_FragmentMaterial.diffuse        md[4]
            // dmp_FragmentLightSource[i].diffuse  ld0[4]

            u8 Diffuse[3];
            Diffuse[0] = nn::gd::Utils::FloatToUnsignedByte(ld0[0] * md[0]);
            Diffuse[1] = nn::gd::Utils::FloatToUnsignedByte(ld0[1] * md[1]);
            Diffuse[2] = nn::gd::Utils::FloatToUnsignedByte(ld0[2] * md[2]);
             nn::gd::LightingStage::light[0].SetColorDiffuse(Diffuse[0], Diffuse[1], Diffuse[2]);

            // Specular0 : dmp_FragmentMaterial.specular0×dmp_FragmentLightSource[i].specular0
            u8 Specular0[3];
            Specular0[0] = nn::gd::Utils::FloatToUnsignedByte(ms[0] * ls0[0]);
            Specular0[1] = nn::gd::Utils::FloatToUnsignedByte(ms[1] * ls0[1]);
            Specular0[2] = nn::gd::Utils::FloatToUnsignedByte(ms[2] * ls0[2]);
            nn::gd::LightingStage::light[0].SetColorSpecular0(Specular0[0], Specular0[1], Specular0[2]);

            // Specular1 : When dmp_LightEnv.lutEnabledRefl is GL_TRUE: dmp_FragmentLightSource[i].specular1
            nn::gd::LightingStage::light[0].SetColorSpecular1(nn::gd::Utils::FloatToUnsignedByte(ls1[0]), nn::gd::Utils::FloatToUnsignedByte(ls1[1]), nn::gd::Utils::FloatToUnsignedByte(ls1[2]));
            nn::gd::VertexInputStage::SetIndexBuffer(_objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            //---------------------  gdAPI call ---------------------------
              nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
             nn::gd::System::DrawIndexed(_objs.patch[j].elm_size, 0);
        }
    }
}



static void RenderShadowBuffer(datGD_t** _objs, nn::math::Vector3 lpos)
{
    nn::math::Matrix44 m;
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos, camUp, target;

    //Set the shader
    nn::gd::ShaderStage::SetShaderPipeline(SPSdwProgName);

    /* When using omnidirectional shadow maps, rendering for six sides is needed in the first pass.
     * Note that because the scene is rendered as though seen from the light position, the viewpoint position is the light position.
     * */

    // rendering +x cube
    /* Arguments 4, 5, 6 of the lookAt function indicate the target position seen from the viewpoint.
     * In other words, lpos[0]+1.f, the fourth argument, indicates that the viewer faces toward the positive x-axis.*/

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -r, r, -t, t, n, f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uProjection"), static_cast<f32*>(proj));

    camPos = nn::math::Vector3(lpos[0], lpos[1], lpos[2]);
    camUp = nn::math::Vector3(0.f, -1.f, 0.f);
    target = nn::math::Vector3(lpos[0] + 1.f, lpos[1], lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    /* Note that although shadow texture is attached to the frame buffer object, a different target is specified for each surface.
     * */

    nn::gd::CombinerStage::SetTextureCombinerState(CombinerSdwProg);
        nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT0, 0, 0, 0, 0);

    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_LINEAR, 0.f, f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_SHADOW_MAP);
#define GD_UTL_CLAMP(x)     (((x) <= 0.f) ? 0.f : ((x) >= 1.f) ? 1.f : (x))
    u8 color[] = {GD_UTL_CLAMP(1.f) * 0xff, GD_UTL_CLAMP(1.f) * 0xff, GD_UTL_CLAMP(1.f) * 0xff, GD_UTL_CLAMP(1.f) * 0xff};

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[0]);
    nn::gd::Memory::ClearTargets(s_renderTarget[0], NULL, color, 1.f, 0);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -x cube
    target = nn::math::Vector3(lpos[0] - 1.f, lpos[1], lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[1]);
    nn::gd::Memory::ClearTargets(s_renderTarget[1], NULL, color, 1.f, 0);

    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering +y cube
    camUp = nn::math::Vector3(0.f, 0.f, 1.f);
    target = nn::math::Vector3(lpos[0], lpos[1] + 1.f, lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[2]);
    nn::gd::Memory::ClearTargets(s_renderTarget[2], NULL, color, 1.f, 0);
    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -y cube
    camUp = nn::math::Vector3(0.f, 0.f, -1.f);
    target = nn::math::Vector3(lpos[0], lpos[1] - 1.f, lpos[2]);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[3]);
    nn::gd::Memory::ClearTargets(s_renderTarget[3], NULL, color, 1.f, 0);
    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering +z cube
    camUp = nn::math::Vector3(0.f, -1.f, 0.f);
    target = nn::math::Vector3(lpos[0], lpos[1], lpos[2] + 1.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[4]);
    nn::gd::Memory::ClearTargets(s_renderTarget[4], NULL, color, 1.f, 0);
    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    // rendering -z cube
    target = nn::math::Vector3(lpos[0], lpos[1], lpos[2] - 1.f);
    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPSdwProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPSdwProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::OutputStage::SetRenderTarget(s_renderTarget[5]);
    nn::gd::Memory::ClearTargets(s_renderTarget[5], NULL, color, 1.f, 0);
    for (int i = 0; _objs[i] != 0; i++)
    {
        RenderObjectsShadow(*_objs[i]);
    }

    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);

}


static void RenderLight(datGD_t& _objs)
{

    /* Renders light source using constant color, and not the fragment lighting results.*/
    nn::gd::CombinerStage::SetTextureCombinerState(CombinerLight);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_NONE);
    nn::gd::OutputStage::SetBlendState(EnableBlend);

    for (int i = 0; i < _objs.obj_num; i++)
    {
        nn::gd::VertexInputStage::SetInputLayout(ILLrProgName);

        nn::gd::VertexBufferResource* buffers[] = { _objs.posVB, _objs.normVB };
        u32 offsets[] = { 0, 0 };
        nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);

        for (unsigned j = _objs.obj[i].patch_offset; j < _objs.obj[i].patch_size + _objs.obj[i].patch_offset; j++)
        {
            nn::gd::VertexInputStage::SetIndexBuffer(_objs.idxVB, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

            //---------------------  gdAPI call ---------------------------
              nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLELIST);
             nn::gd::System::DrawIndexed(_objs.patch[j].elm_size, 3);
        }
    }
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);
    nn::gd::OutputStage::SetBlendState(DisableBlend);
}


static int Setup1stPsssShader()
{
    nn::math::Matrix44 proj;

    nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &s_VSSdwProgName);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSSdwProgName, NULL, &SPSdwProgName);

    nn::gd::InputElementDescription descs[] = { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0};
    u32 strides[] = {3 * sizeof(float)} ;
    nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 1, strides, s_VSSdwProgName, &ILSdwProgName);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}


    static nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &CombinerSdwProg); // create dmp_TexEnv[0] combiner object

    /* Note that when the fragment operations are in shadow accumulation mode, the depth value stored in the shadow texture must not be divided by w.
     * 
     * The camera's 'far' value is set as the scaling coefficient.
     *  For details, see DMP shadows in the programming guide.*/

    return 1;
}

static int Setup2ndPsssShader()
{
    nn::math::Matrix44 proj;

    nn::gd::ShaderStage::CreateShader(shaderBinary, 1, &s_VSLrProgName);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSLrProgName, NULL, &SPLrProgName);

    nn::gd::InputElementDescription descs[] = {
        { 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0},
        { 1, "aNormal", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}};

    u32 strides[] = { 3 * sizeof(float), 3 * sizeof(float) };
    nnResult res = nn::gd::VertexInputStage::CreateInputLayout(descs, 2, strides, s_VSLrProgName, &ILLrProgName);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    static nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::SOURCE_FRAGMENT_PRIMARY_COLOR, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &CombinerLrProg); // create dmp_TexEnv[0] combiner object

    return 1;
}

static int SetupShader()
{
    /* The shader setup process with DMPGL2.0 is performed using the same mechanism as OpenGLES2.0.
     *Because DMPGL2.0 only allows the vertex shader to be user-defined, only a vertex shader object is created.
     * 
     * This sample creates two shader objects because it performes different vertex processing in the first pass and the second pass.
     * */

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);
    DEMO_ASSERT_GL_ERROR();

    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);

    file.Finalize();
    s_AppHeap.Free(buf);

    Setup1stPsssShader();
    Setup2ndPsssShader();

    return 1;
}

static void LoadObjects()
{
    loadDAT_GD( "rom:/resources/knot.dat", &knot);
    loadDAT_GD( "rom:/resources/rcube.dat", &cube);
}

/* unload objects */
static void UnloadObjects(void)
{
    unloadDAT_GD(&knot);
    unloadDAT_GD(&cube);
}



static void Draw1stPass()
{
    nn::gd::RasterizerStage::SetViewport(viewPort);

    RenderShadowBuffer(objects, nn::math::Vector3(lpos0));
}

static void  Draw2ndPass()
{
    nn::math::Matrix34 mv;
    nn::math::Vector3 camPos(
        3.f * sinf(static_cast<f32>(s_FrameNo) * static_cast<f32>(M_PI) / 180.f),
        1.f,
        3.f * cosf(static_cast<f32>(s_FrameNo) * static_cast<f32>(M_PI) / 180.f));
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);

    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);
    
    //Set the viewport
    nn::gd::Viewport viewPorta(0, 0, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT);
    nn::gd::RasterizerStage::SetViewport(viewPorta);

    u8 clearColor[4] = {nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f)};
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

    nn::gd::ShaderStage::SetShaderPipeline(SPLrProgName);

    /* The shadow texture created in the first pass is accessed in the second pass.*/
        // Texture setting

    nn::gd::TextureStage::SetTextureUnit0(*texCubeShadow);
    nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, sampler);

    // Texture combiner settings
    nn::gd::CombinerStage::SetTextureCombinerState(CombinerLrProg);
    nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT0,
          nn::gd::Utils::FloatToUnsignedByte(1.0f), nn::gd::Utils::FloatToUnsignedByte(1.0f), nn::gd::Utils::FloatToUnsignedByte(0.f),  nn::gd::Utils::FloatToUnsignedByte(0.5f));

    // Shadow settings
    f32 shadowBias = n * fSZBiasScale[SDW_SETTINGS] / (f - n);
    nn::gd::TextureStage::SetShadowZBias(shadowBias);

    nn::gd::TextureStage::SetPerspectiveShadow(GD_TRUE);
    nn::gd::OutputStage::SetDepthRangeMode(nn::gd::OutputStage::DEPTHBUFFER_RECIPROCAL_FUNCTION, 0.f, 1.f, 0);
    nn::gd::OutputStage::SetFragOpMode(nn::gd::OutputStage::FRAGMENT_OPERATION_MODE_FRAGMENT_COLOR);

    nn::gd::LightingStage::light[0].EnableLight(GD_TRUE);
    nn::gd::LightingStage::light[0].EnableShadowed(GD_TRUE);
    nn::gd::LightingStage::SetLookUpTableInputValue(nn::gd::LightingStage::LOOKUP_TABLE_ID_RR, nn::gd::LightingStage::INPUT_VALUE_NH);
    nn::gd::LightingStage::EnableLookUpTableReflection(GD_TRUE);
    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_1);
    nn::gd::LightingStage::SetShadowPrimary(GD_TRUE);

    // Upload Lut table at the first time
    static int lightLUT=0;
    if (!lightLUT)
    {
        lightLUT = 1;

        GLfloat lut[512];
        memset(lut, 0, sizeof(lut));

        for (int j = 1; j < 128; j++)
        {
            lut[j] = powf( (float)j / 127.f, 30.f);
            lut[j + 255] = lut[j] - lut[j - 1];
        }
        nn::gd::LightingStage::UploadLookUpTableFloat(nn::gd::LightingStage::LOOKUP_TABLE_UPLOAD_ID_RR, 0, lut, lut+256, 256);
    }

    nn::math::Matrix44 proj;

    nn::math::MTX44Frustum(&proj, -0.1f, 0.1f, -0.1f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH,
                  0.1f * nn::gx::DISPLAY0_HEIGHT / nn::gx::DISPLAY0_WIDTH, 0.2f, 10.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLrProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPLrProgName, "uProjection"), static_cast<f32*>(proj));

    nn::math::MTX34LookAt(&mv, &camPos, &camUp, &target);
    nn::math::Matrix44 m(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLrProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPLrProgName, "uModelView"), static_cast<f32*>(m));

    nn::math::Vector4 vlpos0(lpos0);
    nn::math::Vector4 mv0(mv.m[0][0], mv.m[0][1], mv.m[0][2], mv.m[0][3]);
    nn::math::Vector4 mv1(mv.m[1][0], mv.m[1][1], mv.m[1][2], mv.m[1][3]);
    nn::math::Vector4 mv2(mv.m[2][0], mv.m[2][1], mv.m[2][2], mv.m[2][3]);
    nn::math::Vector4 lpos(nn::math::VEC4Dot(&mv0, &vlpos0), nn::math::VEC4Dot(&mv1, &vlpos0), nn::math::VEC4Dot(&mv2, &vlpos0), vlpos0.w);
    nn::gd::LightingStage::light[0].SetPosition(lpos.x, lpos.y, lpos.z);

    /* The appropriate texture matrix must be set in the second pass. For details, see DMP shadows in the programming guide.
     * */
    nn::math::Matrix34 tex;
    nn::math::Vector3 scale(1.f / (f - n), 1.f / (f - n), 1.f / (f - n));
    nn::math::Vector3 trans(-lpos0[0], -lpos0[1], -lpos0[2]);
    nn::math::MTX34Scale(&tex, &scale);
    nn::math::MTX34MultTranslate(&tex, &tex, &trans);

    m = nn::math::Matrix44(tex);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLrProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPLrProgName, "uTexture"), static_cast<f32*>(m));

    RenderObjectsShadowed(cube);
    RenderObjectsShadowed(knot);

    /* Renders the light source.*/
    trans = nn::math::Vector3(lpos0[0], lpos0[1], lpos0[2]);
    scale = nn::math::Vector3( 0.02f, 0.02f, 0.02f );

    nn::math::MTX34MultTranslate(&mv, &mv, &trans);
    nn::math::MTX34MultScale(&mv, &mv, &scale);
    m = nn::math::Matrix44(mv);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPLrProgName, nn::gd::ShaderStage::GetShaderUniformLocation(SPLrProgName, "uModelView"), static_cast<f32*>(m));

    nn::gd::LightingStage::SetLayerConfiguration(nn::gd::LightingStage::LAYER_CONFIGURATION_0);

    RenderLight(cube);
}

int DrawFrame(void)
{
    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;
    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
    }

    /* Light position */
    lpos0[0] = 0.9f;
    lpos0[1] = 0.5f * sinf(s_FrameNo * 5.f * (float)M_PI / 180.f);
    lpos0[2] = 0.5f * cosf(s_FrameNo * 5.f * (float)M_PI / 180.f);
    lpos0[3] = 1.f;

    /* shadow accumulation pass */
    Draw1stPass();
    /* shadow test pass */
    Draw2ndPass();

    //Transfer framebuffer to display buffer
    int dstAddr;
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]); 
    nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
    nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
        s_texture2DResource_MainColorBuffer,                    //source
        0,                                                      //sourceMipLevelIndex
        0,                                                      //srcOffsetY
        (u8*)dstAddr,                                           //dstAddr
        disp0_Width,                                            //dstWidth
        disp0_Height,                                           //dstHeight
        nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
        nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
        GD_FALSE                                                //yFlip
    );

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (s_FrameNo == 0)
    {
        nngxStartLcdDisplay();
    }
    s_FrameNo++;

    return !glGetError();
}


int Initialize()
{
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */

    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);
    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);
 /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    //---------------------  gdAPI call ---------------------------
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return -1;
     nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);
    //---------------------  gdAPI call ---------------------------

    LoadObjects();

   /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    nnResult res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    if (!SetupShader())
        return -1;

    /* Because a color (1.f, 1.f, 1.f, 1.f) must be used to clear the shadow texture in the first pass, the color buffer's clear color is set as follows.
     * */

    u8 clearColor[4] = {nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f), nn::gd::Utils::FloatToUnsignedByte(1.f)};
     nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, clearColor, 1.f, 0);

    // Create Fragment objects
    nn::gd::DepthStencilStateDescription DSdesc;

    DSdesc.ToDefault();
    DSdesc.m_DepthEnable = GD_TRUE;
    DSdesc.m_DepthFunc = nn::gd::OutputStage::DEPTH_FUNCTION_LESS;
    nn::gd::OutputStage::CreateDepthStencilState(&DSdesc, &depthStencil_Enabled);

    nn::gd::BlendStateDescription blenddesc;

    blenddesc.ToDefault();
    blenddesc.m_BlendType = nn::gd::OutputStage::BLEND_TYPE_BLENDING;
    blenddesc.SetBlendFunc(nn::gd::OutputStage::BLEND_FUNCTION_SRC_ALPHA, nn::gd::OutputStage::BLEND_FUNCTION_ONE_MINUS_SRC_ALPHA, nn::gd::OutputStage::BLEND_EQUATION_ADD);
    nn::gd::OutputStage::CreateBlendState(&blenddesc, &EnableBlend);

    blenddesc.SetBlendMode_NoBlend();
    nn::gd::OutputStage::CreateBlendState(&blenddesc, &DisableBlend);

    // Set Depth test and cull face
    nn::gd::OutputStage::SetDepthStencilState(depthStencil_Enabled, 0);
    nn::gd::RasterizerStage::SetCulling(nn::gd::RasterizerStage::CULLING_CLOCKWISE);

    static nn::gd::Texture2DResourceDescription Text2DResDesc = {SBWIDTHHEIGHT, SBWIDTHHEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_SHADOW, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
//    static nn::gd::RenderTargetDescription descRenderTarget = {0};
    for (unsigned int face = 0; face < 6; face++) {
        res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc, NULL, GD_FALSE, &s_texture2DRes[face]);
        res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DRes[face], &descRenderTarget, &s_renderTarget[face]);
    }

    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::TextureCubeDescription desc(0, 0);
    res = nn::gd::TextureStage::CreateTextureCube(s_texture2DRes, &desc, &texCubeShadow);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.SetShadowCube();
    samplerdesc.m_BorderColor [0] = samplerdesc.m_BorderColor [1] = samplerdesc.m_BorderColor [2] = samplerdesc.m_BorderColor [3] = 255;
    res = nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &sampler);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}


    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &CombinerLight); // create dmp_TexEnv[0] combiner object

    viewPort.m_X = 0;
    viewPort.m_Y = 0;
    viewPort.m_Width = SBWIDTHHEIGHT;
    viewPort.m_Height = SBWIDTHHEIGHT;

    return 0;
}

static void Finalize()
{
    nngxStopCmdlist();

    //Release GD objects
    for (int i = 0; i < 6; i++)
    {
        if (s_texture2DRes[i]) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes[i]);
        if (s_renderTarget[i]) nn::gd::OutputStage::ReleaseRenderTarget(s_renderTarget[i]);
    }
    if (texCubeShadow)   nn::gd::TextureStage::ReleaseTextureCube(texCubeShadow);
    if (sampler)         nn::gd::TextureStage::ReleaseSamplerState(sampler);
    if (CombinerSdwProg) nn::gd::CombinerStage::ReleaseTextureCombinerState(CombinerSdwProg);
    if (CombinerLrProg)  nn::gd::CombinerStage::ReleaseTextureCombinerState(CombinerLrProg);
    if (CombinerLight)   nn::gd::CombinerStage::ReleaseTextureCombinerState(CombinerLight);
    if (depthStencil_Enabled) nn::gd::OutputStage::ReleaseDepthStencilState(depthStencil_Enabled);
    if (EnableBlend)     nn::gd::OutputStage::ReleaseBlendState(EnableBlend);
    if (DisableBlend)    nn::gd::OutputStage::ReleaseBlendState(DisableBlend);   
    if (SPSdwProgName)   nn::gd::ShaderStage::ReleaseShaderPipeline(SPSdwProgName);
    if (SPLrProgName)    nn::gd::ShaderStage::ReleaseShaderPipeline(SPLrProgName);
    if (s_VSSdwProgName) nn::gd::ShaderStage::ReleaseShader(s_VSSdwProgName);
    if (s_VSLrProgName)  nn::gd::ShaderStage::ReleaseShader(s_VSLrProgName);
    if (ILSdwProgName)   nn::gd::VertexInputStage::ReleaseInputLayout(ILSdwProgName);
    if (ILLrProgName)    nn::gd::VertexInputStage::ReleaseInputLayout(ILLrProgName);
    if (shaderBinary)    nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);

    if (s_MainRenderTarget)                         nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer)        nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget)                   nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();
}

void nnMain(void)
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* initialization */
    if (Initialize() >= 0)
    {
        /* Enter loop */
        while (1)
        {
            (void)DrawFrame();
        }
    }

    UnloadObjects();

    Finalize();
    /* shutdown_display */
//    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void loadDAT_GD(const char* _name, datGD_t* _dat)
{
    if (!_name || !_dat)
        return;

    char* fname = 0;
    wchar_t* wfname = 0;
    unsigned char* dat = 0;
    long long int filesize;
    nn::fs::FileReader file;

    do
    {
        fname = (char*)s_AppHeap.Allocate(strlen(_name) + 1);
        if (!fname)
            break;

        int dotpos = 0;
        strcpy(fname, _name);
        for (int i = (int)strlen(fname) - 1; i >= 0; i--)
        {
            if (fname[i] == '.')
            {
                dotpos = i;
                break;
            }
        }
        if (strncmp(&fname[dotpos + 1], "dat", 3) != 0)
            break;

        wfname = static_cast<wchar_t*>(s_AppHeap.Allocate( (strlen(fname)+1)*2 ));
        if (!wfname)
            break;
        (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
        file.Initialize(wfname);
        filesize = file.GetSize();
        dat = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
        if (!dat)
            break;
        filesize = file.Read(dat, filesize);

        dat_header_t* header = (dat_header_t*)dat;
        dat_obj_t* obj = (dat_obj_t*)&dat[sizeof(dat_header_t)];
        dat_patch_t* patch = (dat_patch_t*)&dat[sizeof(dat_header_t) + sizeof(dat_obj_t) * header->obj_num];

        nn::fs::FileReader attribfile;
        unsigned char* attribdata = 0;
        nnResult res;
        nn::gd::VertexBufferResourceDescription desc;
        memset(_dat, 0, sizeof(dat_t));

        if (header->vtx_size)
        {
            memcpy(&fname[dotpos + 1], "vtx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->vtx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->posVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->nrm_size)
        {
            memcpy(&fname[dotpos + 1], "nrm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->nrm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->normVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tex_size)
        {
            memcpy(&fname[dotpos + 1], "tex", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tex_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->texVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->tgt_size)
        {
            memcpy(&fname[dotpos + 1], "tgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->tgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->tangVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->wgt_size)
        {
            memcpy(&fname[dotpos + 1], "wgt", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->wgt_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->wghtVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->idx_size)
        {
            memcpy(&fname[dotpos + 1], "idx", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->idx_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->midxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }
        if (header->elm_size)
        {
            memcpy(&fname[dotpos + 1], "elm", 3);
            (void)std::mbstowcs( wfname, fname, strlen(fname)+1 );
            attribfile.Initialize(wfname);
            filesize = attribfile.GetSize();
            attribdata = static_cast<unsigned char*>(s_AppHeap.Allocate(filesize));
            if (!attribdata)
                break;
            (void)attribfile.Read(attribdata, filesize);
            desc.m_ByteSize = header->elm_size;
            desc.m_MemLocation = nn::gd::Memory::FCRAM;
            res = nn::gd::Resource::CreateVertexBufferResource(&desc, attribdata, &_dat->idxVB);
            if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
            s_AppHeap.Free(attribdata);
            attribfile.Finalize();
        }

        _dat->obj = (dat_obj_t*)s_AppHeap.Allocate(sizeof(dat_obj_t) * header->obj_num);
        _dat->patch = (dat_patch_t*)s_AppHeap.Allocate(sizeof(dat_patch_t) * header->patch_num);
        if (!_dat->obj || !_dat->patch)
        {
            if (_dat->obj)
                s_AppHeap.Free(_dat->obj);
            if (_dat->patch)
                s_AppHeap.Free(_dat->patch);
            break;
        }
        memcpy(_dat->obj, obj, sizeof(dat_obj_t) * header->obj_num);
        memcpy(_dat->patch, patch, sizeof(dat_patch_t) * header->patch_num);
        _dat->obj_num = header->obj_num;
    } while (0);

    if (fname)
        s_AppHeap.Free(fname);
    if (wfname)
        s_AppHeap.Free(wfname);
    if (dat)
        s_AppHeap.Free(dat);
    file.Finalize();

    return;
}

void unloadDAT_GD(datGD_t* dat)
{
    if (dat)
    {
        if (dat->posVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->posVB);
        if (dat->normVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->normVB);
        if (dat->texVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->texVB);
        if (dat->tangVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->tangVB);
        if (dat->wghtVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->wghtVB);
        if (dat->midxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->midxVB);
        if (dat->idxVB)
            nn::gd::Resource::ReleaseVertexBufferResource(dat->idxVB);
        if (dat->obj)
            s_AppHeap.Free(dat->obj);
        if (dat->patch)
            s_AppHeap.Free(dat->patch);
    }

    return;
}

