﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     TextureCombinerSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/os.h>
#include <nn/init.h>
#include <nn/applet.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include "Memory.h"
#include <nn/gd.h>

#define WAIT_VSYNC

#include <string.h>
#include "demo.h"

/* buffer id */
nn::gd::VertexBufferResource* s_coords;
nn::gd::VertexBufferResource* s_texcoords;
nn::gd::VertexBufferResource* s_idxs;

/* program id */
static nn::gd::ShaderBinary*   shaderBinary = NULL;
static nn::gd::Shader*   s_VSProgramID = NULL;
static nn::gd::ShaderPipeline* SPProgramID = NULL;
static nn::gd::InputLayout*    ILProgramID = NULL;

static nn::gd::DepthStencilState* depthTestEnable = 0;

/* ExpHeap for app. */
nn::fnd::ExpHeap s_AppHeap;
uptr s_HeapForGx;
uptr s_HeapForMalloc;
const u32 s_GxHeapSize = 0x400000;
const u32 s_HeapSize = 0x200000*2;

// demo::RenderSystem s_RenderSystem;

extern u8 g_TexBitmapUpper[];
extern u8 g_TexBitmapLower[];
extern u8 g_TexAlpha[];
extern u8 g_TexGray[];

static nn::gd::Texture2D*      s_Texture;
static nn::gd::Texture2DResource* s_texture2DRes;
static nn::gd::SamplerState* sampler;

static u8 color[4];
#define PRINT(__msg) NN_LOG __msg

//Display buffers
GLuint s_displaybuffer0[2];
GLuint s_displaybuffer1[2];
GLint s_CurrentDisplaybuffer0 = 0;
GLint s_CurrentDisplaybuffer1 = 0;

//Framebuffers
static nn::gd::Texture2DResource* s_texture2DResource_MainColorBuffer = 0;
static nn::gd::Texture2DResource* s_texture2DResource_MainDepthStencilBuffer = 0;
nn::gd::RenderTarget* s_MainRenderTarget = 0;
nn::gd::DepthStencilTarget* s_MainDepthStencilTarget = 0;
nn::gd::TargetProperties mainRenderTargetProperty;

GLuint s_CmdlistID;
//GD API callback function
void callbackErrorFunction(nnResult result, const char* funcname)
{
    char* str = 0;
    NN_UNUSED_VAR(str);
    NN_UNUSED_VAR(funcname);

    if (funcname != NULL) {NN_LOG("Return from \"%s\"\n",  funcname);}
    if ((str = nn::gd::System::GetErrorStringFromResult(result)) != NULL) {NN_LOG("Error: %s\n", str);}

    NN_ASSERTMSG( 0, "callBack Error function called");
}


nn::gd::CombinerState *texCom0;
nn::gd::CombinerState *texCom1;
nn::gd::CombinerState *texCom2;
nn::gd::CombinerState *texCom3;
nn::gd::CombinerState *texCom4;
nn::gd::CombinerState *texCom5;

/* generate simple object */
void LoadObjects(void)
{
    GLfloat coords[] = {
        -0.5f, 0.5f, 0.f, 1.f,
        -0.5f,-0.5f, 0.f, 1.f,
         0.5f, 0.5f, 0.f, 1.f,
         0.5f,-0.5f, 0.f, 1.f
    };
    GLfloat texcoords[] = {
        0.f, 1.f, 0.f,
        0.f, 0.f, 0.f,
        1.f, 1.f, 0.f,
        1.f, 0.f, 0.f,
    };

    GLushort idxs[] = {0, 1, 2, 3};

    nnResult res;

    nn::gd::VertexBufferResourceDescription desc;

    desc.m_ByteSize = 4 * sizeof(coords);
    desc.m_MemLocation = nn::gd::Memory::FCRAM;
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, coords, &s_coords);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = 3 * sizeof(texcoords);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, texcoords, &s_texcoords);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    desc.m_ByteSize = 4 * sizeof(idxs);
    res = nn::gd::Resource::CreateVertexBufferResource(&desc, idxs, &s_idxs);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}
}

/* initialization */
static int Initialize(void)
{
    /* Initialize display */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));

    /* Initialize display */
    demo::InitializeMemoryManager(s_HeapForGx, s_GxHeapSize);

    /* Create heap for malloc*/
    s_HeapForMalloc = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_HeapSize));
    setMemoryHeap(s_HeapForMalloc, s_HeapSize);

    nngxInitialize(demo::GetAllocator, demo::GetDeallocator);
    
    /******************************************************/
    // Initialize command list
    nngxGenCmdlists(1, &s_CmdlistID);
    nngxBindCmdlist(s_CmdlistID);
    nngxCmdlistStorage(0x40000, 128);

    //We run the command list now so that any new request that is inserted can be executed in parallel
    nngxRunCmdlist();

    /******************************************************/
    // Init display buffers
    nngxGenDisplaybuffers(2, s_displaybuffer0);
    nngxActiveDisplay(NN_GX_DISPLAY0);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer0[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer0[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, NN_GX_MEM_FCRAM);

    nngxGenDisplaybuffers(2, s_displaybuffer1);
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxDisplayEnv(0, 0);
    nngxBindDisplaybuffer(s_displaybuffer1[0]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    nngxBindDisplaybuffer(s_displaybuffer1[1]);
    nngxDisplaybufferStorage(GL_RGB8_OES, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT, NN_GX_MEM_FCRAM);

    //---------------------  gdAPI call ---------------------------
    nnResult result = nn::gd::System::Initialize();
    if ( GD_NNRESULT_IS_FAILURE(result) ) return 0;

    nn::gd::System::SetCallbackFunctionError(callbackErrorFunction);
    //---------------------  gdAPI call ---------------------------

 /******************************************************/
    // Initialize the main framebuffers
    //Color buffer
    nnResult res;

    nn::gd::Texture2DResourceDescription Text2DResDesc_ColorBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_RGBA_8888, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
     res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_ColorBuffer, 0, GD_FALSE, &s_texture2DResource_MainColorBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::RenderTargetDescription descRenderTarget = {0};
    res = nn::gd::OutputStage::CreateRenderTarget(s_texture2DResource_MainColorBuffer, &descRenderTarget, &s_MainRenderTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    //Depth/stencil buffer
    nn::gd::Texture2DResourceDescription Text2DResDesc_DepthStencilBuffer = {nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT, 1, nn::gd::Resource::NATIVE_FORMAT_DEPTH_24_STENCIL_8, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMB};
    res = nn::gd::Resource::CreateTexture2DResource(&Text2DResDesc_DepthStencilBuffer, 0, GD_FALSE, &s_texture2DResource_MainDepthStencilBuffer);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilTargetDescription descDepthStencilTarget = {0};
    res = nn::gd::OutputStage::CreateDepthStencilTarget(s_texture2DResource_MainDepthStencilBuffer, &descDepthStencilTarget, &s_MainDepthStencilTarget);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    nn::gd::ShaderStage::CreateShaderBinary(buf, read, &shaderBinary);
    file.Finalize();
    s_AppHeap.Free(buf);

    nn::gd::ShaderStage::CreateShader(shaderBinary, 0, &s_VSProgramID);
    nn::gd::ShaderStage::CreateShaderPipeline(s_VSProgramID, NULL, &SPProgramID);

    nn::gd::InputElementDescription descs1[] = {{ 0, "aPosition", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 4, 0},
                                               { 1, "aTexCoord", nn::gd::VertexInputStage::STREAM_TYPE_FLOAT, 3, 0}};
    u32 strides1[] = {4 * sizeof(float), 3 * sizeof(float)} ;
    res = nn::gd::VertexInputStage::CreateInputLayout(descs1, 2, strides1, s_VSProgramID, &ILProgramID);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("\n");}

    nn::gd::DepthStencilStateDescription desc;
    desc.ToDefault();
    desc.m_DepthEnable = GD_TRUE;
    nn::gd::OutputStage::CreateDepthStencilState(&desc, &depthTestEnable);

    LoadObjects();

    // Texture combiner
    nn::gd::CombinerDescription combiner;

    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT0,    nn::gd::CombinerStage::SOURCE_CONSTANT,         nn::gd::CombinerStage::SOURCE_CONSTANT,         nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT0,  nn::gd::CombinerStage::SOURCE_CONSTANT,         nn::gd::CombinerStage::SOURCE_CONSTANT,         nn::gd::CombinerStage::SOURCE_CONSTANT);
    combiner.SetOperandRGB(nn::gd::CombinerStage::UNIT0,   nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR,   nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR,   nn::gd::CombinerStage::OPERAND_RGB_SRC_COLOR);
    combiner.SetOperandAlpha(nn::gd::CombinerStage::UNIT0, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA, nn::gd::CombinerStage::OPERAND_ALPHA_SRC_ALPHA);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT0, GD_TRUE);
    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom0); // create dmp_TexEnv[0] combiner object

    nn::gd::Resource::NativeFormat nativeFormat;
    res = nn::gd::Resource::Helper::ConvertTextureResourceToNativeFormat(nn::gd::Resource::FORMAT_RGB_888, 32, 32, (u8*)&g_TexBitmapUpper, (u8*)&g_TexBitmapUpper, &nativeFormat);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DResourceDescription ResDesc = {32, 32, 1, nativeFormat, nn::gd::Memory::LAYOUT_BLOCK_8, nn::gd::Memory::VRAMA};
    res = nn::gd::Resource::CreateTexture2DResource(&ResDesc, &g_TexBitmapUpper, GD_FALSE, &s_texture2DRes);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::Texture2DDescription Texdesc = {0, 0};
    res = nn::gd::TextureStage::CreateTexture2D(s_texture2DRes, &Texdesc, &s_Texture);
    if ( GD_NNRESULT_IS_FAILURE(res) ) {NN_PANIC("PANIC!!\n");}

    nn::gd::SamplerStateDescription samplerdesc;
    samplerdesc.ToDefault();
    samplerdesc.m_MinFilter = nn::gd::TextureStage::SAMPLER_MIN_FILTER_LINEAR;
    samplerdesc.m_MagFilter = nn::gd::TextureStage::SAMPLER_MAG_FILTER_LINEAR;
    nn::gd::TextureStage::CreateSamplerState(&samplerdesc, &sampler);

    // texture combiner 1
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT1,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS_BUFFER, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT1, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS,        nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT1,  nn::gd::CombinerStage::COMBINE_RGB_ADD);
    combiner.m_BufferColor[0] = nn::gd::Utils::FloatToUnsignedByte(0.7f);
    combiner.m_BufferColor[1] = nn::gd::Utils::FloatToUnsignedByte(1.0f);
    combiner.m_BufferColor[2] = nn::gd::Utils::FloatToUnsignedByte(0.0f);
    combiner.m_BufferColor[3] = nn::gd::Utils::FloatToUnsignedByte(0.0f);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT1, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom1);

    // texture combiner 2
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT2,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_CONSTANT, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT2, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT2,  nn::gd::CombinerStage::COMBINE_RGB_ADD_SIGNED);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT2, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom2);

    // texture combiner 3
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT1,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS_BUFFER, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT1, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT1,  nn::gd::CombinerStage::COMBINE_RGB_ADD);

    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT2,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT2, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT2,  nn::gd::CombinerStage::COMBINE_RGB_MODULATE);

    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT3,   nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS_BUFFER, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT3, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS_BUFFER, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT3,  nn::gd::CombinerStage::COMBINE_RGB_ADD_SIGNED);
    combiner.m_BufferColor[0] = nn::gd::Utils::FloatToUnsignedByte(0.7f);
    combiner.m_BufferColor[1] = nn::gd::Utils::FloatToUnsignedByte(1.0f);
    combiner.m_BufferColor[2] = nn::gd::Utils::FloatToUnsignedByte(0.0f);
    combiner.m_BufferColor[3] = nn::gd::Utils::FloatToUnsignedByte(0.0f);

    combiner.m_BufferInput.m_Rgb2 = nn::gd::CombinerStage::INPUT_PREVIOUS_UNIT;
    combiner.m_BufferInput.m_Alpha2 = nn::gd::CombinerStage::INPUT_PREVIOUS_UNIT;

    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT1, GD_TRUE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT2, GD_TRUE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT3, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom3);

    // texture combiner 4
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT4,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT4, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT4,  nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.SetScaleRGB(nn::gd::CombinerStage::UNIT4,    nn::gd::CombinerStage::SCALE_2);
    combiner.SetScaleAlpha(nn::gd::CombinerStage::UNIT4,  nn::gd::CombinerStage::SCALE_2);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT4, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom4);

    // texture combiner 5
    combiner.ToDefault();
    combiner.SetSourceRGB(nn::gd::CombinerStage::UNIT5,   nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetSourceAlpha(nn::gd::CombinerStage::UNIT5, nn::gd::CombinerStage::SOURCE_TEXTURE0, nn::gd::CombinerStage::SOURCE_PREVIOUS, nn::gd::CombinerStage::SOURCE_PREVIOUS);
    combiner.SetCombineRGB(nn::gd::CombinerStage::UNIT5,  nn::gd::CombinerStage::COMBINE_RGB_REPLACE);
    combiner.SetCombinerInUse(nn::gd::CombinerStage::UNIT5, GD_TRUE);

    nn::gd::CombinerStage::CreateTextureCombinerState(&combiner, &texCom5);

   #define GD_UTL_CLAMP(x)     (((x) <= 0.f) ? 0.f : ((x) >= 1.f) ? 1.f : (x))
   color[0] = GD_UTL_CLAMP(0.36f) * 0xff;
   color[1] = GD_UTL_CLAMP(0.42f) * 0xff;
   color[2] = GD_UTL_CLAMP(0.5f) * 0xff;
   color[3] = GD_UTL_CLAMP(1.f) * 0xff;

   return 0;
}

static void Finalize()
{ 
    nngxStopCmdlist();

    if (texCom0)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom0);
    if (texCom1)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom1);
    if (texCom2)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom2);
    if (texCom3)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom3);
    if (texCom4)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom4);
    if (texCom5)         nn::gd::CombinerStage::ReleaseTextureCombinerState(texCom5);
    if (s_coords)        nn::gd::Resource::ReleaseVertexBufferResource(s_coords);
    if (s_texcoords)     nn::gd::Resource::ReleaseVertexBufferResource(s_texcoords);
    if (s_idxs)          nn::gd::Resource::ReleaseVertexBufferResource(s_idxs);
    if (ILProgramID)     nn::gd::VertexInputStage::ReleaseInputLayout(ILProgramID);
    if (SPProgramID)     nn::gd::ShaderStage::ReleaseShaderPipeline(SPProgramID);
    if (s_VSProgramID)   nn::gd::ShaderStage::ReleaseShader(s_VSProgramID);
    if (shaderBinary)    nn::gd::ShaderStage::ReleaseShaderBinary(shaderBinary);
    if (depthTestEnable) nn::gd::OutputStage::ReleaseDepthStencilState(depthTestEnable);
    if (s_texture2DRes)  nn::gd::Resource::ReleaseTexture2DResource(s_texture2DRes);
    if (s_Texture)       nn::gd::TextureStage::ReleaseTexture2D(s_Texture);
    if (sampler)         nn::gd::TextureStage::ReleaseSamplerState(sampler);

    if (s_MainRenderTarget) nn::gd::OutputStage::ReleaseRenderTarget(s_MainRenderTarget);
    if (s_texture2DResource_MainColorBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainColorBuffer);
    if (s_MainDepthStencilTarget) nn::gd::OutputStage::ReleaseDepthStencilTarget(s_MainDepthStencilTarget);
    if (s_texture2DResource_MainDepthStencilBuffer) nn::gd::Resource::ReleaseTexture2DResource(s_texture2DResource_MainDepthStencilBuffer);

    //Release GD API
    nn::gd::System::Finalize();

    //Release display buffers
    nngxBindDisplaybuffer(0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer0);
    nngxDeleteDisplaybuffers(2, s_displaybuffer1);
    s_displaybuffer0[0] = s_displaybuffer0[1] = 0;
    s_displaybuffer1[0] = s_displaybuffer1[1] = 0;

    //Release command list
    nngxDeleteCmdlists(1, &s_CmdlistID);

    // Finalize DMPGL
    nngxFinalize();

    //Release memory
    demo::FinalizeMemoryManager();



}

int DrawFrame(void)
{
    static int f = 0;
    static int cnt = 0;

    //Get display and framebuffer info
    static int disp0_Format = 0, disp0_Width = 0, disp0_Height = 0;
    static int disp1_Format = 0, disp1_Width = 0, disp1_Height = 0;

    if (disp0_Width == 0)
    {   //Done only once
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp0_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp0_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp0_Height);
        nngxActiveDisplay(NN_GX_DISPLAY1);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_FORMAT, &disp1_Format);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_WIDTH, &disp1_Width);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_HEIGHT, &disp1_Height);
        nn::gd::OutputStage::GetRenderTargetProperties(s_MainRenderTarget, &mainRenderTargetProperty);

        nn::gd::ShaderStage::SetShaderPipeline(SPProgramID);
        nn::gd::OutputStage::SetDepthStencilState(depthTestEnable, 0);
        nn::gd::TextureStage::SetTexture(nn::gd::TextureStage::TEXTURE_UNIT_0, *s_Texture);
        nn::gd::TextureStage::SetSamplerState(nn::gd::TextureStage::TEXTURE_UNIT_0, sampler);
    }

    nn::gd::OutputStage::SetRenderTarget(s_MainRenderTarget);
    nn::gd::OutputStage::SetDepthStencilTarget(s_MainDepthStencilTarget);
    nn::gd::Memory::ClearTargets(s_MainRenderTarget, s_MainDepthStencilTarget, color, 1.f, 0);
    nn::gd::Viewport viewPort(0, 0, mainRenderTargetProperty.m_Width, mainRenderTargetProperty.m_Height);
    nn::gd::RasterizerStage::SetViewport(viewPort);

    nn::math::Matrix44 proj;

    nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f* static_cast<f32>(nn::gx::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::DISPLAY0_WIDTH),
                 0.02f* static_cast<f32>(nn::gx::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::DISPLAY0_WIDTH), 0.2f, 10.f);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uProjection"), static_cast<f32*>(proj));

    nn::math::Matrix34 eye;
    nn::math::Vector3 camPos(0.f, 0.f, 5.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);

    nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);
    nn::math::Matrix44 mv(eye);
    nn::gd::ShaderStage::SetShaderPipelineConstantFloat(SPProgramID, nn::gd::ShaderStage::GetShaderUniformLocation(SPProgramID, "uModelView"), static_cast<f32*>(mv));

    if (cnt == 0)  {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom0);
        nn::gd::CombinerStage::SetTextureCombinerUnitConstantColor(nn::gd::CombinerStage::UNIT0, nn::gd::Utils::FloatToUnsignedByte(0.0f), nn::gd::Utils::FloatToUnsignedByte(1.0f), nn::gd::Utils::FloatToUnsignedByte(0.7f),  nn::gd::Utils::FloatToUnsignedByte(0.0f));
    }
    else if (cnt == 1)  {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom1);
    }
    else if (cnt == 2) {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom2);
    }
    else if (cnt == 3)  {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom3);
    }
    else if (cnt == 4)  {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom4);
    }
    if (cnt == 5)  {
        nn::gd::CombinerStage::SetTextureCombinerState(texCom5);
    }

    if (f % 200 == 0) cnt++;
    if (cnt > 5) cnt = 1;

    nn::gd::VertexInputStage::SetInputLayout(ILProgramID);

    nn::gd::VertexBufferResource* buffers[] = { s_coords,  s_texcoords};
    u32 offsets[] = { 0, 0 };
    nn::gd::VertexInputStage::SetVertexBuffers(0, 2, buffers, offsets);
    nn::gd::VertexInputStage::SetIndexBuffer(s_idxs, nn::gd::VertexInputStage::INDEX_FORMAT_USHORT, 0);

    nn::gd::VertexInputStage::SetPrimitiveTopology(nn::gd::VertexInputStage::PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
    nn::gd::System::DrawIndexed(4, 0);

    /* swap buffer */
    //Execute the 3d command buffer (add the buffer into the command list)
       // nngxSplitDrawCmdlist();

        //Transfer framebuffer to display buffer
        int dstAddr;
        nngxActiveDisplay(NN_GX_DISPLAY0);
        nngxBindDisplaybuffer(s_displaybuffer0[s_CurrentDisplaybuffer0]);
        nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &dstAddr);
        nn::gd::Memory::CopyTexture2DResourceBlockToLinear(
            s_texture2DResource_MainColorBuffer,                    //source
            0,                                                      //sourceMipLevelIndex
            0,                                                      //srcOffsetY
            (u8*)dstAddr,                                           //dstAddr
            disp0_Width,                                            //dstWidth
            disp0_Height,                                           //dstHeight
            nn::gd::Resource::NATIVE_FORMAT_RGB_888,                //dstFormat
            nn::gd::Memory::DOWNSCALING_NONE,                       //DownScalingMode
            GD_FALSE                                                //yFlip
        );

    // Wait until all command requests are processed
    nngxWaitCmdlistDone();

    // Swap the display buffers
    nngxActiveDisplay(NN_GX_DISPLAY1);
    nngxBindDisplaybuffer(s_displaybuffer1[s_CurrentDisplaybuffer1]);
    nngxSwapBuffers(NN_GX_DISPLAY_BOTH);
    s_CurrentDisplaybuffer0 = (s_CurrentDisplaybuffer0 == 0 ? 1 : 0);
    s_CurrentDisplaybuffer1 = (s_CurrentDisplaybuffer1 == 0 ? 1 : 0);

#ifdef WAIT_VSYNC
    nngxWaitVSync(NN_GX_DISPLAY_BOTH);
#endif

    //Clear the command list
    nngxClearCmdlist();
    nngxRunCmdlist();

    if (f == 0)
    {
        nngxStartLcdDisplay();
    }

    f++;
    return !glGetError();
}


#if 1 // debug time print
nn::os::Tick s_StartTick;
nn::os::Tick s_EndTick;
f32 tmp=0;

#define MEASURSPEED_GETICK s_StartTick = nn::os::Tick::GetSystemCurrent();
#define MEASURSPEED_MICROTIME(a) {  \
     nn::fnd::TimeSpan TimeSpan; \
     f32 Times; \
     s_EndTick = nn::os::Tick::GetSystemCurrent(); \
     TimeSpan = s_EndTick - s_StartTick; \
     Times = TimeSpan.GetMicroSeconds(); \
     NN_UNUSED_VAR(Times); \
     a += Times; \
}
#endif

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = (1024 * 64) * 2;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* initialization */
    if (Initialize() >= 0)
    {
        while (1)
        {
            (void)DrawFrame();
        }
    }

    /* shutdown_display */
    Finalize();

    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForMalloc));
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}


