﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_DrawPrimitiveScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_DRAW_PRIMITVE_SCENE_H_
#define GR_DRAW_PRIMITVE_SCENE_H_

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_PrimitiveVertexStream.h"

class IAllocator;

//------------------------------------------------------------------------------
class DrawPrimitiveScene : public IScene
{
public:
    DrawPrimitiveScene( void );
    virtual ~DrawPrimitiveScene();
    
public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void    UpdateScene_( void );
    void    RenderSceneGr_( const nn::math::MTX44& proj, const nn::math::MTX34& view );

public:
    IAllocator*                     allocator;

    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx;
    nn::math::MTX34                 viewMtx;
    nn::ulcd::StereoCamera          stereo;

    u8*                             texture;

    nn::gr::BindSymbolVSFloat       bindSymbolWorldMtx;
    nn::gr::BindSymbolVSFloat       bindSymbolProjMtx;
    nn::gr::BindSymbolVSFloat       bindSymbolViewMtx;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];
    
    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;

    PrimitiveBox*                   m_PrimitiveBox;
    nn::gr::Vertex                  m_VertexBox;
    nn::gr::Vertex::IndexStream     m_IndexStreamBox;
        
    PrimitiveInterleavedSphere *    m_PrimitiveInterleavedSphere;
    nn::gr::Vertex                  m_VertexSphere;
    nn::gr::Vertex::IndexStream     m_IndexStreamSphere;
    
    nn::gr::FragmentLight           m_Light;

    nn::gr::LookUpTable             m_SpotTable;
    nn::gr::LookUpTable             m_SpecularTable;
    nn::gr::LookUpTable             m_DistAttnTable;

    nn::gr::Fog                     m_Fog;
    
private :
    u32                             m_Counter;
    f32                             m_Sin, m_Cos;
    f32                             m_Sin2, m_Cos2;
    f32                             m_Sin01, m_Cos01;

    enum LightType
    {
        LIGHT_TYPE_SPOT,
        LIGHT_TYPE_POINT,
        LIGHT_TYPE_DIRECTIONAL,
        LIGHT_TYPE_NUM
    };
    
    u32                             m_LightType;
};

#endif // GR_DRAW_PRIMITVE_SCENE_H_
