/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PrimitiveVertexStream.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include "gr_PrimitiveVertexStream.h"

namespace
{
    // Vertex stream (Position)
    const f32 s_BoxVtxStreamPos[ PrimitiveBox::VERTEX_NUM * 3 ] =
    {
        -0.5f,  0.5f,  0.5f,   -0.5f,  0.5f, -0.5f,   -0.5f, -0.5f, -0.5f,
        -0.5f, -0.5f,  0.5f,   -0.5f,  0.5f,  0.5f,   -0.5f, -0.5f, -0.5f,
         0.5f,  0.5f,  0.5f,   -0.5f,  0.5f,  0.5f,   -0.5f, -0.5f,  0.5f,
         0.5f, -0.5f,  0.5f,    0.5f,  0.5f,  0.5f,   -0.5f, -0.5f,  0.5f,
         0.5f,  0.5f, -0.5f,    0.5f,  0.5f,  0.5f,    0.5f, -0.5f, -0.5f,
         0.5f, -0.5f, -0.5f,    0.5f,  0.5f,  0.5f,    0.5f, -0.5f,  0.5f,
        -0.5f,  0.5f, -0.5f,    0.5f,  0.5f, -0.5f,   -0.5f, -0.5f, -0.5f,
         0.5f,  0.5f, -0.5f,    0.5f, -0.5f, -0.5f,   -0.5f, -0.5f, -0.5f,
        -0.5f, -0.5f, -0.5f,    0.5f, -0.5f, -0.5f,   -0.5f, -0.5f,  0.5f,
        -0.5f, -0.5f,  0.5f,    0.5f, -0.5f, -0.5f,    0.5f, -0.5f,  0.5f,
        -0.5f,  0.5f, -0.5f,   -0.5f,  0.5f,  0.5f,    0.5f,  0.5f, -0.5f,
        -0.5f,  0.5f,  0.5f,    0.5f,  0.5f,  0.5f,    0.5f,  0.5f, -0.5f,
     };

    // Vertex stream (TexCoord0)
    const f32 s_BoxVtxStreamTex[ PrimitiveBox::VERTEX_NUM * 2 ] =
    {
        1.f, 1.f,  1.f, 0.f,  0.f, 0.f,
        0.f, 1.f,  1.f, 1.f,  0.f, 0.f,
        1.f, 1.f,  0.f, 1.f,  0.f, 0.f,
        1.f, 0.f,  1.f, 1.f,  0.f, 0.f,
        1.f, 0.f,  1.f, 1.f,  0.f, 0.f,
        0.f, 0.f,  1.f, 1.f,  0.f, 1.f,
        0.f, 1.f,  1.f, 1.f,  0.f, 0.f,
        1.f, 1.f,  1.f, 0.f,  0.f, 0.f,
        0.f, 0.f,  1.f, 0.f,  0.f, 1.f,
        0.f, 1.f,  1.f, 0.f,  1.f, 1.f,
        0.f, 0.f,  0.f, 1.f,  1.f, 0.f,
        0.f, 1.f,  1.f, 1.f,  1.f, 0.f,
    };

    // Vertex stream (Normal)
    const f32 s_BoxVtxStreamNrm[ PrimitiveBox::VERTEX_NUM * 3 ] =
    {
        -1.f,  0.f,  0.f,  -1.f,  0.f,  0.f,  -1.f,  0.f,  0.f,
        -1.f,  0.f,  0.f,  -1.f,  0.f,  0.f,  -1.f,  0.f,  0.f,
         0.f,  0.f,  1.f,   0.f,  0.f,  1.f,   0.f,  0.f,  1.f,
         0.f,  0.f,  1.f,   0.f,  0.f,  1.f,   0.f,  0.f,  1.f,
         1.f,  0.f,  0.f,   1.f,  0.f,  0.f,   1.f,  0.f,  0.f,
         1.f,  0.f,  0.f,   1.f,  0.f,  0.f,   1.f,  0.f,  0.f,
         0.f,  0.f, -1.f,   0.f,  0.f, -1.f,   0.f,  0.f, -1.f,
         0.f,  0.f, -1.f,   0.f,  0.f, -1.f,   0.f,  0.f, -1.f,
         0.f, -1.f,  0.f,   0.f, -1.f,  0.f,   0.f, -1.f,  0.f,
         0.f, -1.f,  0.f,   0.f, -1.f,  0.f,   0.f, -1.f,  0.f,
         0.f,  1.f,  0.f,   0.f,  1.f,  0.f,   0.f,  1.f,  0.f,
         0.f,  1.f,  0.f,   0.f,  1.f,  0.f,   0.f,  1.f,  0.f,
    } ;

     // Vertex stream (Color)
    const f32 s_BoxVtxStreamClr[ PrimitiveBox::VERTEX_NUM * 4 ] =
    {
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,
        1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f,  1.0f, 1.0f, 1.0f, 1.0f
    } ;
        
    // hexahedron vertex index
    const u8 s_BoxVtxStreamIdx[ PrimitiveBox::VERTEX_NUM * 1 ] =
    {
         0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,
        12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,
        24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35
    };
}


//------------------------------------------------------------------------------
void PrimitiveBox::SetupVtxStream()
{
    std::memcpy( streamPos, s_BoxVtxStreamPos, sizeof( s_BoxVtxStreamPos ) );
    std::memcpy( streamTex, s_BoxVtxStreamTex, sizeof( s_BoxVtxStreamTex ) );
    std::memcpy( streamNrm, s_BoxVtxStreamNrm, sizeof( s_BoxVtxStreamNrm ) );
    std::memcpy( streamClr, s_BoxVtxStreamClr, sizeof( s_BoxVtxStreamClr ) );
    std::memcpy( streamIdx, s_BoxVtxStreamIdx, sizeof( s_BoxVtxStreamIdx ) );

    nngxUpdateBuffer( this, sizeof( PrimitiveBox ));
}
