﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_Common.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_COMMON_H_
#define GR_COMMON_H_

//------------------------------------------------------------------------------
//
// Macro definitions
//

#define PURE                            = 0

// Simple singleton
#define SMP_SINGLETON_DEF( name )      SimpleSingleton##name
#define SMP_SINGLETON_GET( name )      SimpleSingleton##name::Get()
#define SMP_SINGLETON_IMPL( name )     name SMP_SINGLETON_DEF( name )::m_Instance

#define SMP_SINGLETON_DECL( name )                  \
    struct SimpleSingleton##name                    \
    {                                               \
    public:                                         \
        static name *Get() { return &m_Instance;  } \
    private:                                        \
        static name  m_Instance;                    \
    }

// Simple singleton accessor
#define MEMMGR                          SMP_SINGLETON_GET( MemoryMgr )



//------------------------------------------------------------------------------
//
// Enumerated type
//

// Definition that represents screen types
enum ScreenKind
{
    UPPER_SCREEN        = 0x1 << 0,   //
    LOWER_SCREEN        = 0x1 << 1,   //
    EXTENSION_SCREEN    = 0x1 << 2,   //
    BOTH_SCREENS        = UPPER_SCREEN | LOWER_SCREEN | EXTENSION_SCREEN
};

// Definitions that represent transfer modes of the display buffer
enum BufferTransferMode
{
    TRANSFER_MODE_ANTIALIASE_NOT_USED,  //
    TRANSFER_MODE_ANTIALIASE_2x1,       //
    TRANSFER_MODE_ANTIALIASE_2x2        //
};

// Definitions that represent formats of the display buffer
enum DisplayBufferFormat
{
    DISPLAY_BUFFER_RGB8_OES,
    DISPLAY_BUFFER_RGBA4,
    DISPLAY_BUFFER_RGB5_A1,
    DISPLAY_BUFFER_RGB565
};

//------------------------------------------------------------------------------
//
// Global functions
//

void *operator new     ( size_t size, const std::nothrow_t& ) throw();
void *operator new[]   ( size_t size, const std::nothrow_t& ) throw();
void *operator new     ( size_t size, int alignment ) throw();
void *operator new[]   ( size_t size, int alignment ) throw();
void  operator delete  ( void* ptr ) throw();
void  operator delete[]( void* ptr ) throw();

extern void *Alloc            ( size_t size, u8 alignment = 4 );
extern void *AllocDeviceMemory( size_t size, u8 alignment = 4 );
extern void  Free             ( void *memory );



#endif	// GR_COMMON_H_
