﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicSystem.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_GRAPHICS_SYSTEM_H_
#define GR_GRAPHICS_SYSTEM_H_

#include <nn/gr.h>

#pragma diag_suppress 2530  // #2530: padding added to end of struct GraphicSystem

class GraphicSystem
{
public:
    typedef GfxAllocator *( *GfxAllocatorFactory )( MemoryMgr::AllocatorDevice *dev );

public:
    static void Initialize( GfxAllocatorFactory factory = NULL );
    static void Finalize( void );

    GraphicSystem( void )
        : m_UpperSwapper( NULL )
        , m_LowerSwapper( NULL )
        , m_ExtensionSwapper( NULL )
        , m_CommandIndex( 0 )
        , m_DoubleCommandList( 0 )
    {
        m_CommandList[0] = m_CommandList[1] = 0;
    }

    void Create( size_t commandBufferSize = 512 * 1024, u32 reqCount = 128 );
    void Destroy( void );

    void SetDisplayBuffer( DisplayBuffer *top, DisplayBuffer *bottom, DisplayBuffer *extension );

    void TransferBuffer( const s32 screenKind,
                         const nn::gr::CTR::FrameBuffer::ColorBuffer* pColorBuffer,
                         const bool isAddSplitDrawCmdlist = true );

    void RunCommandList( void );
    void StopCommandList( void );
    void WaitVSync( s32 screenKind );
    void SwapBuffer( s32 screenKind );
    void WaitCommandList( const bool isAddSplitDrawCmdlist = true );
    void SwapCommandList( void );
    void SetUseDoubleCommandList(const bool flag){m_DoubleCommandList = flag;}

private:
    static void *AllocateGraphicsMemory( GLenum area, GLenum aim, GLuint id, GLsizei size );
    static void  DeallocateGraphicsMemory( GLenum area, GLenum aim, GLuint id, void *addr );

private:
    static GfxAllocator * s_GfxAllocator;
    static bool           s_GetGlSystemFlag;
    
    DisplayBuffer *       m_UpperSwapper;
    DisplayBuffer *       m_LowerSwapper;
    DisplayBuffer *       m_ExtensionSwapper;

    u32                   m_CommandList[2];
    u32                   m_CommandIndex;
    bool                  m_DoubleCommandList;

};

#pragma diag_default 2530

#endif // GR_GRAPHICS_SYSTEM_H_
