﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_MemoryMgr.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_MEMORY_MGR_H_
#define GR_MEMORY_MGR_H_

#include <nn.h>

class IAllocator
{
public:
    static const int DEFAULT_ALIGNMENT = 4;

    virtual ~IAllocator()
    {
    }

    virtual void* Alloc( size_t size, u8 alignment ) = 0;
    virtual void  Free( void* memory ) = 0;
    void* Alloc( size_t size )
    {
        return this->Alloc( size, DEFAULT_ALIGNMENT );
    }

    template<typename TObject>
    TObject* Alloc( size_t size, u8 alignment = DEFAULT_ALIGNMENT )
    {
        return static_cast<TObject*>( this->Alloc( sizeof( TObject ) * size, alignment ) );
    }

private:
    template<typename TObject>
    void Construct( TObject* object )
    {
        new( static_cast<void*>( object ) ) TObject();
    }

    template<typename TObject>
    void Destruct( TObject* object )
    {
        object->~TObject();
    }
};

class MemoryMgr
{
public:
    class AllocatorFCRAM : public IAllocator
    {
    public:
        virtual void *Alloc( size_t size, u8 alignment )
        {   
            return ::Alloc( size, alignment );
        }
        virtual void  Free ( void* memory )
        {
            ::Free( memory );
        }
    };

    class AllocatorDevice : public IAllocator
    {
    public:
        virtual void *Alloc( size_t size, u8 alignment )
        {
            return ::AllocDeviceMemory( size, alignment );
        }
        virtual void  Free ( void* memory )
        {
            ::Free( memory );
        }
    };

public:
    MemoryMgr()
    {
    }

    ~MemoryMgr()
    {
    }

    bool    Initialize( size_t mainMemorySize = 20 * 1024 * 1024 );

    void *  AllocateFCRAM( size_t size, u8 alignment );
    void *  AllocateDevice( size_t size, u8 alignment );
    void    Free( void *memory );

    AllocatorFCRAM  *GetDefaultAllocatorMain( void )
    {
        return &m_DefaultAllocatorMain;
    }

    AllocatorDevice *GetDefaultAllocatorDevice( void )
    {   
        return &m_DefaultAllocatorDevice;
    }

private:
    void *AllocateFCRAM( size_t size )
    {
        return m_MainHeap.Allocate( size );
    }

    void *AllocateDevice( size_t size )
    {
        return m_DeviceHeap.Allocate( size );
    }

private:
    nn::os::MemoryBlock     m_MemoryBlock;
    nn::fnd::ExpHeap        m_MainHeap;
    nn::fnd::ExpHeap        m_DeviceHeap;

    AllocatorFCRAM          m_DefaultAllocatorMain;
    AllocatorDevice         m_DefaultAllocatorDevice;
};

SMP_SINGLETON_DECL( MemoryMgr );



#endif // GR_MEMORY_MGR_H_
