﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_Framework.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46775 $
 *---------------------------------------------------------------------------*/

#include <nn/applet.h>

#include "gr_Inner.h"
#include "gr_Framework.h"

//------------------------------------------------------------------------------
extern IScene *                             CreateScene( void );
extern GraphicSystem::GfxAllocatorFactory   GetGraphicsMemoryMgrFunctor( void );

namespace
{

    //------------------------------------------------------------------------------
    GraphicSystem                   s_System;

    static s32 s_CmdStart,s_CmdEnd;
    static nn::os::Tick s_GpuStartTime;
    static nn::fnd::TimeSpan s_GpuSpan;

    //------------------------------------------------------------------------------
    void CmdlistCallback(s32 id)
    {
        if ( s_CmdStart == id )
        {
            s_GpuStartTime = nn::os::Tick::GetSystemCurrent();
        }
        if ( s_CmdEnd == id )
        {
            s_GpuSpan = nn::os::Tick::GetSystemCurrent() - s_GpuStartTime;
        }
    }

    //------------------------------------------------------------------------------
    void Run( IScene *scene )
    {
        scene->Initialize();
        scene->Enter();

        nn::gr::CTR::FrameBuffer::ColorBuffer* pGrColorBuffer =
            scene->GetGrColorBufferPointer();

        nn::hid::PadReader pad;
        nn::hid::PadStatus pad_status;

        u32 cpu_usec = 0;
        u32 gpu_usec = 0;
        u64 total_cpu_usec = 0;
        u64 total_gpu_usec = 0;
        u32 counter = 0;
        const u32 counter_thre = 120;

        bool cmdlistflg = 0;
        bool waitvsyncflg = 1;

        u32 cmdBufferSize    = 0;
        u32 cmdBufferMaxSize = 0;
        
        u32 cmdRequestNum    = 0;
        u32 cmdRequestMaxNum = 0;
        {
            GLint param = 0;
            nngxGetCmdlistParameteri( NN_GX_CMDLIST_MAX_BUFSIZE, & param );
            cmdBufferMaxSize = param;
                
            nngxGetCmdlistParameteri( NN_GX_CMDLIST_MAX_REQCOUNT, & param );
            cmdRequestMaxNum = param;
        }

        NN_UNUSED_VAR( cpu_usec );
        NN_UNUSED_VAR( gpu_usec );

        NN_UNUSED_VAR( total_cpu_usec );
        NN_UNUSED_VAR( total_gpu_usec );
        
        NN_UNUSED_VAR( cmdBufferSize );
        NN_UNUSED_VAR( cmdBufferMaxSize );

        NN_UNUSED_VAR( cmdRequestNum );
        NN_UNUSED_VAR( cmdRequestMaxNum );

        // Display controls
        NN_LOG("START: Output Log\n");
        NN_LOG("A: CommandList Single/Double Swap\n");
        NN_LOG("B: VSync Wait/NoWait Swap\n");
        
        for(;;)
        {
            nngxSetCmdlistCallback(CmdlistCallback);

            u32 swapflg = 0;
            GLuint target = 0;
            nn::os::Tick cpu_begin;
            nn::fnd::TimeSpan cpu_span;

            cpu_begin = nn::os::Tick::GetSystemCurrent();

            // Dump of measurement results
            bool dumpResult = false;
            pad.ReadLatest( &pad_status );
            if ( pad_status.trigger & nn::hid::BUTTON_START )
            {
                dumpResult = true;
            }

            // Toggles double/single for the command list
            if ( pad_status.trigger & nn::hid::BUTTON_A )
            {
                cmdlistflg = 1 - cmdlistflg;
                s_System.SetUseDoubleCommandList(cmdlistflg);

                if(cmdlistflg)
                {
                    NN_LOG("DoubleCommandList\n");
                }
                else
                {
                    NN_LOG("SingleCommandList\n");
                }

                total_cpu_usec = 0;
                total_gpu_usec = 0;
                cpu_usec = 0;
                gpu_usec = 0;
                counter = 0;
            }

            // Toggle VSync waits
            if ( pad_status.trigger & nn::hid::BUTTON_B )
            {
                waitvsyncflg = 1 - waitvsyncflg;

                if(waitvsyncflg)
                {
                    NN_LOG("WaitVSync\n");
                }
                else
                {
                    NN_LOG("NoWaitVSync\n");
                }

                total_cpu_usec = 0;
                total_gpu_usec = 0;
                cpu_usec = 0;
                gpu_usec = 0;
                counter = 0;
            }

            // Calculations
            if( !scene->Calc() )
            {
                break;
            }

            // GPU measurement start point
            {
                s32 requestCount;
                nngxSplitDrawCmdlist();
                nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &requestCount);
                s_CmdStart = requestCount;

                nngxEnableCmdlistCallback( s_CmdStart );
            }

            // Pre-render
            scene->PreDraw();

            // Render
            {
                target = scene->DrawTop();
                if ( target )
                {
                    s_System.TransferBuffer( UPPER_SCREEN,
                        pGrColorBuffer, true );
                    swapflg |= UPPER_SCREEN;
                }

                target = scene->DrawExt();
                if ( target )
                {
                    s_System.TransferBuffer( EXTENSION_SCREEN,
                        pGrColorBuffer, true );
                    swapflg |= EXTENSION_SCREEN;
                }

                target = scene->DrawBtm();
                if ( target )
                {
                    s_System.TransferBuffer( LOWER_SCREEN,
                        pGrColorBuffer, true );
                    swapflg |= LOWER_SCREEN;
                }
            }
            
            // Post-render processing
            scene->PostDraw();

            // GPU measurement end point
            {
                s32 requestCount;
                nngxGetCmdlistParameteri(NN_GX_CMDLIST_USED_REQCOUNT, &requestCount);
                s_CmdEnd = requestCount;

                nngxEnableCmdlistCallback( s_CmdEnd );
            }
            
            cpu_span = nn::os::Tick::GetSystemCurrent() - cpu_begin;

            if ( dumpResult )
            {
                GLint param = 0;
                nngxGetCmdlistParameteri( NN_GX_CMDLIST_USED_BUFSIZE, & param );
                cmdBufferSize = param;
               
                nngxGetCmdlistParameteri( NN_GX_CMDLIST_USED_REQCOUNT, & param );
                cmdRequestNum = param;
            }

            if( swapflg )
            {              

                if(cmdlistflg)
                {
                    s_System.SwapBuffer( swapflg );

                    if(waitvsyncflg)
                    {
                        s_System.WaitVSync ( swapflg );
                    }
                }

                if(cmdlistflg)
                {
                    s_System.WaitCommandList( false );
                    s_System.RunCommandList();
                    s_System.SwapCommandList();
                }
                else
                {
                    s_System.RunCommandList();
                    s_System.WaitCommandList( false );
                    s_System.StopCommandList();
                }

                if(!cmdlistflg)
                {
                    s_System.SwapBuffer( swapflg );

                    if(waitvsyncflg)
                    {
                        s_System.WaitVSync ( swapflg );
                    }
                }
            }

            if ( counter < counter_thre )
            {
                total_cpu_usec += cpu_span.GetMicroSeconds();
                total_gpu_usec += s_GpuSpan.GetMicroSeconds();
               
                counter += 1;
            }
            else
            {
                cpu_usec = 
                    static_cast<u32>( static_cast<f32>( total_cpu_usec ) / static_cast<f32>( counter_thre ) );
                gpu_usec = 
                    static_cast<u32>( static_cast<f32>( total_gpu_usec ) / static_cast<f32>( counter_thre ) );

                total_cpu_usec = 0;
                total_gpu_usec = 0;
                counter = 0;
            }

            if ( dumpResult )
            {
                if ( ( cpu_usec == 0 ) || ( gpu_usec == 0 ) )
                {
                    NN_LOG( "Please wait and push START again.\n" );
                }
                else
                {
                    f32 cpu_per = cpu_usec * 60.0f * 100.0f / ( 1000.0f * 1000.0f );
                    f32 gpu_per = gpu_usec * 60.0f * 100.0f / ( 1000.0f * 1000.0f );

                    NN_UNUSED_VAR( cpu_per );
                    NN_UNUSED_VAR( gpu_per );
    
                    NN_LOG( "CPU   : %8d (usec), %.2f (%%)\n",
                        cpu_usec, cpu_per );
                    NN_LOG( "GPU   : %8d (usec), %.2f (%%)\n",
                        gpu_usec, gpu_per );
                    if(!cmdlistflg)
                    {
                        NN_LOG( "Total : %8d (usec), %.2f (%%)\n",
                            cpu_usec + gpu_usec, cpu_per + gpu_per );
                    }
                    
                    NN_LOG( "CMD : used size = %7d, max size = %7d, used = %.2f (%%)\n",
                        static_cast<u32>( cmdBufferSize) , static_cast<u32>( cmdBufferMaxSize),
                        ( static_cast<f32>( cmdBufferSize ) / static_cast<f32>( cmdBufferMaxSize ) ) * 100.0f );

                    NN_LOG( "REQ : used num  = %7d, max num  = %7d, used = %.2f (%%)\n",
                        cmdRequestNum, cmdRequestMaxNum,
                        ( static_cast<f32>( cmdRequestNum ) / static_cast<f32>( cmdRequestMaxNum ) ) * 100.0f );
                }
            }
        }

        scene->Finalize();

        delete scene;
    }

} // Namespace

//------------------------------------------------------------------------------
void nnMain()
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // Initialization
    {
        GraphicSystem::Initialize( GetGraphicsMemoryMgrFunctor() );

        s_System.Create();
        
        DisplayBufferFormat format = DISPLAY_BUFFER_RGB8_OES;
        // DisplayBufferFormat format = DISPLAY_BUFFER_RGBA4;
        // DisplayBufferFormat format = DISPLAY_BUFFER_RGB5_A1;
        // DisplayBufferFormat format = DISPLAY_BUFFER_RGB565;

        IAllocator* pAllocator = MEMMGR->GetDefaultAllocatorDevice();

        DisplayBuffer *upper = 
            DisplayBuffer::Builder().pAllocator( pAllocator ).screenKind( UPPER_SCREEN ).bufferSize( 400, 240 ).format( format ).Create();
        DisplayBuffer *lower = 
            DisplayBuffer::Builder().pAllocator( pAllocator).screenKind( LOWER_SCREEN ).bufferSize( 320, 240 ).format( format ).Create();
        DisplayBuffer *exten = 
            DisplayBuffer::Builder().pAllocator( pAllocator ).screenKind( EXTENSION_SCREEN ).bufferSize( 400, 240 ).format( format ).Create();

        s_System.SetDisplayBuffer( upper, lower, exten );
    }
    
    // Create an instance of the scene
    Run( CreateScene() );

    // Post-processing
    {
        s_System.Destroy();
        GraphicSystem::Finalize();
    }
}
