﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_GraphicsDrawingScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_DEMO_GRAPHICS_DRAWING_SCENE_H_
#define GR_DEMO_GRAPHICS_DRAWING_SCENE_H_

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_GraphicsDrawing.h"
#include "gr_PrimitiveVertexStream.h"

class IAllocator;

//------------------------------------------------------------------------------

const f32 WALL_THICKNESS                  = 10.0f;
const f32 WALL_X_LENGTH                   = 400.0f;
const f32 WALL_Y_LENGTH                   = 240.0f;

const u32 MAX_WALL_PRIMITIVE_NUM          = 4;

const u32 TRIANGLES_NUM                   = 2;
const u32 MAX_TRIANGLES_NUM               = 1;

const u32 TRIANGLE_FAN_TRIANGLES_NUM      = 6;
const u32 MAX_TRIANGLE_FAN_NUM            = 1;

const u32 TRIANGLE_STRIP_TRIANGLES_NUM    = 4;
const u32 MAX_TRIANGLE_STRIP_NUM          = 1;
const f32 TRIANGLE_STRIP_LENGTH           = 40.0f;

const u32 TRIANGLES_VERTEX_NUM            = 3;
const u32 WALL_TRIANGLES_NUM              = 2;

const u32 MAX_TRIANGLES_VERTEX_COUNT      = TRIANGLES_VERTEX_NUM * TRIANGLES_NUM;
const u32 MAX_TRIANGLE_FAN_VERTEX_COUNT   = TRIANGLES_VERTEX_NUM * TRIANGLE_FAN_TRIANGLES_NUM;
const u32 MAX_TRIANGLE_STRIP_VERTEX_COUNT = TRIANGLES_VERTEX_NUM * TRIANGLE_STRIP_TRIANGLES_NUM;

const u32 DRAW_TRIANGLE_VERTEX_COUNT      = TRIANGLES_VERTEX_NUM * WALL_TRIANGLES_NUM * MAX_WALL_PRIMITIVE_NUM +
                                            MAX_TRIANGLES_VERTEX_COUNT      * MAX_TRIANGLES_NUM                +
                                            MAX_TRIANGLE_FAN_VERTEX_COUNT   * MAX_TRIANGLE_FAN_NUM             +
                                            MAX_TRIANGLE_STRIP_VERTEX_COUNT * MAX_TRIANGLE_STRIP_NUM ;

const u32 AABB_VERTEX_COUNT               = 24;
const u32 MAX_PRIMITIVE_NUM               = MAX_WALL_PRIMITIVE_NUM + MAX_TRIANGLES_NUM + MAX_TRIANGLE_FAN_NUM + MAX_TRIANGLE_STRIP_NUM;

const u32 MAX_SHAPE_VERTEX_COUNT          = DRAW_TRIANGLE_VERTEX_COUNT + AABB_VERTEX_COUNT * MAX_PRIMITIVE_NUM;

const nn::math::VEC2 AABB_DRAW_LINE_OFFSET( 0.5f, 0.5f );

f32 s_TimeStep                            = 1.0f;
f32 s_VelocityDamping                     = 1.0f;
u32 s_AABBCollisionNum                    = 0;
u32 s_TriangleCollisionNum                = 0;

f32 GetRandomFloat( const f32 rangeBegin = 0.0f, const f32 rangeEnd = 1.0f );

//------------------------------------------------------------------------------
class GraphicsDrawingScene : public IScene
{
public:
    GraphicsDrawingScene( void );
    virtual ~GraphicsDrawingScene();
    
public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw(void);
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );
    
private:
    void UpdateScene_( void );
    void DetectAndSolveCollision_( void );
    void SolveBroadCollisionPhase_( PrimitiveBase* pPrimitiveBase0, PrimitiveBase* pPrimitiveBase1,
                                    nn::math::VEC2 normal );
    void DetectNarrowCollisionPhase_( PrimitiveBase* pPrimitiveBase0, PrimitiveBase* pPrimitiveBase1 );

    void RenderSceneGr_( void );
    void RenderSceneResult_( void );    
    void RenderSceneFillPolygons_( void );
    void RenderSceneDrawAABBs_( void );
    void RenderSceneFillIntersectTriangles_( void );

    void RegisterPrimitive( PrimitiveBase* pPrimitiveBase );

public:
    IAllocator*                     allocator;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx;
    nn::math::MTX34                 viewMtx;
    nn::ulcd::StereoCamera          stereo;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];
    
    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;

    u32                             m_Counter;

    PrimitiveTriangles              m_PrimitiveTrianglesWallArray[ MAX_WALL_PRIMITIVE_NUM ];

    PrimitiveTriangles              m_PrimitiveTrianglesArray[ MAX_TRIANGLES_NUM ];
    PrimitiveFan                    m_PrimitiveFanArray[ MAX_TRIANGLE_FAN_NUM ];
    PrimitiveStrip                  m_PrimitiveStripArray[ MAX_TRIANGLE_STRIP_NUM ];
    
    u32                             m_PrimitiveNum;
    PrimitiveBase*                  m_pPrimitiveBaseArray[ MAX_PRIMITIVE_NUM ];
};

#endif // GR_DEMO_GRAPHICS_DRAWING_SCENE_H_
