﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PointSpriteScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_POINTSPRITE_SCENE_H_
#define GR_POINTSPRITE_SCENE_H_

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_PrimitiveVertexStream.h"

class IAllocator;

//------------------------------------------------------------------------------

class PointSpriteScene : public IScene
{
public:
    PointSpriteScene( void );
    virtual ~PointSpriteScene();
    
public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void    UpdateScene_( void );
    void    RenderSceneGr_( const s32 display, const nn::math::MTX44& proj, const nn::math::MTX34& view );
    bit32*  MakeTextureCommand_( bit32* command );

private:
    bit32* DrawPointSpriteCommand_( bit32* command, const s32 display, const nn::math::MTX44& proj, const nn::math::MTX34& view );

public:
    IAllocator*                     allocator;

    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx44;
    nn::math::MTX34                 modelViewMtx34;
    nn::ulcd::StereoCamera          stereo;

    // Vertex shader uniforms
    nn::gr::BindSymbolVSFloat       bindSymbolProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolModelViewMtx44;

    // Geometry shader uniforms
    nn::gr::BindSymbolGSFloat       twnBindSymbolViewPortVec2;
    nn::gr::BindSymbolGSBool        twnBindSymbolDistanceAttenuationBool;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];

    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;

    nn::gr::CTR::RenderState        m_RenderState;
    nn::gr::Combiner                m_Combiner;

    nn::gr::Texture                 m_Texture;

    PrimitivePoint*                 m_pPrimitivePoint;
    nn::gr::Vertex                  m_Vertex;
    nn::gr::Vertex::IndexStream     m_IndexStream;

    u32                             m_Counter;
};

#endif // GR_POINTSPRITE_SCENE_H_
