﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ProcedureTextureScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_PROCEDURE_TEXTURE_SCENE_H_
#define GR_PROCEDURE_TEXTURE_SCENE_H_

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_ProcedureTextureData.h"
#include "gr_PrimitiveVertexStream.h"

class IAllocator;

//------------------------------------------------------------------------------

enum
{
    PROCTEX_FIRE_PATTERN    = 0,
    PROCTEX_CLOUD_PATTERN   = 1,
    PROCTEX_THUNDER_PATTERN = 2,
    PROCTEX_SEA_PATTERN     = 3,
    PROCTEX_MUD_PATTERN     = 4,
    PROCTEX_PATTERN_NUM     = 5
};

wchar_t* s_ProcedureTextureDataFileNameArray[ ] =
{
    L"rom:/proctex_fire01.pctxbin",
    L"rom:/proctex_cloud01.pctxbin",
    L"rom:/proctex_thunder01.pctxbin",
    L"rom:/proctex_sea01.pctxbin",
    L"rom:/proctex_mud01.pctxbin"    
};

class ProcedureTextureScene : public IScene
{
public:
    ProcedureTextureScene( void );
    virtual ~ProcedureTextureScene();
    
public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void    UpdateScene_( void );
    void    RenderSceneGr_( const s32 display, const nn::math::MTX44& proj, const nn::math::MTX34& view );

private:
    void InitializeProcedureTextureArray( void );
    void UpdateProcedureTextureArray( void );
    bit32* DrawProcedureTexture( bit32* command,
        const s32 display, const nn::math::MTX44& proj, const nn::math::MTX34& view );
   
public:
    IAllocator*                     allocator;

    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx;
    nn::math::MTX34                 viewMtx;
    nn::ulcd::StereoCamera          stereo;

    nn::gr::BindSymbolVSFloat       bindSymbolWorldMtx;
    nn::gr::BindSymbolVSFloat       bindSymbolProjMtx;
    nn::gr::BindSymbolVSFloat       bindSymbolViewMtx;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];

    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;

    nn::gr::CTR::ProcedureTexture   m_ProcedureTextureArray[ PROCTEX_PATTERN_NUM ];

    bool                            m_IsEnableNoiseUArray[ PROCTEX_PATTERN_NUM ];
    bool                            m_IsEnableNoiseVArray[ PROCTEX_PATTERN_NUM ];
    NN_PADDING2;

    f32                             m_NoiseUPhaseStepArray[ PROCTEX_PATTERN_NUM ];
    f32                             m_NoiseVPhaseStepArray[ PROCTEX_PATTERN_NUM ];

    f32                             m_TexCoordUArray[ PROCTEX_PATTERN_NUM ][ 2 ];
    f32                             m_TexCoordVArray[ PROCTEX_PATTERN_NUM ][ 2 ];

    f32                             m_ClearColorArray[ PROCTEX_PATTERN_NUM ][ 4 ];

    u32                             m_UpperProcedurePatternIndex;
    PrimitiveBox*                   m_UpperPrimitiveBox;
    nn::gr::Vertex                  m_UpperVertexBox;
    nn::gr::Vertex::IndexStream     m_UpperIndexStreamBox;
    
    u32                             m_LowerProcedurePatternIndex;
    PrimitiveBox*                   m_LowerPrimitiveBox;
    nn::gr::Vertex                  m_LowerVertexBox;
    nn::gr::Vertex::IndexStream     m_LowerIndexStreamBox;

    nn::math::VEC3                  m_Pos;
    nn::math::VEC3                  m_EulerAngleDeg0;
    nn::math::VEC3                  m_EulerAngleDeg1;

    u32                             m_Counter;
};

#endif // GR_PROCEDURE_TEXTURE_SCENE_H_
