﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_ShadowScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef GR_SHADOW_SCENE_H_
#define GR_SHADOW_SCENE_H_

#include <nn/math.h>

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_PrimitiveVertexStream.h"

const u32 SHADOW_TEXTURE_WIDTH = 512;
const u32 SHADOW_TEXTURE_HEIGHT = 512;
const f32 LIGHT_POS[] = {2.0f, 2.0f, 2.0f, 0.0f};

class IAllocator;

class ShadowScene : public IScene
{
public:
    ShadowScene( void );
    virtual ~ShadowScene();

public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit, const s32 index = 0 );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void InitializeTexture( void );
    void UpdateScene_( void );
    void RenderShadowTexture_( const nn::math::MTX44& proj, const nn::math::MTX34& view );
    void RenderSceneGr_( const s32 display, 
                         const nn::math::MTX44& proj,
                         const nn::math::MTX34& view );
    bit32* MakeShadowCommand_( bit32* command,
                                  const nn::math::MTX44& proj,
                                  const nn::math::MTX34& view,
                                  const bool isMakeFullShaderCommand = true );
    bit32* MakeDrawModelCommand_( bit32* command,
                                  const nn::math::MTX44& proj,
                                  const nn::math::MTX34& view,
                                  const bool isMakeFullShaderCommand = true );
    bit32* MakeTextureCommand_( bit32* command );
    bit32* MakeDrawCommand( bit32* command );

public:
    IAllocator*                     allocator;

    // Shader binary for normal mode
    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx44;
    nn::math::MTX34                 modelViewMtx34;
    nn::ulcd::StereoCamera          stereo;

    // Vertex shader uniforms
    nn::gr::BindSymbolVSFloat       bindSymbolProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolModelViewMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolTextureMtx44;

    // Vertex shader uniforms (Shadow)
    nn::gr::BindSymbolVSFloat       bindSymbolShadowProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolShadowModelViewMtx44;

    nn::gr::BindSymbolVSFloat       bindSymbolTexProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolTexModelViewMtx44;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];

    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

    uptr                            shadowColorBufferVirtualAddress;
    uptr                            shadowDepthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    f32 m_Near;
    f32 m_Far;

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;
    nn::gr::FrameBuffer             m_ShadowBuffer;
    
    nn::gr::RenderState             m_RenderState;
    nn::gr::Combiner                m_Combiner;
    nn::gr::FragmentLight           m_Light;
    nn::gr::LookUpTable             m_Table;

    nn::gr::Texture                 m_Texture;

    PrimitiveModel*                 m_pPrimitiveModel;
    nn::gr::Vertex                  m_Vertex;
    nn::gr::Vertex::IndexStream     m_IndexStream;

    u32                             m_Counter;
};

#endif // GR_SHADOW_SCENE_H_
