﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_PrimitiveVertexStream.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef GR_PRIMITIVE_VERTEX_STREAM_H_
#define GR_PRIMITIVE_VERTEX_STREAM_H_

#include <nn.h>
#include <nn/math.h>
#include <nn/fs.h>

#include <nn/gr.h>
#include "gr_Framework.h"

enum PrimitiveType
{
    TRIANGLES                = 0x1,
    TRIANGLES_SILHOUETTE_DMP = 0x2
};

enum VertexAttribute
{
    POSITION_ATTRIBUTE = 0x1,
    COLOR_ATTRIBUTE    = 0x2,
    NORMAL_ATTRIBUTE   = 0x4
};

enum ArrayType
{
    INDEPENDENT_ARRAY,
    INTERLEAVED_ARRAY
};

struct Position3f
{
    f32 x;
    f32 y;
    f32 z;
};

struct Color4f
{
    f32 r;
    f32 g;
    f32 b;
    f32 a;
};

struct Normal3f
{
    f32 x;
    f32 y;
    f32 z;
};

class PrimitiveModel
{
public:
    PrimitiveModel( void );
    ~PrimitiveModel();

    void InitializeTorus( IAllocator* allocator,
                          const GLenum _geometryMode, 
                          const VertexAttribute vertexAttribute,
                          const ArrayType arrayType,
                          const u32 horizontalPolygonCount,
                          const u32 verticalPolygonCount );

    void Finalize( void );

    GLenum geometryMode;

    IAllocator* allocator;

    u8*  pVertexArray;
    uptr pPositionArray;
    uptr pColorArray;
    uptr pNormalArray;

    u16* pIndexArray;
    u16* pTwnIndexArray;

    u32  indexCount;
    u32  twnIndexCount;
};

#endif // GR_PRIMITIVE_VERTEX_STREAM_H_
