﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gr_SilhouetteScene.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef GR_SILHOUETTE_SCENE_H_
#define GR_SILHOUETTE_SCENE_H_

#include <nn/math.h>

#include <nn/gr.h>
#include <nn/ulcd.h>

#include "gr_PrimitiveVertexStream.h"

class IAllocator;



class SilhouetteScene : public IScene
{
public:
    SilhouetteScene( void );
    virtual ~SilhouetteScene();

public:
    virtual bool Initialize( void );
    virtual void Finalize( void );
    virtual void Enter( void );
    virtual bool Calc( void );
    virtual void PreDraw( void );
    virtual GLuint DrawTop( void );
    virtual GLuint DrawBtm( void );
    virtual GLuint DrawExt( void );
    virtual void PostDraw( void );
    virtual void Clear( const u32 bufferBit );
    virtual nn::gr::CTR::FrameBuffer::ColorBuffer* GetGrColorBufferPointer( void );

private:
    void InitializeTexture( void );
    void UpdateScene_( void );
    void RenderSceneGr_( const s32 display, 
                         const nn::math::MTX44& proj,
                         const nn::math::MTX34& view,
                         const PrimitiveType primitiveType );
    bit32* MakeDrawModelCommand_( bit32* command,
                                  const nn::math::MTX44& proj,
                                  const nn::math::MTX34& view,
                                  const bool isMakeFullShaderCommand = true );
    bit32* MakeDrawSilhouetteModelCommand_( bit32* command,
                                            const s32 display,
                                            const nn::math::MTX44& proj,
                                            const nn::math::MTX34& view,
                                            const bool isMakeFullShaderCommand = true );
    bit32* MakeTextureCommand_( bit32* command );
    bit32* MakeDrawCommand( bit32* command );

public:
    IAllocator*                     allocator;

    // Shader binary for normal mode
    u8*                             shaderBinary;
    nn::gr::ShaderLite              shader;
    u32*                            shaderBuffer;

    nn::math::MTX34                 mtx;
    nn::math::MTX44                 projectionMtx44;
    nn::math::MTX34                 modelViewMtx34;
    nn::ulcd::StereoCamera          stereo;

    // Vertex shader uniforms
    nn::gr::BindSymbolVSFloat       bindSymbolProjectionMtx44;
    nn::gr::BindSymbolVSFloat       bindSymbolModelViewMtx44;

    // Geometry shader uniforms for the silhouette shader
    nn::gr::BindSymbolGSFloat       twnBindSymbolWidthOpenEdgeDepthBiasVec3;
    nn::gr::BindSymbolGSFloat       twnBindSymbolColorVec4;
    nn::gr::BindSymbolGSFloat       twnBindSymbolOpenEdgeColorVec4;
    nn::gr::BindSymbolGSFloat       twnBindSymbolOpenEdgeWidthVec4;
    nn::gr::BindSymbolGSBool        twnBindSymbolAcceptEmptyTrianglesBool;
    nn::gr::BindSymbolGSBool        twnBindSymbolScaleByWBool;
    nn::gr::BindSymbolGSBool        twnBindSymbolFrontFaceCCWBool;
    nn::gr::BindSymbolGSBool        twnBindSymbolOpenEdgeWidthScaleByWBool;
    nn::gr::BindSymbolGSBool        twnBindSymbolOpenEdgeDepthBiasScaleByWBool;

    GLuint                          renderTarget;
    GLuint                          renderBuffer[ 2 ];

    uptr                            colorBufferVirtualAddress;
    uptr                            depthBufferVirtualAddress;

private:
    typedef void* (* AllocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, GLsizei size );
    typedef void  (* DeallocateGraphicsMemory)( GLenum area, GLenum aim, GLuint id, void *addr );

    AllocateGraphicsMemory          m_GraphicsMemoryAllocator;
    DeallocateGraphicsMemory        m_GraphicsMemoryDeallocator;

    nn::gr::Viewport                m_Viewport;
    nn::gr::Scissor                 m_Scissor;
    nn::gr::FrameBuffer             m_FrameBuffer;
    
    nn::gr::CTR::RenderState        m_RenderState;
    nn::gr::Combiner                m_Combiner;

    nn::gr::Texture                 m_Texture;

    PrimitiveModel*                 m_pPrimitiveModel;
    nn::gr::Vertex                  m_Vertex;
    nn::gr::Vertex::IndexStream     m_IndexStream;

    u32                             m_Counter;
};

#endif // GR_SILHOUETTE_SCENE_H_
