﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gx_ProcTexEditor.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn/applet.h>

#include "gx_ProcTexEditor.h"


int InitializeGx(void)
{
    const u32 gxHeapSize = 0x800000;
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
        nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>( s_AppHeap.Allocate( gxHeapSize ) );
    
    s_RenderSystem.Initialize( s_HeapForGx, gxHeapSize );
    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aTexCoord");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    // glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.5f, 0.5f, 0.5f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.1f, 0.1f, 0.1f, 1.0f);

    /* Blending */
    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
 
    /* Setup Polygon */
    InitializeVertexBuffer();

    /* Load LookUpTable */
    InitializeLookUpTable();

    /* Setup Texture */
    SetTextureEnvironment();
    
    /* Setup Capture heap */
    s_CaptureBuffer = s_AppHeap.Allocate( nn::gx::DISPLAY0_WIDTH * nn::gx::DISPLAY0_HEIGHT * 3 );
   
    return 0;
}

void InitializeVertexBuffer(void)
{    
    /* Square */
    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, sizeof(s_PositionCoords) + sizeof(s_TexCoords), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(s_PositionCoords), s_PositionCoords);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(s_PositionCoords), sizeof(s_TexCoords), s_TexCoords);

    /* Index */
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(s_IndexArray), s_IndexArray, GL_STATIC_DRAW);
}

void InitializeLookUpTable(void)
{
    GLfloat table[512];
    int i;

    /* Texture collection */
    glGenTextures(1, &s_TexColl);
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_TexColl);

    /* Texture Data */
    glGenTextures(MAX_TEXTURE_OBJECT_NUM, &s_TexID[0]);

    /* Table7 : const 0.0 */
    memset(table, 0, 512 * sizeof(GLfloat));
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_TexID[7]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, 
        GL_LUMINANCEF_DMP, GL_FLOAT, table);

    /* Table8 : const 1.0 */
    memset(table, 0, 512 * sizeof(GLfloat));
    for(i = 0; i < 128; i++)
    {
        table[i] = 1.0f;
    }
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_TexID[8]);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, 
        GL_LUMINANCEF_DMP, GL_FLOAT, table);

    /* FuncRGB */
    SetPresetData(s_FuncRGBArray, 0); // Preset f(x)=x
    SetLookUpTable_256(s_TexID[0], s_FuncRGBArray);

    /* FuncAlpha */
    SetPresetData(s_FuncAlphaArray, 34); // Preset f(x)=1.0
    SetLookUpTable_256(s_TexID[1], s_FuncAlphaArray);

    /* Noise */
    SetPresetData(s_FuncNoiseArray, 38); // Preset 3x^2-2x^3
    SetLookUpTable_256(s_TexID[2], s_FuncNoiseArray);

    /* SamplerR */
    SetPresetData(s_FuncSamplerRArray, 35); // Preset SamplePattern
    SetLookUpTable_512(s_TexID[3], s_FuncSamplerRArray);

    /* SamplerG */
    SetPresetData(s_FuncSamplerGArray, 36); // Preset SamplePattern
    SetLookUpTable_512(s_TexID[4], s_FuncSamplerGArray);

    /* SamplerB */
    SetPresetData(s_FuncSamplerBArray, 37); // Preset SamplePattern
    SetLookUpTable_512(s_TexID[5], s_FuncSamplerBArray);

    /* SamplerA */
    SetPresetData(s_FuncSamplerAArray, 34); // Preset Const 1.0
    SetLookUpTable_512(s_TexID[6], s_FuncSamplerAArray);

    /* EditView */
    // glGenTextures(1, &s_TexEditView);
    s_TexEditView = 0;
    MakeTextureEditView(s_FuncRGBArray);

    /* ColorView */
    // glGenTextures(1, &s_TexColorView);
    s_TexColorView = 0;
    MakeTextureColorView();

    s_TexAlphaView = 0;
    MakeTextureAlphaView();

    glBindTexture(GL_TEXTURE_COLLECTION_DMP, 0);
}

void InitializeButtons( void )
{
    int i;
    GLfloat colorBlack[4]   = { 0.0f, 0.0f, 0.0f, 1.0f };
    // GLfloat colorWhite[4]   = { 1.0f, 1.0f, 1.0f, 1.0f };
    GLfloat colorDefault[4] = { 0.5f, 0.5f, 0.5f, 1.0f };
    GLfloat colorTouch[4]   = { 0.7f, 0.7f, 0.2f, 1.0f };
    GLfloat colorSelect[4]  = { 0.7f, 0.3f, 0.3f, 1.0f };
    GLfloat colorDisable[4] = { 0.3f, 0.3f, 0.3f, 1.0f };

    /* Initialize all button */
    for (i = 0; i < NUM_BUTTON; i++) 
    {
        s_ButtonArray[i].SetData(-1, 0, 0, 0, 0);
        s_ButtonArray[i].SetDefaultColor(colorDefault[0], colorDefault[1], colorDefault[2], colorDefault[3]);
        s_ButtonArray[i].SetTouchColor(colorTouch[0], colorTouch[1], colorTouch[2], colorTouch[3]);
        s_ButtonArray[i].SetSelectColor(colorSelect[0], colorSelect[1], colorSelect[2], colorSelect[3]);
        s_ButtonArray[i].SetDisableColor(colorDisable[0], colorDisable[1], colorDisable[2], colorDisable[3]);
    }

    /* EditView */
    s_ButtonArray[BUTTON_ID_EDITVIEW].SetData(BUTTON_ID_EDITVIEW, 16, 15, 128, 129);
    s_ButtonArray[BUTTON_ID_EDITVIEW].SetDefaultColor(colorBlack[0], colorBlack[1], colorBlack[2], colorBlack[3]);
    s_ButtonArray[BUTTON_ID_EDITVIEW].SetTouchColor(colorBlack[0], colorBlack[1], colorBlack[2], colorBlack[3]);
    s_ButtonArray[BUTTON_ID_EDITVIEW].SetSelectColor(colorBlack[0], colorBlack[1], colorBlack[2], colorBlack[3]);
    s_ButtonArray[BUTTON_ID_EDITVIEW].SetDisableColor(colorBlack[0], colorBlack[1], colorBlack[2], colorBlack[3]);

    /* TableSelect Fc */
    s_ButtonArray[BUTTON_ID_SELECT_FC].SetData(BUTTON_ID_SELECT_FC, 168, 20, 16, 8);
    /* TableSelect Fa */
    s_ButtonArray[BUTTON_ID_SELECT_FA].SetData(BUTTON_ID_SELECT_FA, 188, 20, 16, 8);

    /* TableSelect R */
    s_ButtonArray[BUTTON_ID_SELECT_R].SetData(BUTTON_ID_SELECT_R, 208, 20, 16, 8);
    /* TableSelect G */
    s_ButtonArray[BUTTON_ID_SELECT_G].SetData(BUTTON_ID_SELECT_G, 228, 20, 16, 8);
    /* TableSelect B */
    s_ButtonArray[BUTTON_ID_SELECT_B].SetData(BUTTON_ID_SELECT_B, 248, 20, 16, 8);
    /* TableSelect A */
    s_ButtonArray[BUTTON_ID_SELECT_A].SetData(BUTTON_ID_SELECT_A, 268, 20, 16, 8);

    /* TableSelect Nz */
    s_ButtonArray[BUTTON_ID_SELECT_NZ].SetData(BUTTON_ID_SELECT_A, 288, 20, 16, 8);

    /* PresetPrev */    
    s_ButtonArray[BUTTON_ID_PRESET_PREV].SetData(BUTTON_ID_PRESET_PREV, 168, 48, 8, 8);
    /* PresetNext */    
    s_ButtonArray[BUTTON_ID_PRESET_NEXT].SetData(BUTTON_ID_PRESET_NEXT, 180, 48, 8, 8);
    /* PresetCopy */    
    s_ButtonArray[BUTTON_ID_PRESET_COPY].SetData(BUTTON_ID_PRESET_COPY, 222, 36, 32, 8);

    /* G-Function Prev */
    s_ButtonArray[BUTTON_ID_RGB_PATTERN_PREV].SetData(BUTTON_ID_RGB_PATTERN_PREV, 200 + 8, 64, 8, 8);
    /* G-Function Next */
    s_ButtonArray[BUTTON_ID_RGB_PATTERN_NEXT].SetData(BUTTON_ID_RGB_PATTERN_NEXT, 212 + 8, 64, 8, 8);
    
    /* G-Function Prev */
    s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_PREV].SetData(BUTTON_ID_ALPHA_PATTERN_PREV, 200 + 8, 64 + 10.0f, 8, 8);
    /* G-Function Next */
    s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_NEXT].SetData(BUTTON_ID_ALPHA_PATTERN_NEXT, 212 + 8, 64 + 10.0f, 8, 8);

    {
        const f32 yOffset = 7.0f;

        /* TableEnable R */
        s_ButtonArray[BUTTON_ID_ENABLE_R].SetData(BUTTON_ID_ENABLE_R, 216, 80 + yOffset, 16, 8);
        /* TableEnable G */
        s_ButtonArray[BUTTON_ID_ENABLE_G].SetData(BUTTON_ID_ENABLE_G, 240, 80 + yOffset, 16, 8);
        /* TableEnable B */
        s_ButtonArray[BUTTON_ID_ENABLE_B].SetData(BUTTON_ID_ENABLE_B, 264, 80 + yOffset, 16, 8);
        /* TableEnable A */
        s_ButtonArray[BUTTON_ID_ENABLE_A].SetData(BUTTON_ID_ENABLE_A, 288, 80 + yOffset, 16, 8);
        

        /* NoiseEnable U */
        s_ButtonArray[BUTTON_ID_ENABLE_N_U].SetData(BUTTON_ID_ENABLE_N_U, 168, 132, 24, 8);
        /* NoiseEnable V */
        s_ButtonArray[BUTTON_ID_ENABLE_N_V].SetData(BUTTON_ID_ENABLE_N_V, 168, 168, 24, 8);

        /* NoiseParameterEnable U F */
        s_ButtonArray[BUTTON_ID_ENABLE_N_U_F].SetData(BUTTON_ID_ENABLE_N_U_F, 200, 120, 8, 8);
        /* NoiseParameterEnable U F Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_U_F].SetData(BUTTON_ID_MINUS_N_U_F, 216, 120, 8, 8);
        /* NoiseParameterEnable U F Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_U_F].SetData(BUTTON_ID_PLUS_N_U_F, 228, 120, 8, 8);
        /* NoiseParameterEnable U P */
        s_ButtonArray[BUTTON_ID_ENABLE_N_U_P].SetData(BUTTON_ID_ENABLE_N_U_P, 200, 132, 8, 8);
        /* NoiseParameterEnable U P Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_U_P].SetData(BUTTON_ID_MINUS_N_U_P, 216, 132, 8, 8);
        /* NoiseParameterEnable U P Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_U_P].SetData(BUTTON_ID_PLUS_N_U_P, 228, 132, 8, 8);

        /* NoiseParameterEnable U P Auto */
        s_ButtonArray[BUTTON_ID_ENABLE_N_U_PA].SetData(BUTTON_ID_ENABLE_N_U_PA, 280 - 8, 132, 40 + 8, 8);
        s_ButtonArray[BUTTON_ID_MINUS_N_U_PA].SetData(BUTTON_ID_MINUS_N_U_PA, 280, 132 + 10, 8, 8);
        s_ButtonArray[BUTTON_ID_PLUS_N_U_PA].SetData(BUTTON_ID_PLUS_N_U_PA, 280 + 10, 132 + 10, 8, 8);

        /* NoiseParameterEnable U A */
        s_ButtonArray[BUTTON_ID_ENABLE_N_U_A].SetData(BUTTON_ID_ENABLE_N_U_A, 200, 144, 8, 8);
        /* NoiseParameterEnable U A Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_U_A].SetData(BUTTON_ID_MINUS_N_U_A, 216, 144, 8, 8);
        /* NoiseParameterEnable U A Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_U_A].SetData(BUTTON_ID_PLUS_N_U_A, 228, 144, 8, 8);

        /* NoiseParameterEnable V F */
        s_ButtonArray[BUTTON_ID_ENABLE_N_V_F].SetData(BUTTON_ID_ENABLE_N_V_F, 200, 156, 8, 8);
        /* NoiseParameterEnable V F Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_V_F].SetData(BUTTON_ID_MINUS_N_V_F, 216, 156, 8, 8);
        /* NoiseParameterEnable V F Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_V_F].SetData(BUTTON_ID_PLUS_N_V_F, 228, 156, 8, 8);
        /* NoiseParameterEnable V P */
        s_ButtonArray[BUTTON_ID_ENABLE_N_V_P].SetData(BUTTON_ID_ENABLE_N_V_P, 200, 168, 8, 8);
        /* NoiseParameterEnable V P Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_V_P].SetData(BUTTON_ID_MINUS_N_V_P, 216, 168, 8, 8);
        /* NoiseParameterEnable V P Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_V_P].SetData(BUTTON_ID_PLUS_N_V_P, 228, 168, 8, 8);

        /* NoiseParameterEnable V P Auto */
        s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].SetData(BUTTON_ID_ENABLE_N_V_PA, 280 - 8, 168, 40 + 8, 8);
        s_ButtonArray[BUTTON_ID_MINUS_N_V_PA].SetData(BUTTON_ID_MINUS_N_V_PA, 280, 168 + 10, 8, 8);
        s_ButtonArray[BUTTON_ID_PLUS_N_V_PA].SetData(BUTTON_ID_PLUS_N_V_PA, 280 + 10, 168 + 10, 8, 8);
        
        /* NoiseParameterEnable V A */
        s_ButtonArray[BUTTON_ID_ENABLE_N_V_A].SetData(BUTTON_ID_ENABLE_N_V_A, 200, 180, 8, 8);
        /* NoiseParameterEnable V A Minus */
        s_ButtonArray[BUTTON_ID_MINUS_N_V_A].SetData(BUTTON_ID_MINUS_N_V_A, 216, 180, 8, 8);
        /* NoiseParameterEnable V A Plus */
        s_ButtonArray[BUTTON_ID_PLUS_N_V_A].SetData(BUTTON_ID_PLUS_N_V_A, 228, 180, 8, 8);

        /* Clamp U */
        s_ButtonArray[BUTTON_ID_CLAMP_U].SetData(BUTTON_ID_CLAMP_U, 184, 208, 72, 8);
        /* Shift U */
        s_ButtonArray[BUTTON_ID_SHIFT_U].SetData(BUTTON_ID_SHIFT_U, 264, 208, 32, 8);
        /* Clamp V */
        s_ButtonArray[BUTTON_ID_CLAMP_V].SetData(BUTTON_ID_CLAMP_V, 184, 224, 72, 8);
        /* Shift V */
        s_ButtonArray[BUTTON_ID_SHIFT_V].SetData(BUTTON_ID_SHIFT_V, 264, 224, 32, 8);
       
        /* s_MinTexCoordT Minus */
        s_ButtonArray[BUTTON_ID_MINUS_T_MIN].SetData(BUTTON_ID_MINUS_T_MIN, 24, 224, 8, 8);       
        /* s_MinTexCoordT Plus */
        s_ButtonArray[BUTTON_ID_PLUS_T_MIN].SetData(BUTTON_ID_PLUS_T_MIN, 36, 224, 8, 8);
        /* s_MaxTexCoordT Minus */
        s_ButtonArray[BUTTON_ID_MINUS_T_MAX].SetData(BUTTON_ID_MINUS_T_MAX, 88, 224, 8, 8);
        /* s_MaxTexCoordT Plus */
        s_ButtonArray[BUTTON_ID_PLUS_T_MAX].SetData(BUTTON_ID_PLUS_T_MAX, 100, 224, 8, 8);

        /* Load WaveData */
        s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_PREV].SetData(BUTTON_ID_LOAD_WAVEDATA_PREV, 168 + 34, 8, 8, 8);
        s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_NEXT].SetData(BUTTON_ID_LOAD_WAVEDATA_NEXT, 168 + 44, 8, 8, 8);

        /* Dump WaveData */    
        s_ButtonArray[BUTTON_ID_SAVE_WAVEDATA].SetData(BUTTON_ID_SAVE_WAVEDATA, 204 + 24, 8, 32, 8);

        /* Capture */
        s_ButtonArray[BUTTON_ID_CAPTURE].SetData(BUTTON_ID_CAPTURE, 264, 8, 7 * 8, 8);

        /* AlphaSeparate */
        s_ButtonArray[BUTTON_ID_SEPARATE].SetData(BUTTON_ID_SEPARATE,
            248 + 8, 8 + 24 + 4, 64, 8);
        
        /* Background Color R */
        s_ButtonArray[BUTTON_ID_BGCOLOR_R].SetData(BUTTON_ID_BGCOLOR_R, 216, 92 + yOffset, 16, 8);
        /* Background Color G */
        s_ButtonArray[BUTTON_ID_BGCOLOR_G].SetData(BUTTON_ID_BGCOLOR_G, 240, 92 + yOffset, 16, 8);
        /* Background Color B */
        s_ButtonArray[BUTTON_ID_BGCOLOR_B].SetData(BUTTON_ID_BGCOLOR_B, 264, 92 + yOffset, 16, 8);
    }

    {
        const f32 yOffset = 3.0f;

        /* s_MinTexCoordS Minus */        
        s_ButtonArray[BUTTON_ID_MINUS_S_MIN].SetData(BUTTON_ID_MINUS_S_MIN, 24, 208 + yOffset, 8, 8);
        /* s_MinTexCoordS Plus */
        s_ButtonArray[BUTTON_ID_PLUS_S_MIN].SetData(BUTTON_ID_PLUS_S_MIN, 36, 208 + yOffset, 8, 8);
        /* s_MaxTexCoordS Minus */
        s_ButtonArray[BUTTON_ID_MINUS_S_MAX].SetData(BUTTON_ID_MINUS_S_MAX, 88, 208 + yOffset, 8, 8);
        /* s_MaxTexCoordS Plus */
        s_ButtonArray[BUTTON_ID_PLUS_S_MAX].SetData(BUTTON_ID_PLUS_S_MAX, 100, 208 + yOffset, 8, 8);
    }

    {
        const f32 yOffset = 8.0f;
        s_ButtonArray[BUTTON_ID_TEX_WIDTH_PREV].SetData(BUTTON_ID_TEX_WIDTH_PREV, 130,      158 + yOffset, 8, 8);
        s_ButtonArray[BUTTON_ID_TEX_WIDTH_NEXT].SetData(BUTTON_ID_TEX_WIDTH_NEXT, 130 + 10, 158 + yOffset, 8, 8);
    }
}

void PrintControl( void )
{    
    NN_LOG("Translate object -X/+X : Slide pad left/right .\n");
    NN_LOG("Translate object -Y/+Y : Slide pad down/up    .\n");
    NN_LOG("Translate object -Z/+Z : Slide pad down/up    + L .\n");

    NN_LOG("Rotate object    -X/+X : Slide pad down/up    + R .\n");
    NN_LOG("Rotate object    -Y/+Y : Slide pad left/right + R .\n");
    NN_LOG("Rotate object    -Z/+Z : Slide pad left/right + L .\n");

    NN_LOG("Scale object     -X/+X : Slide pad left/right + L + R .\n");
    NN_LOG("Scale object     -Y/+Y : Slide pad down/up    + L + R .\n");
}

void SetTextureEnvironment(void)
{
    // Binding LookUpTable
    glBindTexture(GL_LUT_TEXTURE0_DMP, s_TexID[0]);
    glBindTexture(GL_LUT_TEXTURE1_DMP, s_TexID[1]);
    glBindTexture(GL_LUT_TEXTURE2_DMP, s_TexID[2]);
    glBindTexture(GL_LUT_TEXTURE3_DMP, (s_EnableSamplerR ? s_TexID[3] : s_TexID[7]));
    glBindTexture(GL_LUT_TEXTURE4_DMP, (s_EnableSamplerG ? s_TexID[4] : s_TexID[7]));
    glBindTexture(GL_LUT_TEXTURE5_DMP, (s_EnableSamplerB ? s_TexID[5] : s_TexID[7]));
    glBindTexture(GL_LUT_TEXTURE6_DMP, (s_EnableSamplerA ? s_TexID[6] : s_TexID[8]));
    
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerRgbMap"), 0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerAlphaMap"), 1);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerNoiseMap"), 2);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerR"), 3);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerG"), 4);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerB"), 5);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptSamplerA"), 6);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].samplerType"), GL_TEXTURE_PROCEDURAL_DMP);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].texcoord"), GL_TEXTURE0);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptMinFilter"), s_TexFilter);//GL_LINEAR);
    //glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexWidth"), 128);
    //glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexOffset"), 0);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexWidth"), s_TexWidth );
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptTexOffset"), s_TexOffset );

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptRgbMap"), s_PtRgbMapArray[s_RgbMapIndex]);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaSeparate"), 
        (s_EnableAlphaSeparate ? GL_TRUE : GL_FALSE));
    if ( s_EnableAlphaSeparate )
    {
        glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptAlphaMap"), s_PtRgbMapArray[s_AlphaMapIndex]);
    }
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampU"), s_PtClampArray[s_ClampUIndex]);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptClampV"), s_PtClampArray[s_ClampVIndex]);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftU"), s_PtShiftArray[s_ShiftUIndex]);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptShiftV"), s_PtShiftArray[s_ShiftVIndex]);

    // Noise Parameter
    if (s_EnableNoiseU) 
    {
        s_NoiseValueUArray[0] = (s_EnableNoiseF_U ? s_SelectNoiseF_U * 0.05f : 0.0f);
        s_NoiseValueUArray[1] = //(s_EnableNoiseAutoP_U ? f * 0.01f : 
                            (s_EnableNoiseAutoP_U ? s_FrameCount * s_NoiseUPhaseStep : 
                            (s_EnableNoiseP_U ? s_SelectNoiseP_U * 0.05f : 0.0f));
        s_NoiseValueUArray[2] = (s_EnableNoiseA_U ? s_SelectNoiseA_U * 0.05f : 0.0f);
    } 
    else 
    {
        s_NoiseValueUArray[0] = 0.0f;
        s_NoiseValueUArray[1] = 0.0f;
        s_NoiseValueUArray[2] = 0.0f;
    }
    
    if (s_EnableNoiseV) 
    {
        s_NoiseValueVArray[0] = (s_EnableNoiseF_V ? s_SelectNoiseF_V * 0.05f : 0.0f);
        s_NoiseValueVArray[1] = //(s_EnableNoiseAutoP_V ? f * 0.01f : 
                            (s_EnableNoiseAutoP_V ? s_FrameCount * s_NoiseVPhaseStep : 
                            (s_EnableNoiseP_V ? s_SelectNoiseP_V * 0.05f : 0.0f));
        s_NoiseValueVArray[2] = (s_EnableNoiseA_V ? s_SelectNoiseA_V * 0.05f : 0.0f);
    } 
    else 
    {
        s_NoiseValueVArray[0] = 0.0f;
        s_NoiseValueVArray[1] = 0.0f;
        s_NoiseValueVArray[2] = 0.0f;
    }

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseEnable"), 
        ((s_EnableNoiseU || s_EnableNoiseV) ? GL_TRUE : GL_FALSE));
    glUniform3fv(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseU"), 1, s_NoiseValueUArray);
    glUniform3fv(glGetUniformLocation(s_PgID, "dmp_Texture[3].ptNoiseV"), 1, s_NoiseValueVArray);

    // Texture Combiner
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_Texture[0].samplerType"), GL_FALSE);

    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineRgb"), GL_REPLACE);
    glUniform1i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandRgb"), 
        GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].operandAlpha"), 
        GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), 
        GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), 
        GL_TEXTURE3, GL_CONSTANT, GL_CONSTANT);
}

void SetPresetData(u8* data, int mode)
{
    int i;
    memset(data, 0, 128);
    switch(mode)
    {
        case 0:     // x
            for (i = 0; i < 128; i++) data[i] = i;
            break;
        case 1:     // 1-x
            for (i = 0; i < 128; i++) data[i] = 128 - i;
            break;
        case 2:     // x^2
            for (i = 0; i < 128; i++) data[i] = i * i / 128;
            break;
        case 3:     // 1-x^2
            for (i = 0; i < 128; i++) data[i] = 128 - (i * i / 128);
            break;
        case 4:     // (2x-1)^2
            for (i = 0; i < 128; i++) data[i] = (2 * i - 128) * (2 * i - 128) / 128;
            break;
        case 5:     // 1-(2x-1)^2
            for (i = 0; i < 128; i++) data[i] = 128 - ((2 * i - 128) * (2 * i - 128) / 128);
            break;
        case 6:     // Triangle
            for (i = 0; i < 128; i++) data[i] = ((i < 64) ? (i * 2) : (128 - (i - 64) * 2));
            break;
        case 7:     // Triangle Inv
            for (i = 0; i < 128; i++) data[i] = 128 - ((i < 64) ? (i * 2) : (128 - (i - 64) * 2));
            break;
        case 8:     // 2 Triangle
            for (i =  0; i <  64; i++) data[i] = ((i < 32) ? (i        * 4) : (128 - (i - 32) * 4));
            for (i = 64; i < 128; i++) data[i] = ((i < 96) ? ((i - 64) * 4) : (128 - (i - 96) * 4));
            break;
        case 9:     // 2 Triangle Inv
            for (i =  0; i <  64; i++) data[i] = 128 - ((i < 32) ? (i        * 4) : (128 - (i - 32) * 4));
            for (i = 64; i < 128; i++) data[i] = 128 - ((i < 96) ? ((i - 64) * 4) : (128 - (i - 96) * 4));
            break;
        case 10:     // 4 Triangle
            for (i =  0; i <  32; i++) data[i] = ((i <  16) ? (i        * 8) : (128 - (i -  16) * 8));
            for (i = 32; i <  64; i++) data[i] = ((i <  48) ? ((i - 32) * 8) : (128 - (i -  48) * 8));
            for (i = 64; i <  96; i++) data[i] = ((i <  80) ? ((i - 64) * 8) : (128 - (i -  80) * 8));
            for (i = 96; i < 128; i++) data[i] = ((i < 112) ? ((i - 96) * 8) : (128 - (i - 112) * 8));
            break;
        case 11:    // 4 Triangle Inv
            for (i =  0; i <  32; i++) data[i] = 128 - ((i <  16) ? (i        * 8) : (128 - (i -  16) * 8));
            for (i = 32; i <  64; i++) data[i] = 128 - ((i <  48) ? ((i - 32) * 8) : (128 - (i -  48) * 8));
            for (i = 64; i <  96; i++) data[i] = 128 - ((i <  80) ? ((i - 64) * 8) : (128 - (i -  80) * 8));
            for (i = 96; i < 128; i++) data[i] = 128 - ((i < 112) ? ((i - 96) * 8) : (128 - (i - 112) * 8));
            break;
        case 12:    // sin(pi/2)
            for (i = 0; i < 128; i++) data[i] = 128 * nn::math::SinDeg(90.0f * i / 128.0f);
            break;
        case 13:    // cos(pi/2)
            for (i = 0; i < 128; i++) data[i] = 128 * nn::math::CosDeg(90.0f * i / 128.0f);
            break;
        case 14:    // sin(pi)
            for (i = 0; i < 128; i++) data[i] = 128 * nn::math::SinDeg(180.0f * i / 128.0f);
            break;
        case 15:    // 1-sin(pi)
            for (i = 0; i < 128; i++) data[i] = 128 - (128 * nn::math::SinDeg(180.0f * i / 128.0f));
            break;
        case 16:    // (1+sin(2pi))/2
            for (i = 0; i < 128; i++) data[i] = 64 + (64 * nn::math::SinDeg(360.0f * i / 128.0f));
            break;
        case 17:    // (1-sin(2pi))/2
            for (i = 0; i < 128; i++) data[i] = 64 - (64 * nn::math::SinDeg(360.0f * i / 128.0f));
            break;
        case 18:    // (1+cos(2pi))/2
            for (i = 0; i < 128; i++) data[i] = 64 + (64 * nn::math::CosDeg(360.0f * i / 128.0f));
            break;
        case 19:    // (1-cos(2pi))/2
            for (i = 0; i < 128; i++) data[i] = 64 - (64 * nn::math::CosDeg(360.0f * i / 128.0f));
            break;
        case 20:    // stair(2pt)
            for (i =  0; i <  64; i++) data[i] = 0;
            for (i = 64; i < 128; i++) data[i] = 128;
            break;
        case 21:    // stair(2pt) Inv
            for (i =  0; i <  64; i++) data[i] = 128;
            for (i = 64; i < 128; i++) data[i] = 0;
            break;
        case 22:    // stair(4pt)
            for (i =  0; i <  32; i++) data[i] = 0;
            for (i = 32; i <  64; i++) data[i] = 43;
            for (i = 64; i <  96; i++) data[i] = 85;
            for (i = 96; i < 128; i++) data[i] = 128;
            break;
        case 23:    // stair(4pt) Inv
            for (i =  0; i <  32; i++) data[i] = 128;
            for (i = 32; i <  64; i++) data[i] = 85;
            for (i = 64; i <  96; i++) data[i] = 43;
            for (i = 96; i < 128; i++) data[i] = 0;
            break;
        case 24:    // stair(8pt)
            for (i =   0; i <  16; i++) data[i] = 0;
            for (i =  16; i <  32; i++) data[i] = 18;
            for (i =  32; i <  48; i++) data[i] = 37;
            for (i =  48; i <  64; i++) data[i] = 55;
            for (i =  64; i <  80; i++) data[i] = 73;
            for (i =  80; i <  96; i++) data[i] = 91;
            for (i =  96; i < 112; i++) data[i] = 110;
            for (i = 112; i < 128; i++) data[i] = 128;
            break;
        case 25:    // stair(8pt) Inv
            for (i =   0; i <  16; i++) data[i] = 128;
            for (i =  16; i <  32; i++) data[i] = 110;
            for (i =  32; i <  48; i++) data[i] = 91;
            for (i =  48; i <  64; i++) data[i] = 73;
            for (i =  64; i <  80; i++) data[i] = 55;
            for (i =  80; i <  96; i++) data[i] = 37;
            for (i =  96; i < 112; i++) data[i] = 18;
            for (i = 112; i < 128; i++) data[i] = 0;
            break;
        case 26:    // pyram(3pt)
            for (i =  0; i <  43; i++) data[i] = 0;
            for (i = 43; i <  85; i++) data[i] = 128;
            for (i = 85; i < 128; i++) data[i] = 0;
            break;
        case 27:    // pyram(3pt) Inv
            for (i =  0; i <  43; i++) data[i] = 128;
            for (i = 43; i <  85; i++) data[i] = 0;
            for (i = 85; i < 128; i++) data[i] = 128;
            break;
        case 28:    // pyram(7pt)
            for (i =   0; i <  18; i++) data[i] = 0;
            for (i =  18; i <  37; i++) data[i] = 43;
            for (i =  37; i <  55; i++) data[i] = 85;
            for (i =  55; i <  73; i++) data[i] = 128;
            for (i =  73; i <  91; i++) data[i] = 85;
            for (i =  91; i < 110; i++) data[i] = 43;
            for (i = 110; i < 128; i++) data[i] = 0;
            break;
        case 29:    // pyram(7pt) Inv
            for (i =   0; i <  18; i++) data[i] = 128;
            for (i =  18; i <  37; i++) data[i] = 85;
            for (i =  37; i <  55; i++) data[i] = 43;
            for (i =  55; i <  73; i++) data[i] = 0;
            for (i =  73; i <  91; i++) data[i] = 43;
            for (i =  91; i < 110; i++) data[i] = 85;
            for (i = 110; i < 128; i++) data[i] = 128;
            break;
        case 30:    // pyram(15pt)
            for (i =   0; i <   9; i++) data[i] = 0;
            for (i =   9; i <  17; i++) data[i] = 18;
            for (i =  17; i <  26; i++) data[i] = 37;
            for (i =  26; i <  34; i++) data[i] = 55;
            for (i =  34; i <  43; i++) data[i] = 73;
            for (i =  43; i <  51; i++) data[i] = 91;
            for (i =  51; i <  60; i++) data[i] = 110;
            for (i =  60; i <  68; i++) data[i] = 128;
            for (i =  68; i <  77; i++) data[i] = 110;
            for (i =  77; i <  85; i++) data[i] = 91;
            for (i =  85; i <  94; i++) data[i] = 73;
            for (i =  94; i < 102; i++) data[i] = 55;
            for (i = 102; i < 111; i++) data[i] = 37;
            for (i = 111; i < 119; i++) data[i] = 18;
            for (i = 119; i < 128; i++) data[i] = 0;
            break;
        case 31:    // pyram(15pt) Inv
            for (i =   0; i <   9; i++) data[i] = 128;
            for (i =   9; i <  17; i++) data[i] = 110;
            for (i =  17; i <  26; i++) data[i] = 91;
            for (i =  26; i <  34; i++) data[i] = 73;
            for (i =  34; i <  43; i++) data[i] = 55;
            for (i =  43; i <  51; i++) data[i] = 37;
            for (i =  51; i <  60; i++) data[i] = 18;
            for (i =  60; i <  68; i++) data[i] = 0;
            for (i =  68; i <  77; i++) data[i] = 18;
            for (i =  77; i <  85; i++) data[i] = 37;
            for (i =  85; i <  94; i++) data[i] = 55;
            for (i =  94; i < 102; i++) data[i] = 73;
            for (i = 102; i < 111; i++) data[i] = 91;
            for (i = 111; i < 119; i++) data[i] = 110;
            for (i = 119; i < 128; i++) data[i] = 128;
            break;
        case 32:    // (const) 0.0
            for (i = 0; i < 128; i++) data[i] = 0;
            break;
        case 33:    // (const) 0.5
            for (i = 0; i < 128; i++) data[i] = 64;
            break;
        case 34:    // (const) 1.0
            for (i = 0; i < 128; i++) data[i] = 128;
            break;
        case 35:    // Sample Red
            data[ 15] =  64;
            data[ 16] = 128;
            data[ 17] =  64;
            data[ 31] =  64;
            data[ 32] = 128;
            data[ 33] =  64;
            data[ 79] =  64;
            data[ 80] = 128;
            data[ 81] =  64;
            data[ 95] =  64;
            data[ 96] = 128;
            data[ 97] =  64;
            break;
        case 36:    // Sample Green
            data[ 15] =  64;
            data[ 16] = 128;
            data[ 17] =  64;
            data[ 31] =  64;
            data[ 32] = 128;
            data[ 33] =  64;
            data[ 47] =  64;
            data[ 48] = 128;
            data[ 49] =  64;
            data[ 63] =  64;
            data[ 64] = 128;
            data[ 65] =  64;
            break;
        case 37:    // Sample Blue
            data[ 15] =  64;
            data[ 16] = 128;
            data[ 17] =  64;
            data[ 47] =  64;
            data[ 48] = 128;
            data[ 49] =  64;
            data[ 79] =  64;
            data[ 80] = 128;
            data[ 81] =  64;
            data[111] =  64;
            data[112] = 128;
            data[113] =  64;
            break;
        case 38:    // 3x^2-2x^3
            for (i = 0; i < 128; i++) {
                float x = i / 128.0f;
                data[i] = (3 * x * x - 2 * x * x * x) * 128;
            }
            break;
        default:
            break;
    }
}

void SetLookUpTable_256(GLuint texture, u8* data)
{
    GLfloat table[256];
    int i;

    memset(table, 0, 256 * sizeof(GLfloat));

    for (i = 0; i < 128; i++) 
    {
        table[i] = data[i] / 128.0f;
    }

    for (i = 0; i < 127; i++) 
    {
        table[i + 128] = table[i + 1] - table[i];
    }

    table[128 + 127] = 0.0f;

    glBindTexture(GL_LUT_TEXTURE0_DMP, texture);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 256, 0, 
        GL_LUMINANCEF_DMP, GL_FLOAT, table);
}

void SetLookUpTable_512(GLuint texture, u8* data)
{
    GLfloat table[512];
    u32 i;

    memset(table, 0, 512 * sizeof(GLfloat));
    
    u32 incrNum = MAX_TEX_WIDTH / s_TexWidth;

    u32 index = 0;
    for (i = 0; i < 128; )
    {
        table[ index ] = data[ i ] / 128.0f;

        index += 1;
        i += incrNum;
    }
    
    index = 0;
    for (i = 0; i < 127; )
    {
        table[index + 256] = table[i + 1] - table[i];
        index += 1;
        i += incrNum;
    }
    table[256 + 127] = 0.0f;

    glBindTexture(GL_LUT_TEXTURE0_DMP, texture);
    glTexImage1D(GL_LUT_TEXTURE0_DMP, 0, GL_LUMINANCEF_DMP, 512, 0, 
        GL_LUMINANCEF_DMP, GL_FLOAT, table);
}

void ResetCamera( void )
{
   s_CamDistance = 1.5f;
    
   s_WorldPos = nn::math::VEC3( 0.0f, 0.0f, 0.0f );
   s_WorldEulerDeg = nn::math::VEC3( 0.0f, 0.0f, 0.0f );

   s_ObjScalePosX = 1.0f;
   s_ObjScalePosY = 1.0f;
}

void SetDefaultSettings( void )
{ 
    ResetCamera();

    s_SelectTableIndex = BUTTON_ID_SELECT_FC;
    s_SelectButtonIndex = -1;
    s_ButtonDownIndex = s_ButtonUpIndex = -1;
    s_TouchX = s_TouchY = -1;
    s_WavePositionX = 0;
    s_PresetIndex = 0;
    s_EnableAlphaSeparate = false;
    s_RgbMapIndex = 0;
    s_AlphaMapIndex = 0;
    s_EnableSamplerR = s_EnableSamplerG = s_EnableSamplerB = s_EnableSamplerA = true;
    s_ClampUIndex = s_ClampVIndex = 3;
    s_ShiftUIndex = s_ShiftVIndex = 0;
    s_EnableNoiseU = s_EnableNoiseV = false;
    s_EnableNoiseF_U = s_EnableNoiseF_V = true;
    s_EnableNoiseP_U = s_EnableNoiseP_V = true;
    s_EnableNoiseAutoP_U = s_EnableNoiseAutoP_V = false;
    s_EnableNoiseA_U = s_EnableNoiseA_V = true;
    s_NoiseValueUArray[0] = s_NoiseValueUArray[1] = s_NoiseValueUArray[2] = 0.0f;
    s_NoiseValueVArray[0] = s_NoiseValueVArray[1] = s_NoiseValueVArray[2] = 0.0f;
    s_SelectNoiseF_U = s_SelectNoiseP_U = s_SelectNoiseA_U = 0;
    s_SelectNoiseF_V = s_SelectNoiseP_V = s_SelectNoiseA_V = 0;
    s_MinTexCoordS = s_MinTexCoordT = 0;
    s_MaxTexCoordS = s_MaxTexCoordT = 1.f / TEXCOORD_SCALE;//10;
    s_ColorBG_R = s_ColorBG_G = s_ColorBG_B = 5;
    
    UpdateProcedureTextureDataArray();

    LoadProcedureTextureData();
}

void ChangePosArray( void )
{    
    GLfloat tmpCoords[] = 
    {
        -0.5f, 0.5f, 0.0f,
        -0.5f,-0.5f, 0.0f,
         0.5f, 0.5f, 0.0f,
         0.5f,-0.5f, 0.0f
    };

    tmpCoords[ 0 ] *= s_ObjScalePosX;
    tmpCoords[ 1 ] *= s_ObjScalePosY;

    tmpCoords[ 3 ] *= s_ObjScalePosX;
    tmpCoords[ 4 ] *= s_ObjScalePosY;

    tmpCoords[ 6 ] *= s_ObjScalePosX;
    tmpCoords[ 7 ] *= s_ObjScalePosY;

    tmpCoords[ 9  ] *= s_ObjScalePosX;
    tmpCoords[ 10 ] *= s_ObjScalePosY;

    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferSubData( GL_ARRAY_BUFFER, 0, sizeof(tmpCoords), tmpCoords );
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void ChangeTexCoordArray( void )
{    
    s_TexCoords[0] = - s_MinTexCoordS * TEXCOORD_SCALE;
    s_TexCoords[1] = s_MaxTexCoordT * TEXCOORD_SCALE;

    s_TexCoords[2] = - s_MinTexCoordS * TEXCOORD_SCALE;
    s_TexCoords[3] = - s_MinTexCoordT * TEXCOORD_SCALE;

    s_TexCoords[4] = s_MaxTexCoordS * TEXCOORD_SCALE;
    s_TexCoords[5] = s_MaxTexCoordT * TEXCOORD_SCALE;

    s_TexCoords[6] = s_MaxTexCoordS * TEXCOORD_SCALE;
    s_TexCoords[7] = - s_MinTexCoordT * TEXCOORD_SCALE;

    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(s_PositionCoords), sizeof(s_TexCoords), s_TexCoords);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}


void MakeTextureEditView(u8* data)
{
    int pos = 0;
    memset(imageEditView, 0, (EDITVIEW_TEX_WIDTH * EDITVIEW_TEX_HEIGHT * EDITVIEW_TEX_PER_PIXEL));

    for (int i = 0; i < 128; i++) 
    {
        pos = CalcNativeDataPosition(i, data[i], 
            EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT) * EDITVIEW_TEX_PER_PIXEL;
        imageEditView[pos + 0] = 255;
        imageEditView[pos + 1] = 255;
        imageEditView[pos + 2] = 255;
        imageEditView[pos + 3] = 255;
    }

    /* Draw Cursor */
    if ((s_WavePositionX >= 0) && (s_WavePositionX <= 127)) 
    {
        u8 x = s_WavePositionX;
        u8 y = data[s_WavePositionX];
        if (((x - 1) >= 0) && ((y - 1) >= 0)) 
        {
            pos = CalcNativeDataPosition((x - 1), (y - 1), 
                EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT) * EDITVIEW_TEX_PER_PIXEL;
            imageEditView[pos + 0] = 255;
            imageEditView[pos + 1] = 0;
            imageEditView[pos + 2] = 255;
            imageEditView[pos + 3] = 255;
        }

        if (((x - 1) >= 0) && ((y + 1) <= 128)) 
        {
            pos = CalcNativeDataPosition((x - 1), (y + 1), 
                EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT) * EDITVIEW_TEX_PER_PIXEL;
            imageEditView[pos + 0] = 255;
            imageEditView[pos + 1] = 0;
            imageEditView[pos + 2] = 255;
            imageEditView[pos + 3] = 255;
        }

        if (((x + 1) < 128) && ((y - 1) >= 0)) 
        {
            pos = CalcNativeDataPosition((x + 1), (y - 1), 
                EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT) * EDITVIEW_TEX_PER_PIXEL;
            imageEditView[pos + 0] = 255;
            imageEditView[pos + 1] = 0;
            imageEditView[pos + 2] = 255;
            imageEditView[pos + 3] = 255;
        }

        if (((x + 1) < 128) && ((y + 1) <= 128))
        {
            pos = CalcNativeDataPosition((x + 1), (y + 1), 
                EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT) * EDITVIEW_TEX_PER_PIXEL;
            imageEditView[pos + 0] = 255;
            imageEditView[pos + 1] = 0;
            imageEditView[pos + 2] = 255;
            imageEditView[pos + 3] = 255;
        }
    }

    if (s_TexEditView != 0) 
    {
        s_RenderSystem.DeleteTexture(s_TexEditView);
    }

    GLuint texid = 0;
    
    s_RenderSystem.GenerateTexture(GL_TEXTURE_2D, EDITVIEW_TEX_FORMAT, 
        EDITVIEW_TEX_WIDTH, EDITVIEW_TEX_HEIGHT, EDITVIEW_TEX_FORMAT, EDITVIEW_TEX_TYPE, imageEditView, texid);
    
    if (texid != 0)
    {
        s_TexEditView = texid;
    }    
}

int CalcNativeDataPosition(int x, int y, int texWidth, int texHeight)
{
    static u8 index[64] = 
    {
        42, 43, 46, 47, 58, 59, 62, 63,
        40, 41, 44, 45, 56, 57, 60, 61,
        34, 35, 38, 39, 50, 51, 54, 55,
        32, 33, 36, 37, 48, 49, 52, 53,
        10, 11, 14, 15, 26, 27, 30, 31,
         8,  9, 12, 13, 24, 25, 28, 29,
         2,  3,  6,  7, 18, 19, 22, 23,
         0,  1,  4,  5, 16, 17, 20, 21
    };
    int blockX, blockY, indX, indY, blockPerLine;

    blockPerLine = texWidth / 8;
    blockX = x / 8;
    blockY = (texHeight - (y == texHeight ? y : (y + 1))) / 8;
    indX = x % 8;
    indY = y % 8;
    return ((64 * (blockPerLine * blockY + blockX)) + (index[indX + indY * 8]));
}

void MakeTextureColorView( void )
{
    int i, j, pos;
    memset(s_ImageColorViewArray, 0, (COLORVIEW_TEX_WIDTH * COLORVIEW_TEX_HEIGHT * COLORVIEW_TEX_PER_PIXEL));

    u32 incrNum = MAX_TEX_WIDTH / s_TexWidth;

    u32 tableIndex = 0;

    i = 0;
    for ( tableIndex = 0; tableIndex < 128; )
    {
        // 2
        for (u32 index = 0; index < incrNum; index++)
        {
            // 8
            for (j = 0; j < 8; j++) 
            {
                pos = CalcNativeDataPosition( tableIndex, j, 
                    COLORVIEW_TEX_WIDTH, COLORVIEW_TEX_HEIGHT) * COLORVIEW_TEX_PER_PIXEL;

                s_ImageColorViewArray[pos + 0] = 255;
                s_ImageColorViewArray[pos + 1] = ((s_FuncSamplerBArray[i] == 0) ? 0 : (s_FuncSamplerBArray[i] * 2 - 1));
                s_ImageColorViewArray[pos + 2] = ((s_FuncSamplerGArray[i] == 0) ? 0 : (s_FuncSamplerGArray[i] * 2 - 1));
                s_ImageColorViewArray[pos + 3] = ((s_FuncSamplerRArray[i] == 0) ? 0 : (s_FuncSamplerRArray[i] * 2 - 1));
            }

            tableIndex += 1;
        }

        i += incrNum;
    }
    
    if (s_TexColorView != 0)
    {
        s_RenderSystem.DeleteTexture(s_TexColorView);
    }
    GLuint texid = 0;
    
    s_RenderSystem.GenerateTexture(GL_TEXTURE_2D, COLORVIEW_TEX_FORMAT, 
         COLORVIEW_TEX_WIDTH, COLORVIEW_TEX_HEIGHT, COLORVIEW_TEX_FORMAT, COLORVIEW_TEX_TYPE, s_ImageColorViewArray, texid);
    if ( texid != 0 )
    {
        s_TexColorView = texid;
    }
}

void MakeTextureAlphaView( void )
{
    int pos;
    memset(s_ImageAlphaViewArray, 0, 
        (COLORVIEW_TEX_WIDTH * COLORVIEW_TEX_HEIGHT * COLORVIEW_TEX_PER_PIXEL));

    // In independent alpha mode, display the Fa color table
    if ( s_EnableAlphaSeparate )
    {
        for ( u32 i = 0; i < 128; i++ )
        {            
            for ( u32 j = 0; j < 8; j++ ) 
            {
                pos = CalcNativeDataPosition( i, j,
                    COLORVIEW_TEX_WIDTH, COLORVIEW_TEX_HEIGHT) * COLORVIEW_TEX_PER_PIXEL;
                
                s_ImageAlphaViewArray[pos + 0] = 255;
                s_ImageAlphaViewArray[pos + 1] = ((s_FuncAlphaArray[i] == 0) ? 0 : (s_FuncAlphaArray[i] * 2 - 1));
                s_ImageAlphaViewArray[pos + 2] = ((s_FuncAlphaArray[i] == 0) ? 0 : (s_FuncAlphaArray[i] * 2 - 1));
                s_ImageAlphaViewArray[pos + 3] = ((s_FuncAlphaArray[i] == 0) ? 0 : (s_FuncAlphaArray[i] * 2 - 1));                    
            }
        }
    }
    // In shared alpha mode, display the alpha color table
    else
    {
        u32 incrNum = MAX_TEX_WIDTH / s_TexWidth;
        u32 i = 0;
        for ( u32 tableIndex = 0; tableIndex < 128; )
        {
            // 2
            for ( u32 index = 0; index < incrNum; index++ )
            {
                // 8
                for ( u32 j = 0; j < 8; j++) 
                {
                    pos = CalcNativeDataPosition( tableIndex, j, 
                        COLORVIEW_TEX_WIDTH, COLORVIEW_TEX_HEIGHT) * COLORVIEW_TEX_PER_PIXEL;

                
                    s_ImageAlphaViewArray[pos + 0] = 255;
                    s_ImageAlphaViewArray[pos + 1] = ((s_FuncSamplerAArray[i] == 0) ? 0 : (s_FuncSamplerAArray[i] * 2 - 1));
                    s_ImageAlphaViewArray[pos + 2] = ((s_FuncSamplerAArray[i] == 0) ? 0 : (s_FuncSamplerAArray[i] * 2 - 1));
                    s_ImageAlphaViewArray[pos + 3] = ((s_FuncSamplerAArray[i] == 0) ? 0 : (s_FuncSamplerAArray[i] * 2 - 1));                    
                }

                tableIndex += 1;
            }

            i += incrNum;
        }
    }

    if (s_TexAlphaView != 0)
    {
        s_RenderSystem.DeleteTexture(s_TexAlphaView);
    }
    
    GLuint texid = 0;
    
    s_RenderSystem.GenerateTexture(GL_TEXTURE_2D, COLORVIEW_TEX_FORMAT, 
        COLORVIEW_TEX_WIDTH, COLORVIEW_TEX_HEIGHT, COLORVIEW_TEX_FORMAT,
        COLORVIEW_TEX_TYPE, s_ImageAlphaViewArray, texid);

    if ( texid != 0 )
    {
        s_TexAlphaView = texid;
    }
}

void MakeTextureViewers(GLuint texid, u8* data, bool updateColor, bool updateEditView )
{
    if (updateColor) 
    {
        SetLookUpTable_512(texid, data);
        MakeTextureColorView();
    } 
    else 
    {
        SetLookUpTable_256(texid, data);
    }

    MakeTextureAlphaView();

    if ( updateEditView )
    {
        MakeTextureEditView(data);
    }
}

void DrawFrameUpper(void)
{
    nngxUpdateState( NN_GX_STATE_ALL );
    
    /* Start Drawing */
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    glUseProgram(s_PgID);
    glViewport(0, 0, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT );

    const f32 DISPLAY0_ASPECT = static_cast<f32>(nn::gx::CTR::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::CTR::DISPLAY0_WIDTH);

    nn::math::Matrix44 proj;
    nn::math::MTX44PerspectivePivotDeg(&proj, 45.0f, 
        DISPLAY0_ASPECT,        
        0.1f, 20.0f,
        nn::math::PIVOT_UPSIDE_TO_TOP);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    /* VertexBuffer */
    SetVertexBuffer();

    /* Texture */
    glBindTexture(GL_TEXTURE_COLLECTION_DMP, s_TexColl);
    SetTextureEnvironment();

    nn::math::Matrix34 eye;
        
    nn::math::VEC3 camPos(0.f, 0.0f, s_CamDistance);
    nn::math::VEC3 camUp(0.f, 1.f, 0.f);

    nn::math::Vector3 target(0.f, 0.f, 0.f);
    
    nn::math::MTX34Identity(&eye);
    nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);

    nn::math::MTX34 objPos;
    nn::math::MTX34Identity( &objPos );    
    nn::math::MTX34Translate( &objPos, s_WorldPos );
    
    nn::math::MTX34 objRot;
    nn::math::MTX34Identity( &objRot );
    nn::math::MTX34RotXYZDeg( &objRot, 
        s_WorldEulerDeg.x, s_WorldEulerDeg.y, s_WorldEulerDeg.z );
    
    nn::math::MTX34Mult(&eye, &eye, &objPos);
    nn::math::MTX34Mult(&eye, &eye, &objRot);
    nn::math::Matrix44 mv(eye);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, (f32*)(mv));

    /* Blending */
    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    /* Polygon */
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);

    glBindTexture(GL_TEXTURE_COLLECTION_DMP, 0);
    
    u32 strLen = strlen( s_ProcedureTextureFileName );
    f32 x = 200.0f - 0.5f * strLen * 8.0f;

    if ( ( s_ColorBG_R + s_ColorBG_G + s_ColorBG_B ) > 5 )
    {
        s_RenderSystem.SetColor(0.0f, 0.0f, 0.0f, 1.0f);
    }
    else
    {
        s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    }
    s_RenderSystem.DrawText(x, 0.0f, s_ProcedureTextureFileName );

    if ( s_CaptureFlag )
    {    
        s_RenderSystem.DrawText( 80.0f, 220.0f,
            "Saving screen shot to SD-Card.");
    }
}

void SetVertexBuffer(void)
{
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(s_PositionCoords));
}

void DrawFrameLower(void)
{
    int i;
    GLfloat color[4];
    bool bSelect;
    bool bDisable;

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    glViewport(0, 0, nn::gx::DISPLAY1_WIDTH, nn::gx::DISPLAY1_HEIGHT);


    /* DrawButtons */
    for (i = 0; i < NUM_BUTTON; i++) 
    {
        if (s_ButtonArray[i].ID < 0) 
        {
            continue;
        }

        bSelect = bDisable = false;

        switch (s_ButtonArray[i].ID) 
        {
            case BUTTON_ID_SELECT_FC:
            case BUTTON_ID_SELECT_FA:
            case BUTTON_ID_SELECT_R:
            case BUTTON_ID_SELECT_G:
            case BUTTON_ID_SELECT_B:
            case BUTTON_ID_SELECT_NZ:
            case BUTTON_ID_SELECT_A:
                if ( i == BUTTON_ID_SELECT_FA )
                {
                    if ( s_EnableAlphaSeparate )
                    {
                        bSelect = (i == s_SelectTableIndex);
                    }
                    else
                    {
                        bDisable = true;
                    }
                }
                else if ( i == BUTTON_ID_SELECT_A )
                {
                    if ( s_EnableAlphaSeparate )
                    {
                        bDisable = true;
                    }
                    else
                    {
                        bSelect = (i == s_SelectTableIndex);
                    }
                }
                else
                {
                    bSelect = (i == s_SelectTableIndex);
                }
                break;

            case BUTTON_ID_SEPARATE:
                break;
            case BUTTON_ID_ENABLE_R:
                bDisable = !s_EnableSamplerR;
                break;
            case BUTTON_ID_ENABLE_G:
                bDisable = !s_EnableSamplerG;
                break;
            case BUTTON_ID_ENABLE_B:
                bDisable = !s_EnableSamplerB;
                break;
            case BUTTON_ID_ENABLE_A:
                bDisable = !s_EnableSamplerA;
                break;
            case BUTTON_ID_ENABLE_N_U:
                // bDisable = !s_EnableNoiseU;
                break;
            case BUTTON_ID_ENABLE_N_V:
                // bDisable = !s_EnableNoiseV;
                break;
            case BUTTON_ID_ENABLE_N_U_F:
                bDisable = !s_EnableNoiseF_U;
                break;
            case BUTTON_ID_MINUS_N_U_F:
                bDisable = !(s_SelectNoiseF_U > 0);
                break;
            case BUTTON_ID_PLUS_N_U_F:
                bDisable = !(s_SelectNoiseF_U < MAX_NOISE_VALUE);
                break;
            case BUTTON_ID_ENABLE_N_U_P:
                bDisable = !s_EnableNoiseP_U;
                break;
            case BUTTON_ID_MINUS_N_U_P:
                bDisable = !(s_SelectNoiseP_U > 0);
                break;
            case BUTTON_ID_PLUS_N_U_P:
                bDisable = !(s_SelectNoiseP_U < MAX_NOISE_VALUE);
                break;
            case BUTTON_ID_ENABLE_N_U_PA:
                // bDisable = !s_EnableNoiseAutoP_U;
                break;
            case BUTTON_ID_ENABLE_N_U_A:
                bDisable = !s_EnableNoiseA_U;
                break;
            case BUTTON_ID_MINUS_N_U_A:
                bDisable = !(s_SelectNoiseA_U > 0);
                break;
            case BUTTON_ID_PLUS_N_U_A:
                bDisable = !(s_SelectNoiseA_U < MAX_NOISE_VALUE);
                break;
            case BUTTON_ID_ENABLE_N_V_F:
                bDisable = !s_EnableNoiseF_V;
                break;
            case BUTTON_ID_MINUS_N_V_F:
                bDisable = !(s_SelectNoiseF_V > 0);
                break;
            case BUTTON_ID_PLUS_N_V_F:
                bDisable = !(s_SelectNoiseF_V < MAX_NOISE_VALUE);
                break;
            case BUTTON_ID_ENABLE_N_V_P:
                bDisable = !s_EnableNoiseP_V;
                break;
            case BUTTON_ID_ENABLE_N_V_PA:
                // bDisable = !s_EnableNoiseAutoP_V;
                break;
            case BUTTON_ID_MINUS_N_V_P:
                bDisable = !(s_SelectNoiseP_V > 0);
                break;
            case BUTTON_ID_PLUS_N_V_P:
                bDisable = !(s_SelectNoiseP_V < MAX_NOISE_VALUE);
                break;
            case BUTTON_ID_ENABLE_N_V_A:
                bDisable = !s_EnableNoiseA_V;
                break;
            case BUTTON_ID_MINUS_N_V_A:
                bDisable = !(s_SelectNoiseA_V > 0);
                break;
            case BUTTON_ID_PLUS_N_V_A:
                bDisable = !(s_SelectNoiseA_V < MAX_NOISE_VALUE);
                break;

            case BUTTON_ID_MINUS_S_MIN:
                bDisable = !(s_MinTexCoordS < MAX_TEXCOORD_VALUE);
                break;

            case BUTTON_ID_PLUS_S_MIN:
                bDisable = !(s_MinTexCoordS > 0);
                break;

            case BUTTON_ID_MINUS_S_MAX:
                bDisable = !(s_MaxTexCoordS > 0);
                break;

            case BUTTON_ID_PLUS_S_MAX:
                bDisable = !(s_MaxTexCoordS < MAX_TEXCOORD_VALUE);
                break;

            case BUTTON_ID_MINUS_T_MIN:
                bDisable = !(s_MinTexCoordT < MAX_TEXCOORD_VALUE);
                break;
            case BUTTON_ID_PLUS_T_MIN:
                bDisable = !(s_MinTexCoordT > 0);
                break;

            case BUTTON_ID_MINUS_T_MAX:
                bDisable = !(s_MaxTexCoordT > 0);
                break;

            case BUTTON_ID_PLUS_T_MAX:
                bDisable = !(s_MaxTexCoordT < MAX_TEXCOORD_VALUE);
                break;
            case BUTTON_ID_ALPHA_PATTERN_PREV:
                if (! s_EnableAlphaSeparate )
                {
                    bDisable = true;
                }
                break;
            case BUTTON_ID_ALPHA_PATTERN_NEXT:
                if ( ! s_EnableAlphaSeparate )
                {
                    bDisable = true;
                }
                break;

            case BUTTON_ID_TEX_WIDTH_PREV:
                if ( s_TexWidth == MIN_TEX_WIDTH )
                {
                    bDisable = true;
                }
                break;

            case BUTTON_ID_TEX_WIDTH_NEXT:
                if ( s_TexWidth == MAX_TEX_WIDTH )
                {
                    bDisable = true;
                }
                break;

            }

            // Current Button
            if (i == s_SelectButtonIndex) 
            {
                color[0] = s_ButtonArray[i].touchColor[0];
                color[1] = s_ButtonArray[i].touchColor[1];
                color[2] = s_ButtonArray[i].touchColor[2];
                color[3] = s_ButtonArray[i].touchColor[3];
            }
            else if (bSelect) 
            {
                // Select
                color[0] = s_ButtonArray[i].selectColor[0];
                color[1] = s_ButtonArray[i].selectColor[1];
                color[2] = s_ButtonArray[i].selectColor[2];
                color[3] = s_ButtonArray[i].selectColor[3];
            }
            else if (bDisable) 
            {
                // Disable
                color[0] = s_ButtonArray[i].disableColor[0];
                color[1] = s_ButtonArray[i].disableColor[1];
                color[2] = s_ButtonArray[i].disableColor[2];
                color[3] = s_ButtonArray[i].disableColor[3];
            } 
            else 
            {
                // Default
                color[0] = s_ButtonArray[i].defaultColor[0];
                color[1] = s_ButtonArray[i].defaultColor[1];
                color[2] = s_ButtonArray[i].defaultColor[2];
                color[3] = s_ButtonArray[i].defaultColor[3];
            }
        
        s_RenderSystem.SetColor(color[0], color[1], color[2], color[3]);

        // Renders the button.
        if(i != BUTTON_ID_EDITVIEW)
        {
            s_RenderSystem.FillRectangle(s_ButtonArray[i].x * 1.0f, s_ButtonArray[i].y * 1.0f, 
                s_ButtonArray[i].width * 1.0f, s_ButtonArray[i].height * 1.0f);
        }

    }

    /* EditView */    
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_ONE, GL_ZERO);

    s_RenderSystem.FillTexturedRectangle(s_TexEditView, 
        16.0f, 15.0f, 128.0f, 129.0f, 128.0f, 129.0f, 128.0f, 256.0f);

    //glEnable(GL_BLEND);
    //glBlendEquation(GL_FUNC_ADD);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    /* ColorView */
    s_RenderSystem.FillTexturedRectangle(s_TexColorView, 
        16.0f, 144.0f, 128.0f, 8.0f, 128.0f, 8.0f, 128.0f, 8.0f);
    
    /* AlphaView */
    { 
        const f32 yOffset = 11.0f;

        s_RenderSystem.FillTexturedRectangle( s_TexAlphaView, 
            16.0f, 144.0f + yOffset, 128.0f, 8.0f,
            128.0f, 8.0f, 128.0f, 8.0f);
    }

    /* LabelText */
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
      
    s_RenderSystem.DrawText( 16.0f, 166.0f,
        "LUT size : %d", s_TexWidth );

    s_RenderSystem.DrawText(168.0f,   8.0f, "Load");
        
    s_RenderSystem.DrawText(168.0f,  36.0f, "Preset");
    
    s_RenderSystem.DrawText(192.0f,  48.0f, "%s", s_NamePresetArray[s_PresetIndex]);
    
    s_RenderSystem.DrawText(168.0f,  64.0f, "Gc(x)");
    s_RenderSystem.DrawText(224.0f + 8.0f,  64.0f, "%s", s_NameRgbMapArray[s_RgbMapIndex]);

    {
        const f32 yOffset = 10.0f;

        s_RenderSystem.DrawText(168.0f, 64.0f + yOffset, "Ga(x)");
        s_RenderSystem.DrawText(224.0f + 8.0f, 64.0f + yOffset, "%s", s_NameRgbMapArray[s_AlphaMapIndex]);
    }

    {
        const f32 yOffset = 7.0f;

        s_RenderSystem.DrawText(168.0f,  80.0f + yOffset, "Color");
        s_RenderSystem.DrawText(192.0f,  92.0f + yOffset, "BG");
        s_RenderSystem.DrawText(288.0f,  92.0f + yOffset, "x.1");
    }

    s_RenderSystem.DrawText(168.0f, 108.0f, "Noise");
    s_RenderSystem.DrawText(168.0f, 120.0f, "U");
    s_RenderSystem.DrawText(168.0f, 156.0f, "V");
    s_RenderSystem.DrawText(240.0f, 120.0f, "%1.2f", (s_SelectNoiseF_U * 0.05f));
    s_RenderSystem.DrawText(240.0f, 132.0f, "%1.2f", (s_SelectNoiseP_U * 0.05f));
    s_RenderSystem.DrawText(240.0f, 144.0f, "%1.2f", (s_SelectNoiseA_U * 0.05f));
    s_RenderSystem.DrawText(240.0f, 156.0f, "%1.2f", (s_SelectNoiseF_V * 0.05f));
    s_RenderSystem.DrawText(240.0f, 168.0f, "%1.2f", (s_SelectNoiseP_V * 0.05f));
    s_RenderSystem.DrawText(240.0f, 180.0f, "%1.2f", (s_SelectNoiseA_V * 0.05f));

    s_RenderSystem.DrawText(184.0f, 196.0f, "Clamp");
    s_RenderSystem.DrawText(264.0f, 196.0f, "Shift");
    s_RenderSystem.DrawText(168.0f, 208.0f, "U");
    s_RenderSystem.DrawText(168.0f, 224.0f, "V");

        
    {
        const f32 yOffset = 3.0f;

        s_RenderSystem.DrawText( 24.0f, 196.0f + yOffset, "Min");
        s_RenderSystem.DrawText( 88.0f, 196.0f + yOffset, "Max");

        s_RenderSystem.DrawText(  8.0f, 208.0f + yOffset, "S");

        s_RenderSystem.DrawText( 48.0f, 208.0f + yOffset, "%2.2f", (s_MinTexCoordS * - TEXCOORD_SCALE));
        s_RenderSystem.DrawText(112.0f, 208.0f + yOffset, "%2.2f", (s_MaxTexCoordS * TEXCOORD_SCALE));
    }
    s_RenderSystem.DrawText(  8.0f, 224.0f, "T");
    s_RenderSystem.DrawText( 48.0f, 224.0f, "%2.2f", (s_MinTexCoordT * - TEXCOORD_SCALE));
    s_RenderSystem.DrawText(112.0f, 224.0f, "%2.2f", (s_MaxTexCoordT * TEXCOORD_SCALE));

    /* ButtonText */
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_FC].x, s_ButtonArray[BUTTON_ID_SELECT_FC].y, "Fc");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_FA].x, s_ButtonArray[BUTTON_ID_SELECT_FA].y, "Fa");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_R].x, s_ButtonArray[BUTTON_ID_SELECT_R].y, "R");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_G].x, s_ButtonArray[BUTTON_ID_SELECT_G].y, "G");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_B].x, s_ButtonArray[BUTTON_ID_SELECT_B].y, "B");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_A].x, s_ButtonArray[BUTTON_ID_SELECT_A].y, "A");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SELECT_NZ].x, s_ButtonArray[BUTTON_ID_SELECT_NZ].y, "Nz");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SEPARATE].x, s_ButtonArray[BUTTON_ID_SEPARATE].y, (s_EnableAlphaSeparate ? "Separate" : "Unified"));

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PRESET_PREV].x, s_ButtonArray[BUTTON_ID_PRESET_PREV].y, "<");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PRESET_NEXT].x, s_ButtonArray[BUTTON_ID_PRESET_NEXT].y, ">");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PRESET_COPY].x, s_ButtonArray[BUTTON_ID_PRESET_COPY].y, "Copy");

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_RGB_PATTERN_PREV].x, s_ButtonArray[BUTTON_ID_RGB_PATTERN_PREV].y, "<");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_RGB_PATTERN_NEXT].x, s_ButtonArray[BUTTON_ID_RGB_PATTERN_NEXT].y, ">");

    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_PREV].x, s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_PREV].y, "<");
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_NEXT].x, s_ButtonArray[BUTTON_ID_ALPHA_PATTERN_NEXT].y, ">");
    }

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_R].x, s_ButtonArray[BUTTON_ID_ENABLE_R].y, "R");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_G].x, s_ButtonArray[BUTTON_ID_ENABLE_G].y, "G");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_B].x, s_ButtonArray[BUTTON_ID_ENABLE_B].y, "B");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_A].x, s_ButtonArray[BUTTON_ID_ENABLE_A].y, "A");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U].y, (s_EnableNoiseU ? "ON" : "OFF"));
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V].y, (s_EnableNoiseV ? "ON" : "OFF"));
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U_F].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U_F].y, "F");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_U_F].x, s_ButtonArray[BUTTON_ID_MINUS_N_U_F].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_U_F].x, s_ButtonArray[BUTTON_ID_PLUS_N_U_F].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U_P].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U_P].y, "P");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_U_P].x, s_ButtonArray[BUTTON_ID_MINUS_N_U_P].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_U_P].x, s_ButtonArray[BUTTON_ID_PLUS_N_U_P].y, "+");

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_TEX_WIDTH_PREV].x, s_ButtonArray[BUTTON_ID_TEX_WIDTH_PREV].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_TEX_WIDTH_NEXT].x, s_ButtonArray[BUTTON_ID_TEX_WIDTH_NEXT].y, "+");

    if ( s_EnableNoiseAutoP_U )
    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U_PA].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U_PA].y, 
            "%1.3f", s_NoiseUPhaseStep );
    }
    else
    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U_PA].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U_PA].y, 
            "Stop");
    }
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_U_PA].x, s_ButtonArray[BUTTON_ID_MINUS_N_U_PA].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_U_PA].x, s_ButtonArray[BUTTON_ID_PLUS_N_U_PA].y, "+");

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_U_A].x, s_ButtonArray[BUTTON_ID_ENABLE_N_U_A].y, "A");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_U_A].x, s_ButtonArray[BUTTON_ID_MINUS_N_U_A].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_U_A].x, s_ButtonArray[BUTTON_ID_PLUS_N_U_A].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_F].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_F].y, "F");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_V_F].x, s_ButtonArray[BUTTON_ID_MINUS_N_V_F].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_V_F].x, s_ButtonArray[BUTTON_ID_PLUS_N_V_F].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_P].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_P].y, "P");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_V_P].x, s_ButtonArray[BUTTON_ID_MINUS_N_V_P].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_V_P].x, s_ButtonArray[BUTTON_ID_PLUS_N_V_P].y, "+");

    // s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].y, (s_EnableNoiseAutoP_V ? "Auto" : "Stop"));
    if ( s_EnableNoiseAutoP_V )
    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].y,
            "%1.3f", s_NoiseVPhaseStep );
    }
    else
    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_PA].y, 
            "Stop");
    }
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_V_PA].x, s_ButtonArray[BUTTON_ID_MINUS_N_V_PA].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_V_PA].x, s_ButtonArray[BUTTON_ID_PLUS_N_V_PA].y, "+");

    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_ENABLE_N_V_A].x, s_ButtonArray[BUTTON_ID_ENABLE_N_V_A].y, "A");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_N_V_A].x, s_ButtonArray[BUTTON_ID_MINUS_N_V_A].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_N_V_A].x, s_ButtonArray[BUTTON_ID_PLUS_N_V_A].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_CLAMP_U].x, s_ButtonArray[BUTTON_ID_CLAMP_U].y, "%s", s_NameClampArray[s_ClampUIndex]);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SHIFT_U].x, s_ButtonArray[BUTTON_ID_SHIFT_U].y, "%s", s_NameShiftArray[s_ShiftUIndex]);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_CLAMP_V].x, s_ButtonArray[BUTTON_ID_CLAMP_V].y, "%s", s_NameClampArray[s_ClampVIndex]);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SHIFT_V].x, s_ButtonArray[BUTTON_ID_SHIFT_V].y, "%s", s_NameShiftArray[s_ShiftVIndex]);

    {
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_S_MIN].x, s_ButtonArray[BUTTON_ID_MINUS_S_MIN].y, "-");
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_S_MIN].x, s_ButtonArray[BUTTON_ID_PLUS_S_MIN].y, "+");
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_S_MAX].x, s_ButtonArray[BUTTON_ID_MINUS_S_MAX].y, "-");
        s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_S_MAX].x, s_ButtonArray[BUTTON_ID_PLUS_S_MAX].y, "+");
    }
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_T_MIN].x, s_ButtonArray[BUTTON_ID_MINUS_T_MIN].y, "-");    
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_T_MIN].x, s_ButtonArray[BUTTON_ID_PLUS_T_MIN].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_MINUS_T_MAX].x, s_ButtonArray[BUTTON_ID_MINUS_T_MAX].y, "-");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_PLUS_T_MAX].x, s_ButtonArray[BUTTON_ID_PLUS_T_MAX].y, "+");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_PREV].x, s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_PREV].y, "<");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_NEXT].x, s_ButtonArray[BUTTON_ID_LOAD_WAVEDATA_NEXT].y, ">");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_SAVE_WAVEDATA].x, s_ButtonArray[BUTTON_ID_SAVE_WAVEDATA].y, "Save");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_CAPTURE].x, s_ButtonArray[BUTTON_ID_CAPTURE].y, "Capture");
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_BGCOLOR_R].x, s_ButtonArray[BUTTON_ID_BGCOLOR_R].y, "%2d", s_ColorBG_R);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_BGCOLOR_G].x, s_ButtonArray[BUTTON_ID_BGCOLOR_G].y, "%2d", s_ColorBG_G);
    s_RenderSystem.DrawText(s_ButtonArray[BUTTON_ID_BGCOLOR_B].x, s_ButtonArray[BUTTON_ID_BGCOLOR_B].y, "%2d", s_ColorBG_B);

    /* Cursor */
    float centerX, centerY, left, right, top, bottom;
    if (s_TouchX >= 0 && s_TouchY >= 0) 
    {
        s_RenderSystem.SetColor(1.0f, 0.0f, 0.0f, 1.0f);
        s_RenderSystem.SetLineWidth(1.0f);
        centerX = s_TouchX * 1.0f;
        centerY = s_TouchY * 1.0f;
        left = ((s_TouchX > 2) ? ((s_TouchX - 2) * 1.0f) : 0.0f);
        right = ((s_TouchX < 318) ? ((s_TouchX + 2) * 1.0f) : 319.0f);
        top = ((s_TouchY > 2) ? ((s_TouchY - 2) * 1.0f) : 0.0f);
        bottom = ((s_TouchY < 238) ? ((s_TouchY + 2) * 1.0f) : 239.0f);
        s_RenderSystem.DrawLine(left, centerY, right, centerY);
        s_RenderSystem.DrawLine(centerX, top, centerX, bottom);
    }
    
    if ((s_WavePositionX >= 0) && (s_WavePositionX <= 127)) 
    {
        u8* data = NULL;
        switch (s_SelectTableIndex) 
        {
            case BUTTON_ID_SELECT_FC:
                data = s_FuncRGBArray;
                break;
            case BUTTON_ID_SELECT_FA:
                data = s_FuncAlphaArray;
                break;
            case BUTTON_ID_SELECT_NZ:
                data = s_FuncNoiseArray;
                break;
            case BUTTON_ID_SELECT_R:
                data = s_FuncSamplerRArray;
                break;
            case BUTTON_ID_SELECT_G:
                data = s_FuncSamplerGArray;
                break;
            case BUTTON_ID_SELECT_B:
                data = s_FuncSamplerBArray;
                break;
            case BUTTON_ID_SELECT_A:
                data = s_FuncSamplerAArray;
                break;
            default:
                break;
        }

        if (data != NULL) 
        {
            s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);

            f32 yOffset = 4.0f;
            s_RenderSystem.DrawText( 24.0f, 168.0f + 4.0f + yOffset,
                "In (%3d)%1.3f", s_WavePositionX, (s_WavePositionX / 128.0f));
            s_RenderSystem.DrawText( 24.0f, 180.0f + 4.0f + yOffset,
                "Out(%3d)%1.3f", data[s_WavePositionX], (data[s_WavePositionX] / 128.0f));
        }
    }
}

void SetClearColor(void)
{        
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, s_ColorBG_R * 0.1f, s_ColorBG_G * 0.1f, s_ColorBG_B * 0.1f, 1.0f);
}

int DrawFrame(void)
{        
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);

    SetClearColor();
    
    s_RenderSystem.Clear();
    DrawFrameUpper();

    s_RenderSystem.SwapBuffers();

    // If the capture flag is set, capture the image
    if ( s_CaptureFlag )
    {       
        CaptureScreen();
    }
    
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
    DrawFrameLower();
    s_RenderSystem.SwapBuffers();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
    
    return !glGetError();
}

void ChangeWaveValue( int value, bool overwrite )
{
    int tmp;

    switch (s_SelectTableIndex) 
    {
    case BUTTON_ID_SELECT_FC:
        tmp = (overwrite ? 0 : s_FuncRGBArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncRGBArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[0], s_FuncRGBArray, false);
        break;

    case BUTTON_ID_SELECT_FA:
        tmp = (overwrite ? 0 : s_FuncAlphaArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncAlphaArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[1], s_FuncAlphaArray, false);
        break;

    case BUTTON_ID_SELECT_NZ:
        tmp = (overwrite ? 0 : s_FuncNoiseArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncNoiseArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[2], s_FuncNoiseArray, false);
        break;

    case BUTTON_ID_SELECT_R:
        tmp = (overwrite ? 0 : s_FuncSamplerRArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncSamplerRArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[3], s_FuncSamplerRArray, true);
        break;

    case BUTTON_ID_SELECT_G:
        tmp = (overwrite ? 0 : s_FuncSamplerGArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncSamplerGArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[4], s_FuncSamplerGArray, true);
        break;

    case BUTTON_ID_SELECT_B:
        tmp = (overwrite ? 0 : s_FuncSamplerBArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncSamplerBArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[5], s_FuncSamplerBArray, true);
        break;

    case BUTTON_ID_SELECT_A:
        tmp = (overwrite ? 0 : s_FuncSamplerAArray[s_WavePositionX]) + value;
        tmp = (tmp < 0 ? 0 : tmp > 128 ? 128 : tmp);
        s_FuncSamplerAArray[s_WavePositionX] = tmp;
        MakeTextureViewers(s_TexID[6], s_FuncSamplerAArray, true);
        break;

    default:
        return;
    }
}

void CopyPresetTable()
{
    switch (s_SelectTableIndex) {
    case BUTTON_ID_SELECT_FC:
        SetPresetData(s_FuncRGBArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[0], s_FuncRGBArray, false);
        break;
    case BUTTON_ID_SELECT_FA:
        SetPresetData(s_FuncAlphaArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[1], s_FuncAlphaArray, false);
        break;
    case BUTTON_ID_SELECT_NZ:
        SetPresetData(s_FuncNoiseArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[2], s_FuncNoiseArray, false);
        break;
    case BUTTON_ID_SELECT_R:
        SetPresetData(s_FuncSamplerRArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[3], s_FuncSamplerRArray, true);
        break;
    case BUTTON_ID_SELECT_G:
        SetPresetData(s_FuncSamplerGArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[4], s_FuncSamplerGArray, true);
        break;
    case BUTTON_ID_SELECT_B:
        SetPresetData(s_FuncSamplerBArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[5], s_FuncSamplerBArray, true);
        break;
    case BUTTON_ID_SELECT_A:
        SetPresetData(s_FuncSamplerAArray, s_PresetIndex);
        MakeTextureViewers(s_TexID[6], s_FuncSamplerAArray, true);
        break;
    default:
        return;
    }
}

void SetProcedureTextureData( const ProcedureTextureData& procedureTextureData )
{        
    s_RgbMapIndex   = GetProcTexMapIndex( procedureTextureData.rgbMap );

    s_EnableAlphaSeparate = procedureTextureData.alphaSeparate;

    if ( s_EnableAlphaSeparate )
    {
        s_AlphaMapIndex = GetProcTexMapIndex( procedureTextureData.alphaMap );
    }
    else
    {
        s_AlphaMapIndex = GetProcTexMapIndex( procedureTextureData.rgbMap );
    }    

    s_ClampUIndex = GetProcTexClampIndex( procedureTextureData.clampU );
    s_ClampVIndex = GetProcTexClampIndex( procedureTextureData.clampV );
    
    s_ShiftUIndex = GetProcTexShiftIndex( procedureTextureData.shiftU );
    s_ShiftVIndex = GetProcTexShiftIndex( procedureTextureData.shiftV );
                   
    // Replace the default value
    s_TexFilter = GL_LINEAR;
    //procedureTextureData.texBias   = 0.5f;
    s_TexOffset = procedureTextureData.texOffset;
    s_TexWidth = procedureTextureData.texWidth;
   
    for (u32 index = 0; index < PROCTEX_DATA_LOOKUP_TABLE_ELEMENT_NUM; index++)
    {
        s_FuncRGBArray[ index ] = procedureTextureData.rgbMapLookUpTable[ index ] * 128.0f;
        s_FuncAlphaArray[ index ] = procedureTextureData.alphaMapLookUpTable[ index ] * 128.0f;
        s_FuncNoiseArray[ index ] = procedureTextureData.noiseLookUpTable[ index ] * 128.0f;
    }

    for (u32 index = 0; index < PROCTEX_DATA_LOOKUP_TABLE_ELEMENT_NUM; index++)
    {
        s_FuncSamplerRArray[ index ] = procedureTextureData.redLookUpTable[ index ] * 128.0f;
        s_FuncSamplerGArray[ index ] = procedureTextureData.greenLookUpTable[ index ] * 128.0f;
        s_FuncSamplerBArray[ index ] = procedureTextureData.blueLookUpTable[ index ] * 128.0f;
        s_FuncSamplerAArray[ index ] = procedureTextureData.alphaLookUpTable[ index ] * 128.0f;
    }

    s_EnableNoiseU = procedureTextureData.isEnableNoiseU;
    s_EnableNoiseV = procedureTextureData.isEnableNoiseV;
    
    s_SelectNoiseA_U = procedureTextureData.noiseUAmplitude / 0.05f;
    s_SelectNoiseF_U = procedureTextureData.noiseUFrequency / 0.05f;
    s_SelectNoiseP_U = procedureTextureData.noiseUPhase / 0.05f;
    
    s_SelectNoiseA_V = procedureTextureData.noiseVAmplitude / 0.05f;
    s_SelectNoiseF_V = procedureTextureData.noiseVFrequency / 0.05f;
    s_SelectNoiseP_V = procedureTextureData.noiseVPhase / 0.05f;

    if ( procedureTextureData.noiseUPhaseStep != 0.0f )
    {
        s_EnableNoiseAutoP_U = true;
        s_NoiseUPhaseStep = procedureTextureData.noiseUPhaseStep;
    }
    else
    {
        s_EnableNoiseAutoP_U = false;
    }

    if ( procedureTextureData.noiseVPhaseStep != 0.0f )
    {
        s_EnableNoiseAutoP_V = true;
        s_NoiseVPhaseStep = procedureTextureData.noiseVPhaseStep;
    }
    else
    {
        s_EnableNoiseAutoP_V = false;
    }

    MakeTextureViewers(s_TexID[1], s_FuncAlphaArray, false);
    MakeTextureViewers(s_TexID[2], s_FuncNoiseArray, false);

    MakeTextureViewers(s_TexID[3], s_FuncSamplerRArray, true);
    MakeTextureViewers(s_TexID[4], s_FuncSamplerGArray, true);
    MakeTextureViewers(s_TexID[5], s_FuncSamplerBArray, true);
    MakeTextureViewers(s_TexID[6], s_FuncSamplerAArray, true);

    MakeTextureViewers(s_TexID[0], s_FuncRGBArray, false);
    s_SelectTableIndex = BUTTON_ID_SELECT_FC;


    s_ColorBG_R = procedureTextureData.clearColor[ 0 ] * 10.0f;
    s_ColorBG_G = procedureTextureData.clearColor[ 1 ] * 10.0f;
    s_ColorBG_B = procedureTextureData.clearColor[ 2 ] * 10.0f;


    s_MinTexCoordS = - procedureTextureData.texCoordU[ 0 ] / TEXCOORD_SCALE;
    s_MaxTexCoordS = procedureTextureData.texCoordU[ 1 ] / TEXCOORD_SCALE;

    s_MinTexCoordT = - procedureTextureData.texCoordV[ 0 ] / TEXCOORD_SCALE;
    s_MaxTexCoordT = procedureTextureData.texCoordV[ 1 ] / TEXCOORD_SCALE;

    ChangeTexCoordArray();
}

void ConvertProcedureTextureData( ProcedureTextureData& procedureTextureData )
{  
    procedureTextureData.header = PROCEDURE_TEXTURE_DATA_HEADER;
    procedureTextureData.version = PROCEDURE_TEXTURE_DATA_VERSION;

    procedureTextureData.rgbMap   = GetPicaProcTexMap( s_PtRgbMapArray[ s_RgbMapIndex ] );

    procedureTextureData.alphaSeparate = s_EnableAlphaSeparate;
    if ( procedureTextureData.alphaSeparate )
    {
        procedureTextureData.alphaMap = GetPicaProcTexMap( s_PtRgbMapArray[ s_AlphaMapIndex ] );
    }
    else
    {
        procedureTextureData.alphaMap = GetPicaProcTexMap( s_PtRgbMapArray[ s_RgbMapIndex ] );
    }

    procedureTextureData.clampU = GetPicaProcTexClamp( s_PtClampArray[ s_ClampUIndex ] );
    procedureTextureData.clampV = GetPicaProcTexClamp( s_PtClampArray[ s_ClampVIndex ] );
    
    procedureTextureData.shiftU = GetPicaProcTexShift( s_PtShiftArray[ s_ShiftUIndex ] );
    procedureTextureData.shiftV = GetPicaProcTexShift( s_PtShiftArray[ s_ShiftVIndex ] );
           
    // Replace the default value
    procedureTextureData.minFilter = PICA_DATA_PROCTEX_LINEAR;
    procedureTextureData.texBias   = 0.5f;
    procedureTextureData.texOffset = s_TexOffset;
    procedureTextureData.texWidth  = s_TexWidth;

    for (u32 index = 0; index < PROCTEX_DATA_LOOKUP_TABLE_ELEMENT_NUM; index++)
    {
        procedureTextureData.rgbMapLookUpTable[ index ]   = s_FuncRGBArray[ index ]    / 128.0f;
        procedureTextureData.alphaMapLookUpTable[ index ] = s_FuncAlphaArray[ index ]  / 128.0f;
        procedureTextureData.noiseLookUpTable[ index ]    = s_FuncNoiseArray[ index ]  / 128.0f;
    }

    // Fill the first 128 pieces of data
    for (u32 index = 0; index < PROCTEX_DATA_COLOR_LOOKUP_TABLE_ELEMENT_HALF_NUM; index++)
    {
        procedureTextureData.redLookUpTable[ index ]   = s_FuncSamplerRArray[ index ] / 128.0f;
        procedureTextureData.greenLookUpTable[ index ] = s_FuncSamplerGArray[ index ] / 128.0f;
        procedureTextureData.blueLookUpTable[ index ]  = s_FuncSamplerBArray[ index ] / 128.0f;
        procedureTextureData.alphaLookUpTable[ index ] = s_FuncSamplerAArray[ index ] / 128.0f;
    }

    // Initialize the last 128 pieces of data to 0.0f
    for (u32 index = 0; index < PROCTEX_DATA_COLOR_LOOKUP_TABLE_ELEMENT_HALF_NUM; index++)
    {        
        u32 offset = index + PROCTEX_DATA_COLOR_LOOKUP_TABLE_ELEMENT_HALF_NUM;
        
        procedureTextureData.redLookUpTable[ offset ] = 0.0f;
        procedureTextureData.greenLookUpTable[ offset ] = 0.0f;
        procedureTextureData.blueLookUpTable[ offset ] = 0.0f;
        procedureTextureData.alphaLookUpTable[ offset ] = 0.0f;
    }
    
    procedureTextureData.isEnableNoise = 
        ( s_EnableNoiseU | s_EnableNoiseV ) ? true : false;

    procedureTextureData.isEnableNoiseU = s_EnableNoiseU;
    procedureTextureData.isEnableNoiseV = s_EnableNoiseV;

    procedureTextureData.noiseUAmplitude = s_SelectNoiseA_U * 0.05f;
    procedureTextureData.noiseUFrequency = s_SelectNoiseF_U * 0.05f;
    procedureTextureData.noiseUPhase     = s_SelectNoiseP_U * 0.05f;

    procedureTextureData.noiseVAmplitude = s_SelectNoiseA_V * 0.05f;
    procedureTextureData.noiseVFrequency = s_SelectNoiseF_V * 0.05f;
    procedureTextureData.noiseVPhase     = s_SelectNoiseP_V * 0.05f;

    if ( s_EnableNoiseAutoP_U )
    {
        procedureTextureData.noiseUPhaseStep = s_NoiseUPhaseStep;
    }
    else
    {
        procedureTextureData.noiseUPhaseStep = 0.0f;
    }

    if ( s_EnableNoiseAutoP_V )
    {
        procedureTextureData.noiseVPhaseStep = s_NoiseVPhaseStep;
    }
    else
    {
        procedureTextureData.noiseVPhaseStep = 0.0f;
    }

    procedureTextureData.clearColor[ 0 ] = s_ColorBG_R * 0.1f;
    procedureTextureData.clearColor[ 1 ] = s_ColorBG_G * 0.1f;
    procedureTextureData.clearColor[ 2 ] = s_ColorBG_B * 0.1f;
    procedureTextureData.clearColor[ 3 ] = 1.0f;

    procedureTextureData.texCoordU[ 0 ] = - s_MinTexCoordS * TEXCOORD_SCALE;
    procedureTextureData.texCoordU[ 1 ] = s_MaxTexCoordS * TEXCOORD_SCALE;

    procedureTextureData.texCoordV[ 0 ] = - s_MinTexCoordT * TEXCOORD_SCALE;
    procedureTextureData.texCoordV[ 1 ] = s_MaxTexCoordT * TEXCOORD_SCALE;
}


PicaDataProcTexMap GetPicaProcTexMap( GLuint rgbMap )
{
    if ( rgbMap == GL_PROCTEX_U_DMP )
    {
        return PICA_DATA_PROCTEX_U_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_V_DMP )
    {
        return PICA_DATA_PROCTEX_V_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_U2_DMP )
    {
        return PICA_DATA_PROCTEX_U2_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_V2_DMP )
    {
        return PICA_DATA_PROCTEX_V2_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_ADD_DMP )
    {
        return PICA_DATA_PROCTEX_ADD_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_ADD2_DMP )
    {
        return PICA_DATA_PROCTEX_ADD2_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_ADDSQRT2_DMP )
    {
        return PICA_DATA_PROCTEX_ADDSQRT2_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_MIN_DMP )
    {
        return PICA_DATA_PROCTEX_MIN_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_MAX_DMP )
    {
        return PICA_DATA_PROCTEX_MAX_DMP;
    }
    else if ( rgbMap == GL_PROCTEX_RMAX_DMP )
    {
        return PICA_DATA_PROCTEX_RMAX_DMP;        
    }
    else
    {
        NN_LOG( "Unsupported rbgMap format = %d\n", rgbMap );
        NN_ASSERT( 0 );

        return PICA_DATA_PROCTEX_U_DMP;
    }
}

PicaDataProcTexClamp GetPicaProcTexClamp( GLuint clamp )
{
    if ( clamp == GL_SYMMETRICAL_REPEAT_DMP )
    {        
        return PICA_DATA_PROCTEX_SYMMETRICAL_REPEAT_DMP;
    }
    else if ( clamp == GL_MIRRORED_REPEAT )
    {
        return PICA_DATA_PROCTEX_MIRRORED_REPEAT;
    }
    else if ( clamp == GL_PULSE_DMP )
    {
        return PICA_DATA_PROCTEX_PULSE_DMP;
    }
    else if ( clamp == GL_CLAMP_TO_EDGE )
    {
        return PICA_DATA_PROCTEX_CLAMP_TO_EDGE;
    }
    else if ( clamp == GL_CLAMP_TO_ZERO_DMP )
    {
        return PICA_DATA_PROCTEX_CLAMP_TO_ZERO_DMP;
    }
    else
    {
        NN_LOG( "Unsupported clamp mode = %d\n", clamp );
        NN_ASSERT( 0 );

        return PICA_DATA_PROCTEX_SYMMETRICAL_REPEAT_DMP;
    }
}

PicaDataProcTexShift GetPicaProcTexShift( GLuint shift )
{
    if ( shift == GL_ODD_DMP )
    {
        return PICA_DATA_PROCTEX_ODD_DMP;
    }
    else if ( shift == GL_EVEN_DMP )
    {
        return PICA_DATA_PROCTEX_EVEN_DMP;
    }
    else if ( shift == GL_NONE_DMP )
    {
        return PICA_DATA_PROCTEX_NONE_DMP;
    }
    else
    {
        NN_LOG( "Unsupported shift mode = %d\n", shift );
        NN_ASSERT( 0 );

        return PICA_DATA_PROCTEX_NONE_DMP;
    }
}

u32 GetProcTexMapIndex( PicaDataProcTexMap rgbMap )
{
    if ( rgbMap == PICA_DATA_PROCTEX_U_DMP)    
    {
        return 0;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_V_DMP )
    {
        return 1;        
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_U2_DMP )
    {
        return 2;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_V2_DMP ) 
    {
        return 3;        
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_ADD_DMP )
    {
        return 4;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_ADD2_DMP )
    {
        return 5;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_ADDSQRT2_DMP )
    {
        return 6;        
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_MIN_DMP )
    {
        return 7;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_MAX_DMP )
    {
        return 8;
    }
    else if ( rgbMap == PICA_DATA_PROCTEX_RMAX_DMP ) 
    {
        return 9;        
    }
    else
    {
        NN_LOG( "Unsupported rbgMap format = %d\n", rgbMap );
        NN_ASSERT( 0 );

        return 0;
    }
}

u32 GetProcTexClampIndex( PicaDataProcTexClamp clamp )
{
    if ( clamp == PICA_DATA_PROCTEX_SYMMETRICAL_REPEAT_DMP )
    {        
        return 0;
    }
    else if ( clamp == PICA_DATA_PROCTEX_MIRRORED_REPEAT )
    {
        return 1;
    }
    else if ( clamp == PICA_DATA_PROCTEX_PULSE_DMP )
    {
        return 2;
    }
    else if ( clamp == PICA_DATA_PROCTEX_CLAMP_TO_EDGE )
    {
        return 3;
    }
    else if ( clamp == PICA_DATA_PROCTEX_CLAMP_TO_ZERO_DMP )
    {
        return 4;
    }
    else
    {
        NN_LOG( "Unsupported clamp mode = %d\n", clamp );
        NN_ASSERT( 0 );

        return 5;
    }
}

u32 GetProcTexShiftIndex( PicaDataProcTexShift shift )
{
    if ( shift == PICA_DATA_PROCTEX_NONE_DMP )
    {
        return 0;
    }
    else if ( shift == PICA_DATA_PROCTEX_ODD_DMP )
    {
        return 1;
    }
    else if ( shift == PICA_DATA_PROCTEX_EVEN_DMP )
    {
        return 2;
    }
    else
    {
        NN_LOG( "Unsupported shift mode = %d\n", shift );
        NN_ASSERT( 0 );

        return 0;
    }
}

void GetFormatCharBuffer( char* char_buffer, const u32 char_buffer_size, const char* format, ...)
{
    va_list va;
    va_start(va, format);
    vsnprintf( char_buffer, char_buffer_size, format, va);
    va_end(va);
}

void GetTimeCharBuffer( char* char_buffer, const u32 char_buffer_size )
{   
    nn::fnd::DateTime dateTime = nn::fnd::DateTime::GetNow();

    GetFormatCharBuffer( char_buffer, char_buffer_size,
        "%4d%02d%02d_%02d%02d%02d",
        dateTime.GetYear(),
        dateTime.GetMonth(),
        dateTime.GetDay(),
        dateTime.GetHour(),
        dateTime.GetMinute(),
        dateTime.GetSecond() );
}

void GetTimeWCharBuffer( wchar_t* wchar_buffer )
{
    const u32 char_buffer_size = 1024;
    char    src_char_buffer[ char_buffer_size ];

    GetTimeCharBuffer( src_char_buffer, char_buffer_size );
    std::mbstowcs( wchar_buffer, src_char_buffer, std::strlen( src_char_buffer ) + 1 );
}

void GetProcedureTextureDataFileName( wchar_t* wchar_buffer, const u32 wchar_buffer_size )
{

    _wcsncpy( wchar_buffer, wchar_buffer_size, L"sdmc:/proctex_data_" );
    
    wchar_t tmpWCharBuffer[ 1024 ];
    GetTimeWCharBuffer( tmpWCharBuffer );

    _wcsncat( wchar_buffer, wchar_buffer_size, tmpWCharBuffer );
    _wcsncat( wchar_buffer, wchar_buffer_size, L".pctxbin" );
}

bool SaveProcedureTextureData( wchar_t* fileName )
{
    nn::Result result;

    nn::fs::FileOutputStream fileOutputStream;
        
    result = fileOutputStream.TryInitialize( fileName , true );
    if ( result.IsFailure() )
    {
        return false;
    }

    s32 saveSize = 0;
    result = fileOutputStream.TryWrite(&saveSize, 
        &s_ProcedureTextureData, sizeof( ProcedureTextureData ), true );

    fileOutputStream.Finalize();

    if ( result.IsFailure() )    
    {
        return false;
    }

    return true;
}

bool LoadProcedureTextureData( void )
{   
    if ( s_ProcedureTextureFileDataNum == 0 )
    {
        return false;
    }

    ProcedureTextureFileData& fileData = 
        s_ProcedureTextureFileDataArray[ s_ProcedureTextureFileDataIndex ];

    s_ProcedureTextureFileName = fileData.name;
    const wchar_t* fileName = fileData.fileName;

    if ( ReadProcedureTextureData( fileName, s_ProcedureTextureData ) )
    { 
        SetProcedureTextureData( s_ProcedureTextureData );

        return true;
    }
    else
    {        
        return false;
    }
}

void WCharBufferToCharBuffer(const wchar_t* src_wchar_buffer, char* dst_char_buffer )
{
    u32 wcs_length = wcslen( src_wchar_buffer );
    const wchar_t* tmp_wchar_ptr = src_wchar_buffer;
    wcsrtombs( dst_char_buffer, &(tmp_wchar_ptr),
               wcs_length + 1,
               NULL );
}


bool ReadProcedureTextureDataArray( const wchar_t* dir )
{    
    nn::fs::Directory directory( dir );
       
    while ( 1 )
    {
        nn::fs::DirectoryEntry entryArray[1];
        s32 entryNum = directory.Read(entryArray,
                                      sizeof(entryArray) / sizeof(*entryArray) );

        if ( entryNum == 0 )
        {            
            break;
        }        

        char nameBuffer[ MAX_PROCTEX_CHAR_NAME_LENGTH ];
        nn::fs::DirectoryEntry& entry = entryArray[0];
        
        WCharBufferToCharBuffer( entry.entryName, nameBuffer );

        u32 wnameLength = wcslen( entry.entryName );

        if (! entry.attributes.isDirectory )
        {
            ProcedureTextureFileData& procTexFileData = 
                s_ProcedureTextureFileDataArray[ s_ProcedureTextureFileDataNum ];
            
            const wchar_t ext[  ] = L".pctxbin";
            
            const u32 extLength = wcslen( ext );
            bool match = true;

            for ( u32 index = 0; index < extLength; index++)
            {
                u32 offset = wnameLength - extLength + index;
                if ( entry.entryName[ offset ] != ext[ index ] )
                {
                    match = false;
                    break;
                }
            }

            if ( match )
            {
                procTexFileData.valid = true;

                _strcpy( procTexFileData.name, MAX_PROCTEX_CHAR_NAME_LENGTH,
                    nameBuffer );

                _wcsncpy( procTexFileData.fileName, MAX_PROCTEX_WCHAR_FILENAME_LENGTH, dir );
                _wcsncat( procTexFileData.fileName, MAX_PROCTEX_WCHAR_FILENAME_LENGTH, entry.entryName );

                s_ProcedureTextureFileDataNum += 1;
            }
        }
    }

    directory.Finalize();
    
    if ( s_ProcedureTextureFileDataNum > 0 )
    {
        return true;
    }
    else
    {
        return false;
    }
}

void MountSdmc( void )
{    
    if ( nn::fs::IsSdmcInserted() )
    {
        if (! s_MountSdmc )
        {
            nn::fs::Unmount( "sdmc:" );

            nn::Result result = nn::fs::MountSdmc();
            if ( result.IsFailure() )
            {
                NN_LOG("[Status] Warning : No SD Card. \n");
                s_MountSdmc = false;
            }
            else
            {
                NN_LOG("[Status] Found SD Card. \n");
                s_MountSdmc = true;
            }
        }
        else
        {
            // Do nothing
        }
    }
    else
    {    
       NN_LOG("[Status] Warning : No SD Card. \n");
       s_MountSdmc = false;
    }
}

bool CheckSdmc( void )
{
    if ( nn::fs::IsSdmcInserted() != s_MountSdmc )
    {
        return true;
    }
    else
    {
        return false;
    }
}

void InitializeProcedureTextureDataArray( void )
{
    s_ProcedureTextureFileDataNum = 0;
    s_ProcedureTextureFileDataIndex = 0;

    for (u32 index = 0; index < MAX_PROCEDURE_TEXTURE_FILE_DATA_SIZE; index++)
    {        
        s_ProcedureTextureFileDataArray[ index ].valid = false;
        
        _strcpy( s_ProcedureTextureFileDataArray[ index ].name,
            MAX_PROCTEX_CHAR_NAME_LENGTH,
            "None" );

        _wcsncpy( s_ProcedureTextureFileDataArray[ index ].fileName,
            MAX_PROCTEX_WCHAR_FILENAME_LENGTH,
            L"None" );
        
    }
}

bool UpdateProcedureTextureDataArray( void )
{
    InitializeProcedureTextureDataArray();

    MountSdmc();
    if ( s_MountSdmc )
    {
        NN_LOG("[Status] Read procedure parameter files from SD Card.\n");
        const wchar_t* dir = L"sdmc:/";
        return ReadProcedureTextureDataArray( dir );
    }
    else
    {
        NN_LOG("[Status] Read procedure parameter files from ROMFS.\n");
        const wchar_t* dir = L"rom:/";
        return ReadProcedureTextureDataArray( dir );
    }
}

bool CaptureScreen(void) 
{
   static u32 fileIndex = 0;

   s_CaptureFlag = false;

   MountSdmc();

   if ( s_MountSdmc )
   {
       GLint addr = 0;
       NN_LOG("[Status] Now saving display buffer to sdmc:/proctex/upper%d.bmp.\n ",fileIndex);
       nngxGetDisplaybufferParameteri(NN_GX_DISPLAYBUFFER_ADDRESS, &addr);
       int chk = SaveDisplayBufferSD(L"sdmc:/proctex", L"upper", fileIndex,
           addr, nn::gx::DISPLAY0_WIDTH, nn::gx::DISPLAY0_HEIGHT,(u8*)s_CaptureBuffer);
        
       if(! chk )
       {
           NN_LOG("[Status] Successfully saved screenshot to the SD Card!\n");
           fileIndex++;

           return true;
       }
       else 
       {
           NN_LOG("[Status] Failed to save screenshot to the SD Card.\n");

           return false;
       }
   }
   else
   {
       return false;
   }
}

void nnMain(void)
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    // fs initialization
    nn::fs::Initialize();
    
    PrintControl();

    nn::Result result = nn::fs::MountSdmc();
    if ( result.IsFailure() )
    {
        NN_LOG("[Status] Warning : No SD Card.\n");
        s_MountSdmc = false;
    }
    else
    {
        NN_LOG("[Status] Found SD Card.\n");
        s_MountSdmc = true;
    }

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ ROMFS_BUFFER_SIZE ];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom( 16, 16, buffer, ROMFS_BUFFER_SIZE ) );

    /* Initialize HID */
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED( result );

    /* Pad Read */
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    /* TouchPanel Read*/
    nn::hid::TouchPanelReader tpReader;
    nn::hid::TouchPanelStatus tpStatus;

    bool checkSdmc = false;

    /* Initialization */
    if ( InitializeGx() >= 0 )
    {
        InitializeButtons();
        SetDefaultSettings();
        ChangeTexCoordArray();
        nn::os::Thread::Sleep( nn::fnd::TimeSpan::FromMilliSeconds( 100 ) );

        while (1)
        {
            /* Pad Reading */
            padReader.ReadLatest(&padStatus);

            /* WaveEdit by Controller */
            int nSpeed = 1;
            if (padStatus.hold & nn::hid::BUTTON_L)
            {
                nSpeed *= 2;
            }
            
            if (padStatus.hold & nn::hid::BUTTON_R)
            {
                nSpeed *= 5;
            }
                        
            if (padStatus.trigger & nn::hid::BUTTON_X)
            {
                ChangeWaveValue(128, true); // Overwrite
            } 
            else if (padStatus.trigger & nn::hid::BUTTON_Y) 
            {
                ChangeWaveValue(0, true);   // Overwrite
            } 
            else if (padStatus.trigger & nn::hid::BUTTON_UP)
            {                
                ChangeWaveValue(nSpeed);
            } 
            else if (padStatus.trigger & nn::hid::BUTTON_DOWN) 
            {                
                ChangeWaveValue(-nSpeed);
            } 
            else if (padStatus.trigger & nn::hid::BUTTON_LEFT) 
            {
                s_WavePositionX -= nSpeed;
                if (s_WavePositionX < 0) s_WavePositionX = 0;
                ChangeWaveValue(0);
            } 
            else if (padStatus.trigger & nn::hid::BUTTON_RIGHT) 
            {
                s_WavePositionX += nSpeed;
                if (s_WavePositionX > 127) s_WavePositionX = 127;
                ChangeWaveValue(0);
            }

            /*Temporary filter support*/
            if (padStatus.trigger & nn::hid::BUTTON_B) 
            {
                if(s_TexFilter == GL_LINEAR)s_TexFilter = GL_NEAREST;
                else s_TexFilter = GL_LINEAR;
            }


            bool pushButtonLR = false;
            if ( ( padStatus.hold & nn::hid::BUTTON_L ) && ( padStatus.hold & nn::hid::BUTTON_R ) )
            {
                pushButtonLR = true;
            }

            bool pushButtonL = false;
            if ( padStatus.hold & nn::hid::BUTTON_L ) 
            {
                pushButtonL = true;
            }

            bool pushButtonR = false;
            if ( padStatus.hold & nn::hid::BUTTON_R ) 
            {
                pushButtonR = true;
            }

            if (padStatus.hold & nn::hid::BUTTON_EMULATION_DOWN)
            {
                if ( pushButtonLR )
                {
                    s_ObjScalePosY -= s_ObjScaleStep;
                    ChangePosArray();
                }                
                else if ( pushButtonL )
                {
                    s_WorldPos[ 2 ] -= s_WorldPosStep;
                }                
                else if ( pushButtonR )
                {
                    s_WorldEulerDeg[ 0 ] -= s_WorldAngleStep;
                }
                else
                {
                    s_WorldPos[ 1 ] -= s_WorldPosStep;
                }
            }

            if (padStatus.hold & nn::hid::BUTTON_EMULATION_UP) 
            {
                if ( pushButtonLR )
                {
                    s_ObjScalePosY += s_ObjScaleStep; 
                    ChangePosArray();
                }   
                else if ( pushButtonL )
                {       
                    s_WorldPos[ 2 ] += s_WorldPosStep;
                }
                else if ( pushButtonR )
                {
                    s_WorldEulerDeg[ 0 ] += s_WorldAngleStep;
                }
                else
                {
                    s_WorldPos[ 1 ] += s_WorldPosStep;
                }
            }


            if (padStatus.hold & nn::hid::BUTTON_EMULATION_LEFT)
            {
                if ( pushButtonLR )
                {
                    s_ObjScalePosX -= s_ObjScaleStep; 
                    ChangePosArray();
                }
                else if ( pushButtonL )
                {
                    s_WorldEulerDeg[ 2 ] -= s_WorldAngleStep;
                }
                else if ( pushButtonR )
                {
                    s_WorldEulerDeg[ 1 ] -= s_WorldAngleStep;
                }
                else
                {    
                    s_WorldPos[ 0 ] -= s_WorldPosStep;
                }
            }
      
            if (padStatus.hold & nn::hid::BUTTON_EMULATION_RIGHT)
            {
                if ( pushButtonLR )
                {
                    s_ObjScalePosX += s_ObjScaleStep; 
                    ChangePosArray();
                }
                else if ( pushButtonL )
                {
                    s_WorldEulerDeg[ 2 ] += s_WorldAngleStep;
                }
                else if ( pushButtonR )
                {
                    s_WorldEulerDeg[ 1 ] += s_WorldAngleStep;
                }
                else
                {
                    s_WorldPos[ 0 ] += s_WorldPosStep;
                }
            }


            if ( s_ObjScalePosX <= 0.0f )
            {
                s_ObjScalePosX = s_ObjScaleStep;
            }

            if ( s_ObjScalePosY <= 0.0f )
            {
                s_ObjScalePosY = s_ObjScaleStep;
            }

            if (padStatus.trigger & nn::hid::DEBUG_PAD_BUTTON_HOME )
            {
                ResetCamera();
            }
            
            /* Reset */
            if (padStatus.trigger & nn::hid::BUTTON_START) 
            {
                // SetDefaultSettings();
                // ChangeWaveValue(0);
                // ChangeTexCoordArray();
                // SetClearColor();
                ResetCamera();
                ChangeTexCoordArray();
                ChangePosArray();
            }

            /* Capture */
            if (padStatus.trigger & nn::hid::BUTTON_A) 
            {
                s_CaptureFlag = true;
            }
            else 
            {
                s_CaptureFlag = false;
            }
            
            /* Touchpanel Reading */
            tpReader.ReadLatest(&tpStatus);
            
            /* Control */
            int nTouchButton = -1;
            if (tpStatus.touch != 0) 
            {
                s_TouchX = tpStatus.x;
                s_TouchY = tpStatus.y;
                for (int i = 0; i < NUM_BUTTON; i++) 
                {
                    if (s_ButtonArray[i].ID < 0) 
                    {
                        continue;
                    }
                    if (s_ButtonArray[i].IsTouch(tpStatus.x, tpStatus.y)) 
                    {
                        nTouchButton = i;
                        break;
                    }
                }
                if (nTouchButton == -1) 
                {
                    // Not touch any button
                    s_ButtonUpIndex = -1;
                } 
                else
                {
                    // Touch
                    if (s_ButtonDownIndex == -1) 
                    {
                        s_ButtonDownIndex = nTouchButton;
                        s_SelectButtonIndex = nTouchButton;
                    }

                    s_ButtonUpIndex = ((nTouchButton == s_SelectButtonIndex) ? nTouchButton : -1);
                    
                    // WaveEdit
                    if (nTouchButton == BUTTON_ID_EDITVIEW) 
                    {
                        s_WavePositionX = s_TouchX - s_ButtonArray[BUTTON_ID_EDITVIEW].x;
                        int value = 128 - (s_TouchY - s_ButtonArray[BUTTON_ID_EDITVIEW].y);
                        ChangeWaveValue(value, true);   // Overwrite
                    }
                }
            } 
            else
            {
                wchar_t wcharBuffer[ MAX_PROCTEX_WCHAR_FILENAME_LENGTH ];
                
                // Click
                switch (s_ButtonUpIndex) 
                {
                case BUTTON_ID_SELECT_FC:
                    s_SelectTableIndex = BUTTON_ID_SELECT_FC;
                    MakeTextureEditView(s_FuncRGBArray);
                    break;

                case BUTTON_ID_SELECT_FA:
                    if ( s_EnableAlphaSeparate )
                    {
                        s_SelectTableIndex = BUTTON_ID_SELECT_FA;
                        MakeTextureEditView(s_FuncAlphaArray);
                        
                        MakeTextureAlphaView();
                    }
                    break;

                case BUTTON_ID_SELECT_R:
                    s_SelectTableIndex = BUTTON_ID_SELECT_R;
                    MakeTextureEditView(s_FuncSamplerRArray);
                    break;

                case BUTTON_ID_SELECT_G:
                    s_SelectTableIndex = BUTTON_ID_SELECT_G;
                    MakeTextureEditView(s_FuncSamplerGArray);
                    break;

                case BUTTON_ID_SELECT_B:
                    s_SelectTableIndex = BUTTON_ID_SELECT_B;
                    MakeTextureEditView(s_FuncSamplerBArray);
                    break;

                case BUTTON_ID_SELECT_A:
                    if ( ! s_EnableAlphaSeparate )
                    {
                        s_SelectTableIndex = BUTTON_ID_SELECT_A;
                        MakeTextureEditView(s_FuncSamplerAArray);

                        MakeTextureAlphaView();
                    }
                    break;

                case BUTTON_ID_SELECT_NZ:
                    s_SelectTableIndex = BUTTON_ID_SELECT_NZ;
                    MakeTextureEditView(s_FuncNoiseArray);
                    break;

                case BUTTON_ID_SEPARATE:
                    s_EnableAlphaSeparate = !s_EnableAlphaSeparate;
                    MakeTextureAlphaView();

                    if ( ( ( s_EnableAlphaSeparate ) && ( s_SelectTableIndex == BUTTON_ID_SELECT_A ) ) ||
                         ( (! s_EnableAlphaSeparate ) && ( s_SelectTableIndex == BUTTON_ID_SELECT_FA ) ) )
                    {
                        s_SelectTableIndex = BUTTON_ID_SELECT_FC;
                        MakeTextureEditView(s_FuncRGBArray);
                    }
                    break;

                case BUTTON_ID_PRESET_PREV:
                    s_PresetIndex--;
                    if (s_PresetIndex < 0) 
                    {
                        s_PresetIndex = (NUM_PRESET - 1);
                    }
                    break;

                case BUTTON_ID_PRESET_NEXT:
                    s_PresetIndex++;
                    if (s_PresetIndex >= NUM_PRESET) 
                    {
                        s_PresetIndex = 0;
                    }
                    break;

                case BUTTON_ID_PRESET_COPY:
                    CopyPresetTable();
                    break;

                case BUTTON_ID_RGB_PATTERN_PREV:
                    s_RgbMapIndex--;
                    if (s_RgbMapIndex < 0) 
                    {
                        s_RgbMapIndex = 9;
                    }

                    if (! s_EnableAlphaSeparate )
                    {
                        s_AlphaMapIndex = s_RgbMapIndex;
                    }
                    break;

                case BUTTON_ID_RGB_PATTERN_NEXT:
                    s_RgbMapIndex++;
                    if (s_RgbMapIndex > 9) 
                    {
                        s_RgbMapIndex = 0;
                    }

                    if (! s_EnableAlphaSeparate )
                    {
                        s_AlphaMapIndex = s_RgbMapIndex;
                    }
                    break;

                case BUTTON_ID_ALPHA_PATTERN_PREV:
                    if ( s_EnableAlphaSeparate )
                    {
                        s_AlphaMapIndex--;
                        if (s_AlphaMapIndex < 0) 
                        {
                            s_AlphaMapIndex = 9;
                        }
                    }
                    break;

                case BUTTON_ID_ALPHA_PATTERN_NEXT:
                    if ( s_EnableAlphaSeparate )
                    {
                        s_AlphaMapIndex++;
                        if (s_AlphaMapIndex > 9) 
                        {
                            s_AlphaMapIndex = 0;
                        }
                    }
                    break;

                case BUTTON_ID_ENABLE_R:
                    s_EnableSamplerR = !s_EnableSamplerR;
                    break;

                case BUTTON_ID_ENABLE_G:
                    s_EnableSamplerG = !s_EnableSamplerG;
                    break;

                case BUTTON_ID_ENABLE_B:
                    s_EnableSamplerB = !s_EnableSamplerB;
                    break;

                case BUTTON_ID_ENABLE_A:
                    s_EnableSamplerA = !s_EnableSamplerA;
                    break;

                case BUTTON_ID_ENABLE_N_U:
                    s_EnableNoiseU = !s_EnableNoiseU;
                    break;

                case BUTTON_ID_ENABLE_N_V:
                    s_EnableNoiseV = !s_EnableNoiseV;
                    break;

                case BUTTON_ID_ENABLE_N_U_F:
                    s_EnableNoiseF_U = !s_EnableNoiseF_U;
                    break;

                case BUTTON_ID_MINUS_N_U_F:
                    s_SelectNoiseF_U--;
                    if (s_SelectNoiseF_U < 0) 
                    {
                        s_SelectNoiseF_U = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_U_F:
                    s_SelectNoiseF_U++;
                    if (s_SelectNoiseF_U > MAX_NOISE_VALUE) 
                    {
                        s_SelectNoiseF_U = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_U_P:
                    s_EnableNoiseP_U = !s_EnableNoiseP_U;
                    break;

                case BUTTON_ID_MINUS_N_U_P:
                    s_SelectNoiseP_U--;
                    if (s_SelectNoiseP_U < 0)
                    {
                        s_SelectNoiseP_U = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_U_P:
                    s_SelectNoiseP_U++;
                    if (s_SelectNoiseP_U > MAX_NOISE_VALUE)
                    {
                        s_SelectNoiseP_U = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_U_PA:
                    s_EnableNoiseAutoP_U = !s_EnableNoiseAutoP_U;
                    break;

                case BUTTON_ID_MINUS_N_U_PA:
                    s_NoiseUPhaseStep -= NOISE_PHASE_STEP;
                    if ( s_NoiseUPhaseStep <= MIN_NOISE_PHASE_STEP )
                    {
                        s_NoiseUPhaseStep = MIN_NOISE_PHASE_STEP;
                    }
                    break;

                case BUTTON_ID_PLUS_N_U_PA:
                    s_NoiseUPhaseStep += NOISE_PHASE_STEP;
                    if ( s_NoiseUPhaseStep >= MAX_NOISE_PHASE_STEP )
                    {
                        s_NoiseUPhaseStep = MAX_NOISE_PHASE_STEP;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_U_A:
                    s_EnableNoiseA_U = !s_EnableNoiseA_U;
                    break;

                case BUTTON_ID_MINUS_N_U_A:
                    s_SelectNoiseA_U--;
                    if (s_SelectNoiseA_U < 0) 
                    {
                        s_SelectNoiseA_U = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_U_A:
                    s_SelectNoiseA_U++;
                    if (s_SelectNoiseA_U > MAX_NOISE_VALUE) 
                    {
                        s_SelectNoiseA_U = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_V_F:
                    s_EnableNoiseF_V = !s_EnableNoiseF_V;
                    break;

                case BUTTON_ID_MINUS_N_V_F:
                    s_SelectNoiseF_V--;
                    if (s_SelectNoiseF_V < 0)
                    {
                        s_SelectNoiseF_V = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_V_F:
                    s_SelectNoiseF_V++;
                    if (s_SelectNoiseF_V > MAX_NOISE_VALUE)
                    {
                        s_SelectNoiseF_V = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_V_P:
                    s_EnableNoiseP_V = !s_EnableNoiseP_V;
                    break;

                case BUTTON_ID_MINUS_N_V_PA:
                    s_NoiseVPhaseStep -= NOISE_PHASE_STEP;
                    if ( s_NoiseVPhaseStep <= MIN_NOISE_PHASE_STEP )
                    {
                        s_NoiseVPhaseStep = MIN_NOISE_PHASE_STEP;
                    }
                    break;

                case BUTTON_ID_PLUS_N_V_PA:
                    s_NoiseVPhaseStep += NOISE_PHASE_STEP;
                    if ( s_NoiseVPhaseStep >= MAX_NOISE_PHASE_STEP )
                    {
                        s_NoiseVPhaseStep = MAX_NOISE_PHASE_STEP;
                    }
                    break;

                case BUTTON_ID_MINUS_N_V_P:
                    s_SelectNoiseP_V--;
                    if (s_SelectNoiseP_V < 0)
                    {
                        s_SelectNoiseP_V = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_V_P:
                    s_SelectNoiseP_V++;
                    if (s_SelectNoiseP_V > MAX_NOISE_VALUE)
                    {
                        s_SelectNoiseP_V = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_ENABLE_N_V_PA:
                    s_EnableNoiseAutoP_V = !s_EnableNoiseAutoP_V;
                    break;

                case BUTTON_ID_ENABLE_N_V_A:
                    s_EnableNoiseA_V = !s_EnableNoiseA_V;
                    break;

                case BUTTON_ID_MINUS_N_V_A:
                    s_SelectNoiseA_V--;
                    if (s_SelectNoiseA_V < 0)
                    {
                        s_SelectNoiseA_V = 0;
                    }
                    break;

                case BUTTON_ID_PLUS_N_V_A:
                    s_SelectNoiseA_V++;
                    if (s_SelectNoiseA_V > MAX_NOISE_VALUE)
                    {
                        s_SelectNoiseA_V = MAX_NOISE_VALUE;
                    }
                    break;

                case BUTTON_ID_CLAMP_U:
                    s_ClampUIndex++;
                    if (s_ClampUIndex > 4)
                    {
                        s_ClampUIndex = 0;
                    }
                    break;

                case BUTTON_ID_SHIFT_U:
                    s_ShiftUIndex++;
                    if (s_ShiftUIndex > 2)
                    {
                        s_ShiftUIndex = 0;
                    }
                    break;

                case BUTTON_ID_CLAMP_V:
                    s_ClampVIndex++;
                    if (s_ClampVIndex > 4)
                    {
                        s_ClampVIndex = 0;
                    }
                    break;

                case BUTTON_ID_SHIFT_V:
                    s_ShiftVIndex++;
                    if (s_ShiftVIndex > 2)
                    {
                        s_ShiftVIndex = 0;
                    }
                    break;

                case BUTTON_ID_MINUS_S_MIN:
                    s_MinTexCoordS += nSpeed;
                    if (s_MinTexCoordS > MAX_TEXCOORD_VALUE)
                    {
                        s_MinTexCoordS = MAX_TEXCOORD_VALUE;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_PLUS_S_MIN:
                    s_MinTexCoordS -= nSpeed;
                    if (s_MinTexCoordS < 0)
                    {
                        s_MinTexCoordS = 0;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_MINUS_S_MAX:
                    s_MaxTexCoordS -= nSpeed;
                    if (s_MaxTexCoordS < 0)
                    {
                        s_MaxTexCoordS = 0;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_PLUS_S_MAX:
                    s_MaxTexCoordS += nSpeed;
                    if (s_MaxTexCoordS > MAX_TEXCOORD_VALUE)
                    {
                        s_MaxTexCoordS = MAX_TEXCOORD_VALUE;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_MINUS_T_MIN:
                    s_MinTexCoordT += nSpeed;
                    if (s_MinTexCoordT > MAX_TEXCOORD_VALUE)
                    {
                        s_MinTexCoordT = MAX_TEXCOORD_VALUE;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_PLUS_T_MIN:
                    s_MinTexCoordT -= nSpeed;
                    if (s_MinTexCoordT < 0)
                    {
                        s_MinTexCoordT = 0;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_MINUS_T_MAX:
                    s_MaxTexCoordT -= nSpeed;
                    if (s_MaxTexCoordT < 0)
                    {
                        s_MaxTexCoordT = 0;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_PLUS_T_MAX:
                    s_MaxTexCoordT += nSpeed;
                    if (s_MaxTexCoordT > MAX_TEXCOORD_VALUE)
                    {
                        s_MaxTexCoordT = MAX_TEXCOORD_VALUE;
                    }
                    ChangeTexCoordArray();
                    break;

                case BUTTON_ID_LOAD_WAVEDATA_PREV:
                    checkSdmc = CheckSdmc();
                    if ( checkSdmc )
                    {
                        UpdateProcedureTextureDataArray();

                        LoadProcedureTextureData();
                    }
                    else
                    {            
                        if ( s_ProcedureTextureFileDataIndex > 0 )
                        {
                            s_ProcedureTextureFileDataIndex -= 1;
                        }

                        LoadProcedureTextureData();
                    }
                    break;

                case BUTTON_ID_LOAD_WAVEDATA_NEXT:         
                    checkSdmc = CheckSdmc();
                    if ( checkSdmc )
                    {
                        UpdateProcedureTextureDataArray();

                        LoadProcedureTextureData();
                    }
                    else
                    {            
                        if ( s_ProcedureTextureFileDataIndex < ( s_ProcedureTextureFileDataNum - 1 ) )
                        {
                            s_ProcedureTextureFileDataIndex += 1;
                        }

                        LoadProcedureTextureData();
                    }
                    break;

                case BUTTON_ID_SAVE_WAVEDATA:
                    ConvertProcedureTextureData( s_ProcedureTextureData );
                                    
                    MountSdmc();

                    GetProcedureTextureDataFileName( wcharBuffer,
                        MAX_PROCTEX_WCHAR_FILENAME_LENGTH );
                    
                    if ( ( s_MountSdmc ) && 
                         ( SaveProcedureTextureData( wcharBuffer ) ) )
                    {
                        NN_LOG("[Status] Saved procedural texture data to the SD Card.\n");
                    }
                    else
                    {
                        NN_LOG("[Status] Failed to save procedural texture data to the SD Card.\n");
                    }

                    UpdateProcedureTextureDataArray();
                    
                    break;

                case BUTTON_ID_CAPTURE:
                    MountSdmc();
                    
                    if ( s_MountSdmc )
                    {
                        s_CaptureFlag = true;
                    }
                    break;

                case BUTTON_ID_TEX_WIDTH_PREV:
                    if ( s_TexWidth > MIN_TEX_WIDTH )
                    {
                        s_TexWidth /= 2;
                        
                        MakeTextureViewers(s_TexID[3], s_FuncSamplerRArray, true, false );
                        MakeTextureViewers(s_TexID[4], s_FuncSamplerGArray, true, false );
                        MakeTextureViewers(s_TexID[5], s_FuncSamplerBArray, true, false );
                        MakeTextureViewers(s_TexID[6], s_FuncSamplerAArray, true, false );
                    }
                    break;

                case BUTTON_ID_TEX_WIDTH_NEXT:
                    if ( s_TexWidth < MAX_TEX_WIDTH )
                    {
                        s_TexWidth *= 2;

                        MakeTextureViewers(s_TexID[3], s_FuncSamplerRArray, true, false );
                        MakeTextureViewers(s_TexID[4], s_FuncSamplerGArray, true, false );
                        MakeTextureViewers(s_TexID[5], s_FuncSamplerBArray, true, false );
                        MakeTextureViewers(s_TexID[6], s_FuncSamplerAArray, true, false );
                     }
                    break;

                case BUTTON_ID_BGCOLOR_R:
                    s_ColorBG_R++;
                    if (s_ColorBG_R > 10)
                    {
                        s_ColorBG_R = 0;
                    }
                    SetClearColor();
                    break;

                case BUTTON_ID_BGCOLOR_G:
                    s_ColorBG_G++;
                    if (s_ColorBG_G > 10) 
                    {
                        s_ColorBG_G = 0;
                    }
                    SetClearColor();
                    break;

                case BUTTON_ID_BGCOLOR_B:
                    s_ColorBG_B++;
                    if (s_ColorBG_B > 10)
                    {
                        s_ColorBG_B = 0;
                    }
                    SetClearColor();
                    break;

                default:
                    break;
                }
                s_TouchX = s_TouchY = -1;
                s_ButtonDownIndex = s_ButtonUpIndex = -1;
                s_SelectButtonIndex = -1;
            }
            
            (void)DrawFrame();
            s_FrameCount++;
            s_FrameCount = s_FrameCount % 1000;
        }
    }
    
    /* Finalize */
    s_RenderSystem.Finalize();

    if ( s_CaptureBuffer != NULL )
    {
        s_AppHeap.Free( s_CaptureBuffer );
        s_CaptureBuffer = NULL;
    }
    
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();

    nn::hid::Finalize();
}
