﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gx_ProcTexEditor.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>
#include <nn/fs.h>
#include <nn/fnd.h>
#include <nn/init.h>
#include <nn/fnd/fnd_ExpHeap.h>

#include <nn/hid.h>

#include <nn/math.h>
#include <nn/gx.h>

#include <string>
#include <string.h>
#include <math.h>
#include <wchar.h>

#include "demo.h"

#include "gx_ProcedureTextureData.h"



// When RECTANGLE is defined, the polygon becomes a 2:1 rectangle
// #define RECTANGLE

namespace
{

    // ExpHeap for app.
    nn::fnd::ExpHeap s_AppHeap;
    uptr s_HeapForGx;

    // Framework for rendering
    demo::RenderSystemDrawing s_RenderSystem;

    // Capture flag
    bool s_CaptureFlag = false;
    
    // Capture buffer
    void* s_CaptureBuffer = NULL;

    // Enumerated type for button IDs
    enum
    {
        BUTTON_ID_EDITVIEW           =  0,
        BUTTON_ID_SELECT_FC          =  1,
        BUTTON_ID_SELECT_FA          =  2,
        BUTTON_ID_SELECT_R           =  3,
        BUTTON_ID_SELECT_G           =  4,
        BUTTON_ID_SELECT_B           =  5,
        BUTTON_ID_SELECT_A           =  6,

        BUTTON_ID_SEPARATE           =  7,

        BUTTON_ID_PRESET_PREV        =  8,
        BUTTON_ID_PRESET_NEXT        =  9,
        BUTTON_ID_PRESET_COPY        = 10,

        BUTTON_ID_RGB_PATTERN_PREV   = 11,
        BUTTON_ID_RGB_PATTERN_NEXT   = 12,

        BUTTON_ID_ENABLE_R           = 13,
        BUTTON_ID_ENABLE_G           = 14,
        BUTTON_ID_ENABLE_B           = 15,
        BUTTON_ID_ENABLE_A           = 16,

        BUTTON_ID_ENABLE_N_U         = 17,
        BUTTON_ID_ENABLE_N_V         = 18,
        
        BUTTON_ID_ENABLE_N_U_F       = 19,
        BUTTON_ID_MINUS_N_U_F        = 20,
        BUTTON_ID_PLUS_N_U_F         = 21,

        BUTTON_ID_ENABLE_N_U_P       = 22,
        BUTTON_ID_MINUS_N_U_P        = 23,
        BUTTON_ID_PLUS_N_U_P         = 24,
        
        BUTTON_ID_ENABLE_N_U_PA      = 25,
        BUTTON_ID_ENABLE_N_U_A       = 26,
        BUTTON_ID_MINUS_N_U_A        = 27,
        BUTTON_ID_PLUS_N_U_A         = 28,
        
        BUTTON_ID_ENABLE_N_V_F       = 29,
        BUTTON_ID_MINUS_N_V_F        = 30,
        BUTTON_ID_PLUS_N_V_F         = 31,
        BUTTON_ID_ENABLE_N_V_P       = 32,
        BUTTON_ID_MINUS_N_V_P        = 33,

        BUTTON_ID_PLUS_N_V_P         = 34,
        BUTTON_ID_ENABLE_N_V_PA      = 35,
        BUTTON_ID_ENABLE_N_V_A       = 36,
        BUTTON_ID_MINUS_N_V_A        = 37,
        BUTTON_ID_PLUS_N_V_A         = 38,

        BUTTON_ID_CLAMP_U            = 39,
        BUTTON_ID_SHIFT_U            = 40,
        BUTTON_ID_CLAMP_V            = 41,
        BUTTON_ID_SHIFT_V            = 42,

        BUTTON_ID_MINUS_S_MIN        = 43,
        BUTTON_ID_PLUS_S_MIN         = 44,
        BUTTON_ID_MINUS_S_MAX        = 45,
        BUTTON_ID_PLUS_S_MAX         = 46,
        BUTTON_ID_MINUS_T_MIN        = 47,
        BUTTON_ID_PLUS_T_MIN         = 48,
        BUTTON_ID_MINUS_T_MAX        = 49,
        BUTTON_ID_PLUS_T_MAX         = 50,

        BUTTON_ID_SAVE_WAVEDATA      = 51,

        BUTTON_ID_BGCOLOR_R          = 52,
        BUTTON_ID_BGCOLOR_G          = 53,
        BUTTON_ID_BGCOLOR_B          = 54,
        
        BUTTON_ID_SELECT_NZ          = 55,
        
        BUTTON_ID_LOAD_WAVEDATA_PREV = 56,
        BUTTON_ID_LOAD_WAVEDATA_NEXT = 57,
        
        BUTTON_ID_MINUS_N_U_PA       = 58,
        BUTTON_ID_PLUS_N_U_PA        = 59,
        BUTTON_ID_MINUS_N_V_PA       = 60,
        BUTTON_ID_PLUS_N_V_PA        = 61,
        
        BUTTON_ID_CAPTURE            = 62,
        
        BUTTON_ID_ALPHA_PATTERN_PREV = 63,
        BUTTON_ID_ALPHA_PATTERN_NEXT = 64,

        BUTTON_ID_TEX_WIDTH_PREV     = 65,
        BUTTON_ID_TEX_WIDTH_NEXT     = 66,

        NUM_BUTTON                   = 67
    };

    // Structure for buttons
    struct ButtonControl
    {
        int     ID;
        int     x;
        int     y;
        int     width;
        int     height;
        GLfloat defaultColor[4];
        GLfloat touchColor[4];
        GLfloat selectColor[4];
        GLfloat disableColor[4];

        /* Function */
        void SetData(int id, int xx, int yy, int w, int h)
        { 
            ID = id;
            x = xx;
            y = yy;
            width = w;
            height = h; 
        }
        
        void SetDefaultColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
        {
            defaultColor[0] = r;
            defaultColor[1] = g;
            defaultColor[2] = b; 
            defaultColor[3] = a;
        }

        void SetTouchColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
        {
            touchColor[0] = r;
            touchColor[1] = g;
            touchColor[2] = b;
            touchColor[3] = a;
        }

        void SetSelectColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
        {
            selectColor[0] = r;
            selectColor[1] = g;
            selectColor[2] = b;
            selectColor[3] = a;
        }

        void SetDisableColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
        {
            disableColor[0] = r;
            disableColor[1] = g;
            disableColor[2] = b;
            disableColor[3] = a;
        }

        bool IsTouch(int xx, int yy)
        { 
            return ( (xx >= x) && (xx < (x + width)) && 
                     (yy >= y) && (yy < (y + height)) ); 
        }
    };

    // Button array
    ButtonControl s_ButtonArray[ NUM_BUTTON ];

    // Total number of preset data items
    const u32 NUM_PRESET = 39;

    // Array of preset data strings
    static char* s_NamePresetArray[] = 
    {
        "x",
        "1-x",
        "x^2",
        "1-x^2",
        "(2x-1)^2",
        "1-(2x-1)^2",
        "Triangle",
        "Triangle Inv",
        "2 Triangle",
        "2 Triangle Inv",
        "4 Triangle",
        "4 Triangle Inv",
        "sin(pi/2)",
        "cos(pi/2)",
        "sin(pi)",
        "1-sin(pi)",
        "(1+sin(2pi))/2",
        "(1-sin(2pi))/2",
        "(1+cos(2pi))/2",
        "(1-cos(2pi))/2",
        "stair(2pt)",
        "stair(2pt) Inv",
        "stair(4pt)",
        "stair(4pt) Inv",
        "stair(8pt)",
        "stair(8pt) Inv",
        "pyram(3pt)",
        "pyram(3pt) Inv",
        "pyram(7pt)",
        "pyram(7pt) Inv",
        "pyram(15pt)",
        "pyram(15pt) Inv",
        "(const) 0.0",
        "(const) 0.5",
        "(const) 1.0",
        "Sample Red",
        "Sample Green",
        "Sample Blue",
        "3x^2-2x^3"
    };

    // G function
    static GLuint s_PtRgbMapArray[] = 
    {
        GL_PROCTEX_U_DMP,
        GL_PROCTEX_V_DMP,
        GL_PROCTEX_U2_DMP,
        GL_PROCTEX_V2_DMP,
        GL_PROCTEX_ADD_DMP,
        GL_PROCTEX_ADD2_DMP,
        GL_PROCTEX_ADDSQRT2_DMP,
        GL_PROCTEX_MIN_DMP,
        GL_PROCTEX_MAX_DMP,
        GL_PROCTEX_RMAX_DMP
    };

    static char* s_NameRgbMapArray[] = 
    {
        "U",
        "V",
        "U2",
        "V2",
        "ADD",
        "ADD2",
        "ADDSQRT2",   
        "MIN",
        "MAX",
        "RMAX"
    };

    // Clamp mode
    static GLuint s_PtClampArray[] = 
    {
        GL_SYMMETRICAL_REPEAT_DMP,
        GL_MIRRORED_REPEAT,
        GL_PULSE_DMP,
        GL_CLAMP_TO_EDGE,
        GL_CLAMP_TO_ZERO_DMP
    };

    static char* s_NameClampArray[] = 
    {
        "SYMMETRIC",
        "MIRRORED",
        "PULSE",
        "TO_EDGE",
        "TO_ZERO"
    };

    // Shift mode
    static GLuint s_PtShiftArray[] = 
    {
        GL_NONE_DMP,
        GL_ODD_DMP,
        GL_EVEN_DMP
    };

    static char* s_NameShiftArray[] = 
    {
        "NONE",
        "ODD",
        "EVEN"
    };
    
    // Amount of change in object's position coordinates
    const f32 s_WorldPosStep   = 0.01f;

    // Amount of change in object's attitude angle
    const f32 s_WorldAngleStep = 1.0f;

    // Object's position coordinates
    nn::math::VEC3 s_WorldPos( 0.0f, 0.0f, 0.0f );

    // Object's Euler angle
    nn::math::VEC3 s_WorldEulerDeg( 0.0f, 0.0f, 0.0f );

    // Object's scale
    f32 s_ObjScaleStep = 0.01f;
    f32 s_ObjScalePosX = 1.0f;
    f32 s_ObjScalePosY = 1.0f;

    // Amount of change in camera's translation
    const f32 s_CamDistanceStep = 0.01f;

    // Camera distance
    GLfloat s_CamDistance = 1.5f;
    
    int s_PresetIndex;
    int s_SelectTableIndex;
    int s_SelectButtonIndex, s_ButtonDownIndex, s_ButtonUpIndex;
    int s_TouchX, s_TouchY;
    int s_WavePositionX;
    u8 s_ColorBG_R, s_ColorBG_G, s_ColorBG_B;
    
    // Frame to P-Parameter
    int s_FrameCount = 0;

    // Whether in alpha independent mode
    bool s_EnableAlphaSeparate = false;

    // G(u,v) of RGB
    int s_RgbMapIndex = 0;
    
    // G(u,v) of Alpha
    int s_AlphaMapIndex = 0;
    
    // F function table of RGB
    u8 s_FuncRGBArray[128];

    // F function table of Alpha
    u8 s_FuncAlphaArray[128];

    // Noise function table
    u8 s_FuncNoiseArray[128];

    // Color lookup table
    u8 s_FuncSamplerRArray[128], s_FuncSamplerGArray[128], s_FuncSamplerBArray[128], s_FuncSamplerAArray[128];

    // Color lookup table to enable
    bool s_EnableSamplerR, s_EnableSamplerG, s_EnableSamplerB, s_EnableSamplerA;

    // Clamp mode to use
    int s_ClampUIndex, s_ClampVIndex;

    // Shift mode to use
    int s_ShiftUIndex, s_ShiftVIndex;

    // Offset to the color lookup table to use
    u32 s_TexOffset = 0;

    // Size of the color lookup table to use
    const u32 MIN_TEX_WIDTH = 2;
    const u32 MAX_TEX_WIDTH = 128;
    u32 s_TexWidth = 128;

    // Noise arguments to enable
    bool s_EnableNoiseU, s_EnableNoiseV;
    bool s_EnableNoiseF_U, s_EnableNoiseP_U, s_EnableNoiseA_U;
    bool s_EnableNoiseF_V, s_EnableNoiseP_V, s_EnableNoiseA_V;
    bool s_EnableNoiseAutoP_U, s_EnableNoiseAutoP_V;

    // Amount of change in noise phase (value of increase or decrease)
    const f32 NOISE_PHASE_STEP = 0.001f;
    
    // Amount of change in noise phase (minimum value)
    const f32 MIN_NOISE_PHASE_STEP = - 0.2f;

    // Amount of change in noise phase (maximum value)
    const f32 MAX_NOISE_PHASE_STEP = 0.2f;

    // Amount of change in noise phase
    f32  s_NoiseUPhaseStep = 0.01f;
    f32  s_NoiseVPhaseStep = 0.01f;

    // Maximum value of noise
    const u32 MAX_NOISE_VALUE = 100;
    
    // Index of noise argument in U direction
    int s_SelectNoiseF_U, s_SelectNoiseP_U, s_SelectNoiseA_U;
    
    // Index of noise argument in V direction
    int s_SelectNoiseF_V, s_SelectNoiseP_V, s_SelectNoiseA_V;

    // Index of noise argument in U direction
    GLfloat s_NoiseValueUArray[3] = { 0.0f, 0.0f, 0.0f };

    // Index of noise argument in V direction
    GLfloat s_NoiseValueVArray[3] = { 0.0f, 0.0f, 0.0f };

    // Maximum value of texture coordinates
    const u32 MAX_TEXCOORD_VALUE = 4000;

    // Amount of change in texture coordinates
    const f32 TEXCOORD_SCALE = 0.5f;

    // Minimum and maximum value of texture coordinates
    int s_MinTexCoordS, s_MinTexCoordT, s_MaxTexCoordS, s_MaxTexCoordT;

    // Filter mode of the texture
    int s_TexFilter = GL_LINEAR;

    // Rectangle's position coordinates
    #ifdef RECTANGLE
    GLfloat s_PositionCoords[] = 
    {
        -1.0f, 0.5f, 0.0f,
        -1.0f, -0.5f, 0.0f,
         1.0f, 0.5f, 0.0f,
         1.0f, -0.5f, 0.0f
    };
    #else 
    GLfloat s_PositionCoords[] = 
    {
        -0.5f, 0.5f, 0.0f,
        -0.5f,-0.5f, 0.0f,
         0.5f, 0.5f, 0.0f,
         0.5f,-0.5f, 0.0f
    };
    #endif

    // Rectangle's texture coordinates
    GLfloat t1 = 0.0f, t2 = 1.0f;
    GLfloat s_TexCoords[] = 
    {
        t1, t2,
        t1, t1,
        t2, t2,
        t2, t1
    };

    // Index array
    GLushort s_IndexArray[] = {0, 1, 2, 2, 1, 3};

    // Vertex buffer ID
    GLuint s_ArrayBufferID = 0;

    // Index array ID
    GLuint s_ElementArrayBufferID = 0;

    // Program ID
    GLuint s_PgID;

    // Shader ID.    
    GLuint s_ShID;

    // Texture object ID array.
    const u32 MAX_TEXTURE_OBJECT_NUM = 10;
    GLuint  s_TexColl;
    GLuint  s_TexID[MAX_TEXTURE_OBJECT_NUM];

    // Texture(EditView/ColorView) 
    const u32 EDITVIEW_TEX_WIDTH = 128;
    const u32 EDITVIEW_TEX_HEIGHT = 256;
    const GLenum EDITVIEW_TEX_FORMAT = GL_RGBA_NATIVE_DMP;
    const GLenum EDITVIEW_TEX_TYPE = GL_UNSIGNED_BYTE;
    const u32 EDITVIEW_TEX_PER_PIXEL  = 4;
    GLuint  s_TexEditView;
    u8      imageEditView[EDITVIEW_TEX_WIDTH * EDITVIEW_TEX_HEIGHT * EDITVIEW_TEX_PER_PIXEL];
    const u32 COLORVIEW_TEX_WIDTH = 128;
    const u32 COLORVIEW_TEX_HEIGHT = 8;
    const GLenum COLORVIEW_TEX_FORMAT = GL_RGBA_NATIVE_DMP;
    const GLenum COLORVIEW_TEX_TYPE = GL_UNSIGNED_BYTE;
    const u32 COLORVIEW_TEX_PER_PIXEL = 4;
    
    // Texture object ID for displaying RGB gradations
    GLuint  s_TexColorView;
    // An array of data for displaying RGB gradations
    u8      s_ImageColorViewArray[COLORVIEW_TEX_WIDTH * COLORVIEW_TEX_HEIGHT * COLORVIEW_TEX_PER_PIXEL];

    // Texture object ID for displaying alpha gradations
    GLuint  s_TexAlphaView;
    // An array of data for displaying alpha gradations
    u8      s_ImageAlphaViewArray[COLORVIEW_TEX_WIDTH * COLORVIEW_TEX_HEIGHT * COLORVIEW_TEX_PER_PIXEL];



    // "true" when SD Card is mounted
    bool s_MountSdmc = false;

    const u32 MAX_PROCTEX_CHAR_NAME_LENGTH      = 128;
    const u32 MAX_PROCTEX_WCHAR_FILENAME_LENGTH = 128;

    // Structure to save procedural texture arguments
    struct ProcedureTextureFileData
    {
        bool valid;
        NN_PADDING3;        
        char name[ MAX_PROCTEX_CHAR_NAME_LENGTH ];        
        wchar_t fileName[ MAX_PROCTEX_WCHAR_FILENAME_LENGTH ];
    };
    
    // Total number of procedural texture argument files
    u32 s_ProcedureTextureFileDataNum = 0;

    // Index of procedural texture argument files.
    u32 s_ProcedureTextureFileDataIndex = 0;

    // Maximum number of procedural texture argument files.
    const u32 MAX_PROCEDURE_TEXTURE_FILE_DATA_SIZE = 64;

    // Array that stores the contents of procedural texture argument files
    ProcedureTextureFileData s_ProcedureTextureFileDataArray[ MAX_PROCEDURE_TEXTURE_FILE_DATA_SIZE ];

    // Name of the current procedural texture argument file.
    const char* s_ProcedureTextureFileName = "None";

    // Structure for procedural texture arguments.
    ProcedureTextureData s_ProcedureTextureData;
} // Namespace

// Initialize the graphics library
int InitializeGx( void );

// Initialize the vertex buffer
void InitializeVertexBuffer( void );

// Initialize the lookup table
void InitializeLookUpTable( void );

// Initialize buttons
void InitializeButtons( void );

// Send operation method to debug output
void PrintControl( void );

// Function declarations that edit procedural texture arguments
void SetTextureEnvironment(void);
void SetPresetData(u8* data, int mode);
void SetLookUpTable_256(GLuint texture, u8* data);
void SetLookUpTable_512(GLuint texture, u8* data);
void ResetCamera( void );
void SetDefaultSettings(void);
void ChangePosArray(void);
void ChangeTexCoordArray(void);
void MakeTextureEditView(u8* data);
int CalcNativeDataPosition(int x, int y, int texWidth, int texHeight);
void MakeTextureColorView(void);
void MakeTextureAlphaView(void);
void MakeTextureViewers(GLuint texid, u8* data, bool updateColor, bool updateEditView = true );

// Function declarations of rendering functions
void DrawFrameUpper(void);
void SetClearColor(void);
void SetVertexBuffer(void);
void DrawFrameLower(void);
void SetClearColor(void);
int DrawFrame(void);

void ChangeWaveValue(int value, bool overwrite = false);
void CopyPresetTable( void );

// Convert the ProcedureTextureData structure to the demo's arguments
void SetProcedureTextureData( const ProcedureTextureData& procedureTextureData );


// Convert the edited procedural texture arguments to the ProcedureTextureData structure
// 
void ConvertProcedureTextureData( ProcedureTextureData& procedureTextureData );
// The following are conversion functions
PicaDataProcTexMap GetPicaProcTexMap( GLuint rgbMap );
PicaDataProcTexClamp GetPicaProcTexClamp( GLuint clamp );
PicaDataProcTexShift GetPicaProcTexShift( GLuint shift );
u32 GetProcTexMapIndex( PicaDataProcTexMap rgbMap );
u32 GetProcTexClampIndex( PicaDataProcTexClamp clamp );
u32 GetProcTexShiftIndex( PicaDataProcTexShift shift );

// Save procedural texture arguments to the SD Card
bool SaveProcedureTextureData( wchar_t* fileName );

// Load the procedural texture argument file
bool LoadProcedureTextureData( void );

// Load into the array all procedural texture argument files that are in the directory
bool ReadProcedureTextureDataArray( const wchar_t* dir );

// Mount the SD Card
void MountSdmc( void );

// Check whether the SD Card state has changed
bool CheckSdmc( void );

void InitializeProcedureTextureDataArray( void );
bool UpdateProcedureTextureDataArray( void );
bool ReadProcedureTextureDataArray( const wchar_t* dir );

bool CaptureScreen(void);
// Saves the display buffer for the upper screen to the SD Card
// Defined in gx_CaptureUtil.cpp
extern int SaveDisplayBufferSD( const wchar_t *dirname, const wchar_t *filename, u32 index,
                                GLuint addr, const u32 width, const u32 height, u8 *work = NULL);


std::size_t _Min( std::size_t size0, std::size_t size1 )
{
    if ( size0 < size1 )
    {
        return size0;
    }
    else
    {
        return size1;
    }
}

/* Please see man pages for details











*/
template<typename TChar>
size_t
_strncpy_t( TChar* dest, std::size_t destCount,
    const TChar* src, std::size_t srcCount )
{
    size_t length = 0;
    
    while ( length < _Min((destCount - 1), srcCount) )
    {
        *dest = *src;
        ++dest;
        ++src;
        ++length;
        
        if (*src == std::char_traits<TChar>::to_char_type('\0'))
        {
            break;
        }
    }

    *dest = std::char_traits<TChar>::to_char_type('\0');

    return length;
}

/* Please see man pages for details











*/
size_t _strcpy( char* dest, std::size_t destCount,
                const char* src)
{
    return _strncpy_t( dest, destCount, src, destCount - 1 );
}

/* Please see man pages for details











*/
size_t _strncpy( char* dest, std::size_t destCount,
                 const char* src, std::size_t srcCount)
{
    return _strncpy_t( dest, destCount, src, srcCount );
}

/* Please see man pages for details










*/

size_t _wcsncpy( wchar_t* dest, std::size_t destCount, const wchar_t* src)
{
    return _strncpy_t( dest, destCount, src, destCount - 1 );
}

/* Please see man pages for details











*/
size_t _wcsncpy( wchar_t* dest, std::size_t destCount,
                 const wchar_t* src, std::size_t srcCount )
{
    return _strncpy_t( dest, destCount, src, srcCount );
}

/* Please see man pages for details










*/
size_t _wcsncat( wchar_t* dest, std::size_t destCount, const wchar_t* src)
{
    u32 len = wcslen( dest );
    return _strncpy_t( dest + len, destCount, src, destCount - 1 );
}
