﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SaveData.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef SAVEDATA_H_
#define SAVEDATA_H_

#include <nn/os.h>
#include <nn/Result.h>

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Save data index
enum SaveDataId
{
    SAVE_DATA_1,
    SAVE_DATA_2,
    SAVE_DATA_3,
    SAVE_DATA_4,

    SAVE_DATA_MAX
};

class SaveData
{
public:
    // Structure of information to save to save data
    struct SaveDataStruct{
        // Information related to save data
        nn::os::Tick playTimeTick;   // Total play time (ticks)
        bool isNewSaveData;          // Flag indicating whether new data
        bool isUseExtraPad;          // Set whether to use Circle Pad Pro
        NN_PADDING2;

        // Information of object to manipulate
        f32 objectPositionX;  // X coordinate of object to manipulate.
        f32 objectPositionY;  // Y coordinate of object to manipulate.
        f32 objectLineLength; // End of line
        f32 objectDegree;     // angle
        NN_PADDING4;
    };

public:
    /* Please see man pages for details 
        
    */
    void Initialize();

    /* Please see man pages for details 
        
    */
    void Finalize();


    /* Please see man pages for details 
        
    */
    void MountSaveDataArchive();

    /* Please see man pages for details 
        
    */
    nn::Result FormatSaveDataArchive();

    /* Please see man pages for details 
        
    */
    void UnmountSaveDataArchive();


    /* Please see man pages for details 
        
    */
    void ReadSaveData(SaveDataStruct* saveDataSet, SaveDataId saveDataId);

    /* Please see man pages for details 
        
    */
    void WriteSaveData(SaveDataStruct* saveDataSet, SaveDataId saveDataId);

    /* Please see man pages for details 
        
    */
    void OpenSaveDataFile(SaveDataId saveDataId);

    /* Please see man pages for details 
        
    */
    void DeleteSaveDataFile(SaveDataId saveDataId);

    static const wchar_t* SAVEDATA_FILENAME[];   // Name of save data to use for display
    nn::os::Tick startTick;                      // Tick at time of save data load
    NN_PADDING4;


private:
    // Save data region file operations
    nn::fs::FileStream m_SaveDataFile;

    // Save data-related constants
    static const size_t SAVEDATA_SIZE = 128;   // Save data maximum size
    static const char*  SAVEDATA_ARCHIVE;      // Path of archive to mount
    static const char*  SAVEDATA_FILEPATH[];   // Path name of save data file

    // Default value for new save data
    SaveDataStruct m_SaveDataSetDefault;
    static const bool IS_SHOWN_MESSAGE_DEFAULT = false;
    static const bool IS_USE_EXTRA_PAD_DEFAULT = true;
    static const bool IS_NEW_SAVE_DATA_DEFAULT = true;

    bool m_IsFileOpen;      // File operations state
    bool m_IsMounted;       // Mount state
    NN_PADDING2;
    NN_PADDING4;
};

}}}}

#endif // SAVEDATA_H_
