﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     SceneSelectSaveData.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/


#include <nn.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include "SceneSelectSaveData.h"



//////////////////////////////////////////////////////////////////////////////////////////
// Constant definitions
//////////////////////////////////////////////////////////////////////////////////////////

namespace
{
    enum State
    {
        STATE_SELECT_SAVE_DATA,  // Select save data

        STATE_MAX
    };

    enum ControlId
    {
        CONTROL_ID_SAVE_1,
        CONTROL_ID_SAVE_2,
        CONTROL_ID_SAVE_3,
        CONTROL_ID_SAVE_4,

        CONTROL_ID_MAX
    };

    // Button size
    const u16 BUTTON_WIDTH  = 180;                   // Button width
    const u16 BUTTON_HEIGHT = 30;                    // Button height
    const u16 BUTTON_SPACE  = 16;                    // Button spacing
    const u16 BUTTON_X      = (320-BUTTON_WIDTH)/2;  // Button X coordinate
    const u16 BUTTON_Y_TOP  = 50;                    // Top button Y coordinate
}

//////////////////////////////////////////////////////////////////////////////////////////

namespace SampleDemos {
namespace hid {
namespace ExtraPad {
namespace AppRecommendedFlow {

// Initialize a scene.
bool SceneSelectSaveData::Initialize()
{
    // Register control callback event
    m_ControlManager.RegisterControlEventCallback(MyControlEventCallback, this);

    // Application
    Application* pApp = Application::GetInstance();

    // Configure save data button label according to state
    wchar_t buttonTitle[32];
    u16 buttonHeight = BUTTON_Y_TOP;

    for(int i=0; i<SAVE_DATA_MAX; i++)
    {
        switch(i)
        {
        case SAVE_DATA_1: pApp->m_SelectedSaveData = SAVE_DATA_1; break;
        case SAVE_DATA_2: pApp->m_SelectedSaveData = SAVE_DATA_2; break;
        case SAVE_DATA_3: pApp->m_SelectedSaveData = SAVE_DATA_3; break;
        case SAVE_DATA_4: pApp->m_SelectedSaveData = SAVE_DATA_4; break;
        }
        // Load save data information
        pApp->m_saveData.ReadSaveData(&(pApp->m_saveDataSet), pApp->m_SelectedSaveData);

        // Play time
        if(pApp->m_saveDataSet.isNewSaveData)
        {
            std::wcscpy(buttonTitle, L"New data");
        }
        else
        {
            // Convert ticks to play time display (hours:minutes:seconds)
            s64 playTimeSeconds = pApp->m_saveDataSet.playTimeTick.ToTimeSpan().GetSeconds();
            s64 playTimeHours   = playTimeSeconds / 3600;
            playTimeSeconds -= playTimeHours * 3600;
            s64 playTimeMinutes = playTimeSeconds / 60;
            playTimeSeconds -= playTimeMinutes * 60;

            std::swprintf(buttonTitle, 20, L"%4lld:%02lld:%02lld",playTimeHours, playTimeMinutes, playTimeSeconds);
        }
        m_ControlManager.Add(new scene::Button(i, BUTTON_X, buttonHeight,
            BUTTON_WIDTH, BUTTON_HEIGHT, buttonTitle));
        buttonHeight += BUTTON_HEIGHT + BUTTON_SPACE;
    }

    m_State   = STATE_SELECT_SAVE_DATA;

    return true;
}

// Finalize the scene.
void SceneSelectSaveData::Finalize()
{
}

// Update the scene.
void SceneSelectSaveData::Update()
{
    m_ControlManager.Update();
}

// Render the scene.
void SceneSelectSaveData::Draw()
{
    // Application
    Application* pApp = Application::GetInstance();
    // Renderer
    demo::RenderSystemDrawing* pRenderSystem = pApp->GetRenderSystem();
    // Text writer
    scene::TextWriter* pTextWriter = pApp->GetTextWriter();

    // Upper screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();

    pTextWriter->BeginRender(NN_GX_DISPLAY0);
    {
        scene::DrawTitle(pTextWriter);

        pTextWriter->SetTextColor(255, 255, 255);
        pTextWriter->Printf(2, 20, L"[Scene] Select save data");
    }
    pTextWriter->EndRender();
    pRenderSystem->SwapBuffers();


    // Lower screen
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();

    u32 textWidth;       // Width of one character
    f32 defaultFontScale = pTextWriter->GetFontScale(); // Default font scale

    pTextWriter->BeginRender(NN_GX_DISPLAY1);
    {
        pTextWriter->SetTextColor(255, 255, 255);
        textWidth = pTextWriter->CalculateStringWidth(L"Select save data");
        pTextWriter->Printf((320 - textWidth)/2, 5, L"Select save data");

        pTextWriter->SetFontScale(defaultFontScale*1.5f); //Set font scale
        pTextWriter->Printf(BUTTON_X-20, BUTTON_Y_TOP, L"1");
        pTextWriter->Printf(BUTTON_X-20, BUTTON_Y_TOP + BUTTON_HEIGHT + BUTTON_SPACE, L"2");
        pTextWriter->Printf(BUTTON_X-20, BUTTON_Y_TOP + 2*(BUTTON_HEIGHT + BUTTON_SPACE), L"3");
        pTextWriter->Printf(BUTTON_X-20, BUTTON_Y_TOP + 3*(BUTTON_HEIGHT + BUTTON_SPACE), L"4");
        pTextWriter->SetFontScale(defaultFontScale); //Reset font scale
        m_ControlManager.Draw();
    }
    pTextWriter->EndRender();

    pRenderSystem->SwapBuffers();

    pRenderSystem->WaitVsync(NN_GX_DISPLAY_BOTH);
}

// Event callback
void SceneSelectSaveData::MyControlEventCallback(scene::ControlEvent event, scene::ControlBase* pControl, s16 x, s16 y, void* pParam)
{
    NN_UNUSED_VAR(x);
    NN_UNUSED_VAR(y);


    if (event == scene::CONTROL_EVENT_TOUCH)
    {
        // Scene
        SceneSelectSaveData* pScene = (SceneSelectSaveData*)pParam;

        // Application
        Application* pApp = Application::GetInstance();

            // Save selected save data
        switch(pControl->GetId())
        {
        case CONTROL_ID_SAVE_1:
            pApp->m_SelectedSaveData = SAVE_DATA_1;
            break;

        case CONTROL_ID_SAVE_2:
            pApp->m_SelectedSaveData = SAVE_DATA_2;
            break;

        case CONTROL_ID_SAVE_3:
            pApp->m_SelectedSaveData = SAVE_DATA_3;
            break;

        case CONTROL_ID_SAVE_4:
            pApp->m_SelectedSaveData = SAVE_DATA_4;
            break;
        }
        pScene->SetNextSceneId(SCENE_ID_EP_OPERATE_SAVE_DATA);
    }

}}}}}

