/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     gx.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

/*
 *------------------------------------------------------------
 * Copyright(c) 2009-2010 by Digital Media Professionals Inc.
 * All rights reserved.
 *------------------------------------------------------------
 * This source code is the confidential and proprietary
 * of Digital Media Professionals Inc.
 *------------------------------------------------------------
 */

#include <nn/gx.h>
#include <nn/fs.h>
#include "gx.h"

extern "C" {
    extern u8* SHADER_BEGIN[];
    extern u8* SHADER_END[];
}

/* buffer id */
GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

/* program id */
GLuint s_PgID;

/* shader id */
GLuint s_ShID;

/* generate like CTR object */
GLfloat s_Color[] = {
    1.f, 0.0f, 0.0f,
    0.f, 1.0f, 0.0f,
    0.f, 0.0f, 1.0f,
    1.f, 0.0f, 0.0f,
    0.f, 1.0f, 0.0f,
    0.f, 0.0f, 1.0f
};

GLfloat s_Coords[24];
GLushort s_Idxs[] = {0, 2, 1, 3, 2, 4, 3, 5, 4};


void GraphicsDemo::Initialize(demo::RenderSystemDrawing* p_RenderSystem, nn::hid::GyroscopeReader* p_GyroscopeReader, nn::hid::AccelerometerReader* p_AccReader)
{
    mp_RenderSystem = p_RenderSystem;
    mp_GyroReader   = p_GyroscopeReader;
    mp_AccReader    = p_AccReader;
}

void GraphicsDemo::LoadObjects(void)
{

    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, sizeof(s_Coords) + sizeof(s_Color), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(s_Coords), s_Coords);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(s_Coords), sizeof(s_Color), s_Color);

    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(s_Idxs), s_Idxs, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(s_Coords));
}

void GraphicsDemo::UpdateObject(void)
{
    nn::hid::GyroscopeStatus gyroStatus;
    mp_GyroReader->ReadLatest(&gyroStatus);
    nn::hid::Direction dir = gyroStatus.direction;


    nn::math::VEC3 points[6] = {
        dir.x - dir.z,
        -dir.x - dir.z,
        dir.x,
        -dir.x,
        dir.y + dir.x,
        dir.y - dir.x
    };

    for(s32 i=0;i<6;i++)
    {
        s_Coords[4*i+0] = points[i].x;
        s_Coords[4*i+1] = points[i].y;
        s_Coords[4*i+2] = points[i].z;
        s_Coords[4*i+3] = 1.0f;
    }

    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(s_Coords), s_Coords);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(s_Coords));
}


void GraphicsDemo::Start()
{
    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);

    // Shader binary setting
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, SHADER_BEGIN, reinterpret_cast<int>(SHADER_END) - reinterpret_cast<int>(SHADER_BEGIN));

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aColor");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    mp_RenderSystem->SetClearColor(NN_GX_DISPLAY0, 0.f, 0.f, 0.f, 1.0f);
    mp_RenderSystem->SetClearColor(NN_GX_DISPLAY1, 0.f, 0.f, 0.f, 1.0f);

    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    LoadObjects();

    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
}

void GraphicsDemo::End(void)
{

}

void GraphicsDemo::DrawUpperFrame(void)
{
    glUseProgram(s_PgID);
    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aColor");
    UpdateObject();

    nn::math::Matrix44 proj;
    nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH,
                 0.02f*nn::gx::DISPLAY0_HEIGHT/nn::gx::DISPLAY0_WIDTH, 0.2f, 20.f);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    nn::math::Matrix34 eye, rot;
    nn::math::Vector3 camPos(0.f, 0.0f, -16.0f);
    nn::math::Vector3 camUp(1.f, 0.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 rotAxis(0.f, 1.f, 0.f);

    nn::math::MTX34Identity(&eye);
    nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);

    nn::math::Matrix44 mv(eye);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, (f32*)(mv));

    glDrawElements(GL_TRIANGLE_STRIP, 9, GL_UNSIGNED_SHORT, 0);
    glFinish();
}

void GraphicsDemo::DrawLowerFrame(void)
{
    nn::hid::GyroscopeStatus     gyroStatus;
    nn::hid::AccelerometerStatus accStatus;

    // Gets most recent gyro sensor value
    if(mp_GyroReader->ReadLatest(&gyroStatus))
    {
        // Display Gyro sensor 
        mp_RenderSystem->SetColor(1.0f, 1.0f, 1.0f, 1.0f);

        mp_RenderSystem->DrawText(0,   0, "<Gyroscope>");

        mp_RenderSystem->DrawText(0,  20, "speed    x = % .2lf", gyroStatus.speed.x);
        mp_RenderSystem->DrawText(0,  30, "         y = % .2lf", gyroStatus.speed.y);
        mp_RenderSystem->DrawText(0,  40, "         z = % .2lf", gyroStatus.speed.z);

        mp_RenderSystem->DrawText(0,  60, "angle    x = % .2lf", gyroStatus.angle.x);
        mp_RenderSystem->DrawText(0,  70, "         y = % .2lf", gyroStatus.angle.y);
        mp_RenderSystem->DrawText(0,  80, "         z = % .2lf", gyroStatus.angle.z);

        mp_RenderSystem->DrawText(0, 100, "dir X    x = % .2lf", gyroStatus.direction.x.x);
        mp_RenderSystem->DrawText(0, 110, "         y = % .2lf", gyroStatus.direction.x.y);
        mp_RenderSystem->DrawText(0, 120, "         z = % .2lf", gyroStatus.direction.x.z);

        mp_RenderSystem->DrawText(0, 140, "dir Y    x = % .2lf", gyroStatus.direction.y.x);
        mp_RenderSystem->DrawText(0, 150, "         y = % .2lf", gyroStatus.direction.y.y);
        mp_RenderSystem->DrawText(0, 160, "         z = % .2lf", gyroStatus.direction.y.z);

        mp_RenderSystem->DrawText(0, 180, "dir Z    x = % .2lf", gyroStatus.direction.z.x);
        mp_RenderSystem->DrawText(0, 190, "         y = % .2lf", gyroStatus.direction.z.y);
        mp_RenderSystem->DrawText(0, 200, "         z = % .2lf", gyroStatus.direction.z.z);

    }

    // Acquires most recent accelerometer value
    if(mp_AccReader->ReadLatest(&accStatus))
    {
        // Display Accelerometer
        mp_RenderSystem->SetColor(1.0f, 1.0f, 1.0f, 1.0f);

        // Converts accelerometer value units to "G"
        nn::hid::AccelerationFloat accFloat;
        mp_AccReader->ConvertToAcceleration(&accFloat, 1, &accStatus);

        mp_RenderSystem->DrawText(200, 180, "<Accelerometer>", accFloat.x);

        mp_RenderSystem->DrawText(200, 200, "Gx = % .2lf", accFloat.x);
        mp_RenderSystem->DrawText(200, 210, "Gy = % .2lf", accFloat.y);
        mp_RenderSystem->DrawText(200, 220, "Gz = % .2lf", accFloat.z);
        mp_RenderSystem->DrawText(200, 230, "|G|= % .2lf", nn::math::FSqrt(accFloat.x*accFloat.x+accFloat.y*accFloat.y+accFloat.z*accFloat.z));
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
