﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ir_util.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef IR_UTIL_H_
#define IR_UTIL_H_

#include <nn/ir.h>

const char* STATUS_LABEL[] = {
    "STOPPED",
    "TRYING_TO_CONNECT",
    "CONNECTED",
    "DISCONNECTING",
    "FATAL_ERROR"
};

const char* GetStatusLabel(nn::ir::ConnectionStatus status)
{
    return STATUS_LABEL[status];
}

const s32 BAUD_RATE_NUM = 16;

const nn::ir::BaudRate BAUD_RATE_LIST[BAUD_RATE_NUM] = {
    nn::ir::BAUD_RATE_3000,
    nn::ir::BAUD_RATE_4800,
    nn::ir::BAUD_RATE_6000,
    nn::ir::BAUD_RATE_7200,
    nn::ir::BAUD_RATE_9600,
    nn::ir::BAUD_RATE_12000,
    nn::ir::BAUD_RATE_18000,
    nn::ir::BAUD_RATE_19200,
    nn::ir::BAUD_RATE_24000,
    nn::ir::BAUD_RATE_36000,
    nn::ir::BAUD_RATE_38400,
    nn::ir::BAUD_RATE_48000,
    nn::ir::BAUD_RATE_57600,
    nn::ir::BAUD_RATE_72000,
    nn::ir::BAUD_RATE_96000,
    nn::ir::BAUD_RATE_115200
};

const s32 BAUD_RATE_VALUES[BAUD_RATE_NUM] = {
    3000,
    4800,
    6000,
    7200,
    9600,
    12000,
    18000,
    19200,
    24000,
    36000,
    38400,
    48000,
    57600,
    72000,
    96000,
    115200
};

inline s32 GetBaudRateIndex(nn::ir::BaudRate baudrate)
{
    s32 idx = -1;
    for(; idx < BAUD_RATE_NUM; idx++)
    {
        if(BAUD_RATE_LIST[idx] == baudrate)
        {
            break;
        }
    }

    NN_TASSERT_(idx >= 0 && idx < BAUD_RATE_NUM);
    return idx;
}

inline nn::ir::BaudRate GetNextBaudRate(nn::ir::BaudRate baudrate, s32 offset)
{
    s32 idx = GetBaudRateIndex(baudrate) + offset;
    if(idx < 0)
    {
        idx = 0;
    }
    else if(idx >= BAUD_RATE_NUM)
    {
        idx = BAUD_RATE_NUM - 1;
    }

    return BAUD_RATE_LIST[idx];
}

inline nn::ir::BaudRate GetNextBiggerBaudRate(nn::ir::BaudRate baudrate)
{
    return GetNextBaudRate(baudrate,1);
}

inline nn::ir::BaudRate GetNextSmallerBaudRate(nn::ir::BaudRate baudrate)
{
    return GetNextBaudRate(baudrate,-1);
}

inline s32 GetBaudRateValue(nn::ir::BaudRate baudrate)
{
    return BAUD_RATE_VALUES[GetBaudRateIndex(baudrate)];
}




#endif /* IR_UTIL_H_ */
