﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     common.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <string.h>
#include <sstream>

#include <nn/pl/CTR/pl_ExBanner.h>

#include "demo.h"
#include "common.h"


// Specify whether to replace the texture data.
#define EXBANNER_USE_TEXTURE_DATA


extern "C" {
    extern u8* SHADER_BEGIN[];
    extern u8* SHADER_END[];
}


namespace
{

    /* buffer id */
    GLuint s_ArrayBufferID;
    GLuint s_ElementArrayBufferID;

    /* program id */
    GLuint s_PgID;

    /* shader id */
    GLuint s_ShID;

    /* generate simple object */
    GLfloat s_Coords[] = {
         0.5f, 0.0f, 0.f, 1.f,
        -0.5f, 0.5f, 0.f, 1.f,
        -0.5f,-0.5f, 0.f, 1.f
    };


    GLfloat s_Color[] = {
        1.f, 0.0f, 0.0f,
        0.f, 1.0f, 0.0f,
        0.f, 0.0f, 1.0f
    };

    GLushort s_Idxs[] = {0, 1, 2};

    // Expanded save data-related code
    const std::wstring  s_PathNameExbannerBinFileW       = L"extdata:/ExBanner/COMMON.bin";
    const std::wstring  s_PathNameExbannerRootDirW       = L"extdata:/ExBanner";
    // Use the default icon file.
    const std::wstring  s_PathNameIcnFileW               = L"rom:/Default.icn";
    const char*         s_ArchiveNameRom                 = "rom:";

    const        nn::fs::ExtSaveDataId s_ExtSaveDataId   = 0xFF3F5;
    const char*  s_ArchiveNameExtSaveData                = "extdata:";

    void* s_WorkingMemory                                = NULL;

    nn::fnd::ExpHeap   s_ExtHeap;
    nn::os::LightEvent s_AwakeEvent;

    const u32 TEXTURE_ALIGHMENT  = 128;

    u32   s_TextureWidth          = 0;
    u32   s_TextureHeight         = 0;

    const u32 TEXTURE_PIXEL_DEPTH = 4;
    const u32 DEFAULT_ALIGN_SIZE  = 4;

    // Start address of the texture data
    void * s_TexDataPtr = NULL;

    // Various flags
    bool s_IsWaitingOneFrame     = false;
    bool s_IsMountRom            = false;
    bool s_IsMountExtSaveData    = false;

    // Sample type
    u32  s_SampleType            = 0;

    //Set a fixed time for successfully saving/deleting the contextual banner.
    //This is used to display this fact on the upper screen.
    u32  s_ExBnrSaveCnt     = 0;
    u32  s_ExBnrDeleteCnt   = 0;
    
    nn::applet::AppletQueryReply mySleepQueryCallback( uptr arg )
    {
        NN_UNUSED_VAR(arg);
        return nn::applet::CTR::REPLY_LATER;
    }

    void myAwakeCallback( uptr arg )
    {
        NN_UNUSED_VAR(arg);
        s_AwakeEvent.Signal();
    }
}

bool ProcessExBanner();
bool MountRom();
bool MountExtSaveData();
bool MyDeleteExtSaveData();

void LoadObjects(void)
{
    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, sizeof(s_Coords) + sizeof(s_Color), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(s_Coords), s_Coords);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(s_Coords), sizeof(s_Color), s_Color);

    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(s_Idxs), s_Idxs, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(s_Coords));
}

void UpdateObject(void)
{
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(s_Coords));
}

void Start(demo::RenderSystemDrawing* pRenderSystem)
{
    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);

    // Shader binary setting
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, SHADER_BEGIN, reinterpret_cast<int>(SHADER_END) - reinterpret_cast<int>(SHADER_BEGIN));

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aColor");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    pRenderSystem->SetClearColor(NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f);
    pRenderSystem->SetClearColor(NN_GX_DISPLAY1, 0.f, 0.f, 0.f, 1.0f);

    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    LoadObjects();

    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
}


// isDrawExtSaveData = true  : Buffer for writing expanded save data
// isDrawExtSaveData = false : Render to the screen
void DrawFrame(bool isDrawExtSaveData)
{
    static u32 f = 0;

    glUseProgram(s_PgID);
    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aColor");
    UpdateObject();

    nn::math::Matrix44 proj;
    nn::math::Matrix34 view, rot;

    nn::math::Vector3 camPos(0.f, 0.4f, 9.5f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    nn::math::Vector3 rotAxis(0.f, 1.f, 0.f);

    if(!isDrawExtSaveData)
    {
        nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f*nn::gx::DISPLAY1_HEIGHT/nn::gx::DISPLAY1_WIDTH,
                               0.02f*nn::gx::DISPLAY1_HEIGHT/nn::gx::DISPLAY1_WIDTH, 0.2f, 10.f);
    }
    else// For saving to expanded save data
    {
        glViewport(0, 0, s_TextureWidth, s_TextureHeight);

        switch(s_SampleType)
        {
          case SAMPLE_TYPE_CAMERA_ROTATE:
            //90 degree rotation
            camUp.x =  1.f;
            camUp.y =  0.f;
            camUp.z =  0.f;
            // Create a projection matrix based on the view frustum at the near clipping plane.
            nn::math::MTX44Frustum(&proj, -0.02f*s_TextureWidth/s_TextureHeight, 0.02f*s_TextureWidth/s_TextureHeight,
                                   -0.02f, 0.02f, 0.2f, 10.f);
            break;
          case SAMPLE_TYPE_TEX_ROTATE:
            nn::math::MTX44Frustum(&proj, -0.02f, 0.02f,
                                   -0.02f*s_TextureHeight/s_TextureWidth,0.02f*s_TextureHeight/s_TextureWidth, 0.2f, 10.f);
            break;
          default:
            NN_ASSERT(false);
        }
    }

    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));

    nn::math::MTX34Identity(&view);
    nn::math::MTX34LookAt(&view, &camPos, &camUp, &target);
    nn::math::MTX34RotAxisDeg(&rot, &rotAxis, -6.f * (f%60));

    nn::math::MTX34Mult(&view, &view, &rot);
    nn::math::Matrix44 mv(view);
    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, (f32*)(mv));

    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, 0);
    glFinish();

    f++;
}

void PrepareTexture()
{
    GLuint mTexId;
    GLuint frameBuffer;

    // Create a frame buffer object
    glGenFramebuffers( 1, &frameBuffer );
    glBindFramebuffer( GL_FRAMEBUFFER, frameBuffer );
    DEMO_ASSERT_GL_ERROR();

    // Create a texture object
    glActiveTexture(GL_TEXTURE0);

    glGenTextures(1, &mTexId);
    glBindTexture(GL_TEXTURE_2D, mTexId);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

    DEMO_ASSERT_GL_ERROR();

    // Clear the buffer.
    memset(s_TexDataPtr, 0x000000FF, s_TextureWidth*s_TextureHeight*TEXTURE_PIXEL_DEPTH);

    
    // You can also specify a texture allocated in main memory by specifying NN_GX_MEM_FCRAM to glTexImage2D, but you cannot clear it using glClear due to hardware specifications.
    // 
    // Rendering main memory can easily result in bus conflicts and is inefficient comparted to textures allocated in VRAM.
    // 

    // Main memory is allocated for s_TexDataPtr in order to write buffer data after rendering to expanded save data.
    // 
    glTexImage2D(GL_TEXTURE_2D | NN_GX_MEM_FCRAM | GL_NO_COPY_FCRAM_DMP,
                 0,
                 GL_RGBA_NATIVE_DMP,
                 s_TextureWidth,
                 s_TextureHeight,
                 0,
                 GL_RGBA_NATIVE_DMP,
                 GL_UNSIGNED_BYTE,
                 static_cast<u8*>(s_TexDataPtr));
    DEMO_ASSERT_GL_ERROR();

    glFramebufferTexture2D( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, mTexId, 0 );

    DEMO_ASSERT_GL_ERROR();

    // glClear() cannot be called for the reason given above.

    glDepthMask( GL_FALSE );
    glStencilMask( GL_FALSE );

    // Delete the frame buffer and texture buffer objects
    glBindFramebuffer( GL_FRAMEBUFFER, frameBuffer );
    DEMO_ASSERT_GL_ERROR();

    // Specifying true renders to the buffer to be written in expanded save data.
    DrawFrame(true);
    DEMO_ASSERT_GL_ERROR();
    glDeleteFramebuffers(1,&frameBuffer);
    DEMO_ASSERT_GL_ERROR();
    glDeleteTextures(1,&mTexId);
    DEMO_ASSERT_GL_ERROR();
}


//---------------
void* AllocateMemory(size_t size, u32 align = DEFAULT_ALIGN_SIZE)
{
    return s_ExtHeap.Allocate(size,align);
}

void FreeMemory(void* p)
{
    s_ExtHeap.Free(p);
}



void Display(demo::RenderSystemDrawing* pRenderSystem)
{
    u16 x = 1;
    u16 y = 2;
    u16 step = 16;

    // DISPLAY 0
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY0);
    pRenderSystem->Clear();
    pRenderSystem->SetColor(1.f, 1.f, 1.f);
    pRenderSystem->DrawText(step * x, step * y++, "EXBanner SampleDemo.");

    y++;

    pRenderSystem->DrawText(step * x, step * y++, "Press A to Save ExBannerData.");
    pRenderSystem->DrawText(step * x, step * y++, "Press B to Delete ExtSaveData.");

    y++;
    if(s_ExBnrSaveCnt)
    {
        pRenderSystem->DrawText(step * x, step * y++, "ExBannerData Saved.");
        s_ExBnrSaveCnt--;
    }
    if(s_ExBnrDeleteCnt)
    {
        pRenderSystem->DrawText(step * x, step * y++, "ExBannerData Deleted.");
        s_ExBnrDeleteCnt--;
    }
    
    
    pRenderSystem->SwapBuffers();

    // DISPLAY 1
    pRenderSystem->SetRenderTarget(NN_GX_DISPLAY1);
    pRenderSystem->Clear();
    // Specifying false is for display rendering
    DrawFrame(false);
    pRenderSystem->SwapBuffers();

}

void Update(nn::hid::PadStatus padStatus)
{
    //Wait one loop for rendering.
    if(s_IsWaitingOneFrame)
    {
        s_IsWaitingOneFrame = false;
        if(!ProcessExBanner())
        {
            NN_ASSERT(false);
        }
    }
    
    if(padStatus.trigger & nn::hid::BUTTON_A)
    {
        s_IsWaitingOneFrame = true;

        if(!s_IsMountRom && !MountRom())
        {
            NN_ASSERT(false);
        }
        if(!s_IsMountExtSaveData && !MountExtSaveData())
        {
            NN_ASSERT(false);
        }

        // Render the lower screen to the texture.
        // Because we must wait for rendering, expanded save data is written in the next loop.
        // 
        PrepareTexture();
    }

    if(padStatus.trigger & nn::hid::BUTTON_B)
    {
        MyDeleteExtSaveData();
    }

}


bool FsSampleReadIcnData(void** ppData, size_t* pSize)
{
    NN_LOG("FsSampleReadIcnData\n");

    // Open the *.icn file
    nn::fs::FileInputStream f(s_PathNameIcnFileW.c_str());

    // Allocate buffer
    size_t size = f.GetSize();
    void* p = AllocateMemory(size);
    if(!p)
    {
        NN_LOG("FsSampleReadIcnData failed.\n");
        return false;
    }

    // Load data
    f.Read(p, size);

    *pSize  = size;
    *ppData = p;
    return true;
}


bool MountRom()
{
    NN_LOG("MountRom\n");

    const size_t    maxFiles        = 8;
    const size_t    maxDirectories  = 8;
    const bool      useCache        = true;

    // Gets the working memory size
    //   - This function does not return invalid values
    size_t workingMemorySize = nn::fs::GetRomRequiredMemorySize(maxFiles, maxDirectories, useCache);

    if(!s_WorkingMemory)
    {
        // Allocate working memory
        s_WorkingMemory = AllocateMemory(workingMemorySize);
        if(!s_WorkingMemory)
        {
            NN_LOG("Cannot allocate working memory!\n");
            return false;
        }
    }

    // Mount ROM archive
    // This function returns only success
    nn::fs::MountRom(s_ArchiveNameRom, maxFiles, maxDirectories, s_WorkingMemory, workingMemorySize, useCache);
    s_IsMountRom = true;
    return true;
}


bool MountExtSaveData()
{
    NN_LOG("MountExtSaveData\n");

    const size_t    maxFiles        = 8;
    const size_t    maxDirectories  = 8;
    void*           pIcnData;
    size_t          size;

    nn::Result result;

    result = nn::fs::MountExtSaveData(s_ArchiveNameExtSaveData, s_ExtSaveDataId);
    if(result.IsFailure())
    {
        //NN_DBG_PRINT_RESULT(result);

        if(result <= nn::fs::ResultNotFound())
        {
            if(result <= nn::fs::ResultMediaNotFound())
            {
                // SD Card cannot be recognized
                NN_LOG("MountExtSaveData() failed : No SD card!\n");
                return false;
            }
        }
        else if((result <= nn::fs::ResultNotFormatted()) ||
                (result <= nn::fs::ResultVerificationFailed()))
        {
            // Expanded save data must be recreated
            result = nn::fs::DeleteExtSaveData(s_ExtSaveDataId);
            if(result.IsFailure())
            {
                NN_LOG("DeleteExtSaveData() failed.\n");
                return false;
            }
        }
        else if(result <= nn::fs::ResultBadFormat())
        {
            // SD Card must be formatted
            NN_LOG("MountExtSaveData() failed : Bad format!\n");
            return false;
        }
        else if(result <= nn::fs::ResultOperationDenied())
        {
            if(result <= nn::fs::ResultWriteProtected())
            {
                // SD Card is write protected
                NN_LOG("MountExtSaveData() failed : SD card is write protected!\n");
                return false;
            }
            else if(result <= nn::fs::ResultMediaAccessError())
            {
                // Hardware-related factors by bad contact or other causes
                NN_LOG("MountExtSaveData() failed : Access error!\n");
                return false;
            }
            else
            {
                // Unexpected errors
                // Can be caused by the files or directories on the SD Card being read-only
                NN_LOG("MountExtSaveData() failed : Unknown operation denied error!\n");
                return false;
            }
        }
        else
        {
            // Unexpected errors
            NN_LOG("MountExtSaveData() failed : Unknown error!\n");
            return false;
        }
    }
    else
    {
        s_IsMountExtSaveData = true;
        return true;
    }

    // Expanded save data must be created.

    // Load ICN file from ROM archive
    if(!FsSampleReadIcnData(&pIcnData, &size))
    {
        NN_LOG("Cannot read icn file!\n");
        return false;
    }

    // Create the expanded save data
    result = nn::fs::CreateExtSaveData(s_ExtSaveDataId, pIcnData, size, maxDirectories, maxFiles);
    FreeMemory(pIcnData);

    // Detailed errors are not displayed.
    // See the fs sample.
    if(result.IsFailure())
    {
        NN_DBG_PRINT_RESULT(result);
        NN_LOG("CreateExtSaveData() failed.\n");
    }

    result = nn::fs::MountExtSaveData(s_ArchiveNameExtSaveData, s_ExtSaveDataId);
    if(result.IsFailure())
    {
        NN_DBG_PRINT_RESULT(result);
        NN_LOG("MountExtSaveData() failed.\n");
        return false;
    }

    s_IsMountExtSaveData = true;
    return true;
}

bool UnmountRom()
{

    if(s_IsMountRom)
    {
        nn::Result result = nn::fs::Unmount(s_ArchiveNameRom);

        if(result.IsFailure())
        {
            NN_DBG_PRINT_RESULT(result);
            NN_LOG("UnmountRom failed.\n");
            return false;
        }

        if(s_WorkingMemory)
        {
            FreeMemory(s_WorkingMemory);
            s_WorkingMemory = 0;
        }
        s_IsMountRom = false;
    }

    if(s_IsMountExtSaveData)
    {
        nn::Result result = nn::fs::Unmount(s_ArchiveNameExtSaveData);

        if(result.IsFailure())
        {
            NN_DBG_PRINT_RESULT(result);
            NN_LOG("UnmountExtSaveData failed.\n");
            return false;
        }

        s_IsMountExtSaveData = false;
    }

    return true;
}

bool SaveExBannerData(nn::pl::CTR::ExBanner exbanner)
{
    NN_LOG("SaveExBannerData\n");

    nn::Result result;
    nn::fs::TryCreateDirectory(s_PathNameExbannerRootDirW.c_str());

    // Expanded save data must be created using TryCreateFile.
    // Expanded save data cannot be created using TryInitialize, which does not allow a size to be specified.
    const s64 size = sizeof(exbanner) - sizeof(void*) + s_TextureWidth*s_TextureHeight*TEXTURE_PIXEL_DEPTH;

    nn::fs::TryCreateFile(s_PathNameExbannerBinFileW.c_str(), size);

    bool createIfNotExist = false;
    nn::fs::FileOutputStream f;
    result = f.TryInitialize(s_PathNameExbannerBinFileW.c_str(), createIfNotExist);
    if(result.IsFailure())
    {
        NN_DBG_PRINT_RESULT(result);
        NN_LOG("TryInitialize() failed.\n");
        return false;
    }

     s32 out = 0;

    // Write data other than binary texture data.
    // The value given by sizeof(void*) must therefore be subtracted.
    result = f.TryWrite(&out, &exbanner, sizeof(exbanner) - sizeof(void*), false);
    if(result.IsFailure())
    {
        NN_DBG_PRINT_RESULT(result);
        NN_LOG("TryWrite() ExBanner failed.\n");
        f.Finalize();
        return false;
    }
    if(s_TexDataPtr)
    {
        // Write data binary texture data (tex_data).
        result = f.TryWrite(&out, s_TexDataPtr, s_TextureWidth*s_TextureHeight*TEXTURE_PIXEL_DEPTH, true);
        if(result.IsFailure())
        {
            NN_DBG_PRINT_RESULT(result);
            NN_LOG("TryWrite() ExBannerTexData failed.\n");
            f.Finalize();
            return false;
        }
    }
    // Close the file
    f.Finalize();

    return true;
}


void CreateExbannerData(nn::pl::CTR::ExBanner* pExbanner)
{
    NN_LOG("CreateExbannerData\n");

#ifdef EXBANNER_USE_TEXTURE_DATA
    pExbanner->tex_width    = s_TextureWidth;
    pExbanner->tex_height   = s_TextureHeight;
    pExbanner->tex_format = nn::pl::CTR::EXBANNER_TEX_FORMAT_RGBA8;
    strncpy(pExbanner->texture_name, "COMMON1", sizeof(pExbanner->texture_name));
#else
    // Pad with zeros if not using texture data.
    pExbanner->tex_width    = 0;
    pExbanner->tex_height   = 0;
    pExbanner->tex_format   = 0;
    memset(pExbanner->texture_name, 0, sizeof(pExbanner->texture_name));
    pExbanner->tex_data     = NULL;
#endif

    // These values are used only for download expanded banners.
    // Pad with zeros.
    pExbanner->limit_year  = 0;
    pExbanner->limit_month = 0;
    pExbanner->limit_date  = 0;

    // Specify text.
    // When using expandedl banners, you can replace the texture, replace the text, or replace both.
    // 
    nn::fnd::DateTime dateTime = nn::fnd::DateTime::GetNow();
    memset(pExbanner->text, 0, nn::pl::CTR::EXBANNER_MESSAGE_LENGTH*sizeof(wchar_t));

    std::swprintf( pExbanner->text, nn::pl::CTR::EXBANNER_MESSAGE_LENGTH,
             L"Exbanner Sample  %4d年%02d月%02d日%02d時%02d分%02d秒に作成しました。",
             dateTime.GetYear(),
             dateTime.GetMonth(),
             dateTime.GetDay(),
             dateTime.GetHour(),
             dateTime.GetMinute(),
             dateTime.GetSecond());
    
    // TextureData is written into s_TexDataPtr.
    // For more details, see the PrepareTexture().
}

bool MyDeleteExtSaveData()
{
    UnmountRom();
    
    NN_LOG("DeleteExtSaveData()\n");
    nn::Result result;
    result = nn::fs::DeleteExtSaveData(s_ExtSaveDataId);
    if(result <= nn::fs::ResultNotFound())
    {
        //If the file does not exist at first, a message indicating it was deleted is not displayed.
        //
        return true;
    }
    else if(result.IsFailure())
    {
        NN_DBG_PRINT_RESULT(result);
        NN_LOG("DeleteExtSaveData() failed.\n");
        NN_ASSERT(false);
    }

    s_ExBnrDeleteCnt = 120;//Approx. 2 seconds
        
    return true;
}

bool ProcessExBanner()
{
    nn::pl::CTR::ExBanner exbanner;
    CreateExbannerData(&exbanner);
    if(!SaveExBannerData(exbanner))
    {
        return false;
    }

    //Flag for displaying that saving was successful on the screen
    s_ExBnrSaveCnt = 120;//Approx. 2 seconds
    
    return true;
}

void Init()
{
    // Set sleep callback, sleep recovery callback
    nn::applet::SetSleepQueryCallback(mySleepQueryCallback, 0);
    nn::applet::SetAwakeCallback(myAwakeCallback, 0);
    nn::applet::Enable();
    s_AwakeEvent.Initialize(false);
    
    nn::fs::Initialize();

    // Allocate heap
    s_ExtHeap.Initialize(
        nn::os::GetDeviceMemoryAddress(),
        nn::os::GetDeviceMemorySize(),
        nn::os::ALLOCATE_OPTION_LINEAR);

    // TexData must have 128-byte alignment
    s_TexDataPtr = AllocateMemory(s_TextureWidth*s_TextureHeight*TEXTURE_PIXEL_DEPTH,TEXTURE_ALIGHMENT);
    memset(s_TexDataPtr, 0, s_TextureWidth*s_TextureHeight*TEXTURE_PIXEL_DEPTH);

    // Initialize hid
    NN_UTIL_PANIC_IF_FAILED(nn::hid::Initialize());

}


bool WaitAndCheck()
{
    nn::applet::WaitForStarting();

    // Recover the GPU register settings
    nngxUpdateState(NN_GX_STATE_ALL);
    nngxValidateState(NN_GX_STATE_ALL,GL_TRUE);

    if ( nn::applet::IsExpectedToCloseApplication() )
    {
        return true;
    }
    return false;
}

bool ProcessApplet()
{
    //Sleep support
    if ( nn::applet::IsExpectedToReplySleepQuery() )
    {
        nn::applet::ReplySleepQuery( nn::applet::REPLY_ACCEPT );
        s_AwakeEvent.Wait();
        
        // Processing after restart
        nn::gx::StartLcdDisplay();
    }

    // HOME Button support
    if (nn::applet::IsExpectedToProcessHomeButton())
    {
        nn::applet::ProcessHomeButton();
        if(WaitAndCheck())
        {
            return true;
        }
    }

    // POWER Button support
    if ( nn::applet::IsExpectedToProcessPowerButton() )
    {
        nn::applet::ProcessPowerButton();
        if(WaitAndCheck())
        {
            return true;
        }
    }

    return false;
}

void SampleMain(u32 sampleType)
{
    switch(sampleType)
    {
      case SAMPLE_TYPE_CAMERA_ROTATE:
        s_TextureWidth          = 256;
        s_TextureHeight         = 128;
        s_SampleType = sampleType;
        break;
      case SAMPLE_TYPE_TEX_ROTATE:
        s_TextureWidth          = 128;
        s_TextureHeight         = 256;
        s_SampleType = sampleType;
        break;
      default:
        NN_ASSERT(false);
        break;
    }
    
    bool isExpectedToCloseApplication = false;

    Init();

    // Prepare pad input
    nn::hid::PadReader        padReader;
    nn::hid::PadStatus        padStatus;

    // Prepare RenderSystem
    uptr heapForGx = reinterpret_cast<uptr>(s_ExtHeap.Allocate(0x800000));
    demo::RenderSystemDrawing renderSystem;
    renderSystem.Initialize(heapForGx, 0x800000);
    renderSystem.SetClearColor(NN_GX_DISPLAY_BOTH, 0.f, 0.f, 0.f, 0.f);
    Start(&renderSystem);

    while(!isExpectedToCloseApplication)
    {
        padReader.ReadLatest(&padStatus);
        Display(&renderSystem);
        Update(padStatus);

        renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
        
        isExpectedToCloseApplication = ProcessApplet();

    }

    renderSystem.Finalize();
    nn::hid::Finalize();
    nn::applet::SetSleepQueryCallback(NULL, 0);
    nn::applet::SetAwakeCallback(NULL, 0);
    s_AwakeEvent.Finalize();

    if(s_TexDataPtr)
    {
        FreeMemory(s_TexDataPtr);
        s_TexDataPtr = NULL;
    }

    UnmountRom();
    s_ExtHeap.Free(reinterpret_cast<void*>(heapForGx));
    s_ExtHeap.Finalize();

    nn::applet::CloseApplication();

}


