﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/types.h>
#include <nn/init.h>
#include <nn/os.h>
#include <nn/ndm.h>
#include <nn/util.h>
#include <nn/applet.h>

extern "C" void nnMain( void )
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    nn::Result result;

    // Initializes the daemon manager control library.
    result = nn::ndm::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    // Stops all network daemons running in the background.
    // Note: Ideally, only the daemons that need to be stopped would be stopped, but at this stage there is not sufficient information to select them.
    //   
    // Blocks until daemons are completely stopped.
    result = nn::ndm::SuspendScheduler();
    NN_UTIL_PANIC_IF_FAILED(result);

    {
        // Perform processing that relies on network and NAND performance.
    }

    // Resumes the daemons.
    // When daemons no longer need to be stopped, resume them.
    result = nn::ndm::ResumeScheduler();
    NN_UTIL_PANIC_IF_FAILED(result);

    // Terminates the daemon manager control library.
    result = nn::ndm::Finalize();
    NN_UTIL_PANIC_IF_FAILED(result);

    NN_LOG("END\n");
}

