﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     news_test.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/news.h>

extern "C" void nninitStartUp(void)
{
    static const size_t HEAP_SIZE = 512 * 1024;
    nn::os::SetupHeapForMemoryBlock(HEAP_SIZE + 32 * 4096);
    nn::init::InitializeAllocator(HEAP_SIZE);
}

extern "C" void nninitSetup(void)
{
    // Overwrite default with dummy processing
}

extern "C" void nnMain(void)
{
    // Initializes the news library.
    NN_LOG("Initialize news library.\n");

    nn::Result r = nn::news::Initialize();
    NN_UTIL_PANIC_IF_FAILED(r);

    // Determines if the application was started from the News List.
    nn::news::JumpArg arg;
    bool boot_newslist = nn::news::IsFromNewsList( &arg );

    NN_LOG("Boot from newslist : %d\n", boot_newslist);
    if( boot_newslist ) NN_LOG("JumpParam : %lld\n", arg.jumpParam);

    // Posts news data.
    NN_LOG("Post news data.\n");

    wchar_t *subject = L"サンプル表題";
    wchar_t *message = L"サンプル本文";

    // u64 type values can be passed to jumpParam and can be obtained from JumpArg.jumpParam of IsFromNewsList.
    // 
    // Does not need to be specified unless the value must be passed.
    r = nn::news::PostNews(subject, message, NULL, 0, 0, 0, 0x10);
    NN_UTIL_PANIC_IF_FAILED(r);

    // Posts notification data that includes a URL.
    NN_LOG("Post news data with URL.\n");

    wchar_t *url_subject = L"URLつきサンプル表題";
    wchar_t *url_message = L"URLつきサンプル本文";
    const u8* url = "testtesttest";

    size_t work_buf_size = nn::news::GetWorkBufferSizeForNewsUrl( url_message, url );
    u8* work_buf = new u8[work_buf_size];

    r = nn::news::PostNewsUrl(url_subject, url_message, url, work_buf);
    NN_UTIL_PANIC_IF_FAILED(r);

    delete [] work_buf;

    // Cleans up the news library.
    NN_LOG("Finalize news library.\n");

    r = nn::news::Finalize();
    NN_UTIL_PANIC_IF_FAILED(r);

    NN_LOG("End.\n");
    for(;;){}
}


