﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     os_DemoUtil.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

#include "os_DemoUtil.h"

namespace DemoUtil {

// Thread stack used with CreateThreads
nn::os::StackBuffer<NN_DEMO_OS_THREAD_STACK_SIZE> s_ThreadStacks[NN_DEMO_OS_MAX_THREAD_NUM];

/* Please see man pages for details 
  

                
                

                
                
                

  
  
  

  
*/
void CreateThreads(nn::os::Thread* pThreads, nn::os::ThreadFunc func, u32 num)
{
    NN_ASSERT(num <= NN_DEMO_OS_MAX_THREAD_NUM);

    for(u32 i = 0; i < num; ++i)
    {
        pThreads[i].Start( func, i, s_ThreadStacks[i], nn::os::DEFAULT_THREAD_PRIORITY - 1 );
    }
}

/* Please see man pages for details 
  

                

  
  

  
*/
void JoinThreads(nn::os::Thread* pThreads, u32 num)
{
    NN_ASSERT(num <= NN_DEMO_OS_MAX_THREAD_NUM);

    for(u32 i = 0; i < num; ++i)
    {
        pThreads[i].Join();
    }
}


#define PRINT_COUNT(obj)                \
    NN_LOG("  %-10s %3d/%3d\n",         \
        #obj,                           \
        nn::os::obj::GetCurrentCount(), \
        nn::os::obj::GetMaxCount() )

void DumpObjectCounts()
{
    NN_LOG("----\n");
    NN_LOG("DumpObjectCounts\n");
    PRINT_COUNT(Event);
    PRINT_COUNT(Mutex);
    PRINT_COUNT(Semaphore);
    PRINT_COUNT(Thread);
    PRINT_COUNT(Timer);
    NN_LOG("----\n");
}

}
