﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     os_EventDemo.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

#include "os_DemoUtil.h"

/* Please see man pages for details 
    
    
*/
/* Please see man pages for details 
    
    
*/
/* Please see man pages for details 
    
    
    
    
    
*/

// Events used in the demo
nn::os::Event s_Event[NN_DEMO_OS_MAX_THREAD_NUM];


/* Please see man pages for details 
  

                
                
                
                
                
                

  

  
*/
void EventDemoFunc(uptr param)
{
    u32 threadNum = static_cast<int>(param);

    // Wait for an event
    NN_LOG(" Thread %d, Waiting event\n", threadNum);
    s_Event[threadNum].Wait();

    // Generate an event
    NN_LOG(" Thread %d, Signaled event\n", threadNum);
    for(int i = 0; i < 5; ++i)
    {
        NN_LOG("  Thread %d, i = %d\n", threadNum, i);
        nn::os::Thread::Yield();
    }

    // Clear the event
    s_Event[threadNum].ClearSignal();

    // Trigger the next event
    NN_LOG(" Thread %d, Signal next event\n", threadNum);
    s_Event[(threadNum + 1) % NN_DEMO_OS_MAX_THREAD_NUM].Signal();

    // Wait for another event
    NN_LOG(" Thread %d, Waiting event again\n", threadNum);
    s_Event[threadNum].Wait();
    NN_LOG(" Thread %d, Signaled event again\n", threadNum);

    // Trigger the next event
    NN_LOG(" Thread %d, Signal next event\n", threadNum);
    s_Event[(threadNum + 1) % NN_DEMO_OS_MAX_THREAD_NUM].Signal();

    NN_LOG(" Thread %d, END\n", threadNum);
}

/* Please see man pages for details 
  

                

  

  
*/
void SimpleEventFunc(nn::os::Event* pEvent)
{
    // Wait for an event
    NN_LOG("Waiting event\n");
    pEvent->Wait();

    NN_LOG("Thread END\n");
}

/* Please see man pages for details 
  

                

  
*/
void EventDemo1()
{
    NN_LOG("-- Event Demo1 --\n");

    // Creates events
    for(u32 i = 0; i < NN_DEMO_OS_MAX_THREAD_NUM; ++i)
    {
        s_Event[i].Initialize(true);
    }

    // Creates thread
    nn::os::Thread threads[NN_DEMO_OS_MAX_THREAD_NUM];
    DemoUtil::CreateThreads(threads, EventDemoFunc, NN_DEMO_OS_MAX_THREAD_NUM);

    DemoUtil::DumpObjectCounts();

    // Send notification only of the first event.
    s_Event[0].Signal();

    // Wait for thread termination
    DemoUtil::JoinThreads(threads, NN_DEMO_OS_MAX_THREAD_NUM);

    // Discard event
    for(u32 i = 0; i < NN_DEMO_OS_MAX_THREAD_NUM; ++i)
    {
        s_Event[i].Finalize();
    }
}

/* Please see man pages for details 
  

                

  
*/
void EventDemo0()
{
    NN_LOG("-- Event Demo0 --\n");

    // Creates events
    nn::os::Event event0(false);

    // Create thread that waits for events
    const size_t STACK_SIZE = 4096;
    nn::os::Thread thread;
    nn::os::StackMemoryBlock stack(STACK_SIZE);
    thread.Start(SimpleEventFunc, &event0, stack);

    // Wait for thread creation
    nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(1));

    // Notify the thread waiting for the event
    NN_LOG("Signal event.\n");
    event0.Signal();

    // Wait for thread termination
    thread.Join();
}

/* Please see man pages for details 
  

                

  
*/
void EventDemo()
{
    NN_LOG("*** Event Demo starts. ***\n");

    EventDemo0();
    EventDemo1();

    NN_LOG("*** End of Event Demo ***\n");
}

/*  

*/

/*  

*/

/*  

*/

