/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     os_MemoryDemo.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

/*!
    @addtogroup Demo Demo
    @{
*/
/*!
    @addtogroup os_Demo OS
    @{
*/
/*!
    @defgroup os_MemoryDemo Memory
    
    Demo program for memory functions.
    
    @{
*/

/*!
  @brief        Demo that allocates, accesses, and frees a memory block.

                Allocates a memory block, and displays the size and start address of the allocated block. 
                Loads to and reads from the allocated memory block. 

  @return  None.
*/
void MemoryDemo1()
{
    NN_LOG("-- Memory Demo1 --\n");

    const int SIZE = 16 * 1024 * 1024;
    nn::os::MemoryBlock block(SIZE);
    NN_LOG("* Allocate Size:%08x\n", block.GetSize());
    NN_LOG("* Allocate Addr:%08x\n", block.GetAddress());

    // Writes a value to memory.
    int* p = reinterpret_cast<int*>(block.GetAddress());
    NN_LOG("* Write From   :%08x\n", p);
    for(int i = 0; i < 4096; i++)
    {
        *p = i;
        p++;
    }
    NN_LOG("*       To     :%08x\n", p);

    // Load a value from memory
    p = reinterpret_cast<int*>(block.GetAddress());
    NN_LOG("* Read From    :%08x\n", p);
    for(int i = 0; i < 4096; i++)
    {
        NN_ASSERT(*p == i);
        p++;
    }
    NN_LOG("*      To      :%08x\n", p);

    // Frees the memory block
    block.Finalize();
}

/*!
  @brief        This is a memory block demo. 

                Runs the memory block demo. 

  @return  None.
*/
void MemoryDemo()
{
    NN_LOG("*** Memory Demo starts. ***\n");

    MemoryDemo1();

    NN_LOG("*** End of Memory Demo ***\n");
}

/*!
    @}
*/

/*!
    @}
*/

/*!
    @}
*/

