﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     os_MutexDemo.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46695 $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>

#include "os_DemoUtil.h"

/* Please see man pages for details 
    
    
*/
/* Please see man pages for details 
    
    
*/
/* Please see man pages for details 
    
    
    
    
    
*/

using nn::os::Thread;
using nn::os::Mutex;

namespace {

    volatile int g_Counter;
    Mutex g_Lock;
    Thread g_Threads1[NN_DEMO_OS_MAX_THREAD_NUM];
    Thread g_Threads2[NN_DEMO_OS_MAX_THREAD_NUM];

    void IncrementWithoutMutex(uptr)
    {
        // Although it is contrived, immediately after loading a variable in the register, goes to sleep and control is transferred to another thread.
        // 
        int tmp = g_Counter;
        nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(100));
        ++tmp;
        g_Counter = tmp;
    }

    void IncrementWithMutex(uptr)
    {
        g_Lock.Lock();
    
        // Since the overall addition is made exclusive, g_Counter can read or write to only one thread at a time.
        // 
        IncrementWithoutMutex(NULL);
    
        g_Lock.Unlock();
    }

}


/* Please see man pages for details 
  

                
                

  
*/

void MutexDemo()
{
    NN_LOG("*** Mutex Demo starts. ***\n");
    g_Lock.Initialize();

    NN_LOG("g_Counter is to be incremented by %d threads\n", NN_DEMO_OS_MAX_THREAD_NUM);
    // Updates g_Counter after exclusive control is made.
    g_Counter = 0;

    // Creates thread
    DemoUtil::CreateThreads(g_Threads1, IncrementWithMutex, NN_DEMO_OS_MAX_THREAD_NUM);

    // Wait for thread termination
    DemoUtil::JoinThreads(g_Threads1, NN_DEMO_OS_MAX_THREAD_NUM);

    NN_LOG("IncrementWithMutex: g_Counter=%d\n", g_Counter);

    // Updates g_Counter without exclusive control.
    g_Counter = 0;
    // Creates thread
    DemoUtil::CreateThreads(g_Threads2, IncrementWithoutMutex, NN_DEMO_OS_MAX_THREAD_NUM);

    // Wait for thread termination
    DemoUtil::JoinThreads(g_Threads2, NN_DEMO_OS_MAX_THREAD_NUM);
    NN_LOG("IncrementWithoutMutex: g_Counter=%d\n", g_Counter);

    NN_LOG("*** End of Mutex Demo ***\n");
}


/*  

*/

/*  

*/

/*  

*/

