﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/applet.h>
#include <nn/ptm/CTR/ptm_Api.h>

#include "demo.h"

namespace {

    const u8 TEXT_LEN = 16;
    const char ADAPTER_STATE_TEXT[][TEXT_LEN] =
    {
        "NOT CONNECTED",
        "CONNECTED"
    };
    const char BATTERY_CHARGE_STATE_TEXT[][TEXT_LEN] =
    {
        "NOT CHARGING",
        "CHARGING"
    };
    const char BATTERY_LEVEL_TEXT[][TEXT_LEN] =
    {
        "LEVEL 0",
        "LEVEL 1",
        "LEVEL 2",
        "LEVEL 3",
        "LEVEL 4",
        "LEVEL 5"
    };

}

extern "C" void nnMain()
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("Battery Demo\n");

    const u32 gxHeapSize = 0x800000;

    // Allocate heap
    nn::fnd::ExpHeap appHeap;
    appHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );

    // Prepare RenderSystem
    uptr heapForGx = reinterpret_cast<uptr>(appHeap.Allocate(gxHeapSize));
    demo::RenderSystemDrawing renderSystem;
    renderSystem.Initialize(heapForGx, gxHeapSize);
    renderSystem.SetColor(1.f, 1.f, 1.f);

    nn::ptm::CTR::Initialize();

    while(1)
    {        
        renderSystem.SetRenderTarget(NN_GX_DISPLAY0);
        renderSystem.Clear();

        nn::ptm::CTR::AdapterState adapterState = nn::ptm::CTR::GetAdapterState();
        nn::ptm::CTR::BatteryChargeState batteryChargeState = nn::ptm::CTR::GetBatteryChargeState();
        nn::ptm::CTR::BatteryLevel batteryLevel = nn::ptm::CTR::GetBatteryLevel();

        renderSystem.DrawText(20, 20, "BATTERY DEMO");
        renderSystem.DrawText(40, 50, "ADAPTER STATE        : %s", ADAPTER_STATE_TEXT[adapterState]);
        renderSystem.DrawText(40, 70, "BATTERY CHARGE STATE : %s", BATTERY_CHARGE_STATE_TEXT[batteryChargeState]);
        renderSystem.DrawText(40, 90, "BATTERY LEVEL        : %s", BATTERY_LEVEL_TEXT[batteryLevel]);

        renderSystem.SwapBuffers();
        renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
    }

    // nn::ptm::CTR::Finalize();
}
