﻿#ifndef NN_SAMPLE_RDT_DEMO_RDT_COMMON_H_
#define NN_SAMPLE_RDT_DEMO_RDT_COMMON_H_

#include <nn/types.h>
#include <nn/Result.h>

#define SUPPORT_UDS_CLIENT

#define DEBUG_PRINTF(format, args...)     NN_LOG(format, ## args)
#define TRACE()       DEBUG_PRINTF("[%s(%d)] TRACE\n", __FUNCTION__, __LINE__)
#define RESULT_TRACE(result)     do{\
    if(result.IsFailure()) DEBUG_PRINTF("[%s(%d)] ## result %08x \n",__FUNCTION__, __LINE__, result.GetPrintableBits()); \
    }while(0) 

#define MSLEEP(msec) nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMilliSeconds(msec))
#define PUSHED_A(padStatus)     (padStatus.hold & nn::hid::BUTTON_A)
#define PUSHED_B(padStatus)     (padStatus.hold & nn::hid::BUTTON_B)
#define PUSHED_X(padStatus)     (padStatus.hold & nn::hid::BUTTON_X)
#define PUSHED_Y(padStatus)     (padStatus.hold & nn::hid::BUTTON_Y)
#define PUSHED_L(padStatus)     (padStatus.hold & nn::hid::BUTTON_L)
#define PUSHED_R(padStatus)     (padStatus.hold & nn::hid::BUTTON_R)

/*

#define PUSHED_A(padStatus)     (padStatus & PAD_BUTTON_A)
#define PUSHED_B(padStatus)     (padStatus & PAD_BUTTON_B)
#define PUSHED_X(padStatus)     (padStatus & PAD_BUTTON_X)
#define PUSHED_Y(padStatus)     (padStatus & PAD_BUTTON_Y)
#define PUSHED_L(padStatus)     (padStatus & PAD_BUTTON_L)
#define PUSHED_R(padStatus)     (padStatus & PAD_BUTTON_R)
*/
#define MEMCPY      std::memcpy
#define MEMCMP      std::memcmp
#define MEMSET      std::memset

static const size_t RDT_COMMAND_SIZE = 128;
static const size_t RDT_SENDDATA_SIZE = 512*1024;

static const size_t RDT_BUF_SIZE = 16*1024;
static const u8     RDT_DATA_PORT_NUMBER = 5;
static const u8     RDT_COMMAND_PORT_NUMBER_C2S = 10;
static const u8     RDT_COMMAND_PORT_NUMBER_S2C = 9;

static const u16    RDT_DEMO_HEADER_REQ = 0x5555;
static const u16    RDT_DEMO_HEADER_ACK = 0x5757;
static const u16    RDT_DEMO_HEADER_CANCEL = 0x9999;

// Constant that specifies how many times to call the Process function in each game frame.
// In this sample demo, the sending side uses the Send function to write 4KB of data to the send buffer all at once in each game frame.
//
// Around 1400 bytes of data are sent every time the Process function is called, so if the Process function is called three times, that is 4200 bytes (1400 x 3), which balances out with the 4KB written to the send buffer by the Send function.
//
//
static const int REPEAT_PROCESS_FUNCTION = 3;


#endif
