﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_DataServer.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_RDT_DEMO_RDT_DATA_SERVER_H_
#define NN_SAMPLE_RDT_DEMO_RDT_DATA_SERVER_H_

#include <nn.h>
#include <nn/rdt.h>
/* RdtServer 
    - Sender for sending data
    - Sender for sending commands
    - Receiver for receiving commands
    
        STATE_WAIT
            The Receiver for receiving commands waits for REQUEST to be sent from Client.
        STATE_OPEN
            The Sender for sending data sends data to Client.
        STATE_WAITING_FINISHED
            The Receiver for receiving commands waits for ACK to be sent from Client.
        STATE_FINISHED
            In this state, ACK has been received and data has finished being sent.

        STATE_CANCEL_SEND
            The Sender for sending commands sends a CANCEL command.
            End the Cancel process.
        STATE_CANCEL_RECEIVED
            The Receiver for receiving commands receives the CANCEL command.
            The Sender for sending data is closed by the Cancel function.
        STATE_CANCEL_CLOSED
            The cancel process ends and the Sender and Receiver are stopped.

*/
class RdtServer
{
public:
    RdtServer();
    ~RdtServer();

    nn::Result Initialize(nn::rdt::SenderConfig& dataSenderConfig, nn::rdt::SenderConfig& commandSenderConfig, nn::rdt::ReceiverConfig& commandReceiverConfig);
    void Finalize();
    
    /* Set data that the server will send via the data port. */
    void        SetSendData(u8* data, size_t size);

    /* Start Server processes. Transition to STATE_WAIT. */
    nn::Result  Start();
    
    /* Stop Server processes. Transition to STATE_CANCEL_SEND. */
    void        Cancel();
    void        ForceClose();

    /* Call the Process functions for Sender and Receiver. */
    nn::Result  Process();

    /* Perform the Sender and Receiver processes and transition the state. */
    nn::Result  Update();
    
    /* True when STATE_FINISHED. In this state, the sending of data has completed. */
    bool IsFinished();

    /* True when stopped due to a cancel or error. */
    bool IsCanceled();
    
private:
    enum ServerState
    {
        STATE_CLOSE,
        STATE_WAIT,
        STATE_OPEN,
        STATE_WAITING_FINISHED,
        STATE_FINISHED,
        STATE_CANCEL_SEND,
        STATE_CANCEL_RECEIVED,
        STATE_CANCEL_CLOSED,
        STATE_ERROR,
        STATE_EXIT,
        
        STATE_MAX
    };

    bool IsReqCommand(u8* data);
    bool IsAckCommand(u8* data);
    bool IsCancelCommand(u8* data);
    static void SetCommandCancel(u8* data);
    static void initClose();
    static void initWait();
    static void initOpen();
    static void initWaitingFinished();
    static void initFinished();
    static void initCancelSend();
    static void initCancelReceived();
    static void initCancelClosed();
    static void initError();
    static void initExit();

    void ChangeState(ServerState state);
    nn::Result  UpdateRdt();

    nn::rdt::Sender*    m_pCommandSender;
    nn::rdt::Receiver*  m_pCommandReceiver;
    nn::rdt::Sender*    m_pDataSender;
    ServerState    m_State;

    NN_PADDING3;

};

#endif

