﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_ReceiverUtil.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_RDT_DEMO_RDT_RECEIVER_UTIL_H_
#define NN_SAMPLE_RDT_DEMO_RDT_RECEIVER_UTIL_H_

#include <nn.h>
#include <nn/rdt.h>

enum ReceiverUtilState
{
    RECEIVER_CLOSE,
    RECEIVER_OPENING,               // Receiver::Wait() through RECEIVER_STATE_OPENED
    RECEIVER_OPENED,                // RECEIVER_STATE_OPENED through RECEIVER_STATE_FINISHED through Receiver::Close
    RECEIVER_WAITING_FINISHED,      // RECEIVER_STATE_FINISHED through RECEIVER_STATE_CLOSED
    RECEIVER_FINISHED,              // RECEIVER_STATE_CLOSED
    
    RECEIVER_STATE_MAX
};


class ReceiverUtil
{
public:

    void Initialize(nn::rdt::Receiver* receiver);

    /* Start Receiver operations. Proceed through functions: Receiver::Wait -> Receiver::Receive -> Receiver::Close */
    void Open(s32 timeout = 0);

    /* Stop Receiver operations.*/
    void Close();

    /* Call the Receiver::Cancel function. */
    void Cancel();
    
    /* Call the Receiver::Process function and perform ReceiverUtil state transitions.
       In this demo, processes like the Receiver::Process function and the Receiver::Receive function are called at the same time. */
    nn::Result Process();
    
    nn::Result Update();
    
    void SetReceiveBuf(u8* data, size_t size);

    bool IsReceiveFinished();
    bool IsClosed();
private:

    /*
    bool finished       True if the processes that the various functions are supposed to do are completed.
    bool success        True if the processes of the various functions have ended normally.
    s32  timeout        Cancel if there is still no change to the process after the number of times specified by this value.
                        (This is not used in this demo because there is upper-level monitoring.)
    */

    nn::Result Ready( bool* finished, bool* success, s32 timeout = 0);

    nn::Result Receive( u8* dataBuf, size_t dataBufSize, bool* finished, bool* success, s32 timeout = 0);

    nn::Result Finish( bool* finished, bool* success, s32 timeout = 0);

   
    nn::rdt::Receiver*      m_pReceiver;
    u8*                     m_recvData;
    size_t                  m_recvDataSize;
    ReceiverUtilState       m_ReceiverUtilState;
    NN_PADDING3;
    size_t                  m_TotalRecv;
    s32                     m_Timeout;
};

#endif
