﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     rdt_SenderUtil.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev:$
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_RDT_DEMO_RDT_SENDER_UTIL_H_
#define NN_SAMPLE_RDT_DEMO_RDT_SENDER_UTIL_H_

#include <nn.h>
#include <nn/rdt.h>

enum SenderUtilState
{
    SENDER_CLOSE,
    SENDER_OPENING,             // Sender::Open() through SENDER_STATE_OPENED
    SENDER_OPENED,              // SENDER_STATE_OPENED through Sender::Close()
    SENDER_WAITING_FINISHED,    // SENDER_STATE_WAITING_FINISHED through SENDER_STATE_CLOSED
    SENDER_FINISHED,            // SENDER_STATE_CLOSED
    
    SENDER_STATE_MAX
};


class SenderUtil
{
public:
    nn::Result Initialize(nn::rdt::Sender* sender);
    void SetSendData(u8* data, size_t size);
    
    /* Start Sender operations. Proceed through functions: Sender::Open -> Sender::Send -> Sender::Close */
    void Open(s32 timeout = 0);
    
    /* Stop Sender operations.*/
    void Close();
    
    /* Call the Sender::Cancel function. */
    void Cancel();
    
    /* Call the Sender::Process function and perform SenderUtil state transitions.
       In this demo, processes like the Sender::Process function and the Sender::Send function are called at the same time. */
    nn::Result Process();

    nn::Result Update();
    
    /* True if the state has transitioned to the sending-completed state. */
    bool IsSendFinished();
    
    bool IsClosed();

private:

    /*
    bool finished       True if the processes that the various functions are supposed to do are completed.
    bool success        True if the processes of the various functions have ended normally.
    s32  timeout        Cancel if there is still no change to the process after the number of times specified by this value.
                        (This is not used in this demo because there is upper-level monitoring.)
    */

    nn::Result Ready( bool* finished, bool* success, s32 timeout = 0);

    nn::Result Send( u8* data, size_t dataSize, bool* finished, bool* success, s32 timeout = 0);

    nn::Result Finish( bool* finished, bool* success, s32 timeout = 0);

    
    nn::rdt::Sender*        m_pSender;
    u8*                     m_sendData;
    size_t                  m_sendDataSize;
    SenderUtilState         m_SenderUtilState;
    NN_PADDING3;
    size_t                  m_TotalSent;
    s32                     m_Timeout;
};

#endif
