﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     demo1.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47355 $
 *---------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <nn.h>
#include "applet.h"
#include "../device.h"
#include "../ModuleGx/gx.h"
#include "../ModuleRtc/rtc.h"
#include "../ModulePad/pad.h"
#include "../ModuleTp/tp.h"
#include "../ModuleAcc/acc.h"
#include "../ModuleMic/mic.h"
#include "../ModuleSnd/snd.h"
#include "../ModuleCamera/camera.h"
#include "../ModuleFs/fs.h"
#include "../ModuleSavedata/savedata.h"

#include "demo.h"


//----------------------------------------------------------
// CRO
namespace
{
    const size_t HEAP_SIZE = 20 * 1024 * 1024;
    const size_t HEAP_SIZE_DLL = 2 * 1024 * 1024;

    const char* const MODULE_NAMES[] =
    {
        "ModuleAcc",
        "ModuleCamera",
        "ModuleFs",
        "ModuleGx",
        "ModuleMic",
        "ModulePad",
        "ModuleRtc",
        "ModuleSavedata",
        "ModuleSnd",
        "ModuleTp",
        "ModuleDemo",
    };
    const int NUM_MODULES = sizeof(*nn::util::NumOfElementsT(MODULE_NAMES));

    nn::ro::Module* s_pModules[NUM_MODULES];
    nn::ro::RegistrationList* s_pRl;

    bit8  s_Buffer[1* 1024 * 1024] NN_ATTRIBUTE_ALIGN(0x1000);

    void* s_pHeapBuffer = NULL;
    bit8* s_pHeapPointer = NULL;
}

nn::ro::Module* LoadCro(const char* path)
{
    s64 croSize;

    // Load CRO from ROM-FS
    nn::fs::FileInputStream f;
    f.Initialize(path);
    croSize = f.GetSize();
    f.Read(s_pHeapPointer, croSize);
    f.Finalize();

    void* pRo = s_pHeapPointer;

    // Allocate buffer for CRO
    nn::ro::SizeInfo sizeInfo;
    nn::ro::GetSizeInfo(&sizeInfo, pRo);

    s_pHeapPointer = reinterpret_cast<bit8*>(sizeInfo.fix3End);

    void* pBuffer = s_pHeapPointer;
    const size_t bufferSize = sizeInfo.bufferSize;
    s_pHeapPointer += nn::math::RoundUp(bufferSize, 0x1000);

    // Load CRO
    nn::ro::Module* pModule = nn::ro::LoadModule(pRo, croSize, pBuffer, bufferSize, true, nn::ro::FIX_LEVEL_3);
    NN_NULL_ASSERT(pModule);

    NN_LOG("cro=%08x-%08x(fix3End=%08x)  buf=%08x-%08x\n", 
        pRo,
        reinterpret_cast<uptr>(pRo) + static_cast<s32>(croSize),
        sizeInfo.fix3End,
        pBuffer, 
        reinterpret_cast<uptr>(pBuffer) + bufferSize );
    
    return pModule;
}
void SetupModules(nn::fnd::ExpHeap* pHeap)
{
    NN_LOG("Demo1 App: SetupModules begin\n");

    nn::fs::FileInputStream f;
    bit8* pStaticBss = s_Buffer;
    s_pHeapBuffer = pHeap->Allocate(HEAP_SIZE_DLL);

    s_pHeapPointer = reinterpret_cast<bit8*>(nn::math::RoundUp(reinterpret_cast<uptr>(s_pHeapBuffer), 0x1000));

    NN_LOG("Demo1 App: load crs\n");
    s64 crsSize;
    {
        f.Initialize(L"rom:/static.crs");
        crsSize = f.GetSize();
        f.Read(pStaticBss, crsSize);
        f.Finalize();
    }
    void* pRs = pStaticBss;
    pStaticBss += nn::math::RoundUp(crsSize, 0x1000);
    
    NN_LOG("Demo1 App: load crr\n");
    s64 crrSize;
    {
        f.Initialize(L"rom:/.crr/static.crr");
        crrSize = f.GetSize();
        f.Read(pStaticBss, crrSize);
        f.Finalize();
    }
    void* pRr = pStaticBss;
    pStaticBss += crrSize;

    const size_t crsSize4k = nn::math::RoundUp(crsSize, 0x1000);
    
    NN_LOG("Demo1 App: ro initialize\n");
    nn::ro::Initialize(pRs, crsSize4k);

    NN_LOG("Demo1 App: register crr\n");
    s_pRl = nn::ro::RegisterList(pRr, crrSize);
    NN_NULL_ASSERT(s_pRl);


    // Load
    for( int i = 0; i < NUM_MODULES; ++i )
    {
        char path[32];
        std::snprintf(path, sizeof(*nn::util::NumOfElementsT(path)), "rom:/%s.cro", MODULE_NAMES[i]);
        NN_LOG("Demo1 App: Load %-16s ", MODULE_NAMES[i]);
        s_pModules[i] = LoadCro(path);
    }

    // Check
    bool isAllModuleResolved = true;
    for( int i = 0; i < NUM_MODULES; ++i )
    {
        bool isAllSymbolResolved = s_pModules[i]->IsAllSymbolResolved();
        NN_LOG("Demo1 App: %-16s->IsAllSymbolResolved() = %d\n", MODULE_NAMES[i], isAllSymbolResolved);
        isAllModuleResolved = (isAllModuleResolved && isAllSymbolResolved);
    }
    NN_ASSERTMSG(isAllModuleResolved, "some modules are not resolved.");

    // Initialize
    for( int i = 0; i < NUM_MODULES; ++i )
    {
        NN_LOG("Demo1 App: %-16s->DoInitialize()\n", MODULE_NAMES[i]);
        s_pModules[i]->DoInitialize();
    }

    NN_LOG("Demo1 App: SetupModules end\n");
}
void CleanupModules(nn::fnd::ExpHeap* pHeap)
{
    NN_LOG("Demo1 App: CleanupModules begin\n");

    s_pRl->Unregister();

    // Finalize
    for( int i = 0; i < NUM_MODULES; ++i )
    {
        NN_LOG("Demo1 App: %-16s->DoFinalize()\n", MODULE_NAMES[i]);
        s_pModules[i]->DoFinalize();
    }
    
    // Unload
    for( int i = 0; i < NUM_MODULES; ++i )
    {
        NN_LOG("Demo1 App: %-16s->Unload()\n", MODULE_NAMES[i]);
        s_pModules[i]->Unload();
    }

    pHeap->Free(s_pHeapBuffer);

    NN_LOG("Demo1 App: CleanupModules end\n");
}
// CRO
//----------------------------------------------------------

bool StartDemo()
{
    // Description of Operations
    NN_LOG("X + A : Run Optional Device Demo (Ex. Card Backup Device..)\n");

    // #######################################################################################################
    // Application initialization sequence
    // #######################################################################################################

    const u32 s_GxHeapSize = 0x800000;
    bool restartFlag = false;

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    //----------------------------------------------------------
    // CRO
    nn::os::MemoryBlock mb;
    mb.Initialize(HEAP_SIZE);
    nn::fnd::ExpHeap roHeap;
    roHeap.Initialize(mb.GetAddress(), mb.GetSize() );
    SetupModules(&roHeap);
    // CRO
    //----------------------------------------------------------
    {
        // Initialize hid
        nn::hid::Initialize();

        // Allocate heap
        nn::fnd::ExpHeap appHeap;
        appHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );

        // Prepare RenderSystem
        uptr heapForGx = reinterpret_cast<uptr>(appHeap.Allocate(s_GxHeapSize));
        demo::RenderSystemDrawing s_RenderSystem;
        s_RenderSystem.Initialize(heapForGx, s_GxHeapSize);


        // Graphics demo initialization
        GraphicsDemo graphicsDemo;
        graphicsDemo.Initialize(&s_RenderSystem);

        // Camera demo initialization
        CameraDemo cameraDemo;
        cameraDemo.Initialize(&s_RenderSystem, &appHeap);

        // System clock demo initialization
        RtcDemo rtcDemo;
        rtcDemo.Initialize(&s_RenderSystem);

        // Pad demo initialization
        PadDemo padDemo;
        padDemo.Initialize(&s_RenderSystem);

        // Touch panel demo initialization
        TouchPanelDemo touchPanelDemo;
        touchPanelDemo.Initialize(&s_RenderSystem);

        // Accelerometer demo initialization
        AccDemo accDemo;
        accDemo.Initialize(&s_RenderSystem);

        // Mic demo initialization
        MicDemo micDemo;
        micDemo.Initialize(&s_RenderSystem);

        // Sound demo initialization
        SndDemo sndDemo;
        sndDemo.Initialize(&s_RenderSystem, &appHeap);

        // File system demo initialization
        FsDemo fsDemo;
        fsDemo.Initialize(&s_RenderSystem, &appHeap);

        // Save data
        SaveDataDemo saveDataDemo;

        // Device list
        Device* deviceList[] = {
                                    &graphicsDemo,
                                    &cameraDemo,
                                    &rtcDemo,
                                    &padDemo,
                                    &touchPanelDemo,
                                    &accDemo,
                                    &micDemo,
                                    &sndDemo,
                                    &fsDemo
                               };
        int deviceNum = sizeof(deviceList) / sizeof(Device*);

        Device* optionalDeviceList[] =
                               {
                                    &saveDataDemo
                               };
        int optionalDeviceNum = sizeof(optionalDeviceList) / sizeof(Device*);

        bool isOptionalDeviceTest = false;

        // Start each device
        for(int i = 0; i < deviceNum; i++)
        {
            deviceList[i]->Start();
        }

        // After this, the application itself handles sleep
        TransitionHandler::EnableSleep();

        // #######################################################################################################
        // Main loop
        // #######################################################################################################
        while(1)
        {
            s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
            s_RenderSystem.Clear();
            deviceList[0]->Draw();
            s_RenderSystem.SwapBuffers();

            s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
            s_RenderSystem.Clear();
            for(int i = 1; i < deviceNum; i++)
            {
                deviceList[i]->Draw();
            }

            s_RenderSystem.SwapBuffers();

            // Restart if START + L + R is pressed
            if ( padDemo.GetLatestPadStatus().hold == (nn::hid::BUTTON_START | nn::hid::BUTTON_L | nn::hid::BUTTON_R) )
            {
                restartFlag = true;
                break;
            }

            if(padDemo.GetLatestPadStatus().hold == (nn::hid::BUTTON_X | nn::hid::BUTTON_A) && !isOptionalDeviceTest)
            {
                //Additional device demo initialization
                saveDataDemo.Initialize(&s_RenderSystem, &appHeap);

                for(int i = 0; i < optionalDeviceNum; i++)
                {
                    optionalDeviceList[i]->Start();
                }
                isOptionalDeviceTest = true;
                NN_LOG("Run Optional Device Demo.\n");

            }
            for(int i = 0; i < optionalDeviceNum&&isOptionalDeviceTest; i++)
            {
                optionalDeviceList[i]->Draw();
            }

            s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

            // Run sleep, HOME Menu, and POWER Menu transitions.
            TransitionHandler::Process();
            // Determine whether there has been an exit notification.
            if (TransitionHandler::IsExitRequired())
            {
                break;
            }
        }

        // #######################################################################################################
        // Application termination processing
        // #######################################################################################################
        // When nn::applet::IsExpectedToCloseApplication() == true happens, the sleep-type callbacks cannot be called automatically.
        // 

        //Stop and terminate added devices
        if( isOptionalDeviceTest == true )
        {
            for(int i = 0; i < optionalDeviceNum; i++)
            {
                optionalDeviceList[i]->End();
                optionalDeviceList[i]->Finalize();
            }
        }

        // Stop each device
        for(int i = 0; i < deviceNum; i++)
        {
            deviceList[i]->End();
        }

        // Terminate each device
        for(int i = 0; i < deviceNum; i++)
        {
            deviceList[i]->Finalize();
        }

        // Exit RenderSystem
        s_RenderSystem.Finalize();

        // Destroy heap
        appHeap.Free(reinterpret_cast<void*>(heapForGx));
        appHeap.Finalize();
    }
    //----------------------------------------------------------
    // CRO
    CleanupModules(&roHeap);
    roHeap.Finalize();
    mb.Finalize();
    // CRO
    //----------------------------------------------------------

    return restartFlag;
}

void nnMain()
{
    NN_LOG("Demo1 Start\n");

    bool restartFlag = false;
    u8 restartCount = 0;

    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();

    // Receive values passed with RestartApplication
    {
        u8 dBuf[ NN_APPLET_PARAM_BUF_SIZE ];
        bool r = nn::applet::GetStartupArgument( dBuf );
        if ( r )
        {
            if (dBuf[NN_APPLET_PARAM_BUF_SIZE - 1] == 0xFF)
            {
                NN_LOG("----demo1 received StartupArgument.\n");
                NN_LOG("    jump count = %d\n", dBuf[0] );
                restartCount = dBuf[0];
            }
        }
        else
        {
            NN_LOG("----demo1 cannot receive startupArgument.\n");
        }
    }

    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        restartFlag = StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    NN_LOG("Demo1 End\n");

    // Performs application restart or finalization.
    if ( restartFlag )
    {
        u8 startupArgBuf[NN_APPLET_PARAM_BUF_SIZE];
        startupArgBuf[0] = restartCount + 1;
        startupArgBuf[NN_APPLET_PARAM_BUF_SIZE - 1] = 0xFF;
        // Restarts the application. The call to nn::applet::RestartApplication does not return.
        nn::applet::RestartApplication(startupArgBuf);
    }
    else
    {
        // Finalize the application. The call to nn::applet::CloseApplication does not return.
        nn::applet::CloseApplication();
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
