﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46894 $
 *---------------------------------------------------------------------------*/

#include <nn.h>

#include <nn/fnd.h>
#include <nn/os.h>
#include <nn/fs.h>
#include <nn/ro.h>
#include <nn/applet.h>

#include "module.h"
#include "module2.h"

void nnMain()
{
    // Call only the nn::applet::Enable function to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    NN_LOG("App: nnMain\n");

    NN_LOG("App: initialize\n");

    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    nn::fs::FileInputStream f;

    nn::os::MemoryBlock mb;
    mb.Initialize(8 * 1024 * 1024);
    bit8* pHeap = reinterpret_cast<bit8*>(mb.GetAddress());

    NN_LOG("App: load crs\n");
    s64 crsSize;
    {
        f.Initialize(L"rom:/static.crs");
        crsSize = f.GetSize();
        f.Read(pHeap, crsSize);
        f.Finalize();
    }
    void* pRs = pHeap;
    pHeap += nn::math::RoundUp(crsSize, 0x1000);

    NN_LOG("App: load crr\n");
    s64 crrSize;
    {
        f.Initialize(L"rom:/.crr/static.crr");
        crrSize = f.GetSize();
        f.Read(pHeap, crrSize);
        f.Finalize();
    }
    void* pRr = pHeap;
    pHeap += crrSize;


    NN_LOG("App: load Module\n", pHeap);
    s64 cro1Size;
    {
        f.Initialize(L"rom:/Module.cro");
        cro1Size = f.GetSize();
        f.Read(pHeap, cro1Size);
        f.Finalize();
    }
    void* pRo1 = pHeap;
    pHeap += nn::math::RoundUp(cro1Size, 0x1000);


    NN_LOG("App: load Module2\n", pHeap);
    s64 cro2Size;
    {
        f.Initialize(L"rom:/Module2.cro");
        cro2Size = f.GetSize();
        f.Read(pHeap, cro2Size);
        f.Finalize();
    }
    void* pRo2 = pHeap;
    pHeap += nn::math::RoundUp(cro2Size, 0x1000);

    nn::ro::SizeInfo sizeInfo1;
    nn::ro::GetSizeInfo(&sizeInfo1, pRo1);
    void* pBuffer1 = pHeap;
    pHeap += nn::math::RoundUp(sizeInfo1.bufferSize, 0x1000);

    nn::ro::SizeInfo sizeInfo2;
    nn::ro::GetSizeInfo(&sizeInfo2, pRo2);
    void* pBuffer2 = pHeap;
    pHeap += nn::math::RoundUp(sizeInfo2.bufferSize, 0x1000);

    NN_LOG("App:    ro initialize\n");
    nn::ro::Initialize(pRs, crsSize);

    NN_LOG("App:    register crr\n");
    nn::ro::RegistrationList* pRl = nn::ro::RegisterList(pRr, crrSize);
    NN_NULL_ASSERT(pRl);
    {
        NN_LOG("App:    ro load\n");
        nn::ro::Module* pModule1 = nn::ro::LoadModule(pRo1, cro1Size, pBuffer1, sizeInfo1.bufferSize);
        nn::ro::Module* pModule2 = nn::ro::LoadModule(pRo2, cro2Size, pBuffer2, sizeInfo2.bufferSize);
        NN_NULL_ASSERT(pModule1);
        NN_NULL_ASSERT(pModule2);

        NN_LOG("App:    ro initialize\n");
        pModule1->DoInitialize();
        pModule2->DoInitialize();

        NN_LOG("App:    call\n");
        Test();

        NN_LOG("App:    ro unload Module2\n");
        pModule2->DoFinalize();
        pModule2->Unload();

        NN_LOG("App:    call\n");
        Test();

        NN_LOG("App:    ro unload Module\n");
        pModule1->DoFinalize();
        pModule1->Unload();
    }
    NN_LOG("App:    unregister crr\n");
    pRl->Unregister();
    NN_LOG("App:    ro finalize\n");
    nn::ro::Finalize();

    NN_LOG("App: end\n");
}
