﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     EffectSample.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_DEMOS_SND_EFFECT_SAMPLE_H_
#define NN_SAMPLE_DEMOS_SND_EFFECT_SAMPLE_H_

#include <nn/snd.h>
#include "SndDemoBase.h"

class EffectSample : public SndDemoBase
{
protected:
    virtual void OnInitialize();
    virtual void OnFinalize();
    virtual void OnUpdate(nn::hid::PadStatus&);
    virtual void OnDraw(demo::RenderSystemDrawing&);
    virtual void SoundThreadFuncImpl();

public:
    enum EffectType
    {
        EFFECT_TYPE_NONE = 0,
        EFFECT_TYPE_CPU_DELAY = 1,
        EFFECT_TYPE_CPU_REVERB = 2,
        EFFECT_TYPE_DSP_DELAY = 3,
        EFFECT_TYPE_DSP_REVERB = 4,
        EFFECT_TYPE_NUM = 5
    };
    const char*& GetEffectTypeString()
    {
        static const char* STRING[] =
        {
            "None",
            "Delay (CPU)", "Reverb (CPU)",
            "Delay (DSP)", "Reverb (DSP)"
        };
        return STRING[m_FxType];
    }

private:
    void InitializeEffect(EffectType type);
    void FinalizeEffect();

private:
    static const int SOUND_THREAD_PRIORITY = 2;
    static const size_t SOUND_THREAD_STACK_SIZE = 4096;
    nn::os::Thread m_SoundThread;
    bool m_SoundThreadFlag;
    NN_PADDING3;
    nn::os::CriticalSection m_CriticalSection;

    nn::util::SizedEnum1<EffectType> m_FxType;
    NN_PADDING3;
    uptr                        m_FxWorkBuffer;
    nn::snd::FxDelay            m_FxDelay;
    nn::snd::FxDelay::Param     m_FxDelayParam;
    nn::snd::FxReverb           m_FxReverb;
    nn::snd::FxReverb::Param    m_FxReverbParam;
    nn::snd::DspFxDelay         m_DspFxDelay;
    nn::snd::DspFxDelay::Param  m_DspFxDelayParam;
    nn::snd::DspFxReverb        m_DspFxReverb;
    nn::snd::DspFxReverb::Param m_DspFxReverbParam;
};

#endif  // NN_SAMPLE_DEMOS_SND_EFFECT_SAMPLE_H_
