﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     FilterSample.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/os.h>
#include <nn/hid.h>
#include <nn/snd.h>

#include "FilterSample.h"

#include "WaveFileReader.h"
#include "WaveFileWriter.h"

namespace
{
    const int FILE_NUM = 1;
    const char* FILE_NAME[FILE_NUM] =
    {
        "rom:/sinesweep.wav"
    };

    // Voice
    int s_CurrentVoice;
    nn::snd::Voice* s_pVoice[FILE_NUM];

    const int FILTER_NUM = 24;
    int s_FilterIndex = 0;

    enum MyFilterType
    {
        MY_FILTER_TYPE_NONE,
        MY_FILTER_TYPE_MONOPOLE,
        MY_FILTER_TYPE_BIQUAD_LOWPASS,
        MY_FILTER_TYPE_BIQUAD_HIGHPASS,
        MY_FILTER_TYPE_BIQUAD_BANDPASS
    };
    MyFilterType s_MyFilterType = MY_FILTER_TYPE_NONE;

    struct MonoFilterInfo
    {
        char* text;
        nn::snd::MonoFilterCoefficients coef;
    };

    MonoFilterInfo monoFilter[FILTER_NUM] =
    {
        { "16000 Hz", { 0x6a09, 0x15f6 } }, { "12800 Hz", { 0x6871, 0x178e } },
        { "10240 Hz", { 0x6463, 0x1b9c } }, { " 8000 Hz", { 0x5db3, 0x224c } },
        { " 6400 Hz", { 0x5618, 0x29e7 } }, { " 5120 Hz", { 0x4d7a, 0x3285 } },
        { " 4000 Hz", { 0x4367, 0x3c98 } }, { " 3200 Hz", { 0x3a5a, 0x45a5 } },
        { " 2560 Hz", { 0x31c5, 0x4e3a } }, { " 2000 Hz", { 0x2924, 0x56db } },
        { " 1600 Hz", { 0x2244, 0x5dbb } }, { " 1280 Hz", { 0x1c50, 0x63af } },
        { " 1000 Hz", { 0x16c0, 0x693f } }, { "  800 Hz", { 0x1292, 0x6d6d } },
        { "  640 Hz", { 0x0f18, 0x70e7 } }, { "  500 Hz", { 0x0bf5, 0x740a } },
        { "  400 Hz", { 0x09a9, 0x7656 } }, { "  320 Hz", { 0x07ca, 0x7835 } },
        { "  256 Hz", { 0x0646, 0x79b9 } }, { "  200 Hz", { 0x04ed, 0x7b12 } },
        { "  160 Hz", { 0x03f5, 0x7c0a } }, { "  128 Hz", { 0x032d, 0x7cd2 } },
        { "  100 Hz", { 0x027d, 0x7d82 } }, { "   80 Hz", { 0x01fe, 0x7e01 } },
    };


    struct BiquadFilterInfo
    {
        char* text;
        nn::snd::BiquadFilterCoefficients coef;
        NN_PADDING2;
    };

    // low-pass
    BiquadFilterInfo biquadLowPassFilter[FILTER_NUM] =
    {
        { "16000 Hz", { 0x36fa, 0x6df4, 0x36fa, 0x8c47, 0xcaca } },  // actually 15000 Hz
        { "12800 Hz", { 0x2bf0, 0x57e0, 0x2bf0, 0xa967, 0xdc70 } },
        { "10240 Hz", { 0x2071, 0x40e3, 0x2071, 0xcd87, 0xe904 } },
        { " 8000 Hz", { 0x173d, 0x2e7b, 0x173d, 0xef3d, 0xee4c } },
        { " 6400 Hz", { 0x110d, 0x221a, 0x110d, 0x0903, 0xeec0 } },
        { " 5120 Hz", { 0x0c59, 0x18b3, 0x0c59, 0x1eee, 0xecbf } },
        { " 4000 Hz", { 0x087e, 0x10fc, 0x087e, 0x332c, 0xe8d8 } },
        { " 3200 Hz", { 0x05f5, 0x0bea, 0x05f5, 0x423b, 0xe487 } },
        { " 2560 Hz", { 0x041f, 0x083d, 0x041f, 0x4e96, 0xdff6 } },
        { " 2000 Hz", { 0x02b3, 0x0565, 0x02b3, 0x598e, 0xdb05 } },
        { " 1600 Hz", { 0x01d1, 0x03a3, 0x01d1, 0x616d, 0xd6df } },
        { " 1280 Hz", { 0x0137, 0x026e, 0x0137, 0x67b7, 0xd325 } },
        { " 1000 Hz", { 0x00c5, 0x018a, 0x00c5, 0x6d2e, 0xcf90 } },
        { "  800 Hz", { 0x0082, 0x0103, 0x0082, 0x710d, 0xccce } },
        { "  640 Hz", { 0x0055, 0x00a9, 0x0055, 0x741f, 0xca7b } },
        { "  500 Hz", { 0x0035, 0x0069, 0x0035, 0x76c7, 0xc85a } },
        { "  400 Hz", { 0x0022, 0x0044, 0x0022, 0x78a9, 0xc6c7 } },
        { "  320 Hz", { 0x0016, 0x002c, 0x0016, 0x7a27, 0xc57c } },
        { "  256 Hz", { 0x000e, 0x001d, 0x000e, 0x7b57, 0xc46d } },
        { "  200 Hz", { 0x0009, 0x0012, 0x0009, 0x7c5f, 0xc37c } },
        { "  160 Hz", { 0x0006, 0x000b, 0x0006, 0x7d1b, 0xc2ce } },
        { "  128 Hz", { 0x0004, 0x0007, 0x0004, 0x7db0, 0xc241 } },
        { "  100 Hz", { 0x0002, 0x0004, 0x0002, 0x7e32, 0xc1c5 } },
        { "   80 Hz", { 0x0001, 0x0003, 0x0001, 0x7e8f, 0xc16b } },
    };

    // high-pass
    BiquadFilterInfo biquadHighPassFilter[FILTER_NUM] =
    {
        { "   80 Hz", { 0x3bf9, 0x880e, 0x3bf9, 0x7f0d, 0xc0f1 } },
        { "  100 Hz", { 0x3bdd, 0x8846, 0x3bdd, 0x7ed0, 0xc12c } },
        { "  128 Hz", { 0x3bb5, 0x8896, 0x3bb5, 0x7e7b, 0xc17f } },
        { "  160 Hz", { 0x3b88, 0x88f1, 0x3b88, 0x7e19, 0xc1dd } },
        { "  200 Hz", { 0x3b4f, 0x8962, 0x3b4f, 0x7d9e, 0xc252 } },
        { "  256 Hz", { 0x3b00, 0x8a00, 0x3b00, 0x7cf1, 0xc2f5 } },
        { "  320 Hz", { 0x3aa6, 0x8ab4, 0x3aa6, 0x7c2b, 0xc3ac } },
        { "  400 Hz", { 0x3a36, 0x8b95, 0x3a36, 0x7b31, 0xc48f } },
        { "  500 Hz", { 0x39aa, 0x8cac, 0x39aa, 0x79f8, 0xc5a5 } },
        { "  640 Hz", { 0x38e8, 0x8e31, 0x38e8, 0x783d, 0xc722 } },
        { "  800 Hz", { 0x380b, 0x8fea, 0x380b, 0x763e, 0xc8ca } },
        { " 1000 Hz", { 0x36fa, 0x920c, 0x36fa, 0x73b9, 0xcaca } },
        { " 1280 Hz", { 0x3580, 0x9500, 0x3580, 0x7027, 0xcd77 } },
        { " 1600 Hz", { 0x33d7, 0x9852, 0x33d7, 0x6c03, 0xd05c } },
        { " 2000 Hz", { 0x31ce, 0x9c64, 0x31ce, 0x66c2, 0xd3bc } },
        { " 2560 Hz", { 0x2f06, 0xa1f5, 0x2f06, 0x5f4a, 0xd80d } },
        { " 3200 Hz", { 0x2bf0, 0xa821, 0x2bf0, 0x569a, 0xdc70 } },
        { " 4000 Hz", { 0x2836, 0xaf95, 0x2836, 0x4b8b, 0xe12e } },
        { " 5120 Hz", { 0x2334, 0xb998, 0x2334, 0x3bb7, 0xe68e } },
        { " 6400 Hz", { 0x1dbf, 0xc483, 0x1dbf, 0x2913, 0xeb0d } },
        { " 8000 Hz", { 0x173d, 0xd186, 0x173d, 0x10c3, 0xee4c } },
        { "10240 Hz", { 0x0eab, 0xe2ab, 0x0eab, 0xec30, 0xee0b } },
        { "12800 Hz", { 0x05f5, 0xf416, 0x05f5, 0xbdc5, 0xe487 } },
        { "16000 Hz", { 0x00c5, 0xfe77, 0x00c5, 0x92d2, 0xcf90 } },  // actually 15000 Hz
    };

    // band-pass
    BiquadFilterInfo biquadBandPassFilter[FILTER_NUM] =
    {
        { "   0 - 16000 Hz", { 0x3ff5, 0x0000, 0xc00c, 0x0000, 0x3fea } },
        { "  80 - 14464 Hz", { 0x3c9d, 0x0000, 0xc364, 0x061c, 0x393a } },
        { "  95 - 12928 Hz", { 0x398b, 0x0000, 0xc676, 0x0c2a, 0x3315 } },
        { " 101 - 11520 Hz", { 0x36c4, 0x0000, 0xc93d, 0x11b5, 0x2d88 } },
        { " 113 - 10240 Hz", { 0x3427, 0x0000, 0xcbd9, 0x16e1, 0x284e } },
        { " 127 -  9152 Hz", { 0x31cc, 0x0000, 0xce35, 0x1b88, 0x2398 } },
        { " 143 -  8128 Hz", { 0x2f65, 0x0000, 0xd09b, 0x2044, 0x1eca } },
        { " 160 -  7232 Hz", { 0x2d16, 0x0000, 0xd2eb, 0x24d3, 0x1a2b } },
        { " 180 -  6464 Hz", { 0x2ae2, 0x0000, 0xd51f, 0x2927, 0x15c3 } },
        { " 202 -  5760 Hz", { 0x28a0, 0x0000, 0xd760, 0x2d96, 0x1141 } },
        { " 226 -  5120 Hz", { 0x2652, 0x0000, 0xd9af, 0x3220, 0x0ca3 } },
        { " 254 -  4576 Hz", { 0x2418, 0x0000, 0xdbe9, 0x367d, 0x082f } },
        { " 286 -  4064 Hz", { 0x21b7, 0x0000, 0xde49, 0x3b26, 0x036f } },
        { " 320 -  3616 Hz", { 0x1f59, 0x0000, 0xe0a7, 0x3fcd, 0xfeb3 } },
        { " 360 -  3232 Hz", { 0x1d03, 0x0000, 0xe2fd, 0x445e, 0xfa07 } },
        { " 404 -  2880 Hz", { 0x1a92, 0x0000, 0xe56e, 0x4927, 0xf525 } },
        { " 452 -  2560 Hz", { 0x1807, 0x0000, 0xe7fa, 0x4e25, 0xf00e } },
        { " 508 -  2288 Hz", { 0x157d, 0x0000, 0xea84, 0x531c, 0xeafa } },
        { " 572 -  2032 Hz", { 0x12b6, 0x0000, 0xed4a, 0x588a, 0xe56d } },
        { " 640 -  1808 Hz", { 0x0fdf, 0x0000, 0xf022, 0x5e1d, 0xdfbe } },
        { " 720 -  1616 Hz", { 0x0ce7, 0x0000, 0xf31a, 0x63e8, 0xd9ce } },
        { " 808 -  1440 Hz", { 0x09aa, 0x0000, 0xf657, 0x6a3e, 0xd354 } },
        { " 904 -  1280 Hz", { 0x061f, 0x0000, 0xf9e2, 0x7130, 0xcc3e } },
        { "1016 -  1144 Hz", { 0x0239, 0x0000, 0xfdc8, 0x78cc, 0xc472 } },
    };
}

void FilterSample::OnInitialize()
{
    // Initialize DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::Initialize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::LoadDefaultComponent());
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Initialize());

    m_CriticalSection.Initialize();
    s_CurrentVoice = 0;
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i] = NULL;
    }

    // Start sound thread
    m_SoundThreadFlag = true;
    m_SoundThread.StartUsingAutoStack(
        SoundThreadFunc,
        reinterpret_cast<uptr>(this),
        SOUND_THREAD_STACK_SIZE,
        SOUND_THREAD_PRIORITY
    );
}

void FilterSample::OnFinalize()
{
    // Destroy the sound thread
    m_SoundThreadFlag = false;
    m_SoundThread.Join();
    m_SoundThread.Finalize();

    m_CriticalSection.Finalize();

    // Shut down the DSP and SND libraries
    NN_UTIL_PANIC_IF_FAILED(nn::snd::Finalize());
    NN_UTIL_PANIC_IF_FAILED(nn::dsp::UnloadComponent());
    nn::dsp::Finalize();
}

void FilterSample::OnUpdate(nn::hid::PadStatus& padStatus)
{
    int voice = s_CurrentVoice;
    nn::snd::Voice* pVoice = s_pVoice[voice];

    if (padStatus.trigger & nn::hid::BUTTON_R)
    {
        s_MyFilterType = MY_FILTER_TYPE_NONE;
    }
    if (padStatus.trigger & nn::hid::BUTTON_A)
    {
        s_MyFilterType = MY_FILTER_TYPE_MONOPOLE;
    }
    if (padStatus.trigger & nn::hid::BUTTON_B)
    {
        s_MyFilterType = MY_FILTER_TYPE_BIQUAD_LOWPASS;
    }
    if (padStatus.trigger & nn::hid::BUTTON_X)
    {
        s_MyFilterType = MY_FILTER_TYPE_BIQUAD_HIGHPASS;
    }
    if (padStatus.trigger & nn::hid::BUTTON_Y)
    {
        s_MyFilterType = MY_FILTER_TYPE_BIQUAD_BANDPASS;
    }
    if (padStatus.trigger & nn::hid::BUTTON_UP)
    {
        s_FilterIndex = nn::math::Min(s_FilterIndex + 1, FILTER_NUM - 1);
    }
    if (padStatus.trigger & nn::hid::BUTTON_DOWN)
    {
        s_FilterIndex = nn::math::Max(s_FilterIndex - 1, 0);
    }

    if (padStatus.trigger)
    {
        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        switch (s_MyFilterType)
        {
        case MY_FILTER_TYPE_NONE:
            pVoice->SetFilterType(nn::snd::FILTER_TYPE_NONE);
            break;
        case MY_FILTER_TYPE_MONOPOLE:
            pVoice->SetMonoFilterCoefficients(monoFilter[s_FilterIndex].coef);
            pVoice->SetFilterType(nn::snd::FILTER_TYPE_MONOPOLE);
            break;
        case MY_FILTER_TYPE_BIQUAD_LOWPASS:
            pVoice->SetBiquadFilterCoefficients(biquadLowPassFilter[s_FilterIndex].coef);
            pVoice->SetFilterType(nn::snd::FILTER_TYPE_BIQUAD);
            break;
        case MY_FILTER_TYPE_BIQUAD_HIGHPASS:
            pVoice->SetBiquadFilterCoefficients(biquadHighPassFilter[s_FilterIndex].coef);
            pVoice->SetFilterType(nn::snd::FILTER_TYPE_BIQUAD);
            break;
        case MY_FILTER_TYPE_BIQUAD_BANDPASS:
            pVoice->SetBiquadFilterCoefficients(biquadBandPassFilter[s_FilterIndex].coef);
            pVoice->SetFilterType(nn::snd::FILTER_TYPE_BIQUAD);
            break;
        }
    }
}

void FilterSample::OnDraw(demo::RenderSystemDrawing& renderSystem)
{
    int line = 1;

    renderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    renderSystem.Clear();

    renderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    renderSystem.DrawText(10, 10 * line++, "CTR SampleDemos - snd/FilterSample");
    line++;

    renderSystem.DrawText(10, 10 * line++, "[R]          Use no filter");
    renderSystem.DrawText(10, 10 * line++, "[A]          Use monopole (lowpass)");
    renderSystem.DrawText(10, 10 * line++, "[B]          Use biquad (lowpass)");
    renderSystem.DrawText(10, 10 * line++, "[X]          Use biquad (highpass)");
    renderSystem.DrawText(10, 10 * line++, "[Y]          Use biquad (bandpass)");
    renderSystem.DrawText(10, 10 * line++, "[Up/Down]    Change filter index");
    line++;

    switch (s_MyFilterType)
    {
    case MY_FILTER_TYPE_NONE:
        renderSystem.DrawText(10, 10 * line++, "Current:     none");
        break;
    case MY_FILTER_TYPE_MONOPOLE:
        renderSystem.DrawText(10, 10 * line++, "Current:     monopole");
        renderSystem.DrawText(10, 10 * line++, "             ( lowpass, %s)",
                              monoFilter[s_FilterIndex].text);
        break;
    case MY_FILTER_TYPE_BIQUAD_LOWPASS:
        renderSystem.DrawText(10, 10 * line++, "Current:     biquad");
        renderSystem.DrawText(10, 10 * line++, "             ( lowpass, %s)",
                              biquadLowPassFilter[s_FilterIndex].text);
        break;
    case MY_FILTER_TYPE_BIQUAD_HIGHPASS:
        renderSystem.DrawText(10, 10 * line++, "Current:     biquad");
        renderSystem.DrawText(10, 10 * line++, "             (highpass, %s)",
                              biquadHighPassFilter[s_FilterIndex].text);
        break;
    case MY_FILTER_TYPE_BIQUAD_BANDPASS:
        renderSystem.DrawText(10, 10 * line++, "Current:     biquad");
        renderSystem.DrawText(10, 10 * line++, "             (bandpass, %s)",
                              biquadBandPassFilter[s_FilterIndex].text);
        break;
    }

    renderSystem.SwapBuffers();
    renderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

void FilterSample::SoundThreadFuncImpl()
{
    // Set sound output to stereo
    nn::snd::SetSoundOutputMode(nn::snd::OUTPUT_MODE_STEREO);

    // Set master volume
    nn::snd::SetMasterVolume(1.0);

    NN_LOG("Loading wav files...\n");

    u8* pMemory[FILE_NUM];
    nn::snd::WaveBuffer waveBuffer[FILE_NUM];
    // Open the file
    for (int i = 0; i < FILE_NUM; i++)
    {
        WaveFileReader reader(FILE_NAME[i]);
        int dataSize = reader.GetDataSize();
        int channelCount = reader.GetChannelCount();
        int sampleRate = reader.GetSampleRate();
        int quantumBits = reader.GetQuantumBits();

        // Get and check device memory
        pMemory[i] = reinterpret_cast<u8*>(AllocateMemory(dataSize, 32));
        NN_ASSERT(pMemory[i]);

        // Load wave data and invalidate cache
        reader.ReadData(pMemory[i], dataSize);
        nn::snd::FlushDataCache(reinterpret_cast<uptr>(pMemory[i]), dataSize);

        nn::snd::Voice* pVoice = nn::snd::AllocVoice(128, NULL, NULL);
        NN_ASSERT(pVoice);
        s_pVoice[i] = pVoice;

        nn::snd::SampleFormat format = (quantumBits == 8) ?
            nn::snd::SAMPLE_FORMAT_PCM8 : nn::snd::SAMPLE_FORMAT_PCM16;

        pVoice->SetChannelCount(channelCount);
        pVoice->SetSampleFormat(format);

        nn::snd::InitializeWaveBuffer(&waveBuffer[i]);
        waveBuffer[i].bufferAddress = pMemory[i];
        waveBuffer[i].sampleLength  = nn::snd::GetSampleLength(dataSize, format, channelCount);
        waveBuffer[i].loopFlag      = true;

        // Set volume
        nn::snd::MixParam mix;
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_LEFT ] = 0.707f; // Main volume (L)
        mix.mainBus[nn::snd::CHANNEL_INDEX_FRONT_RIGHT] = 0.707f; // Main volume (R)
        pVoice->SetMixParam(mix);
        pVoice->SetVolume(1.0f);

        // Set pitch
        pVoice->SetSampleRate(sampleRate);
        pVoice->SetPitch(1.0f);

        // Add buffer
        pVoice->AppendWaveBuffer(&waveBuffer[i]);
    }

    // Start playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        s_pVoice[i]->SetState(nn::snd::Voice::STATE_PLAY);
    }

    // The sound thread stops automatically during Sleep Mode
    // (It waits in nn::snd::WaitForDspSync).
    // However, if strict synchronization with graphics is required, the application must control it.
    // 
    while (m_SoundThreadFlag)
    {
        nn::snd::WaitForDspSync();      // Wait to receive data from DSP

        nn::os::CriticalSection::ScopedLock lock(m_CriticalSection);
        nn::snd::SendParameterToDsp();  // Send parameters to the DSP.
    }

    // End playback
    for (int i = 0; i < FILE_NUM; i++)
    {
        if (s_pVoice[i])
        {
            nn::snd::FreeVoice(s_pVoice[i]);
        }
        FreeMemory(pMemory[i]);
    }
}

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
