﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ReadIcon.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/tpl.h>
#include "demo.h"
#include "applet.h"

/* Please see man pages for details














*/

namespace
{
    demo::RenderSystemDrawing s_RenderSystem;           /* Framework for rendering */

    nn::fnd::ExpHeap s_AppHeap;                         /* Heap */
    uptr s_HeapForGx;
    const u32 s_GxHeapSize = 0x400000;
}

wchar_t* ctpkRomFilename0 = L"rom:/Icon48x48.ctpk"; /* Texture package file to load (upper screen) */
wchar_t* ctpkRomFilename1 = L"rom:/Icon24x24.ctpk"; /* Texture package file to load (lower screen) */

namespace
{
    const int GL_MEMORY_SIZE = 0x800000;
}

/* Variables for storing information about the loaded texture */
GLuint s_CPTK_TexID0 = 0;
u32 s_CTPKW0 = 0;
u32 s_CTPKH0 = 0;
u32 s_TexW0  = 0;
u32 s_TexH0  = 0;
GLuint s_CPTK_TexID1 = 0;
u32 s_CTPKW1 = 0;
u32 s_CTPKH1 = 0;
u32 s_TexW1  = 0;
u32 s_TexH1  = 0;

/* Function Prototypes */
void Initialize(void);
void DrawFrame(void);
void DrawDisplay0(void);
void DrawDisplay1(void);
void Finalize(void);

void LoadTexturePackage(void);
void DeleteTexture(void);
bool GetGLFormat(const nn::tpl::CTR::CtrTexFormat texformat, GLenum& internalFormat, GLenum& format, GLenum& type);
u8* GetTextureData(const wchar_t* ctpkFilename,
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight);
u8* AddPadding(const u32& iconWidth, const u32& iconHeight, const u32& textureWidth, const u32& textureHeight, u8* dataBuf);
u32 GetTextureLength(const u32& imageLength);



/* Please see man pages for details





*/
void Initialize(void)
{
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* Create heap */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(),
        nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(GL_MEMORY_SIZE));

    /* Allocates 8 MB memory in FCRAM for graphics */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    LoadTexturePackage();
}


/* Please see man pages for details





*/
void LoadTexturePackage()
{
    GLenum target = GL_TEXTURE_2D;
    GLenum internalFormat = GL_RGB_NATIVE_DMP;
    GLenum format = GL_RGB_NATIVE_DMP;
    GLenum type = GL_UNSIGNED_SHORT_5_6_5;


    /* Loads texture data from texture package (for upper screen) */
    u8* textureDataPtr = GetTextureData(ctpkRomFilename0,
                                        s_CTPKW0,
                                        s_CTPKH0,
                                        s_TexW0,
                                        s_TexH0);
    if(!textureDataPtr)
    {
        NN_LOG("LoadTexturePackage() GetTextureData TexPath = %s failed.\n", "Icon48x48.ctpk");
        NN_ASSERT(0);
    }
    NN_LOG("GenerateTexture() width = %d, height = %d \n", s_CTPKW0, s_CTPKH0);

    /* Generates a texture in the render framework. */
    GLuint textureId = 0;
    s_RenderSystem.GenerateTexture(target, internalFormat, s_TexW0, s_TexH0,
        format, type, textureDataPtr, textureId);

    if ( textureId != 0 )
    {
        s_CPTK_TexID0 = textureId;
        NN_LOG("  New Texture0Id = %d\n", textureId);
    }
    delete [] textureDataPtr;
    GLenum error = glGetError();
    if ( error != GL_NO_ERROR )
    {
        NN_LOG("LoadTexturePackage() failed.\n");
        NN_ASSERT(0);
    }

    /* Loads texture data from texture package (for upper screen) */
    u8* textureDataPtr1 = GetTextureData(ctpkRomFilename1,
                                        s_CTPKW1,
                                        s_CTPKH1,
                                        s_TexW1,
                                        s_TexH1);
    if(!textureDataPtr1)
    {
        NN_LOG("LoadTexturePackage() GetTextureData TexPath = %s failed.\n", "Icon24x24.ctpk");
        NN_ASSERT(0);
    }
    NN_LOG("GenerateTexture() width = %d, height = %d \n", s_CTPKW1, s_CTPKH1);

    /* Generates a texture in the render framework. */
    GLuint textureId1 = 0;
    s_RenderSystem.GenerateTexture(target, internalFormat, s_TexW1, s_TexH1,
        format, type, textureDataPtr1, textureId1);

    if ( textureId1 != 0 )
    {
        s_CPTK_TexID1 = textureId1;
        NN_LOG("  New Texture1Id = %d\n", textureId1);
    }
    delete [] textureDataPtr1;
    error = glGetError();
    if ( error != GL_NO_ERROR )
    {
        NN_LOG("LoadTexturePackage() failed.\n");
        NN_ASSERT(0);
    }

}

/* Please see man pages for details




*/
void DeleteTexture(void)
{
    if ( s_CPTK_TexID0 != 0 )
    {
        bool flag = s_RenderSystem.DeleteTexture(s_CPTK_TexID0);
        if ( flag )
        {
            NN_LOG("Successfully deleted texture(%d)\n", s_CPTK_TexID0);
            s_CPTK_TexID0 = 0;
        }
        else
        {
            NN_LOG("Failed to delete texture(%d)\n", s_CPTK_TexID0);
            NN_ASSERT(0);
        }
    }

    GLenum error = glGetError();
    if ( error != GL_NO_ERROR )
    {
        NN_LOG("DeleteTexture() failed.\n");
        NN_ASSERT(0);
    }

    if ( s_CPTK_TexID1 != 0 )
    {
        bool flag = s_RenderSystem.DeleteTexture(s_CPTK_TexID1);
        if ( flag )
        {
            NN_LOG("Successfully deleted texture(%d)\n", s_CPTK_TexID1);
            s_CPTK_TexID1 = 0;
        }
        else
        {
            NN_LOG("Failed to delete texture(%d)\n", s_CPTK_TexID1);
            NN_ASSERT(0);
        }
    }

    error = glGetError();
    if ( error != GL_NO_ERROR )
    {
        NN_LOG("DeleteTexture() failed.\n");
        NN_ASSERT(0);
    }
}

/* Please see man pages for details










*/
u8* GetTextureData(const wchar_t* ctpkRomFilename,
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight)
{
    // Loads package file from ROM.
    NN_LOG("\nReading data from ROMFS...\n");

    nn::fs::FileReader file(ctpkRomFilename);

    size_t fileSize = file.GetSize();
    NN_LOG("  fileSize = %d (Byte)\n", fileSize);
    if ( fileSize == 0 )
    {
        NN_LOG("Failed to open ctpkRomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    void* buf = s_AppHeap.Allocate(fileSize);

    s32 readSize = file.Read(buf, fileSize);
    NN_LOG("  readSize = %d (Byte)\n", readSize);
    if ( readSize == 0 )
    {
        NN_LOG("Failed to read RomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    // Gets the texture data.
    // The package file for the icon has no header and the actual data is directly stored.
    // 
    // The texture width and height are determined from the file size.
    u8* texDataBuffer = reinterpret_cast<u8*>(buf);
    if (fileSize == 48 * 48 * sizeof(u16))
    {
        bmpWidth  = 48;
        bmpHeight = 48;
    }
    else
    {
        bmpWidth  = 24;
        bmpHeight = 24;
    }
    NN_LOG("width = %d, height = %d, size = %d\n", bmpWidth, bmpHeight, fileSize);

    textureWidth = GetTextureLength(bmpWidth);
    textureHeight = GetTextureLength(bmpHeight);
    NN_LOG("TextureInfo Padding width = %d, height = %d\n", textureWidth, textureHeight);

    // Fill with padding (allocate memory).
    u8* textureDataBuffer = AddPadding(bmpWidth, bmpHeight, textureWidth, textureHeight, texDataBuffer);

    file.Finalize();
    s_AppHeap.Free(buf);

    return textureDataBuffer;
}

/* Please see man pages for details










*/
u8* AddPadding(const u32& iconWidth, const u32& iconHeight, const u32& textureWidth, const u32& textureHeight, u8* srcData)
{
    // The icon is RGB565, so the calculations always assume 2 bytes per texel.
    u8* textureDataBuffer = new GLubyte[(textureWidth * textureHeight * 2)];// 1 texel is 2 bytes
    u8* pDestBuf = textureDataBuffer;

    // Send data for each block
    const u32 blockSize = ((8 * 8) * 2);
    for(int v=0; v<iconHeight/8; v++)
    {
        for(int u=0; u<iconWidth/8; u++)
        {
            std::memcpy(pDestBuf, srcData, blockSize);
            pDestBuf += blockSize;
            srcData += blockSize;
        }
        // Insert block for gap in the U direction
        int numUGap = (textureWidth - iconWidth) / 8;
        std::memset(pDestBuf, 0, blockSize*numUGap);
        pDestBuf += blockSize*numUGap;
    }
    // Insert block for gap in the V direction
    int numVGap = (textureHeight - iconHeight) / 8;
    for(int v=0; v<numVGap; v++)
    {
        std::memset(pDestBuf, 0, blockSize * (textureWidth / 8));
        pDestBuf += blockSize * (textureWidth / 8);
    }

    return textureDataBuffer;
}

/* Please see man pages for details







*/
u32 GetTextureLength(const u32& imageLength)
{
    u32 textureLength = 8;

    /* 8, 16, 32, 64, 128, 256, 512, 1024 */
    for (u32 i = 0; i < 7; i++)
    {
        if ( imageLength > textureLength )
        {
            textureLength *= 2;
        }
        else
        {
            return textureLength;
        }
    }

    return 1024;
}

/* Please see man pages for details



*/
void DrawDisplay0(void)
{
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.0f, 1.0f, 0.0f, 1.0f);
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.Clear();

    f32 windowCoordinateX = 0.0f;
    f32 windowCoordinateY = 0.0f;
    f32 rectangleWidth = 400.0f;
    f32 rectangleHeight = 240.0f;
    s_RenderSystem.FillTexturedRectangle(s_CPTK_TexID0,
        windowCoordinateX, windowCoordinateY,
        rectangleWidth, rectangleHeight,
        s_TexW0, s_TexH0,
        s_TexW0, s_TexH0);

    s_RenderSystem.SwapBuffers();
}

/* Please see man pages for details



*/
void DrawDisplay1(void)
{
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.0f, 0.0f, 1.0f, 0.0f);
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.Clear();

    f32 windowCoordinateX = 0.0f;
    f32 windowCoordinateY = 0.0f;
    f32 rectangleWidth = 320.0f;
    f32 rectangleHeight = 240.0f;
    s_RenderSystem.FillTexturedRectangle(s_CPTK_TexID1,
        windowCoordinateX, windowCoordinateY,
        rectangleWidth, rectangleHeight,
        s_TexW1, s_TexH1,
        s_TexW1, s_TexH1 );

    s_RenderSystem.SwapBuffers();
}


/* Please see man pages for details



*/
void DrawFrame(void)
{
    DrawDisplay0();
    DrawDisplay1();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

/* Please see man pages for details




*/
void Finalize(void)
{
    DeleteTexture();
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));

    s_AppHeap.Finalize();
}

void StartDemo()
{
    /* Initialization */
    Initialize();

    // After this, the application itself handles sleep
    TransitionHandler::EnableSleep();

    while ( 1 )
    {
        /* Rendering */
        DrawFrame();

        // Run sleep, HOME Menu, and POWER Menu transitions.
        TransitionHandler::Process();
        // Determine whether there has been an exit notification.
        if (TransitionHandler::IsExitRequired())
        {
            break;
        }
    }

    /* End processing */
    Finalize();
}

/* Please see man pages for details




*/
void nnMain()
{
    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    // Finalize the application. The call to nn::applet::CloseApplication does not return.
    nn::applet::CloseApplication();
}
