﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ReadTexturePackage.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/tpl.h>
#include "demo.h"
#include "applet.h"

/* Please see man pages for details













*/

namespace
{
    demo::RenderSystemDrawing s_RenderSystem;           /* Framework for rendering */
    
    nn::fnd::ExpHeap s_AppHeap;                         /* Heap */
    uptr s_HeapForGx;
    const u32 s_GxHeapSize = 0x400000;
}

wchar_t* bmpRomFilename = L"rom:/ConvertInput.ctpk"; /* Texture package file to load */

namespace 
{
    const int GL_MEMORY_SIZE = 0x800000;
}

/* Variables for storing information about the loaded texture */
u32 s_TexDrawCount = 0;
u32 s_TexDrawIndex = 0;
GLuint s_CPTK_TexID0[9] = {0,0,0,0,0,0,0,0,0};
GLuint s_CPTK_TexID1[9] = {0,0,0,0,0,0,0,0,0};
u32 s_CTPKW0[9] = {0,0,0,0,0,0,0,0,0};
u32 s_CTPKH0[9] = {0,0,0,0,0,0,0,0,0};
u32 s_TexW0[9]  = {0,0,0,0,0,0,0,0,0};
u32 s_TexH0[9]  = {0,0,0,0,0,0,0,0,0};
u32 s_CTPKW1[9] = {0,0,0,0,0,0,0,0,0};
u32 s_CTPKH1[9] = {0,0,0,0,0,0,0,0,0};
u32 s_TexW1[9]  = {0,0,0,0,0,0,0,0,0};
u32 s_TexH1[9]  = {0,0,0,0,0,0,0,0,0};


/* Function Prototypes */
void Initialize(void);
void DrawFrame(void);
void DrawDisplay0(void);
void DrawDisplay1(void);
void Finalize(void);

void LoadTexturePackage(void);
void DeleteTexture(void);
bool GetGLFormat(const nn::tpl::CTR::CtrTexFormat texformat, GLenum& internalFormat, GLenum& format, GLenum& type);
u8* GetTextureData(const wchar_t* bmpRomFilename, 
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight,
                   const char* texpath, nn::tpl::CTR::CtrTextureInfo& texInfo,
                   int* mipLevel, unsigned int* mipmapSize);
u32 GetTextureLength(const u32& imageLength);



/* Please see man pages for details





*/
void Initialize(void)
{
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* Create heap */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), 
        nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(GL_MEMORY_SIZE));
       
    /* Allocates 8 MB memory in FCRAM for graphics */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);
    
    LoadTexturePackage();
}


/* Please see man pages for details





*/
void LoadTexturePackage()
{
    int mipLevel = 0;
    unsigned int mipSize[NN_TPL_MAX_MIPMAPLEVEL];
    nn::tpl::CTR::CtrTextureInfo texInfo;
    GLenum target = GL_TEXTURE_2D;
    GLenum internalFormat = GL_RGBA_NATIVE_DMP; 
    GLenum format = GL_RGBA_NATIVE_DMP;
    GLenum type = GL_UNSIGNED_SHORT_4_4_4_4;


    const char* texpath0[8] = { "data/rgba2rgba8888.tga", "data/rgba2rgba5551.tga", "data/rgba2rgba4444.tga", "data/rgba2rgb888.tga", 
                                "data/rgba2rgb565.tga",   "data/rgba2LA88.tga",     "data/rgba2L8.tga",
                                "data/rgba2L4.tga" };
    const char* texpath1[8] = { "data/rgb2rgba8888.tga",  "data/rgb2rgba5551.tga",  "data/rgb2rgba4444.tga",  "data/rgb2rgb888.tga", 
                                "data/rgb2rgb565.tga",    "data/rgb2LA88.tga",      "data/rgb2L8.tga",
                                "data/rgb2L4.tga" };

    for(int iCnt =0; iCnt < 8; iCnt++)
    {
        /* Loads texture data from texture package (for upper screen) */
        u8* textureDataPtr = GetTextureData(bmpRomFilename,
                                            s_CTPKW0[iCnt],
                                            s_CTPKH0[iCnt],
                                            s_TexW0[iCnt],
                                            s_TexH0[iCnt],
                                            texpath0[iCnt],
                                            texInfo,
                                            &mipLevel,
                                            mipSize);
        if(!textureDataPtr)
        {
            NN_LOG("LoadTexturePackage() GetTextureData TexNo = %d failed.\n", iCnt);
            NN_ASSERT(0);
        }
        /* Converts TexturePackager internal format to DMPGL format */
        GetGLFormat(texInfo.texFormat, internalFormat, format, type);

        /* Generates a texture in the render framework. */
        GLuint textureId = 0;
        s_RenderSystem.GenerateTexture(target, internalFormat, s_TexW0[iCnt], s_TexH0[iCnt],
            format, type, textureDataPtr, textureId);

        if ( textureId != 0 )
        {
            s_CPTK_TexID0[iCnt] = textureId;
            NN_LOG("  New Texture0Id = %d\n", textureId);
        }
        delete [] textureDataPtr;
        GLenum error = glGetError();
        if ( error != GL_NO_ERROR )
        {
            NN_LOG("LoadTexturePackage() failed.\n");
            NN_ASSERT(0);
        }

        /* Loads texture data from texture package (for lower screen) */
        textureDataPtr = GetTextureData(bmpRomFilename,
                                            s_CTPKW1[iCnt],
                                            s_CTPKH1[iCnt],
                                            s_TexW1[iCnt],
                                            s_TexH1[iCnt],
                                            texpath1[iCnt],
                                            texInfo,
                                            &mipLevel,
                                            mipSize);
        if(!textureDataPtr)
        {
            NN_LOG("LoadTexturePackage() GetTextureData TexNo = %d failed.\n", iCnt);
            NN_ASSERT(0);
        }
        /* Converts TexturePackager internal format to DMPGL format */
        GetGLFormat(texInfo.texFormat, internalFormat, format, type);

        /* Generates a texture in the render framework. */
        textureId = 0;
        s_RenderSystem.GenerateTexture(target, internalFormat, s_TexW1[iCnt], s_TexH1[iCnt],
            format, type, textureDataPtr, textureId);

        if ( textureId != 0 )
        {
            s_CPTK_TexID1[iCnt] = textureId;
            NN_LOG("  New Texture1Id = %d\n", textureId);
        }
        delete [] textureDataPtr;
        error = glGetError();
        if ( error != GL_NO_ERROR )
        {
            NN_LOG("LoadTexturePackage() failed.\n");
            NN_ASSERT(0);
        }
    }
    
}

/* Please see man pages for details




*/
void DeleteTexture(void)
{        
    for(int iCnt=0; iCnt < 8; iCnt++)
    {
        if ( s_CPTK_TexID0[iCnt] != 0 )
        {
            bool flag = s_RenderSystem.DeleteTexture(s_CPTK_TexID0[iCnt]);
            if ( flag )
            {
                NN_LOG("Successfully deleted texture(%d)\n", s_CPTK_TexID0[iCnt]);
                s_CPTK_TexID0[iCnt] = 0;
            }
            else
            {
                NN_LOG("Failed to delete texture(%d)\n", s_CPTK_TexID0[iCnt]);
                NN_ASSERT(0);
            }
        }

        GLenum error = glGetError();
        if ( error != GL_NO_ERROR )
        {
            NN_LOG("DeleteTexture() failed.\n");
            NN_ASSERT(0);
        }

        if ( s_CPTK_TexID1[iCnt] != 0 )
        {
            bool flag = s_RenderSystem.DeleteTexture(s_CPTK_TexID1[iCnt]);
            if ( flag )
            {
                NN_LOG("Successfully deleted texture(%d)\n", s_CPTK_TexID1[iCnt]);
                s_CPTK_TexID1[iCnt] = 0;
            }
            else
            {
                NN_LOG("Failed to delete texture(%d)\n", s_CPTK_TexID1[iCnt]);
                NN_ASSERT(0);
            }
        }

        error = glGetError();
        if ( error != GL_NO_ERROR )
        {
            NN_LOG("DeleteTexture() failed.\n");
            NN_ASSERT(0);
        }
    }
}

/* Please see man pages for details








*/
bool GetGLFormat(const nn::tpl::CTR::CtrTexFormat texformat, GLenum& internalFormat, GLenum& format, GLenum& type)
{
    switch(texformat)
    {
        case nn::tpl::CTR::TEXFMT_RGBA4444:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_4_4_4_4;
            break;
        case nn::tpl::CTR::TEXFMT_RGBA5551:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_5_5_5_1;
            break;
        case nn::tpl::CTR::TEXFMT_RGBA8888:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_RGB565:
            internalFormat = GL_RGB_NATIVE_DMP;
            format = GL_RGB_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_5_6_5;
            break;
        case nn::tpl::CTR::TEXFMT_RGB888:
            internalFormat = GL_RGB_NATIVE_DMP;
            format = GL_RGB_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_ETC1:
            internalFormat = GL_ETC1_RGB8_NATIVE_DMP;
            format = GL_ETC1_RGB8_NATIVE_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_ETC1_A4:
            internalFormat = GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP;
            format = GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_A8:
            internalFormat = GL_ALPHA_NATIVE_DMP;
            format = GL_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_A4:
            internalFormat = GL_ALPHA_NATIVE_DMP;
            format = GL_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_4BITS_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_L8:
            internalFormat = GL_LUMINANCE_NATIVE_DMP;
            format = GL_LUMINANCE_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_L4:
            internalFormat = GL_LUMINANCE_NATIVE_DMP;
            format = GL_LUMINANCE_NATIVE_DMP;
            type = GL_UNSIGNED_4BITS_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_LA88:
            internalFormat = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_LA44:
            internalFormat = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE_4_4_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_HL8:
            internalFormat = GL_HILO8_DMP_NATIVE_DMP;
            format = GL_HILO8_DMP_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_SHADOW:
            internalFormat = GL_SHADOW_DMP;
            format = GL_SHADOW_DMP;
            type = GL_UNSIGNED_INT;
            break;
        case nn::tpl::CTR::TEXFMT_GAS:
            internalFormat = GL_GAS_DMP;
            format = GL_GAS_DMP;
            type = GL_UNSIGNED_SHORT;
            break;
        case nn::tpl::CTR::TEXFMT_REF:
            internalFormat = GL_LUMINANCEF_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_FLOAT;
            break;
        default:
            return false;
    }
    return true;
}


/* Please see man pages for details













*/
u8* GetTextureData(const wchar_t* bmpRomFilename, 
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight, 
                   const char* texpath, nn::tpl::CTR::CtrTextureInfo& texInfo,
                   int* mipLevel, unsigned int* mipmapSize)
{   
    NN_LOG("\nReading data from ROMFS...\n");

    nn::fs::FileReader file(bmpRomFilename);
    
    size_t fileSize = file.GetSize();
    NN_LOG("  fileSize = %d (Byte)\n", fileSize);
    if ( fileSize == 0 )
    {
        NN_LOG("Failed to open bmpRomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    /* Allocate memory for loading */
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 readSize = file.Read(buf, fileSize);        
    NN_LOG("  readSize = %d (Byte)\n", readSize);
    if ( readSize == 0 )
    {
        NN_LOG("Failed to open RomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    /* check packagefile */
    if(!nn::tpl::CTR::CheckTexturePackageHeader(buf))
    {
        NN_LOG("This TexturePackageFile is invalid.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }

    /* get texture index */
    int texindex = nn::tpl::CTR::GetTextureIndex(buf, texpath);
    if(texindex == -1)
    {
        NN_LOG("Failed to get the texture's index.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }
    
    /* get texture */
    u8* texDataBuffer = NULL;
    texDataBuffer = (u8*)nn::tpl::CTR::GetTexture(mipLevel, mipmapSize, buf, texindex);
    if(!texDataBuffer)
    {
        NN_LOG("Failed to get the texture.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }

    /* get texture info */
    bool res = nn::tpl::CTR::GetTextureInfo(&texInfo, buf, texindex);
    if(!res)
    {
        NN_LOG("Failed to get texture information.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }

    NN_LOG("TextureInfo width = %d, height = %d, size = %d\n", texInfo.width, texInfo.height, texInfo.texDataSize);

    bmpWidth = texInfo.width;
    bmpHeight = texInfo.height;

    textureWidth = GetTextureLength(bmpWidth);
    textureHeight = GetTextureLength(bmpHeight);
    /* Allocate texture memory */
    u8* textureDataBuffer = new GLubyte[(texInfo.texDataSize)];
    std::memcpy(textureDataBuffer, texDataBuffer, texInfo.texDataSize);

    file.Finalize();
    s_AppHeap.Free(buf);

    return textureDataBuffer;
}

/* Please see man pages for details







*/
u32 GetTextureLength(const u32& imageLength)
{        
    u32 textureLength = 8;

    /* 8, 16, 32, 64, 128, 256, 512, 1024 */
    for (u32 i = 0; i < 7; i++)
    {
        if ( imageLength > textureLength )
        {
            textureLength *= 2;
        }
        else
        {
            return textureLength;
        }
    }

    return 1024;
}

/* Please see man pages for details



*/
void DrawDisplay0(void)
{
    s_TexDrawCount++;
    if(s_TexDrawCount > 150)
    {
        s_TexDrawCount = 0;
        s_TexDrawIndex++;
        if(s_TexDrawIndex > 7)
        {
            s_TexDrawIndex = 0;
        }
    }
    
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.0f, 1.0f, 0.0f, 1.0f);
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.Clear();

    f32 windowCoordinateX = 0.0f;
    f32 windowCoordinateY = 0.0f;
    f32 rectangleWidth = 400.0f;
    f32 rectangleHeight = 240.0f;
    s_RenderSystem.FillTexturedRectangle(s_CPTK_TexID0[s_TexDrawIndex],
        windowCoordinateX, windowCoordinateY,
        rectangleWidth, rectangleHeight,
        s_CTPKW0[s_TexDrawIndex], s_CTPKH0[s_TexDrawIndex],
        s_TexW0[s_TexDrawIndex], s_TexH0[s_TexDrawIndex]);

    s_RenderSystem.SwapBuffers();
}

/* Please see man pages for details



*/
void DrawDisplay1(void)
{
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.0f, 0.0f, 1.0f, 0.0f);
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.Clear();

    f32 windowCoordinateX = 0.0f;
    f32 windowCoordinateY = 0.0f;
    f32 rectangleWidth = 320.0f;
    f32 rectangleHeight = 240.0f;
    s_RenderSystem.FillTexturedRectangle(s_CPTK_TexID1[s_TexDrawIndex],
        windowCoordinateX, windowCoordinateY,
        rectangleWidth, rectangleHeight,
        s_CTPKW1[s_TexDrawIndex], s_CTPKH1[s_TexDrawIndex],
        s_TexW1[s_TexDrawIndex], s_TexH1[s_TexDrawIndex]);

    s_RenderSystem.SwapBuffers();
}

/* Please see man pages for details



*/
void DrawFrame(void)
{
    DrawDisplay0();    
    DrawDisplay1();    

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}

/* Please see man pages for details




*/
void Finalize(void)
{
    DeleteTexture();
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));

    s_AppHeap.Finalize();    
}

void StartDemo()
{
    /* Initialization */
    Initialize();

    // After this, the application itself handles sleep
    TransitionHandler::EnableSleep();

    while ( 1 )
    {
        /* Rendering */
        DrawFrame();

        // Run sleep, HOME Menu, and POWER Menu transitions.
        TransitionHandler::Process();
        // Determine whether there has been an exit notification.
        if (TransitionHandler::IsExitRequired())
        {
            break;
        }
    }

    /* End processing */
    Finalize();
}

/* Please see man pages for details




*/
void nnMain()
{
    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    // Finalize the application. The call to nn::applet::CloseApplication does not return.
    nn::applet::CloseApplication();
}
