﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ReadTexturePackageCompAndMipmap.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/gx.h>
#include <nn/math.h>
#include <nn/fs.h>
#include <nn/fnd.h>
#include <nn/tpl.h>
#include "demo.h"
#include "applet.h"

/* Please see man pages for details















*/

namespace {

/* Enumerated type */
typedef enum CombinerMode {
    COMB_MODE_ADD = 0,
    COMB_MODE_SUBTRACT,
    COMB_MODE_MULT,
    COMB_MODE_ADDMULT,
    COMB_MODE_DECAL,
    COMB_MODE_DECAL2,
    COMB_MODE_BLEND,
    COMB_MODE_TWOCOLOR_INTERPOLATE,

    COMB_MODE_NUM
} CombinerMode;

static char *g_CombinerModeName[] =
{
    "ADD",
    "SUBTRACT",
    "MULTIPLY",
    "ADD AND MULTIPLY",
    "DECAL",
    "DECAL2",
    "BLENDING",
    "INTERPOLATE BY TWO COLORS"
};

/* buffer id */
GLuint s_ArrayBufferID;
GLuint s_ElementArrayBufferID;

/* program id */
GLuint s_ProgramID;

/* shader id */
GLuint s_ShaderID;

/* Texture */
GLuint s_Texture[2];
GLuint s_TextureGray;
GLuint s_TextureAlpha;

/* Other variables */
int s_Count = 1;
int s_Mode = COMB_MODE_ADD;

GLfloat s_Intensity[4] = {0.0f, 0.0f, 0.0f, 0.0f};

nn::fnd::ExpHeap s_AppHeap;
uptr s_AddrForGxHeap;
const u32 s_GxHeapSize = 0x800000;

demo::RenderSystemDrawing s_RenderSystem;

/* For debugging */
GLfloat s_CamPosZ = 3.0f;
bool s_Plus = true;

} /* namespace */


/* Function Prototype Declarations */
void LoadObjects(void);
void ReadyObjects(void);
void DrawDisplay0(void);
void DrawDisplay1(void);
void SetTextureCombiner(void);
void DrawFrame(void);
int Initialize(void);


int GetImageSize(GLenum internalFormat, u32 TextureWidth, u32 TextureHeight);
bool GetGLFormat(const nn::tpl::CTR::CtrTexFormat texformat, GLenum& internalFormat, GLenum& format, GLenum& type);
u8* GetTextureData(const wchar_t* bmpRomFilename, 
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight,
                   const char* texpath, nn::tpl::CTR::CtrTextureInfo& texInfo,
                   int* mipLevel, unsigned int* mipmapSize);
u32 GetTextureLength(const u32& imageLength);


/* Please see man pages for details



*/
void LoadObjects(void)
{
    GLfloat coords[] = {
        -0.5f, 0.5f, 0.f, 1.f,
        -0.5f,-0.5f, 0.f, 1.f,
         0.5f, 0.5f, 0.f, 1.f,
         0.5f,-0.5f, 0.f, 1.f   
    };
    GLfloat texcoords[] = {
        0.f, 1.f, 0.f,
        0.f, 0.f, 0.f,
        1.f, 1.f, 0.f,
        1.f, 0.f, 0.f,
    };

    GLushort idxs[] = {0, 1, 2, 3};

    glGenBuffers(1, &s_ArrayBufferID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBufferData(GL_ARRAY_BUFFER, sizeof(coords) + sizeof(texcoords), 0, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(coords), coords);
    glBufferSubData(GL_ARRAY_BUFFER, sizeof(coords), sizeof(texcoords), texcoords);
    
    glGenBuffers(1, &s_ElementArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idxs), idxs, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(coords));
}

/* Please see man pages for details



*/
void ReadyObjects(void)
{
    glUseProgram(s_ProgramID);
    glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)(16*sizeof(GLfloat)));
}

/* Please see man pages for details






*/
void DrawDisplay0(void)
{
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();

    ReadyObjects();

    /* Texture combiner settings */
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Texture[0].samplerType"), GL_TEXTURE_2D);
    glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_Texture[1].samplerType"), GL_TEXTURE_2D);

    if (s_Mode == COMB_MODE_TWOCOLOR_INTERPOLATE)
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, s_TextureGray);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    else if (s_Mode == COMB_MODE_ADDMULT || s_Mode == COMB_MODE_DECAL)
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, s_Texture[0]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D , GL_TEXTURE_MIN_LOD, -5.0f);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, s_TextureAlpha);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    else
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, s_Texture[0]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D , GL_TEXTURE_MIN_LOD, -5.0f);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, s_Texture[1]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D , GL_TEXTURE_MIN_LOD, -5.0f);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
    SetTextureCombiner();

    nn::math::Matrix44 proj;

    nn::math::MTX44Frustum(&proj, -0.02f, 0.02f, -0.02f* static_cast<f32>(nn::gx::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::DISPLAY0_WIDTH),
                 0.02f* static_cast<f32>(nn::gx::DISPLAY0_HEIGHT) / static_cast<f32>(nn::gx::DISPLAY0_WIDTH), 0.2f, 30.f);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(proj));


    if(s_Plus){
        s_CamPosZ += 0.1f;
        if(s_CamPosZ >= 30.0f)
        {
            s_Plus = false;
        }
    }
    else
    {
        s_CamPosZ -= 0.1f;
        if(s_CamPosZ <= 3.0f)
        {
            s_Plus = true;
        }
    }

    nn::math::Matrix34 eye;
    nn::math::Vector3 camPos(0.f, 0.f, s_CamPosZ);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, 0.f);
    
    nn::math::MTX34LookAt(&eye, &camPos, &camUp, &target);
    nn::math::Matrix44 mv(eye);
    glUniformMatrix4fv(glGetUniformLocation(s_ProgramID, "uModelView"), 1, GL_TRUE, static_cast<f32*>(mv));

    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, 0);

    s_RenderSystem.SwapBuffers();
}

/* Please see man pages for details



*/
void DrawDisplay1(void)
{
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();
        
    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f);
    s_RenderSystem.DrawText(0.0f, 0.0f, "Count: %d", s_Count);
    s_RenderSystem.DrawText(0.0f, 8.0f, "MODE: %s", g_CombinerModeName[s_Mode]);
    
    s_RenderSystem.SwapBuffers();
}

/* Please see man pages for details



*/
void SetTextureCombiner(void)
{
    switch(s_Mode)
    {
    case COMB_MODE_ADD:
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_TEXTURE0, GL_TEXTURE1, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_ADD);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_SUBTRACT:
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_TEXTURE0, GL_TEXTURE1, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_SUBTRACT);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_MULT:
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_TEXTURE0, GL_TEXTURE1, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_MODULATE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_ADDMULT: /* Requires the texture data with the alpha component */
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE1, GL_TEXTURE1, GL_TEXTURE0);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_ALPHA, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineRgb"), GL_MULT_ADD_DMP);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].scaleAlpha"), 1.0f);

        /* Make sure to initialize the following texture combiner */
        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].constRgba"), 0.0f, 0.0f, 0.0f, 0.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleAlpha"), 1.0f);

        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].constRgba"), 0.0f, 0.0f, 0.0f, 0.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_REPLACE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_DECAL: /* Requires the texture data with the alpha component */
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcRgb"), GL_TEXTURE1, GL_TEXTURE0, GL_TEXTURE1);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].srcAlpha"), GL_TEXTURE0, GL_CONSTANT, GL_CONSTANT);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_ALPHA);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineRgb"), GL_INTERPOLATE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[0].scaleAlpha"), 1.0f);

        /* Make sure to initialize the following texture combiner */
        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].constRgba"), 0.0f, 0.0f, 0.0f, 0.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineRgb"), GL_REPLACE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleAlpha"), 1.0f);

        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].constRgba"), 0.0f, 0.0f, 0.0f, 0.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_PREVIOUS, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_REPLACE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_DECAL2:
        glUniform4fv(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].constRgba"), 1, s_Intensity);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcRgb"), GL_TEXTURE1, GL_CONSTANT, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandRgb"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_COLOR);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineRgb"), GL_MODULATE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleRgb"), 1.0f);

        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_TEXTURE1, GL_TEXTURE0, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_INTERPOLATE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_BLEND:
        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].constRgba"), 0.7f, 0.7f, 0.7f, 0.7f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_TEXTURE0, GL_TEXTURE1, GL_CONSTANT);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_TEXTURE1, GL_CONSTANT);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_INTERPOLATE);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_INTERPOLATE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    case COMB_MODE_TWOCOLOR_INTERPOLATE: /* Requires the grayscale texture */
        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].constRgba"), 1.0f, 1.0f, 0.0f, 1.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].srcRgb"), GL_CONSTANT, GL_TEXTURE0, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].operandRgb"), GL_SRC_COLOR, GL_SRC_COLOR, GL_SRC_COLOR);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].combineRgb"), GL_MODULATE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[1].scaleRgb"), 1.0f);

        glUniform4f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].constRgba"), 0.0f, 0.0f, 1.0f, 1.0f);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcRgb"), GL_CONSTANT, GL_TEXTURE0, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].srcAlpha"), GL_TEXTURE0, GL_PREVIOUS, GL_PREVIOUS);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandRgb"), GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR, GL_SRC_COLOR);
        glUniform3i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].operandAlpha"), GL_SRC_ALPHA, GL_SRC_ALPHA, GL_SRC_ALPHA);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineRgb"), GL_MULT_ADD_DMP);
        glUniform1i(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].combineAlpha"), GL_REPLACE);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleRgb"), 1.0f);
        glUniform1f(glGetUniformLocation(s_ProgramID, "dmp_TexEnv[2].scaleAlpha"), 1.0f);
        break;
    default:
        break;
    }
}

/* Please see man pages for details



*/
void DrawFrame(void)
{
    DrawDisplay0();
    DrawDisplay1();

    s_Count++;
    /* Switches texture combiner setting using s_Count value */
    if (s_Count % 100 == 0)
    {
        s_Mode++;
        if (s_Mode >= COMB_MODE_NUM)
        {
            s_Mode = COMB_MODE_ADD;
        }
    }

    if (s_Mode == COMB_MODE_DECAL2)
    {
        if (s_Count % 10 == 0)
        {
            for (int i=0; i<4; i++)
            {
                s_Intensity[i] += 0.1f;
                if (s_Intensity[i] > 1.0f)
                {
                    s_Intensity[i] = 0.0f;
                }
            }
        }
    }
    
    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);
}


/* Please see man pages for details





*/
/* initialization */
int Initialize(void)
{
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    /* Initialize display */
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_AddrForGxHeap = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    s_RenderSystem.Initialize(s_AddrForGxHeap, s_GxHeapSize);

    s_ProgramID = glCreateProgram();
    s_ShaderID = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShaderID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_ProgramID, s_ShaderID);
    glAttachShader(s_ProgramID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_ProgramID, 0, "aPosition");
    glBindAttribLocation(s_ProgramID, 1, "aTexCoord");

    glLinkProgram(s_ProgramID);
    glValidateProgram(s_ProgramID);
    glUseProgram(s_ProgramID);

    int mipLevel0 = 0; int mipLevel1 = 0; int mipLevelGray = 0; int mipLevelAlpha = 0;
    unsigned int mipSize0[NN_TPL_MAX_MIPMAPLEVEL];
    unsigned int mipSize1[NN_TPL_MAX_MIPMAPLEVEL];
    unsigned int mipSizeGray[NN_TPL_MAX_MIPMAPLEVEL];
    unsigned int mipSizeAlpha[NN_TPL_MAX_MIPMAPLEVEL];
    nn::tpl::CTR::CtrTextureInfo texInfo1; nn::tpl::CTR::CtrTextureInfo texInfo2; nn::tpl::CTR::CtrTextureInfo texInfoGray; nn::tpl::CTR::CtrTextureInfo texInfoAlpha;
    u32 s_CTPK0Width = 0; u32 s_CTPK0Height = 0; u32 s_Texture0Width = 0; u32 s_Texture0Height = 0;
    u32 s_CTPK1Width = 0; u32 s_CTPK1Height = 0; u32 s_Texture1Width = 0; u32 s_Texture1Height = 0;
    u32 s_CTPK2Width = 0; u32 s_CTPK2Height = 0; u32 s_Texture2Width = 0; u32 s_Texture2Height = 0;
    u32 s_CTPK3Width = 0; u32 s_CTPK3Height = 0; u32 s_Texture3Width = 0; u32 s_Texture3Height = 0;
    wchar_t* bmpRomFilename = L"rom:/ConvertInput.ctpk";
    const char* texpath = "data/rgba2rgba4444.tga";
    u8* textureDataPtr_U = GetTextureData(bmpRomFilename,
                                        s_CTPK0Width,
                                        s_CTPK0Height,
                                        s_Texture0Width,
                                        s_Texture0Height,
                                        texpath,
                                        texInfo1,
                                        &mipLevel0,
                                        mipSize0);
    const char* texpathD = "data/rgba2ETC1.tga";
    u8* textureDataPtr_D = GetTextureData(bmpRomFilename,
                                        s_CTPK1Width,
                                        s_CTPK1Height,
                                        s_Texture1Width,
                                        s_Texture1Height,
                                        texpathD,
                                        texInfo2,
                                        &mipLevel1,
                                        mipSize1);

    const char* texpathG = "data/rgba2L4.tga";
    u8* textureDataPtr_G = GetTextureData(bmpRomFilename,
                                        s_CTPK2Width,
                                        s_CTPK2Height,
                                        s_Texture2Width,
                                        s_Texture2Height,
                                        texpathG,
                                        texInfoGray,
                                        &mipLevelGray,
                                        mipSizeGray);

    const char* texpathA = "data/rgba2ETC1_a4.tga";
    u8* textureDataPtr_A = GetTextureData(bmpRomFilename,
                                        s_CTPK3Width,
                                        s_CTPK3Height,
                                        s_Texture3Width,
                                        s_Texture3Height,
                                        texpathA,
                                        texInfoAlpha,
                                        &mipLevelAlpha,
                                        mipSizeAlpha);

    GLenum internalFormat = GL_RGBA_NATIVE_DMP; 
    GLenum format = GL_RGBA_NATIVE_DMP;
    GLenum type = GL_UNSIGNED_SHORT_4_4_4_4;

    /* Load Texture */
    glGenTextures(2, s_Texture);
    glGenTextures(1, &s_TextureGray);
    glGenTextures(1, &s_TextureAlpha);

    /* Converts TexturePackager internal format to GL format */
    GetGLFormat(texInfo1.texFormat, internalFormat, format, type);

    glBindTexture(GL_TEXTURE_2D, s_Texture[0]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, mipLevel0 * -1);
    if( (internalFormat == GL_ETC1_RGB8_NATIVE_DMP) ||
        (internalFormat == GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP))
    {
        unsigned int mipmaptotal = 0;
        for(int i=0; i<mipLevel0; i++)
        {
            mipmaptotal += GetImageSize(internalFormat, s_Texture0Width >> i, s_Texture0Height >> i);
        }
        glCompressedTexImage2D(GL_TEXTURE_2D, mipLevel0 * -1, internalFormat, s_Texture0Width, s_Texture0Height, 0, mipmaptotal, textureDataPtr_U);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glCompressedTexImage2D() failed.\n");
        }
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D, mipLevel0 * -1, internalFormat, s_Texture0Width, s_Texture0Height, 0, format, type, textureDataPtr_U);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glTexImage2D() failed.\n");
        }
    }


    /* Converts TexturePackager internal format to GL format */
    GetGLFormat(texInfo2.texFormat, internalFormat, format, type);

    glBindTexture(GL_TEXTURE_2D, s_Texture[1]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, mipLevel1 * -1);
    if( (internalFormat == GL_ETC1_RGB8_NATIVE_DMP) ||
        (internalFormat == GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP))
    {
        unsigned int mipmaptotal = 0;
        for(int i=0; i<mipLevel1; i++)
        {
            mipmaptotal += GetImageSize(internalFormat, s_Texture0Width >> i, s_Texture0Height >> i);
        }
        glCompressedTexImage2D(GL_TEXTURE_2D, mipLevel1 * -1, internalFormat, s_Texture1Width, s_Texture1Height, 0, mipmaptotal, textureDataPtr_D);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glCompressedTexImage2D() failed.\n");
        }
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D, mipLevel1 * -1, internalFormat, s_Texture1Width, s_Texture1Height, 0, format, type, textureDataPtr_D);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glTexImage2D() failed.\n");
        }
    }



    GetGLFormat(texInfoGray.texFormat, internalFormat, format, type);
    glBindTexture(GL_TEXTURE_2D, s_TextureGray);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, mipLevelGray * -1);
    if( (internalFormat == GL_ETC1_RGB8_NATIVE_DMP) ||
        (internalFormat == GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP))
    {
        unsigned int mipmaptotal = 0;
        for(int i=0; i<mipLevelGray; i++)
        {
            mipmaptotal += GetImageSize(internalFormat, s_Texture2Width >> i, s_Texture2Height >> i);
        }
        glCompressedTexImage2D(GL_TEXTURE_2D, mipLevelGray * -1, internalFormat, s_Texture2Width, s_Texture2Height, 0, mipmaptotal, textureDataPtr_G);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glCompressedTexImage2D() failed.\n");
        }
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D, mipLevelGray * -1, internalFormat, s_Texture2Width, s_Texture2Height, 0, format, type, textureDataPtr_G);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glTexImage2D() failed.\n");
        }
    }

    GetGLFormat(texInfoAlpha.texFormat, internalFormat, format, type);
    glBindTexture(GL_TEXTURE_2D, s_TextureAlpha);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_LOD, mipLevelAlpha * -1);
    if( (internalFormat == GL_ETC1_RGB8_NATIVE_DMP) ||
        (internalFormat == GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP))
    {
        unsigned int mipmaptotal = 0;
        for(int i=0; i<mipLevelAlpha; i++)
        {
            mipmaptotal += GetImageSize(internalFormat, s_Texture3Width >> i, s_Texture3Height >> i);
        }
        glCompressedTexImage2D(GL_TEXTURE_2D, mipLevelAlpha * -1, internalFormat, s_Texture3Width, s_Texture3Height, 0, mipmaptotal, textureDataPtr_A);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glCompressedTexImage2D() failed.\n");
        }
    }
    else
    {
        glTexImage2D(GL_TEXTURE_2D, mipLevelAlpha * -1, internalFormat, s_Texture3Width, s_Texture3Height, 0, format, type, textureDataPtr_A);
        GLenum error_enum = glGetError();
        if ( error_enum != GL_NO_ERROR )
        {    
            NN_TPANIC_("glTexImage2D() Alpha failed.\n");
        }
    }

    glBindTexture(GL_TEXTURE_2D, 0);

    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.36f, 0.42f, 0.5f, 1.0f);

    glClearDepthf(1.f);

    glEnable(GL_CULL_FACE);
    glFrontFace(GL_CCW);
    glCullFace(GL_BACK);

    LoadObjects();

    return 0;
}


void StartDemo()
{
    /* initialization */
    if (Initialize() >= 0)
    {
        // After this, the application itself handles sleep
        TransitionHandler::EnableSleep();

        /* Enter loop */
        while (1)
        {
            DrawFrame();

            // Run sleep, HOME Menu, and POWER Menu transitions.
            TransitionHandler::Process();
            // Determine whether there has been an exit notification.
            if (TransitionHandler::IsExitRequired())
            {
                break;
            }
        }
    }

    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_AddrForGxHeap));
    s_AppHeap.Finalize();
}

/* Please see man pages for details





*/
void nnMain()
{
    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    // Finalize the application. The call to nn::applet::CloseApplication does not return.
    nn::applet::CloseApplication();
}

/* Please see man pages for details









*/
int GetImageSize(GLenum internalFormat, u32 TextureWidth, u32 TextureHeight)
{
    int blocksize = 0;
    if(internalFormat == GL_ETC1_RGB8_NATIVE_DMP)
    {
        blocksize = 8;
    }
    else if(internalFormat == GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP)
    {
        blocksize = 16;
    }

    return TextureWidth/4 * TextureHeight/4 * blocksize;
}

/* Please see man pages for details








*/
bool GetGLFormat(const nn::tpl::CTR::CtrTexFormat texformat, GLenum& internalFormat, GLenum& format, GLenum& type)
{
    switch(texformat)
    {
        case nn::tpl::CTR::TEXFMT_RGBA4444:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_4_4_4_4;
            break;
        case nn::tpl::CTR::TEXFMT_RGBA5551:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_5_5_5_1;
            break;
        case nn::tpl::CTR::TEXFMT_RGBA8888:
            internalFormat = GL_RGBA_NATIVE_DMP;
            format = GL_RGBA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_RGB565:
            internalFormat = GL_RGB_NATIVE_DMP;
            format = GL_RGB_NATIVE_DMP;
            type = GL_UNSIGNED_SHORT_5_6_5;
            break;
        case nn::tpl::CTR::TEXFMT_RGB888:
            internalFormat = GL_RGB_NATIVE_DMP;
            format = GL_RGB_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_ETC1:
            internalFormat = GL_ETC1_RGB8_NATIVE_DMP;
            format = GL_ETC1_RGB8_NATIVE_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_ETC1_A4:
            internalFormat = GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP;
            format = GL_ETC1_ALPHA_RGB8_A4_NATIVE_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_A8:
            internalFormat = GL_ALPHA_NATIVE_DMP;
            format = GL_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_A4:
            internalFormat = GL_ALPHA_NATIVE_DMP;
            format = GL_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_4BITS_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_L8:
            internalFormat = GL_LUMINANCE_NATIVE_DMP;
            format = GL_LUMINANCE_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_L4:
            internalFormat = GL_LUMINANCE_NATIVE_DMP;
            format = GL_LUMINANCE_NATIVE_DMP;
            type = GL_UNSIGNED_4BITS_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_LA88:
            internalFormat = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_LA44:
            internalFormat = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE_4_4_DMP;
            break;
        case nn::tpl::CTR::TEXFMT_HL8:
            internalFormat = GL_HILO8_DMP_NATIVE_DMP;
            format = GL_HILO8_DMP_NATIVE_DMP;
            type = GL_UNSIGNED_BYTE;
            break;
        case nn::tpl::CTR::TEXFMT_SHADOW:
            internalFormat = GL_SHADOW_DMP;
            format = GL_SHADOW_DMP;
            type = GL_UNSIGNED_INT;
            break;
        case nn::tpl::CTR::TEXFMT_GAS:
            internalFormat = GL_GAS_DMP;
            format = GL_GAS_DMP;
            type = GL_UNSIGNED_SHORT;
            break;
        case nn::tpl::CTR::TEXFMT_REF:
            internalFormat = GL_LUMINANCEF_DMP;
            format = GL_LUMINANCE_ALPHA_NATIVE_DMP;
            type = GL_FLOAT;
            break;
        default:
            return false;
    }
    return true;
}

/* Please see man pages for details













*/
u8* GetTextureData(const wchar_t* bmpRomFilename, 
                   u32& bmpWidth, u32& bmpHeight,
                   u32& textureWidth, u32& textureHeight, 
                   const char* texpath, nn::tpl::CTR::CtrTextureInfo& texInfo,
                   int* mipLevel, unsigned int* mipmapSize)
{   
    NN_LOG("\nReading data from ROMFS...\n");

    nn::fs::FileReader file(bmpRomFilename);
    
    size_t fileSize = file.GetSize();
    NN_LOG("  fileSize = %d (Byte)\n", fileSize);
    if ( fileSize == 0 )
    {
        NN_LOG("Failed to open bmpRomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    void* buf = s_AppHeap.Allocate(fileSize);

    s32 readSize = file.Read(buf, fileSize);        
    NN_LOG("  readSize = %d (Byte)\n", readSize);
    if ( readSize == 0 )
    {
        NN_LOG("Failed to open RomFilename.\n");
        NN_ASSERT(0);
        return NULL;
    }

    /* get texture index */
    int texindex = nn::tpl::CTR::GetTextureIndex(buf, texpath);
    if(texindex == -1)
    {
        NN_LOG("Failed to get the texture's index.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }
    
    /* get texture */
    u8* texDataBuffer = NULL;
    texDataBuffer = (u8*)nn::tpl::CTR::GetTexture(mipLevel, mipmapSize, buf, texindex);
    if(!texDataBuffer)
    {
        NN_LOG("Failed to get the texture.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }

    /* get texture info */
    bool res = nn::tpl::CTR::GetTextureInfo(&texInfo, buf, texindex);
    if(!res)
    {
        NN_LOG("Failed to get texture information.\n");
        file.Finalize();
        s_AppHeap.Free(buf);
        return 0;
    }

    NN_LOG("TextureInfo width = %d, height = %d, size = %d\n", texInfo.width, texInfo.height, texInfo.texDataSize);

    bmpWidth = texInfo.width;
    bmpHeight = texInfo.height;

    textureWidth = GetTextureLength(bmpWidth);
    textureHeight = GetTextureLength(bmpHeight);
    u8* textureDataBuffer = new GLubyte[(texInfo.texDataSize)];
    std::memcpy(textureDataBuffer, texDataBuffer, texInfo.texDataSize);

    file.Finalize();
    s_AppHeap.Free(buf);

    return textureDataBuffer;
}

/* Please see man pages for details







*/
u32 GetTextureLength(const u32& imageLength)
{        
    u32 textureLength = 8;

    /* 8, 16, 32, 64, 128, 256, 512, 1024 */
    for (u32 i = 0; i < 7; i++)
    {
        if ( imageLength > textureLength )
        {
            textureLength *= 2;
        }
        else
        {
            return textureLength;
        }
    }

    return 1024;
}
