﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     uds.h

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#ifndef NN_SAMPLE_DEMOS_DEMO1_UDS_H_
#define NN_SAMPLE_DEMOS_DEMO1_UDS_H_

#include <nn.h>
#include <nn/uds.h>

enum ComMode
{
    NOT_INITIALIZED, // Uninitialized state
    INITIALIZING,    // Currently being initialized
    DISCONNECTED,    // Initialized and unconnected state
    SCANNING,        // Scanning
    MASTER,          // Communicating as master
    CLIENT,          // Communicating as Client
    SPECTATOR,       // Connected as Spectator
    ERROR            // Entered sleep or Wi-Fi OFF mode
};

struct ConnectParam
{
    int     targetNum;
    ComMode mode;
    NN_PADDING3;
};

struct ComResult
{
    u32 receivedTime;       //
    f32 troughput;          //
    u64 receivedSize;       //
    u32 errorNum;           //
    NN_PADDING4;
};

struct NodeData
{
    u16     nodeId;     //
    NN_PADDING2;
    char    name[11];   //
    NN_PADDING1;


    u32     startCount;     //
    u32     latestCount;    //
    u64     receivedSize;   //
    u32     errorCount;     //
    NN_PADDING4;

    nn::os::Tick startTick; //
    
    ComResult   totalResult;    //
    ComResult   lapResult;      //
};

const u8 SUB_ID = 0x01;
const int PACKET_SIZE = nn::uds::UDS_PACKET_PAYLOAD_MAX_SIZE;
const int SEND_INTERVAL = 10;
const int SCAN_MAX = 5;
const char PASSPHRASE[] = "CTR_SDK_SAMPLE";

const s32 CONNECTION_MONITOR_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY;
const s32 RESULT_MONITOR_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY-1;
const s32 SEND_THREAD_PRIORITY = nn::os::DEFAULT_THREAD_PRIORITY -1;
const s32 RECEIVE_THREAD_PRIORITY = SEND_THREAD_PRIORITY -1;                    //Give the receiving thread the maximum priority

void UdsInitializeThread(uptr param);               //Initialization thread
void UdsScanThread(uptr param);                     //Scan thread
void ConnectionMonitorThread(ConnectParam param);   //Connection management thread

//Get rendering data
void GetResult(u16* pMyNodeId, int* pSendError, NodeData nodeData[nn::uds::NODE_MAX]);
void GetNetDescList(int* pNum, nn::uds::NetworkDescription netDesc[SCAN_MAX], nn::uds::NodeInformation masterInfo[SCAN_MAX] );


/* NN_SAMPLE_DEMOS_DEMO1_UDS_H_ */
#endif

/*---------------------------------------------------------------------------*
  End of file
 *---------------------------------------------------------------------------*/
