﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ngc_Main.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 46365 $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/cfg.h>
#include <nn/ngc.h>
#include <nn/applet.h>

extern "C" void nnMain()
{
    // Call only nn::applet::Enable to also allow execution from the HOME Menu
    // HOME Menu transitions, POWER Button presses, and sleep are all unsupported
    nn::applet::Enable();

    nn::fs::Initialize();       // Initializes the libaries needed to use the NGC library
    nn::cfg::Initialize();      // Initializes the libaries needed to use the NGC library
    
    nn::ngc::ProfanityFilter filter;
    nn::Result result = filter.Initialize();
    NN_UNUSED_VAR( result );
    NN_ASSERT_RESULT( result );
    
    // Gets the profanity word pattern version information that was installed on the system.
    u32 nVersion = filter.GetContentVersion();
    NN_ASSERT( nVersion != 0 );
    NN_LOG( "Pattern version = %d\n", nVersion );
    
    // Access the system's current region and language settings to check for profanity, under the assumption that text data can move between regions.
    // 
    const wchar_t* ppWords[] =
    {
        L"mario",
        L"luigi",
        L"badword",
        L"クッパ",
    };
    bit32 bResults[ 4 ];
    
    result = filter.CheckProfanityWords( bResults, true, ppWords, 4 );
    NN_ASSERT_RESULT( result );
    
    for ( int i = 0; i < 4; i++ )
    {
        if ( bResults[ i ] == 0 )
        {
            NN_LOG("id[%d] is not a profanity word.\n", i);
        }
        else
        {
            NN_LOG("id[%d] is a profanity word.\n", i);
        }
    }
    
    nn::cfg::Finalize();
    NN_LOG("*** End of demo ***\n");
    while(true){}
}
