﻿/*---------------------------------------------------------------------------*
  Project:  Horizon
  File:     ulcd_Simple.cpp

  Copyright (C)2009-2012 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 47228 $
 *---------------------------------------------------------------------------*/

#include <nn.h>

#include <nn/fnd.h>
#include <nn/gx.h>
#include <nn/fs.h>
#include <nn/hid.h>
#include <nn/ulcd.h>

#include "demo.h"
#include "applet.h"

namespace
{
    const u32 s_GxHeapSize = 0x800000;
    const s32 MODE_NUM       = 2;

    GLuint s_ArrayBufferID;
    GLuint s_ElementArrayBufferID;

    /* program id */
    GLuint s_PgID;

    /* shader id */
    GLuint s_ShID;

    nn::ulcd::CTR::StereoCamera s_StereoCamera;

    demo::RenderSystemExt s_RenderSystem;
    nn::fnd::ExpHeap s_AppHeap;
    uptr s_HeapForGx;

    int  s_CurrentMode   = 0;
    f32  s_DepthLevel    = 10.0f;
    f32  s_Factor    =  1.0f;

    int  s_Count = 0;
}

static void LoadObjects(void);
static void DrawDisplay0(nn::math::Matrix44* proj, nn::math::Matrix34* view);
static void DrawDisplay1(void);
static int DrawFrame(void);
static void PadReadAndEvent(nn::hid::PadReader& padReader);
static int Initialize(void);

static void LoadObjects(void)
{
    static bool isLoadOnce = false;

    GLfloat coords[] = {
         0.5f, -0.5f,  0.5f,   0.5f, -0.5f, -0.5f,   0.5f,  0.5f, -0.5f,   0.5f,  0.5f,  0.5f, // +X
        -0.5f, -0.5f, -0.5f,  -0.5f, -0.5f,  0.5f,  -0.5f,  0.5f,  0.5f,  -0.5f,  0.5f, -0.5f, // -X
        -0.5f,  0.5f,  0.5f,   0.5f,  0.5f,  0.5f,   0.5f,  0.5f, -0.5f,  -0.5f,  0.5f, -0.5f, // +Y
        -0.5f, -0.5f, -0.5f,   0.5f, -0.5f, -0.5f,   0.5f, -0.5f,  0.5f,  -0.5f, -0.5f,  0.5f, // -Y
        -0.5f, -0.5f,  0.5f,   0.5f, -0.5f,  0.5f,   0.5f,  0.5f,  0.5f,  -0.5f,  0.5f,  0.5f, // +Z
         0.5f, -0.5f, -0.5f,  -0.5f, -0.5f, -0.5f,  -0.5f,  0.5f, -0.5f,   0.5f,  0.5f, -0.5f  // -Z
    };
    GLfloat color[] = {
        1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f,
        0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f,
        0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f,
        1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f, 1.f, 0.0f, 0.0f,
        0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f, 0.f, 1.0f, 0.0f,
        0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f, 0.f, 0.0f, 1.0f
    };

    GLushort idxs[] =
    {
         0,  1,  2,  0,  2,  3,
         4,  5,  6,  4,  6,  7,
         8,  9, 10,  8, 10, 11,
        12, 13, 14, 12, 14, 15,
        16, 17, 18, 16, 18, 19,
        20, 21, 22, 20, 22, 23
    };

    if (isLoadOnce)
    {
        glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
    }
    else
    {
        glGenBuffers(1, &s_ArrayBufferID);
        glBindBuffer(GL_ARRAY_BUFFER, s_ArrayBufferID);
        glBufferData(GL_ARRAY_BUFFER, sizeof(coords) + sizeof(color), 0, GL_STATIC_DRAW);
        glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(coords), coords);
        glBufferSubData(GL_ARRAY_BUFFER, sizeof(coords), sizeof(color), color);

        glGenBuffers(1, &s_ElementArrayBufferID);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_ElementArrayBufferID);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idxs), idxs, GL_STATIC_DRAW);
        isLoadOnce = true;
    }

    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0) ;
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (GLvoid*)sizeof(coords));
}

static void DrawDisplay0(nn::math::Matrix44* proj, nn::math::Matrix34* view)
{
    glUseProgram(s_PgID);
    LoadObjects();

    glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uProjection"), 1, GL_TRUE, static_cast<f32*>(*proj));

    // Render a triangle that has a different position
    nn::math::Matrix34 tmp, rot;
    nn::math::Vector3 pos[3];
    pos[0] = nn::math::Vector3(0.0f, 0.0f, 0.0f);
    pos[1] = nn::math::Vector3(0.7f, 0.5f, -5.0f);
    pos[2] = nn::math::Vector3(-1.2f, -0.9f, -10.0f);

    nn::math::Vector3 axis(0.0f, 1.0f, 0.2f);

    for (int i=0; i<3; ++i)
    {
        nn::math::MTX34MultTranslate(&tmp, view, &pos[i]);

        nn::math::MTX34RotAxisDeg(&rot, &axis, -3.f * s_Count * (i+1));

        nn::math::MTX34Mult(&tmp, &tmp, &rot);

        nn::math::Matrix44 mv(tmp);
        glUniformMatrix4fv(glGetUniformLocation(s_PgID, "uModelView"), 1, GL_TRUE, (f32*)(mv));

        glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_SHORT, 0);

        axis.x = axis.x + 0.5f;
    }
}

static void DrawDisplay1(void)
{
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY1);
    s_RenderSystem.Clear();

    // Render cursor
    s_RenderSystem.SetColor(0.0f, 1.0f, 0.0f, 1.0f);
    s_RenderSystem.DrawText(4,  s_CurrentMode * 8, "*");

    s_RenderSystem.SetColor(1.0f, 1.0f, 1.0f, 1.0f);
    s_RenderSystem.DrawText(16,  0, "Depth Level: %.2f", s_DepthLevel);
    s_RenderSystem.DrawText(16,  8, "Factor     : %.3f", s_Factor);

    // Parallax on the base plane needed to display objects at any distance from the game camera
    s_RenderSystem.DrawText(16, 40, "dist(from CAM) 10.0f: %f", s_StereoCamera.GetParallax(10.0f));
    s_RenderSystem.DrawText(16, 48, "dist(from CAM) 15.0f: %f", s_StereoCamera.GetParallax(15.0f));

    // Parallax on the base plane needed to display objects infinitely far away (the maximum possible depth)
    s_RenderSystem.DrawText(16, 56, "max: %f", s_StereoCamera.GetMaxParallax());

    // Send the lower LCD image to the display buffer
    s_RenderSystem.Transfer();

    s_RenderSystem.SwapBuffers();
}

static int DrawFrame(void)
{
    nn::math::Matrix34 view;
    nn::math::Vector3 camPos(0.f, 0.4f, 10.0f);
    nn::math::Vector3 camUp(0.f, 1.f, 0.f);
    nn::math::Vector3 target(0.f, 0.f, camPos.z - s_DepthLevel);

    nn::math::MTX34LookAt(&view, &camPos, &camUp, &target);

    // Calculate the view matrix and projection matrix for rendering the image that corresponds to the left and right parallax
    nn::math::Matrix44 projL, projR;
    nn::math::Matrix34 viewL, viewR;

    s_StereoCamera.SetBaseCamera(&view);
    s_StereoCamera.CalculateMatrices(&projL, &viewL, &projR, &viewR, s_DepthLevel, s_Factor, nn::math::PIVOT_UPSIDE_TO_TOP);

    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0);
    s_RenderSystem.Clear();
    // Render the image for the left eye
    DrawDisplay0(&projL, &viewL);
    // Send the image for the left eye to the display buffer
    s_RenderSystem.Transfer();

    // The render buffer for the upper screen is shared, so clear it once
    s_RenderSystem.SetRenderTarget(NN_GX_DISPLAY0_EXT);
    s_RenderSystem.Clear();
    // Render the image for the right eye
    DrawDisplay0(&projR, &viewR);
    // Send the image for the right eye to the display buffer
    s_RenderSystem.Transfer();

    // Swap the display buffers for the left and right eyes used for the upper screen
    s_RenderSystem.SwapBuffers();

    // Render the lower screen
    DrawDisplay1();

    s_RenderSystem.WaitVsync(NN_GX_DISPLAY_BOTH);

    ++s_Count;

    return !glGetError();

}

static void PadReadAndEvent(nn::hid::PadReader& padReader)
{
    nn::hid::PadStatus pad;
    padReader.ReadLatest(&pad);

    if (pad.trigger & nn::hid::BUTTON_UP)
    {
        if ( --s_CurrentMode < 0)
            s_CurrentMode = MODE_NUM - 1;
    }

    if (pad.trigger & nn::hid::BUTTON_DOWN)
    {
        if ( ++s_CurrentMode >= MODE_NUM )
            s_CurrentMode = 0;
    }

    if (pad.hold & nn::hid::BUTTON_LEFT)
    {
        switch (s_CurrentMode)
        {
        case 0:     // Depth Level
            s_DepthLevel -= 0.01f;
            if (s_DepthLevel <= 0.0f)
                s_DepthLevel = 0.01f;
            break;
        case 1:     // Factor
            {
                s_Factor -= 0.001f;
                // 0.0f must be less than or equal to s_Factor
                if (s_Factor < 0.0f)
                    s_Factor = 0.0f;
            }
            break;
        }
    }

    if (pad.hold & nn::hid::BUTTON_RIGHT)
    {
        switch (s_CurrentMode)
        {
        case 0:     // Depth Level
            {
                s_DepthLevel += 0.01f;
            }
            break;
        case 1:     // Factor
            {
                s_Factor += 0.001f;
                if (s_Factor >= 1.0f)
                    s_Factor = 1.0f;
            }
            break;
        }
    }
}

static int Initialize(void)
{
    nn::hid::Initialize();

    // fs initialization
    nn::fs::Initialize();

    const size_t ROMFS_BUFFER_SIZE = 1024 * 64;
    static char buffer[ROMFS_BUFFER_SIZE];
    NN_UTIL_PANIC_IF_FAILED(
        nn::fs::MountRom(16, 16, buffer, ROMFS_BUFFER_SIZE));

    s_StereoCamera.Initialize();

    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(), nn::os::GetDeviceMemorySize() );
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(s_GxHeapSize));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, s_GxHeapSize);

    s_PgID = glCreateProgram();
    s_ShID = glCreateShader(GL_VERTEX_SHADER);

    nn::fs::FileReader file(L"rom:/shader.shbin");
    size_t fileSize = file.GetSize();
    void* buf = s_AppHeap.Allocate(fileSize);

    s32 read = file.Read(buf, fileSize);
    glShaderBinary(1, &s_ShID, GL_PLATFORM_BINARY_DMP, buf, read);
    file.Finalize();
    s_AppHeap.Free(buf);

    glAttachShader(s_PgID, s_ShID);
    glAttachShader(s_PgID, GL_DMP_FRAGMENT_SHADER_DMP);

    glBindAttribLocation(s_PgID, 0, "aPosition");
    glBindAttribLocation(s_PgID, 1, "aColor");

    glLinkProgram(s_PgID);
    glValidateProgram(s_PgID);
    glUseProgram(s_PgID);

    glClearColor(0.36f, 0.42f, 0.5f, 1.0f);
    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.36f, 0.42f, 0.5f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.36f, 0.42f, 0.5f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0_EXT, 0.36f, 0.42f, 0.5f, 1.0f);

    LoadObjects();

    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcRgb"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);
    glUniform3i(glGetUniformLocation(s_PgID, "dmp_TexEnv[0].srcAlpha"), GL_PRIMARY_COLOR, GL_PRIMARY_COLOR, GL_PRIMARY_COLOR);

    s_StereoCamera.SetBaseFrustum( -0.04f, 0.04f, -0.04f*nn::gx::DISPLAY0_WIDTH/nn::gx::DISPLAY0_HEIGHT,
                           0.04f*nn::gx::DISPLAY0_WIDTH/nn::gx::DISPLAY0_HEIGHT, 0.2f, 30.f);

    s_RenderSystem.SetLcdMode(NN_GX_DISPLAYMODE_STEREO);

    return 0;
}

void StartDemo()
{
    /* initialization */
    if (Initialize() >= 0)
    {
        // After this, the application itself handles sleep
        TransitionHandler::EnableSleep();

        nn::hid::PadReader padReader;
        while (1)
        {
            PadReadAndEvent(padReader);
            (void)DrawFrame();

            // Run sleep, HOME Menu, and POWER Menu transitions.
            TransitionHandler::Process();
            // Determine whether there has been an exit notification.
            if (TransitionHandler::IsExitRequired())
            {
                break;
            }
        }
    }

    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_StereoCamera.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
    s_AppHeap.Finalize();
}

void nnMain()
{
    // Initialize via the applet.
    // Sleep Mode is rejected automatically until TransitionHandler::EnableSleep is called.
    TransitionHandler::Initialize();
    // Here we need to check for exit notifications.
    if (!TransitionHandler::IsExitRequired())
    {
        StartDemo();
    }
    // Finalize via the applet.
    TransitionHandler::Finalize();

    // Finalize the application. The call to nn::applet::CloseApplication does not return.
    nn::applet::CloseApplication();
}
